/* $Id: e2p_cpbar.c 1861 2006-04-25 08:27:44Z tpgww $
 * **************************************
 *
 * Emelcopybar - a plugin for emelFM2
 * by Florian Zaehringer
 * with help from tooar and tom to port it to Gtk+2
 *
 * source of source-code and inspiration: 
 * emelFM by Michael Clark
 *
 * **************************************
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 ***************************************/
#include "emelfm2.h"
#include <string.h>
#include <pthread.h>
#include "e2_plugins.h"
#include "e2_dialog.h"
#include "e2_option.h"
#include "e2_filelist.h"
#include "e2_task.h"

//max. no. of _bytes_, at the end of a copied item's source-path
//and dest-path, shown in the progress dialog
//paths longer than this will be 'ellipsized'
#define MAX_CHAR 55
//size of buffer for progress dialog
//make this >= (MAX_CHAR+4)*2 + space for rest of message text
#define MAX_MSG 220
//interval (usec) between progress window updates
//for items < and > 10MB resepcively
#define MIN_UPDATE_INTERVAL 100000
#define MAX_UPDATE_INTERVAL 200000
//include things to support pause and resume
#define IS_PAUSABLE

#ifdef IS_PAUSABLE
//these are to support suspend/resume
//#if defined (E2_CURRENTGTK_VERSION) && GTK_CHECK_VERSION (2, 6, 0)
#define SUSICON GTK_STOCK_MEDIA_PAUSE
#define RESICON GTK_STOCK_MEDIA_PLAY
/*#else
#define SUSICON NULL
#define RESICON NULL
#endif */

typedef enum
{
	E2_BARTASK_STOPPED = 1 << 0,
	E2_BARTASK_PAUSED = 1 << 1,
	E2_BARTASK_COMPLETED = 1 << 2,
	E2_BARTASK_SUCCEEDED = 1 << 3,
} E2_BarFlags;
#endif

typedef struct
{
	guint64 count;	//count of items to be processed, in active pane
	guint64 totalsize;	//aggregate apparent size of items to be processed
} E2_BarData;

typedef struct
{
	pthread_mutex_t mutex;  //protects access to these data
	pthread_cond_t cond;   //signals change
	gchar *dlocal;	//localised destination (maybe temp) path of item copied
	guint64 done_size;
	guint64 refresh_interval;
} E2_ProgressData;

typedef struct
{
//	pthread_mutex_t mutex;  //protects access to these data
	gchar *slocal;	//localised source path of item copied
	gchar *dlocal;	//localised destination path of item copied
	gboolean completed;	//set TRUE when the backend task is completed (valid or not)
	gboolean result;	//value returned by backend task function
} E2_ActionData;

typedef struct
{
	GtkWidget *dialog;
	GtkWidget *label;
//	GtkWidget *label2;
	GtkWidget *progbar;
#ifdef IS_PAUSABLE
	GtkWidget *pause_btn;
	GtkWidget *resume_btn;
	GtkWidget *stop_btn;
	E2_BarFlags flags;
#endif
} E2_BarWindowData;

/**
 * @brief update items' count and total size
 * This is a callback for a treewalk function
 * @param local_name path of item reported by the walker, localised string
 * @param statbuf pointer to struct stat with data about @a local_name
 * @param status code from the walker, indicating what type of item is being reported
 * @param twdata pointer to tw data struct
 * @return E2TW_CONTINUE always
 **/
static E2_TwResult
_e2p_cpbar_twcb (const gchar *local_name,
	const struct stat *statbuf, E2_TwStatus status, E2_BarData *twdata)
{
	gchar *utf;
	gchar *msg;
	switch (status)
	{
		case E2TW_DL:	//directory, not opened due to tree-depth limit
		case E2TW_DM:	//directory, not opened due to different file system
		{
			gchar *utf = F_FILENAME_FROM_LOCALE ((gchar *)local_name);
			gchar *msg = g_strdup_printf ("%s %s", _("Cannot open directory"), utf);
			e2_output_print_error (msg, TRUE);
			F_FREE (utf);
		}
		break;
		case E2TW_DNR:	//unreadable directory
			utf = F_FILENAME_FROM_LOCALE ((gchar *)local_name);
			msg = g_strdup_printf ("%s %s - ", _("Cannot open directory"), utf);
			e2_output_print (curr_tab, msg, NULL, FALSE, E2_ERRORTAGS, NULL);
			e2_output_print_errno ();
			F_FREE (utf);
			g_free (msg);
		case E2TW_D:	//directory
		case E2TW_F:	//not directory or link
		case E2TW_SL:	//symbolic link
		case E2TW_SLN:	//symbolic link naming non-existing file
			twdata->count++;
			twdata->totalsize += statbuf->st_size;
			break;
		case E2TW_NS:	//un-statable item
			//WARN ??
			twdata->count++;
//		case E2TW_DP:	//directory, finished
		default:
			break;
	}
	return E2TW_CONTINUE;
}
/**
 * @brief 'shorten' string @a string if it is longer than allowed
 * @a dots is set to "..." if @a string is too long, or else ""
 *
 * @param string utf string which may need to be shortened
 * @param dots pointer to store for utf string which will preceed @a string, when it is displayed
 * @param string_diff pointer to store for the no. of bytes to be omitted from the start of @a string, when it is displayed
 *
 * @return
 **/
static void
_e2p_cpbar_shorten (gchar *string, gchar **dots, gint *string_diff)
{
	if (strlen (string) > MAX_CHAR)
	{
		//nominally we will drop this many bytes off the start of string
		gint trunc = strlen (string) - MAX_CHAR;
		//but make sure we don't cut off in the middle of a char
		gchar *c = g_utf8_find_next_char (string+trunc-1, NULL);
		*string_diff = c - string;
		*dots = "...";
	}
	else
	{
		*string_diff = 0;
		*dots = "";
	}
	return;
}
/**
 * @brief determine how much progress has been made on the current item
 *
 * This is the thread function which provides data for progressive
 * updates during the course of copying an item.
 * The thread doesn't end of its own accord, but continues until
 * it is cancelled by the main thread
 *
 * @param data pointer to struct with data which is needed here
 *
 * @return NULL (never happens)
 **/
static void *
_e2p_cpbar_progress (E2_ProgressData *data)
{
	gchar *local;
	struct stat sb;
//	printd (DEBUG, "monitor thread underway");

	while (TRUE)
	{
		pthread_testcancel ();	//has a cancel-instruction been issued ?
		//prevent the main thread from setting the cancel signal now
		if (pthread_mutex_lock (&(data->mutex)))
		{
			printd (WARN, "Lock error!");
		}
		//get a local copy of the present item's destination path
		//(that will be NULL if the action thread has finished already)
		local = (data->dlocal == NULL) ? NULL : g_strdup (data->dlocal);

		if (pthread_mutex_unlock (&(data->mutex)))
		{
			printd (WARN,"Unlock error!");
		}

		if (local != NULL)
		{	//we're still underway
			//determine how much of current item has been copied
			E2_BarData pdata;
			pdata.totalsize = 0;
//			pdata.count = 0;
			if (!e2_fs_lstat (local, &sb))
			{
				if (S_ISDIR (sb.st_mode))
					e2_fs_tw (local, _e2p_cpbar_twcb, &pdata, -1, E2TW_PHYS);
				else
//			{
				pdata.totalsize = sb.st_size;
//				pdata.count = 1;
//			}
			}
			g_free (local);

			if (pthread_mutex_lock (&(data->mutex)))
			{
				printd (WARN, "Lock error!\n");
			}
			//update size data for the main thread to use
			data->done_size = pdata.totalsize;
//			printd (DEBUG, "progress is %li", pdata.totalsize);

			if (pthread_cond_signal (&data->cond))
			{
				printd (WARN,"Signal error!");
			}

			if (pthread_mutex_unlock (&(data->mutex)))
			{
				printd (WARN,"Unlock error!");
			}
		}
		//usleep is another cancellation point
		usleep (data->refresh_interval);	//delay between updates
	}
	return NULL;
}
/**
 * @brief perform the copy
 *
 * This is the thread function which performs the copy task.
 * The thread simply exits when completed
 *
 * @param data pointer to struct with data which is needed here
 *
 * @return NULL (irrelevant pointer)
 **/
static void *
_e2p_cpbar_action (E2_ActionData *data)
{
//	printd (DEBUG, "action thread underway for %s", data->slocal);
	data->result = e2_task_backend_copy	//no background-copy option in this context
		(data->slocal, data->dlocal, FALSE);	//no time preservation (yet?)
	data->completed = TRUE;
	return NULL;
}
/**
 * @brief update progress window details, and copy item @a src to @a dest
 *
 * @param src path of item to be copied, utf-8 string
 * @param dest new path of copied item, utf-8 string
 * @param dest_dir the directory part of @a dest, utf-8 string
 * @param breakflag pointer to store for abort-flag
 * @param bdata pointer to bar data struct holding progress data
 * @param tdata pointer to bar data struct holding totals data
 * @param wdata pointer to info-window data struct
 *
 * @return
 **/
static void _e2p_cpbar_exec (gchar *src, gchar *dest, gchar *dest_dir,
#ifndef IS_PAUSABLE
	gboolean *breakflag,
#endif
	E2_BarData *bdata, E2_BarData *tdata, E2_BarWindowData *wdata)
{
	gchar labeltext[MAX_MSG];	//utf-8 string, top line of progress dialog
	gchar progresstext[64];	//utf-8 string, middle line of progress dialog
	gchar *slocal, *dlocal;	//localised strings
	gchar *src_dots, *dest_dots;	//set to "...", or "" if name not shortened
	gint src_diff, dest_diff;	//offsets into src, dest when they are shortened
	gfloat fraction;

	_e2p_cpbar_shorten (src, &src_dots, &src_diff);
	_e2p_cpbar_shorten (dest_dir, &dest_dots, &dest_diff);	//FIXME do this only once

	g_snprintf (labeltext, sizeof (labeltext),
		_("copying %s%s\nto %s%s\nthis is item %llu of %llu"),
		src_dots, src+src_diff,
		dest_dots, dest_dir+dest_diff,
		bdata->count, tdata->count);
	gtk_label_set_text (GTK_LABEL (wdata->label), labeltext);

	gchar *progress_format = _("%.2f MB of %.2f MB  (%.0f\%%)");
	
	slocal = F_FILENAME_TO_LOCALE (src);
	dlocal = F_FILENAME_TO_LOCALE (dest);

	//get size of item to be copied
	E2_BarData pdata = { 0,0 };
	e2_fs_tw (slocal, _e2p_cpbar_twcb, &pdata, -1, E2TW_PHYS);

    //create action thread
#ifdef LATE_DELETE
	//work with a temp name so that the backend doesn't create
	//its own (which would prevent destination-monitoring)
	gchar *templocal = e2_utils_get_tempname (dlocal);
	E2_ActionData a_data = { slocal , templocal, FALSE, FALSE };
#else
	E2_ActionData a_data = { slocal , dlocal, FALSE, FALSE };
#endif
	pthread_t action_thread_id;
	gint status = pthread_create (&action_thread_id, NULL,
		(void *) _e2p_cpbar_action, &a_data);
	if (status != 0)
	{
		printd (WARN,"action-thread-create error!");
//		if (errno == EAGAIN)
//			printd (WARN, "not enough system resources");
		return;
	}
	//wait for task to initiate, at least
	g_usleep (50000);
	//create progress monitor, if task not finished already
	if (!a_data.completed)
	{
		gtk_widget_show (wdata->dialog);
		//create monitor thread
		E2_ProgressData m_data;
		pthread_mutex_init (&(m_data.mutex), NULL);
		pthread_cond_init (&(m_data.cond), NULL);
#ifdef LATE_DELETE
		//get the temp name now in use for copying the item
		m_data.dlocal = templocal;
#else
		m_data.dlocal = dlocal;
#endif
		m_data.done_size = 0;
		//rough approach to setting the reporting interval (usec)
//		m_data.refresh_interval = pdata.totalsize * 10;
//		if (m_data.refresh_interval > MAX_UPDATE_INTERVAL)
		if (pdata.totalsize < 10000000)
			m_data.refresh_interval = MIN_UPDATE_INTERVAL;
		else
			m_data.refresh_interval = MAX_UPDATE_INTERVAL;	//with a cap
		pthread_t mon_thread_id;
		status = pthread_create (&mon_thread_id, NULL,
			(void *) _e2p_cpbar_progress, &m_data);
		if (status != 0)
		{
			printd (WARN,"monitor-thread-create error!");
			return;
		}

		//update progress when the monitor thread reports
		guint64 progress;
		while (!a_data.completed)	//loop until the action thread is completed
		{
			if (pthread_mutex_lock (&(m_data.mutex)))
			{
				printd (WARN, "Lock error!");
			}
			while (m_data.done_size == 0)
			{
				status = pthread_cond_wait (&(m_data.cond), &(m_data.mutex));
			}
			if (status != 0)
			{
				printd (WARN,"Wait error!");
			}

			progress = m_data.done_size + bdata->totalsize;
			m_data.done_size = 0;	//reset

			if (pthread_mutex_unlock (&(m_data.mutex)))
			{
				printd (WARN,"Unlock error!");
			}

#ifdef IS_PAUSABLE
			if (wdata->flags & E2_BARTASK_STOPPED)	//user wants to abort
#else
			if (*breakflag) //user wants to abort
#endif
			{
				if (pthread_cancel (mon_thread_id))
				{
					printd (WARN,"Thread cancel error!\n");
				}
				//cancel the task thread
				if (pthread_cancel (action_thread_id))
				{
					printd (WARN,"Thread cancel error!\n");
				}
				//cleanup anything part-copied
#ifdef LATE_DELETE
				e2_task_backend_delete (templocal);
				g_free (templocal);
#else
				e2_task_backend_delete (dlocal);
#endif
				F_FREE (slocal);
				F_FREE (dlocal);
				return;
			}

			fraction =  (gdouble) progress / tdata->totalsize;
			g_snprintf (progresstext, sizeof (progresstext), progress_format,
				progress / 1048576.0, tdata->totalsize / 1048576.0, 
				fraction * 100.0);

//			gtk_label_set_text (GTK_LABEL (wdata->label2), progresstext);
			gtk_progress_bar_set_text (GTK_PROGRESS_BAR (wdata->progbar), progresstext);
			//deal with rounding errors
			if (fraction > 1.0)
				fraction = 1.0;
			gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (wdata->progbar), fraction);
			WAIT_FOR_EVENTS;	//make sure window contents are updated
		}

		//immediate signal to monitor thread to stop processing
		if (pthread_mutex_lock (&(m_data.mutex)))
		{
			printd (WARN, "Lock error!");
		}
		m_data.dlocal = NULL;
		if (pthread_mutex_unlock (&(m_data.mutex)))
		{
			printd (WARN,"Unlock error!");
		}

		//show the full-time score ASAP
		if (a_data.result) //copy succeeded
		{
			progress = pdata.totalsize + bdata->totalsize;
			g_snprintf (progresstext, sizeof (progresstext), progress_format,
				progress / 1048576.0, tdata->totalsize / 1048576.0, 
				100.0);
//			gtk_label_set_text (GTK_LABEL (wdata->label2), progresstext);
			gtk_progress_bar_set_text (GTK_PROGRESS_BAR (wdata->progbar), progresstext);
			gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (wdata->progbar),
				(gdouble) progress / tdata->totalsize);
			//update progressive total
			bdata->totalsize = progress;
		}

		//cancel the monitor thread
		if (pthread_cancel (mon_thread_id))
		{
			printd (WARN,"Thread cancel error!\n");
		}
		//make sure things are finished before we move on
		pthread_join (action_thread_id, NULL);
		pthread_join (mon_thread_id, NULL);

		if (a_data.result) //copy succeeded
			WAIT_FOR_EVENTS;	//make sure window contents are updated

		usleep (100000);	//small delay to (sort of) show the completion status
									//(and maybe to prevent crashes ??)
		//need to slow down ...
		//this approach is empirical, needs to be validated for various systems
//		if (bdata->count > 300 && bdata->count%5 == 0)
//		if (pdata->count > 300 && bdata->count%5 == 0)
//			g_usleep (50000);	//small delay to allow filesystem to complete its business
#ifdef IS_PAUSABLE
//		if (!(wdata->flags & E2_BARTASK_PAUSED))
#endif
//		gtk_widget_hide (wdata->dialog);
	}
	else
		bdata->totalsize += pdata.totalsize;

	if (a_data.result) //copy succeeded
	{
#ifdef LATE_DELETE
		e2_task_backend_rename (templocal, dlocal);
#endif
		//update the progressive totals
//		bdata->count += pdata.count;
/* final display now done earlier
		bdata->count++;
		bdata->totalsize += pdata.totalsize;
		fraction =  (gdouble) bdata->totalsize / tdata->totalsize;
		//and show them
		g_snprintf (progresstext, sizeof (progresstext), progress_format,
			bdata->totalsize / 1048576.0, tdata->totalsize / 1048576.0, 
			fraction * 100.0);

		gtk_label_set_text (GTK_LABEL (wdata->label2), progresstext);
//		gtk_progress_bar_set_text (GTK_PROGRESS_BAR (wdata->progbar), progresstext);
		//deal with rounding errors
		if (fraction > 1.0)
			fraction = 1.0;
		gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (wdata->progbar), fraction);
		WAIT_FOR_EVENTS;	//make sure window contents are updated
*/
	}
	else
	{
		//cleanup anything part-copied
#ifdef LATE_DELETE
		e2_task_backend_delete (templocal);
#else
		e2_task_backend_delete (dlocal);
#endif
	}
#ifdef LATE_DELETE
	g_free (templocal);
#endif
	F_FREE (slocal);
	F_FREE (dlocal);

#ifdef IS_PAUSABLE
	if (wdata->flags & E2_BARTASK_PAUSED)
	{	//if it was paused, the dialog must be still visible
		e2_filelist_enable_refresh ();
		gtk_main ();
//		gtk_widget_hide (wdata->dialog);
	}
#endif		
	return;
}
#ifdef IS_PAUSABLE
/**
 * @brief callback for stop-button press and window-close event
 *
 * @param widget the activated widget which triggered the callback
 * @param flag pointer to store for flag which will abort the process when set to TRUE;
 *
 * @return TRUE always
 **/
static gboolean
_e2p_cpbar_break_cb (GtkWidget *widget, E2_BarWindowData *wdata)
{
	wdata->flags |= E2_BARTASK_STOPPED;
	//handle paused copying
	if (wdata->flags & E2_BARTASK_PAUSED)
	{
		wdata->flags &= ~E2_BARTASK_PAUSED;
		e2_filelist_disable_refresh ();
		gtk_main_quit ();
		//FIXME = cleanups
		// pthread_cancel ();
	}

	return TRUE;
}
/**
 * @brief callback for pause-button press
 *
 * @param widget the button widget which triggered the callback
 * @param resbutton pointer to paired button widget
 *
 * @return TRUE always
 **/
static gboolean
_e2p_cpbar_pause_cb (GtkWidget *widget, E2_BarWindowData *wdata)
{
	gtk_widget_set_sensitive (widget, FALSE);
	gtk_widget_set_sensitive (wdata->resume_btn, TRUE);
	gtk_widget_grab_focus (wdata->resume_btn);
	wdata->flags |= E2_BARTASK_PAUSED;
	//actual pause is started in the monitor loop
//	pthread_cancel (mon_thread_id);
// join act_thread_id
//	pthread_cancel (act_thread_id);
// ?? create a thread with a gtk_main ()
//	gint status = pthread_create (&pausethreadID, NULL,
//		(void *) _e2p_cpbar_paused, NULL);
//	pthread_join (pausethreadID, NULL);
	return TRUE;
}
/**
 * @brief callback for resume-button press
 *
 * @param widget the button widget which triggered the callback
 * @param resbutton pointer to paired button widget
 *
 * @return TRUE always
 **/
static gboolean
_e2p_cpbar_resume_cb (GtkWidget *widget, E2_BarWindowData *wdata)
{
	gtk_widget_set_sensitive (widget, FALSE);
	gtk_widget_set_sensitive (wdata->pause_btn, TRUE);
	gtk_widget_grab_focus (wdata->pause_btn);
	wdata->flags &= ~E2_BARTASK_PAUSED;
	e2_filelist_disable_refresh ();
	gtk_main_quit ();
	//FIXME resume copying
//	pthread_create (mon_thread_id);
//	pthread_create (act_thread_id);
	//= terminate the gtk_main () thread
	return TRUE;
}
#else
/**
 * @brief callback for stop-button press and window-close event
 *
 * @param widget the activated widget which triggered the callback
 * @param flag pointer to store for flag which will abort the process when set to TRUE;
 *
 * @return TRUE always
 **/
static gboolean
_e2p_cpbar_break_cb (GtkWidget *widget,
#ifdef IS_PAUSABLE
	E2_BarFlags *flags
#else
	gboolean *flag
#endif
	)
{
#ifdef IS_PAUSABLE
	*flags |= E2_BARTASK_STOPPED;
#else
	*flag = TRUE;
#endif
	return TRUE;
}
#endif
/**
 * @brief setup progress display window, and process each selected item
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget optional pointer to widget where the action was initiated, generally NULL
 *
 * @return TRUE if action completed successfully, else FALSE
 */
static gboolean
_e2p_cpbar (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
	GtkWidget *vbox;	//, *hbox;
	GList *snapshot_list, *snapshot_base;
	gchar *src_dir, *dest_dir, *src, *dest;	//pointers to utf strings
	gchar *slocal, *dlocal;	//pointers to localised strings

	FileInfo *info = e2_fileview_get_selected_first_local (curr_view);
	if (info == NULL)
		return FALSE;

	dlocal = F_FILENAME_TO_LOCALE (other_view->dir);
	if (e2_fs_access (dlocal, W_OK))
	{
		e2_output_print_default (_("\nI recently spoke to your filesystem and it told me you\n"
			"aren't allowed to write there. You should check that...\n"));
		F_FREE (dlocal);
		return FALSE;
	}
	F_FREE (dlocal);

	//setup the information window
	E2_BarWindowData windowdata;
#ifdef IS_PAUSABLE
	windowdata.flags = 0;
#else
	gboolean breakflag = FALSE;	//this is for aborting the process
#endif
	windowdata.dialog = e2_dialog_create (NULL, NULL, _("copying"), NULL, NULL);
	e2_dialog_setup (windowdata.dialog, app.main_window);
//CHECKME permit Esc-key aborting ?
#ifdef IS_PAUSABLE
	g_signal_connect (G_OBJECT (windowdata.dialog), "delete_event",
		G_CALLBACK (_e2p_cpbar_break_cb), &windowdata);
#else
	g_signal_connect (G_OBJECT (windowdata.dialog), "delete_event",
		G_CALLBACK (_e2p_cpbar_break_cb), &breakflag);
#endif
	gtk_dialog_set_has_separator (GTK_DIALOG (windowdata.dialog), FALSE);

	vbox = GTK_DIALOG (windowdata.dialog)->vbox;
	windowdata.label = e2_widget_add_label (vbox, "", 0.0, FALSE, 0);
	
//	hbox = e2_widget_add_box (vbox, FALSE, 0, TRUE, FALSE, 0);
//	windowdata.label2 = e2_widget_add_label (hbox, "", 0.5, TRUE, 0);

	windowdata.progbar = gtk_progress_bar_new ();
	gtk_box_pack_start (GTK_BOX (vbox), windowdata.progbar, TRUE, TRUE, E2_PADDING_LARGE);
	gtk_widget_show (windowdata.progbar);
#ifdef IS_PAUSABLE
	//buttons in reverse order
	windowdata.resume_btn =
	e2_dialog_add_undefined_button_custom (windowdata.dialog, FALSE,
		E2_RESPONSE_USER1,_("_Resume"), RESICON, _("Resume copying after pause"),
		_e2p_cpbar_resume_cb, &windowdata);
	//this one is disabled for now
	gtk_widget_set_sensitive (windowdata.resume_btn, FALSE);
	windowdata.pause_btn =
	e2_dialog_add_undefined_button_custom (windowdata.dialog, FALSE,
		E2_RESPONSE_USER2,_("_Pause"), SUSICON,
		_("Suspend copying, after the current item"),
		_e2p_cpbar_pause_cb, &windowdata);
	windowdata.stop_btn =
	e2_dialog_add_undefined_button_custom (windowdata.dialog, TRUE,
		E2_RESPONSE_NOTOALL,_("_Stop"), GTK_STOCK_STOP,
		_("Abort the copying"), _e2p_cpbar_break_cb, &windowdata);
#else
	e2_dialog_add_undefined_button_custom (windowdata.dialog, TRUE,
		E2_RESPONSE_NOTOALL,_("_Stop"), GTK_STOCK_STOP,
		_("Abort the copying"), _e2p_cpbar_break_cb, &breakflag);
#endif
	//accumulate non-recursed count and total size of src item(s)
	E2_BarData totaldata = { 0,0 };
	src_dir = F_FILENAME_TO_LOCALE (curr_view->dir);
	gint topcount = 0;
	//prevent loss of info items' data while processing
	e2_filelist_disable_refresh ();
	snapshot_list = snapshot_base = e2_fileview_get_selected_local (curr_view);
	for (; snapshot_list != NULL ; snapshot_list = snapshot_list->next)
	{
		slocal = e2_utils_strcat (src_dir, ((FileInfo *)snapshot_list->data)->filename);
		e2_fs_tw (slocal, _e2p_cpbar_twcb, &totaldata, -1, E2TW_PHYS);
		g_free (slocal); //glib used on localised strings, should be ok
		topcount++;
	}
	totaldata.count = topcount;	//not interested in nested count
	e2_filelist_enable_refresh ();
	F_FREE (src_dir);
	g_list_free (snapshot_base);

	//get a copy of dirs, in case user changes dir
	src_dir = g_strdup (curr_pane->path);
	dest_dir = g_strdup (other_pane->path);

	gboolean check = e2_option_bool_get ("confirm-overwrite");
	E2_BarData progressdata = { 1,0 };

	e2_widget_set_cursor (GDK_WATCH);
	e2_filelist_disable_refresh ();

	GPtrArray *names = e2_fileview_get_selected (curr_view);
	guint count;
	E2_SelectedItemInfo **iterator = (E2_SelectedItemInfo **) names->pdata;
//	const_base = const_list;
	OW_ButtonFlags extras = (topcount > 1) ? BOTHALL : NONE;
//	totaldata.count = topcount;	//report only the count of selected items

	for (count=0; count < names->len; count++, iterator++)
	{
#ifdef IS_PAUSABLE
		if (windowdata.flags & E2_BARTASK_STOPPED)
#else
		if (breakflag)
#endif
			break;	//user pressed stop btn or closed info window
		//src_dir, dest_dir have trailing "/"
		src = g_strconcat (src_dir, (*iterator)->filename, NULL);
		dest = g_strconcat (dest_dir, (*iterator)->filename, NULL);
		dlocal = F_FILENAME_TO_LOCALE (dest);

		if (check && e2_fs_access2 (dlocal) == 0)
		{
			e2_filelist_enable_refresh ();  //allow updates while we wait
			slocal = F_FILENAME_TO_LOCALE (src);
			DialogButtons result = e2_dialog_ow_check (slocal, dlocal, extras);
			F_FREE (slocal);
			e2_filelist_disable_refresh ();
			switch (result)
			{
				case YES_TO_ALL:
					check = FALSE;
				case OK:
					_e2p_cpbar_exec (src, dest, dest_dir,
#ifndef IS_PAUSABLE
						&breakflag,
#endif
						&progressdata, &totaldata, &windowdata);
					break;
				case CANCEL:
					break;
				default:
					result = NO_TO_ALL;
					break;
			}
			if (result == NO_TO_ALL)
			{
				g_free (src);
				g_free (dest);
				F_FREE (dlocal);	//finished with dlocal for this item
				break;
			}
		}
		else  //no overwrite, or don't care
		{
			_e2p_cpbar_exec (src, dest, dest_dir,
#ifndef IS_PAUSABLE
				&breakflag,
#endif
				&progressdata, &totaldata, &windowdata);
		}
		g_free (src);
		g_free (dest);
		F_FREE (dlocal);	//finished with dlocal for this item
		progressdata.count++;
	}
	gtk_widget_destroy (windowdata.dialog);
	e2_fileview_clean_selected (names);
	g_free (src_dir);
	g_free (dest_dir);

	//CHECKME E2_INCLIST - do other view, & current if it's same
	e2_filelist_check_dirty (GINT_TO_POINTER(1));

	e2_widget_set_cursor (GDK_LEFT_PTR);
	e2_filelist_enable_refresh ();
	return TRUE;
}

//aname must be confined to this module
static gchar *aname;
/**
 * @brief initialize this plugin
 *
 * @param p pointer to plugin data struct to be populated
 *
 * @return TRUE if the intialization succeeded
 **/
gboolean
init_plugin (Plugin *p)
{
#define ANAME "cpbar"
	aname = _("cpbar");

	p->signature = ANAME VERSION;
	p->menu_name = _("_Copy");
	p->description = _("Copy selected item(s), with displayed progress details");
	p->icon = "plugin_copy_"E2IP".png";  //use icon file pathname if appropriate

	if (p->plugin_cb == NULL)
	{
		p->plugin_cb = _e2p_cpbar;  //set plugins menu item function
		//no need to free this
		gchar *action_name = g_strconcat (_A(5),".",aname,NULL);
		e2_plugins_action_register (action_name, E2_ACTION_TYPE_ITEM, _e2p_cpbar,
		  NULL, FALSE, 0, NULL);
		return TRUE;
	}
	return FALSE;
}
/**
 * @brief cleanup transient things for this plugin
 *
 * @param p pointer to data struct for the plugin
 *
 * @return TRUE if all cleanups were completed
 **/
gboolean clean_plugin (Plugin *p)
{
  gchar *action_name = g_strconcat (_A(5),".",aname,NULL);
  gboolean ret = e2_plugins_action_unregister (action_name);
  g_free (action_name);
  return ret;
}
