/* $Id: e2p_pack.c 1856 2006-04-24 12:51:28Z tpgww $
 *
 * Copyright (C) 2003-2005 tooar <tooar@gmx.net>
 * Portions copyright (C) 1999 Michael Clark
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "emelfm2.h"
#include <string.h>
#include "e2_plugins.h"
#include "e2_dialog.h"
#include "e2_command.h"
#include "e2_filelist.h"

static GtkWidget *filename_entry;
static GtkWidget *pkgtype_combo;

enum { TAR_GZ, TAR_BZ2, TAR, ZIP, RAR, ARJ, MAXTYPES };

static gboolean _e2p_pack_no_cb (GtkWidget *widget, GtkWidget *dialog)
{
//  gtk_widget_set_sensitive (app.main_window, TRUE);
  gtk_widget_destroy (dialog);
  gtk_main_quit ();
  return TRUE;
}

static gboolean _e2p_pack_yes_cb (GtkWidget *widget, GtkWidget *dialog)
{
  static gchar *name_str [MAXTYPES] =
  {	//these extension strings are in same order as enum
    ".tar.gz",
    ".tar.bz2",
    ".tar",
    ".zip",
	".rar",
	".arj"
  };
  static gchar *cmd_str [MAXTYPES] =
  {	//these command strings are in same order as enum
	  //NOTE that %%f is converted to %f when the command is constructed with g_strdup_printf ()
    ">tar cvf - %%f | gzip - > \"%s\"",	//run in separate shell
    ">tar cvf - %%f | bzip2 - > \"%s\"",	//ditto
    "tar cvf \"%s\" %%f",
    "zip -r \"%s\" %%f",
	"rar a -r -ol \"%s\" %%f",
	"arj a -a1 -r \"%s\" %%f"
  };

  gint pkg_type = gtk_combo_box_get_active (GTK_COMBO_BOX (pkgtype_combo));
  if (pkg_type != -1)
  {
	const gchar *chosen_name = gtk_entry_get_text (GTK_ENTRY (filename_entry));
	gboolean flag = (*chosen_name != '\0');
	gchar *full_name;
	if (flag)
	{
		full_name = g_strconcat (chosen_name, name_str [pkg_type], NULL);
		if (e2_option_bool_get ("confirm-overwrite"))
		{
			gchar *utf = g_strdup_printf ("%s%s", curr_view->dir, full_name);  //separator comes with dir
    		gchar *dlocal = F_FILENAME_TO_LOCALE (utf);
			if (e2_fs_access2 (dlocal) == 0)
			{	//same-named item exists already
				DialogButtons result = e2_dialog_ow_check (NULL, dlocal, NONE);
				if (result != OK)
				{
					flag = FALSE;	//signal that we will not proceed
					g_free (full_name);
				}
			}
			g_free (utf);
			F_FREE (dlocal);
		}
	}
	if (flag)
	{
		gchar *command =  g_strdup_printf (cmd_str [pkg_type], full_name);
		e2_filelist_disable_refresh ();	//prevent changes to selected item data
		e2_command_run (command, E2_COMMAND_RANGE_DEFAULT);
		e2_filelist_enable_refresh ();
    	g_free (command);
		g_free (full_name);
	}
  }

  return _e2p_pack_no_cb (widget, dialog);
}

static gboolean _e2p_pack_delete_cb (GtkWidget *widget,
	GdkEvent *event, GtkWidget *dialog)
{
    return _e2p_pack_no_cb (widget, dialog);
}

static gboolean _e2p_pack_key_press_cb (GtkWidget *widget,
	GdkEventKey *event, GtkWidget *dialog)
{
  if (event->keyval == GDK_Return)
	  return _e2p_pack_yes_cb (widget, dialog);
  else if (event->keyval == GDK_Escape)
    return _e2p_pack_no_cb (widget, dialog);
	  
  return FALSE;
}

/**
 * @brief create and run dialog
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget optional pointer to widget where the action was initiated, generally NULL
 *
 * @return TRUE if action completed successfully, else FALSE
 */
static gboolean _e2p_pack (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
  FileInfo *info = e2_fileview_get_selected_first_local (curr_view);
  if (info == NULL)
	return FALSE;

  GtkWidget *dialog = e2_dialog_create (NULL,
	_("Filename:"), _("archive creation"), NULL, NULL);

  GtkWidget *hbox = g_object_get_data (G_OBJECT (dialog), "e2-dialog-hbox");
  gchar *suggested_name = F_FILENAME_FROM_LOCALE (info->filename);
  filename_entry = e2_widget_add_entry (hbox, suggested_name, TRUE, TRUE);
  F_FREE (suggested_name);

  pkgtype_combo = gtk_combo_box_entry_new_text ();
  //these package name strings are in same order as enum
  gtk_combo_box_append_text (GTK_COMBO_BOX (pkgtype_combo), _(".tar.gz"));  //index 0
  gtk_combo_box_append_text (GTK_COMBO_BOX (pkgtype_combo), _(".tar.bz2"));
  gtk_combo_box_append_text (GTK_COMBO_BOX (pkgtype_combo), _(".tar"));
  gtk_combo_box_append_text (GTK_COMBO_BOX (pkgtype_combo), _(".zip"));
  gtk_combo_box_append_text (GTK_COMBO_BOX (pkgtype_combo), _(".rar"));
  gtk_combo_box_append_text (GTK_COMBO_BOX (pkgtype_combo), _(".arj"));  //index 5
  gtk_combo_box_set_active (GTK_COMBO_BOX (pkgtype_combo), 0);
  gtk_widget_set_size_request (pkgtype_combo, 120, -1);  //'natural' width (= filename entry width) too big
  gtk_box_pack_start (GTK_BOX (hbox), pkgtype_combo, TRUE, TRUE, 0);
  gtk_widget_show (pkgtype_combo);

  g_signal_connect (G_OBJECT (dialog), "delete_event",
	G_CALLBACK (_e2p_pack_delete_cb), dialog);
  g_signal_connect (G_OBJECT (dialog), "key_press_event",
	G_CALLBACK (_e2p_pack_key_press_cb), dialog);

  e2_dialog_add_button_custom (dialog, FALSE, &E2_BUTTON_CANCEL,
	NULL, _e2p_pack_no_cb, dialog);
  e2_dialog_add_button_custom (dialog, TRUE, &E2_BUTTON_OK,
	NULL, _e2p_pack_yes_cb, dialog);

  gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
  gtk_widget_show (dialog);
//  gtk_widget_set_sensitive (app.main_window, FALSE);

  gtk_main();
  return TRUE;
}

//aname must be confined to this module
static gchar *aname;
/**
 * @brief plugin initialization function, called by main program
 *
 * @param p ptr to plugin data struct
 *
 * @return TRUE if the initialization succeeds, else FALSE
 **/
gboolean init_plugin (Plugin *p)
{
#define ANAME "pack"
  aname = _("pack");

  p->signature = ANAME VERSION;
  p->menu_name = _("_Pack..");
  p->description = _("Build an archive containing the selected item(s)");
  p->icon = "plugin_"ANAME"_"E2IP".png";  //use icon file pathname if appropriate

  if (p->plugin_cb == NULL)
  {
	p->plugin_cb = _e2p_pack;  //set plugins menu item function
	//no need to free this
	gchar *action_name = g_strconcat (_A(5),".",aname,NULL);
	e2_plugins_action_register (action_name, E2_ACTION_TYPE_ITEM, _e2p_pack,
	  NULL, FALSE, 0, NULL);
	return TRUE;
  }
  return FALSE;
}
/**
 * @brief cleanup transient things for this plugin
 *
 * @param p pointer to data struct for the plugin
 *
 * @return TRUE if all cleanups were completed
 **/
gboolean clean_plugin (Plugin *p)
{
  gchar *action_name = g_strconcat (_A(5),".",aname,NULL);
  gboolean ret = e2_plugins_action_unregister (action_name);
  g_free (action_name);
  return ret;
}
