/* $Id$
 *
 * Copyright (C) 2006 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file plugins/e2p_times.c 
 * @brief plugin for setting item times, and related functions
 * This is NOT thread-safe
 **/

#include "emelfm2.h"
#include <time.h>
#include <sys/time.h>
#include <string.h>
#include "e2_plugins.h"
#include "e2_filelist.h"
#include "e2_dialog.h"
#include "e2_ownership_dialog.h"
#include "e2_task.h"

//this is the deepest level of nested dirs that will be processed
#define E2_DIRNEST_LIMIT 32

//this is a copy from task backend
typedef struct 
{
	gchar *path;
	mode_t mode;
} E2P_DirEnt;

typedef struct
{
	gboolean continued_after_problem;
	time_t mtime;
	time_t atime;
	time_t ctime;
	GList *dirdata;
} E2_TouchData;

static GtkWidget *mtime_entry;
static GtkWidget *atime_entry;
static GtkWidget *ctime_entry;
static GtkWidget *set_mtime_button;
static GtkWidget *set_atime_button;
static GtkWidget *set_ctime_button;
static GtkWidget *recurse_button = NULL;
static gboolean *recurse;
static gboolean permission_local;
static gint real_response;	//store for over-ride dialog response

//path is localised
static gboolean _e2pt_touch1 (gchar *path, struct stat *sb,
	E2_TouchData *data)
{
	struct utimbuf tb;
	struct timeval systime1, systime2, strt, nd;
	struct timezone tz;
	//old or new mod time
	tb.modtime = (data->mtime == (time_t ) -1) ? sb->st_mtime : data->mtime;
	//old or new access time
	tb.actime = (data->atime == (time_t ) -1) ? sb->st_atime : data->atime;

	gboolean retval = TRUE;
	struct tm *tmptr;
	gboolean fixc = (data->ctime != (time_t) -1);
	if (fixc)
	{	//store current sys time
		gettimeofday (&systime1, &tz);
		//set sys time to required ctime approx.
		time_t _now = time (NULL);
		strt.tv_sec = data->ctime;
		strt.tv_usec = 0;
		tmptr = localtime (&_now);
		if (tmptr->tm_isdst > 0)
		{  //daylight savings is in effect
			strt.tv_sec -= 3600; //FIXME correct offset
		}
		else
			printd (DEBUG, "DST hack NOT done");
			
		settimeofday (&strt, NULL);
	}
	if (fixc || tb.modtime != sb->st_mtime || tb.actime != sb->st_atime)
	{
		if (e2_fs_utime (path, &tb))
		{ //handle error
			retval = FALSE;
		}
	}
	if (fixc)
	{ //find elapsed time since sys time was reset
		gettimeofday (&nd, NULL);
		gint nsec;
/*		if (nd.tv_usec < strt.tv_usec)
		{
		    nsec = (strt.tv_usec - nd.tv_usec) / 1000000 + 1;
    		strt.tv_usec -= 1000000 * nsec;
    		strt.tv_sec += nsec;
		}
		if (nd.tv_usec - strt.tv_usec > 1000000)
		{
			nsec = (strt.tv_usec - nd.tv_usec) / 1000000;
			strt.tv_usec += 1000000 * nsec;
			strt.tv_sec -= nsec;
		} */
		systime2.tv_sec = systime1.tv_sec + nd.tv_sec - strt.tv_sec;
		systime2.tv_usec = systime1.tv_usec + nd.tv_usec - strt.tv_usec;
		if (systime2.tv_usec > 1000000)
		{
			nsec = systime2.tv_usec / 1000000 + 1;
			systime2.tv_usec -= 1000000 * nsec;
			systime2.tv_sec += nsec;
		}
		//set sys time = old sys time + elapsed time
		//no DST correction needed
		settimeofday (&systime2, &tz);
   }
   return retval;
}
/**
 * @brief callback function for recursive directory touches
 *
 * Tree is being walked breadth-first, not physical.
 * Dirs are made accessible and writable if not already
 * so and it's permitted, dirs are added to a list to be
 * processed after all the tree has been traversed, 
 * Other items are changed as requested (if possible)
 *
 * @param filename localised string, absolute path of item to change
 * @param statptr pointer to struct stat with info about the current item
 * @param status one of the treewalker-function status codes
 * @param user_data pointer to user-specified data
 * @return competion code: E2TW_CONTINUE if succeeds, others as appropriate
 */
//this maybe will end up in task backend
static E2_TwResult  _e2_task_twcb_touch (const char *filename,
	const struct stat *statptr, E2_TwStatus status, E2_TouchData *user_data)
{
	E2_TwResult retval = E2TW_CONTINUE;	//default error code = none
	mode_t mode, newmode;
	E2P_DirEnt *data;
	switch (status)
	{
		case E2TW_DNR:
			//try to make the dir accessible
			mode = statptr->st_mode;
			newmode = (mode | S_IWUSR | S_IXUSR) & ALLPERMS;
			if (e2_fs_chmod (filename, newmode))
			{
				e2_task_error_local (_("Cannot access directory %s"), (gchar *)filename);
				retval = E2TW_FIXME;	//E2TW_SKIP_SUBTREE;
				break;
			}
		case E2TW_D:
			if (status == E2TW_D)
			{	//we didn't get here via a _DNR report
				mode = statptr->st_mode;
				//ensure dir is writable, if we can
				if (!(mode & S_IWUSR))
				{
					newmode = (mode | S_IWUSR) & ALLPERMS;
					if (e2_fs_chmod (filename, newmode))
					{
						e2_task_error_local (_("Cannot write to directory %s"), (gchar *)filename);
						retval = E2TW_FIXME;	//E2TW_SKIP_SUBTREE;
						break;
					}
				}
				else
					newmode = mode; //set flag
			}
			if (newmode != mode)
			{
				//add this dir to list of items to restore after whole tree is scanned
				data = g_malloc (sizeof (E2P_DirEnt));
				data->path = g_strdup (filename);
				data->mode = mode & ALLPERMS;	//want to restore the original value
				user_data->dirdata = g_list_append (user_data->dirdata, data);
			}
		case E2TW_SL:
		case E2TW_SLN:
		case E2TW_F:
			_e2pt_touch1 ((gchar *)filename, (struct stat *) statptr, user_data);
		case E2TW_DP:	//ignore dir completion reports
			break;
		case E2TW_DM:
		case E2TW_DL:
			retval = E2TW_FIXME;
			break;
//		case E2TW_NS:	//un-statable item WARN ??
		default:
			retval = E2TW_STOP;
			break;
	}

	if (retval == E2TW_FIXME)
	{
		user_data->continued_after_problem = TRUE;
		retval = E2TW_CONTINUE;	//continue after bad item
	}
	return retval;
}
/**
 * @brief change time(s) of item @a path to values stored in @a statbuf
 *
 * If invoked on a non-dir, or a dir without recursion, it is processed
 * here. If recursion is invoked on a dir, a ntfw funtion is invoked. 
 * By that, all nested dir access permissons will be set to include x,
 * non-dir items will be have their new times set when
 * they are 'reported'. Finally, here, dirs will be set to their original
 * permissions (bottom-up order) at the end of the process
 * Recursive touch works on the host filesystem only.
 * Links are not affected, their target is processed. (OK ??)
 *
 * @param path localised string, absolute path of item to process
 * @param statbuf ptr to data struct with times to set
 * @param recurse TRUE to do recursive change (ignored if @a path is not a dir)
 * @return TRUE if operation succeeds
 */
static gboolean _e2p_touch (gchar *path, E2_TouchData *data, gboolean recurse)
{
	struct stat statbuf;
	if (recurse)
	{
		//confirm whether path is a dir or not
		if (e2_fs_stat (path, &statbuf))  //look _through_ links
		{
			//abort if we can't find the item
			return (e2_task_error_local (_("Cannot get current data for %s"), path));
		}
		if (S_ISDIR (statbuf.st_mode))
		{	//recursive touch

			//no processed dirs yet
			data->dirdata = NULL;

			E2_TwFlags flags = E2TW_MOUNT;	//this filesystem only, breadth-first
			gboolean retval = e2_fs_tw (path, _e2_task_twcb_touch, data, E2_DIRNEST_LIMIT, flags);

			if (data->dirdata != NULL)
			{	//revert dir permissions, LIFO (=bottom-up) order
				GList *base = g_list_last (data->dirdata);
				for (; base!=NULL ; base=base->prev)
				{
					E2P_DirEnt *dirfix = base->data;
					if (e2_fs_chmod (dirfix->path, dirfix->mode) && errno != ENOENT) //FIXME vfs
						e2_task_error_local (_("Cannot change permissions of %s"), path);
					g_free (dirfix->path);
					g_free (dirfix);
				}
				g_list_free (data->dirdata);
			}
//			g_free (data); stacked, not heaped
			return retval;
		}
		else //not dir, handle without recurse
			recurse = FALSE;
	}
	if (!recurse)
	{
		//get current times
		if (e2_fs_lstat (path, &statbuf))  //no link pass-thru
		{
			return (e2_task_error_local (_("Cannot get current times of %s"), path));
		}
		return (_e2pt_touch1 (path, &statbuf, data));
	}
	return FALSE; //warning prevention only
}
/**
 * @brief convert date/time string in @a entry to corresponding time variable
 *
 * @param entry entry widget with string for new date/time
 * @param error pointer to store for error flag
 *
 * @return time value as a time_t, -1 if conversion fails
 */
static time_t _e2p_times_parse_time (GtkWidget *entry, gboolean *error)
{
	struct tm tm;
	//ensure clear result structure
	memset (&tm, '\0', sizeof (tm));
	time_t newtime;

	const gchar *newvalue = gtk_entry_get_text (GTK_ENTRY (entry));
	const gchar *cp = strptime (newvalue, "%x %X", &tm);
	if (cp == NULL)
		cp = strptime (newvalue, "%x", &tm);	//try with date only
/* no point in trying seconds only - mktime() fails in that case, anyway
	if (cp == NULL)
	{
		cp = strptime (newvalue, "%X", &tm);	//try with time only
		//FIXME add the current date from statbuf
	} */
	if (cp != NULL)
	{ //note if *cp != '\0' then there's still a problem
		tm.tm_isdst = 1;	//don't correct for daylight saving
		newtime = mktime (&tm);
		if (newtime != (time_t) -1)
			return newtime; 
	}
	gchar *msg = g_strdup_printf ("%s %s", _("Unable to convert date/time"), newvalue);
	e2_output_print_error (msg, TRUE);
	*error = TRUE;
	return (time_t) -1;
}
/**
 * @brief handle ok button press
 * Converts all entered date/time strings to calendar values
 * If there's a problem, it may set a flag to produce a cancel
 * response for the dialog as a whole
 *
 * @param widget the clicked button widget
 * @param data pointer to data struct to store time values
 *
 * @return
 */
static void _e2p_times_ok_cb (GtkWidget *widget, E2_TouchData *data)
{
  gboolean error = FALSE;
  real_response = GTK_RESPONSE_OK;	//this means the default response will be returned
  if (permission_local)
  {  //store any new times
	  if (GTK_TOGGLE_BUTTON (set_mtime_button)->active)
		  data->mtime = _e2p_times_parse_time (mtime_entry, &error);
	  else
		  data->mtime = (time_t) -1;
	  if (GTK_TOGGLE_BUTTON (set_atime_button)->active)
		  data->atime = _e2p_times_parse_time (atime_entry, &error);
	  else
		  data->atime = (time_t) -1;
	  if (GTK_TOGGLE_BUTTON (set_ctime_button)->active)
	  {
		  data->ctime = _e2p_times_parse_time (ctime_entry, &error);
		  if (data->ctime != (time_t) -1)
		  {  //warn about system clock changes
			  gchar *prompt = _("Changing 'ctime' requires temporary changes to "
			  "the system clock. That is normally unwise, as typically, other things "
			  "rely on system time. Click 'ok' to proceed.");
			  DialogButtons choice = e2_dialog_warning (prompt);
			  if (choice != OK)
				  data->ctime = (time_t) -1;
		  }
	  }
  	  else
		  data->ctime = (time_t) -1;

	  if (recurse_button != NULL)
		  *recurse = GTK_TOGGLE_BUTTON (recurse_button)->active;
    if (error)
	{ //error message ??
		real_response = GTK_RESPONSE_CANCEL;
		 return;
	}
	else
	{	//anything valid to change ?
		if (data->mtime == (time_t) -1
			&& data->atime == (time_t) -1
			&& data->ctime == (time_t) -1 )
			real_response = GTK_RESPONSE_CANCEL;
		return;
	}
  }
  else
  {  //no permisson, the ok click was pointless ...
	  real_response = GTK_RESPONSE_CANCEL;
	  return;
  }
}
/**
 * @brief dialog response callback - make response available to main loop
 * In general, the response of the clicked button is made available.
 * However if the ok/yesall cb generates an error, the return code
 * set there will be used instead
 *
 * @param button the clicked button widget
 * @param response the response returned by the dialog
 * @param dialog_result pointer to store for @a response or error cancel
 *
 * @return
 */
static void _e2p_times_response_decode_cb (GtkWidget *button,
	gint response, gint *dialog_result)
{
//	printd (DEBUG, "response decode cb, response is %d", response);
	*dialog_result = (real_response == GTK_RESPONSE_OK) ?
	response : real_response;
	gtk_main_quit ();
	return;
}
/**
 * @brief create entry widget for the dialog
 *
 * @return the widget
 */
static GtkWidget *_e2p_times_create_entry (void)
{
  GtkWidget *entry = gtk_entry_new ();
//  gtk_entry_set_activates_default (GTK_ENTRY (entry), TRUE);
//  gtk_widget_set_size_request (entry,  40, -1);
  gtk_widget_show (entry);
  return entry;
}
/**
 * @brief create time-change dialog
 *
 * @param info pointer to selection data struct for the item being processed
 * @param recurse_ret pointer to store for user's choice for a recursive change
 * @param statbuf pointer to filesystem data struct for the item being processed
 * @param multi TRUE if the change is part of a multi-item selection
 *
 * @return the dialog widget
 */
static GtkWidget *_e2p_times_setup (
					E2_SelectedItemInfo *info,
					gboolean *recurse_ret,
					E2_TouchData *data,
					gboolean multi )
{
  GtkWidget *permissions_dialog;
  GtkWidget *dialog_vbox, *sub_vbox;
  GtkWidget *hbox;
  GtkWidget *frame;
  GtkWidget *action_area;
  GtkWidget *table;
  struct stat statbuf;
  struct tm *tm_ptr;
  gchar date_string[32];

  // set pointers to the static values
  recurse = recurse_ret;
  gint myuid = getuid ();
  permission_local = e2_fs_check_write_permission (myuid, info->filename);
;
  //get item's current times
  //active pane path = cwd, so no need to prepend path
  gchar *local = F_FILENAME_TO_LOCALE (info->filename);
  e2_fs_lstat (local, &statbuf);
  gboolean thisis_dir = e2_fs_is_dir3 (local);
  F_FREE (local);

  permissions_dialog = gtk_dialog_new ();
  E2_DIALOG_TITLE (GTK_WINDOW (permissions_dialog), _("times"));
  gtk_dialog_set_has_separator (GTK_DIALOG (permissions_dialog), FALSE);  //all info in frames, no need
  dialog_vbox = GTK_DIALOG (permissions_dialog)->vbox;
  action_area = GTK_DIALOG (permissions_dialog)->action_area;
  gtk_container_set_border_width (GTK_CONTAINER (dialog_vbox), E2_PADDING);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX(dialog_vbox), hbox, TRUE, TRUE, E2_PADDING); //top, bottom padding 
  gchar *labeltype = (thisis_dir) ? _("Directory name") : _("Filename") ;
  gchar *label_text = g_strdup_printf ("%s: <b>%s</b>", labeltype, info->filename);
  e2_widget_add_label (hbox, label_text, 0, TRUE, E2_PADDING); //L, R padding
  gtk_widget_show (hbox);
  g_free (label_text);

  frame = gtk_frame_new (_("Times"));
  gtk_box_pack_start (GTK_BOX (dialog_vbox), frame, TRUE, TRUE, E2_PADDING);
  gtk_widget_show (frame);

  sub_vbox = gtk_vbox_new (FALSE, E2_PADDING);
  gtk_container_add (GTK_CONTAINER (frame), sub_vbox);
  gtk_widget_show (sub_vbox);

  gint rows = (thisis_dir) ? 5 : 4;
  table = e2_widget_add_table (sub_vbox, rows, 4, TRUE, TRUE, E2_PADDING);  //4 rows, 4 cols, homogen, fill
  e2_widget_add_label_to_table (table, _("Current value"), 0,		1,2,0,1);
  e2_widget_add_label_to_table (table, _("New value"), 0,	 		2,3,0,1);

  e2_widget_add_label_to_table (table, _("Accessed"), 0,	    0,1,1,2);
  e2_widget_add_label_to_table (table, _("Content Modified"), 0,0,1,2,3);
  e2_widget_add_label_to_table (table, _("Inode Changed"), 0,	0,1,3,4);

  tm_ptr = localtime (&statbuf.st_atime);
  strftime (date_string, sizeof (date_string), "%x %X", tm_ptr);
  e2_widget_add_label_to_table (table, date_string, 0,    1,2,1,2);
  tm_ptr = localtime (&statbuf.st_mtime);
  strftime (date_string, sizeof (date_string), "%x %X", tm_ptr);
  e2_widget_add_label_to_table (table, date_string, 0,    1,2,2,3);
  tm_ptr = localtime (&statbuf.st_ctime);
  strftime (date_string, sizeof (date_string), "%x %X", tm_ptr);
  e2_widget_add_label_to_table (table, date_string, 0,    1,2,3,4);
  
  atime_entry = _e2p_times_create_entry ();
  gtk_table_attach_defaults (GTK_TABLE (table), atime_entry, 2,3,1,2);
  mtime_entry = _e2p_times_create_entry ();
  gtk_table_attach_defaults (GTK_TABLE (table), mtime_entry, 2,3,2,3);
  ctime_entry = _e2p_times_create_entry ();
  gtk_table_attach_defaults (GTK_TABLE (table), ctime_entry, 2,3,3,4);

  set_atime_button = e2_button_add_toggle_to_table (table, 
		    _("_Set"), FALSE, NULL, NULL,
		  	3,4,1,2);
  set_mtime_button = e2_button_add_toggle_to_table (table, 
		    _("_Set"), FALSE, NULL, NULL,
			3,4,2,3 ); //gint left, gint right, gint top, gint bottom
  set_ctime_button = e2_button_add_toggle_to_table (table, 
		    _("_Set"), FALSE, NULL, NULL,
			3,4,3,4);
  if (thisis_dir)  //no longer do this for 'multi'
	 recurse_button = e2_button_add_toggle_to_table (table, _("R_ecurse:"),
			FALSE, NULL, NULL,
			3,4,4,5 );

  if (!permission_local) //no right to change item times
  {
	gtk_widget_set_sensitive (mtime_entry, FALSE);
	gtk_widget_set_sensitive (atime_entry, FALSE);
	gtk_widget_set_sensitive (ctime_entry, FALSE);

	gtk_widget_set_sensitive (set_mtime_button, FALSE);
	gtk_widget_set_sensitive (set_atime_button, FALSE);
	gtk_widget_set_sensitive (set_ctime_button, FALSE);
	gtk_widget_set_sensitive (recurse_button, FALSE);

	e2_ownership_dialog_warn (dialog_vbox);
  }
  else if (myuid != 0)
  {
	 //only root can alter system time, which is needed to change ctime
	gtk_widget_set_sensitive (ctime_entry, FALSE);
	gtk_widget_set_sensitive (set_ctime_button, FALSE);
  }
// add buttons in the order that they will appear
  GtkWidget *btn;
  if (multi)
	{
		e2_dialog_add_defined_button (permissions_dialog, &E2_BUTTON_NOTOALL);
		btn = e2_dialog_add_button_custom (permissions_dialog, FALSE,
			&E2_BUTTON_APPLYTOALL, NULL, _e2p_times_ok_cb, data);
		if (!permission_local)
			gtk_widget_set_sensitive (btn, FALSE);
	}
	if (permission_local || multi)
	{
		btn = e2_dialog_add_defined_button (permissions_dialog, &E2_BUTTON_CANCEL);
		if (!permission_local)
			gtk_widget_set_sensitive (btn, FALSE);
	}
	e2_dialog_add_button_custom (permissions_dialog, TRUE, &E2_BUTTON_OK,
		NULL, _e2p_times_ok_cb, data);

	e2_dialog_setup (permissions_dialog, app.main_window);
	return permissions_dialog;
}

/**
 * @brief show/change time(s) of selected items in active pane
 *
 * The actual operation is performed by a separate back-end function
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget optional pointer to widget where the action was initiated, generally NULL
 *
 * @return TRUE if action completed successfully, else FALSE
 */
static gboolean _e2p_task_times (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
//	printd (DEBUG, "task: permissions");
	GPtrArray *names = e2_fileview_get_selected (curr_view);
	if (names == NULL)	//nothing selected
		return FALSE;
	GString *path = g_string_sized_new (PATH_MAX+NAME_MAX);
	GtkWidget *dialog=NULL;
	guint count;
	gint result;
	E2_SelectedItemInfo **iterator = (E2_SelectedItemInfo **) names->pdata;
	gboolean multisrc = names->len > 1;
	gboolean all = FALSE;
	E2_TouchData data;
	gint myuid = getuid ();  //do this once, for speed

/* out-of-loop setup = FIXME
	GtkWidget *dialog;
	dialog = e2_permissions_dialog_setup (info, &recurse);
	e2_dialog_add_buttons_simple (dialog, buttons ..., NULL);
*/	
//	prepare_dialog();
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, times task");
#endif
	e2_filelist_disable_refresh ();
	e2_widget_set_cursor (GDK_WATCH);

	for (count=0; count < names->len; count++, iterator++)
	{
//		if (g_str_equal((*iterator)->filename, ".."))continue;
		gboolean recurse;
		//path is to current dir
 		gboolean permission = e2_fs_check_write_permission (myuid, (*iterator)->filename);
//FIXME for single-setup: instead of the following, adjust file details in dialog, reset default button
		if (all)
		{
			//check if we have permission to change this item
			//NB this tests _all_ w permissions, touch not so sophisticated ...
//			gchar *itempath = g_strconcat (curr_view->dir, (*iterator)->filename, NULL);
// 			permission = e2_utils_check_write_permission (myuid, itempath);
//			g_free (itempath);
			if (permission)
				result = GTK_RESPONSE_OK;
			else
			{
				result = GTK_RESPONSE_CANCEL;
				gchar *msg = g_strdup_printf ("%s %s", 
					_("You do not have authority to change time(s) for"),
					(*iterator)->filename);
				e2_output_print_error(msg, TRUE);
			}
		}
		else
		{
			recurse=FALSE;
			dialog = _e2p_times_setup (*iterator, &recurse, &data,
				multisrc);
			g_signal_connect (G_OBJECT (dialog), "response",
				G_CALLBACK (_e2p_times_response_decode_cb), &result);
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, times dialog");
#endif
			e2_filelist_enable_refresh ();  //allow updates while we wait
			_e2_dialog_show (dialog, app.main_window, 0);
			gtk_main ();
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, permissions dialog");
#endif
			e2_filelist_disable_refresh ();
//			printd (DEBUG, "dialog result is %d", result);
		}

		switch (result)
		{
		  case E2_RESPONSE_APPLYTOALL:
			all = TRUE;
			result = GTK_RESPONSE_OK;
		  case GTK_RESPONSE_OK:
			if (permission)
			{
				g_string_printf (path, "%s%s", curr_view->dir, (*iterator)->filename); //separator comes with dir
				gchar *slocal = F_FILENAME_TO_LOCALE (path->str);
#ifdef E2_INCLIST
				if (_e2p_touch (slocal, &data, recurse))
				{
					//FIXME update line in treeview
				}
#else
#ifdef E2_FAM
				_e2p_touch (slocal, &data, recurse);
#else
				if (_e2p_touch (slocal, &data, recurse))
					//make the file-list refresher notice successful change
					e2_fs_touchnow (curr_view->dir);
#endif
#endif
				F_FREE (slocal);
			}
		  case GTK_RESPONSE_CANCEL:
			break;
		  default:
			result = E2_RESPONSE_NOTOALL;  // break flag;
			break;
		}
		if (result == E2_RESPONSE_NOTOALL)
			break;
		if (!all && (count < names->len - 1))
			//destroy dialog, ready for next one
			gtk_widget_destroy (dialog);
	}
	e2_fileview_clean_selected (names);
	g_string_free (path, TRUE);

	//FIXME E2_INCLIST - just do other view if it's same
	e2_filelist_check_dirty (GINT_TO_POINTER(1));

	if ((count == names->len) || (result == E2_RESPONSE_NOTOALL))
		gtk_widget_destroy (dialog);
//	gtk_widget_set_sensitive (app.main_window, TRUE);
	e2_widget_set_cursor (GDK_LEFT_PTR);
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, permissions task");
#endif
	e2_filelist_enable_refresh ();
	gtk_widget_grab_focus (curr_view->treeview);
	return TRUE;
}

//aname must be confined to this module
static gchar *aname;
/**
 * @brief plugin initialization function, called by main program
 *
 * @param p ptr to plugin data struct
 *
 * @return TRUE if the initialization succeeds, else FALSE
 **/
gboolean init_plugin (Plugin *p)
{
#define ANAME "timeset"
  aname = _("timeset");

  p->signature = ANAME VERSION;
  p->menu_name = _("Change _times..");
  p->description = _("Change any of the time properties of selected items");
  p->icon = "plugin_"ANAME"_"E2IP".png";  //use icon file pathname if appropriate

  if (p->plugin_cb == NULL)
  {
	p->plugin_cb = _e2p_task_times;  //set plugins menu item function
	//no need to free this
	gchar *action_name = g_strconcat (_A(5),".",aname,NULL);
	e2_plugins_action_register (action_name, E2_ACTION_TYPE_ITEM, _e2p_task_times,
	  NULL, FALSE, 0, NULL);
	return TRUE;
  }
  return FALSE;
}
/**
 * @brief cleanup transient things for this plugin
 *
 * @param p pointer to data struct for the plugin
 *
 * @return TRUE if all cleanups were completed
 **/
gboolean clean_plugin (Plugin *p)
{
  gchar *action_name = g_strconcat (_A(5),".",aname,NULL);
  gboolean ret = e2_plugins_action_unregister (action_name);
  g_free (action_name);
  return ret;
}
