/* $Id: e2_command.c 1861 2006-04-25 08:27:44Z tpgww $
 *
 * Copyright (C) 2004-2006 tooar <tooar@gmx.net>
 * 
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/command/e2_command.c 
 * @brief functions for handling e2 commands
 *
 **/

#include "e2_command.h"
#include <string.h>
#include <sys/wait.h>
#include <signal.h>
#include "e2_alias.h"

GList *children = NULL;
gboolean tochild = FALSE;	//TRUE when sending a command to a child

static gint _e2_command_run_sync (gchar *command,
	gchar **args, gboolean show, gchar *prefix);
static gboolean _e2_command_watch (GIOChannel *ioc,
	GIOCondition cond, E2_CommandRuntime *rt, gboolean error);
//static void _e2_command_sigchld_handler (gint num);
static E2_CommandRuntime *_e2_command_set_child_data
	(gint pid, gint mode, gchar *command, gint exit, GIOChannel *in);

/**
 * @brief block incoming SIGCHLD signals
 * @return
 **/
static void _e2_command_block (void)
{
	sigset_t set;	//, oldset;

	sigemptyset (&set);
	sigaddset (&set, SIGCHLD);
	sigprocmask (SIG_BLOCK, &set, NULL);	//&oldset);
}
/**
 * @brief unblock incoming SIGCHLD signals
 * @return
 **/
static void _e2_command_unblock (void)
{
	sigset_t set;	//, oldset;
	
	//re-install the signal handler (some systems need this) WRONG
//	signal (SIGCHLD, _e2_command_sigchld_handler);
	
	//and unblock it
	sigemptyset (&set);
	sigaddset (&set, SIGCHLD);
	sigprocmask (SIG_UNBLOCK, &set, NULL);	//&oldset);
}
/*
static void signal_handler (gint sig)
{
	printd (DEBUG, "signal: %d", sig);
}
*/
/**
 * @brief clean up data for child processes which are finished
 * Note that opened pipes etc will come here, even though 
 * they're not logged as children
 * @param num the signal number (17)
 * @return
 **/
//copied from gtk faq @ http://gtk.org
static void _e2_command_sigchld_handler (gint num)
{
	printd (DEBUG, "sigchld_handler (num:%d)", num);
	pid_t pid;
	gint status;

//not needed SIGCHLD signals blocked while its handler is in progress
//	_e2_command_block ();

	//get termimated child(ren), if any
	while ((pid = waitpid ((pid_t) -1, &status, WNOHANG | WUNTRACED)) > 0)
	{
#if DEBUG_LEVEL > 2
		gint exitstatus;
		if (WIFEXITED (status))
		{
			exitstatus = WEXITSTATUS (status);
			printd (DEBUG, "child (%d) exit status = %d", (gint) pid,
				exitstatus);
		}
		else if (WIFSIGNALED (status))
		{
			exitstatus = WTERMSIG (status);
			printd (DEBUG, "child (%d) terminated by signal %d",
				(gint) pid, exitstatus);
		}
		else if (WIFSTOPPED (status))
		{
			exitstatus = WSTOPSIG (status);
			printd (DEBUG, "child (%d) stopped by signal %d", (gint) pid, exitstatus);
		}
		else
			printd (DEBUG, "child (%d) exited magically", (gint) pid);
#endif
//		if (pid > 0)
//		{
//			E2_CommandRuntime *rt;
			//keep trying until fn not "busy"
			while (_e2_command_set_child_data (pid, 0, NULL, status, NULL) == NULL);
//		}
	}
//see comment above
//	_e2_command_unblock ();
}
/**
 * @brief set handler for SIGCHLD signals
 * @return
 **/
static void _e2_command_set_sigchild_handler (void)
{
	struct sigaction sigdata;
	sigdata.sa_handler = _e2_command_sigchld_handler;
	sigemptyset (&sigdata.sa_mask);
	sigdata.sa_flags = SA_RESTART;
	sigaction (SIGCHLD, &sigdata, NULL);
/*	signal(SIGCHLD, signal_handler);
	signal(SIGPOLL, signal_handler);
	signal(SIGABRT,signal_handler);
	signal(SIGALRM,signal_handler);
	signal(SIGHUP,signal_handler);
	signal(SIGINT,signal_handler);
	signal(SIGPIPE,signal_handler);
	signal(SIGQUIT,signal_handler);
	signal(SIGTERM,signal_handler);
	signal(SIGUSR1,signal_handler);
	signal(SIGUSR2,signal_handler); */
}
/**
 * @brief process non-error-iochannel data
 * @param ioc the channel event source
 * @param cond flags indicating the condition(s) that have been satisfied
 * @param rt pointer to data struct for the command to which the channel applies
 * @return FALSE if the event source should be removed
 **/
static gboolean _e2_command_watch_std (GIOChannel *ioc,
	GIOCondition cond, E2_CommandRuntime *rt)
{
	return _e2_command_watch (ioc, cond, rt, FALSE);
}
/**
 * @brief process error-iochannel data
 * @param ioc the channel event source
 * @param cond flags indicating the condition(s) that have been satisfied
 * @param rt pointer to data struct for the command to which the channel applies
 * @return FALSE if the event source should be removed
 **/
static gboolean _e2_command_watch_err (GIOChannel *ioc,
	GIOCondition cond, E2_CommandRuntime *rt)
{
	return _e2_command_watch (ioc, cond, rt, TRUE);
}
/**
 * @brief process iochannel data
 *
 * Data are grabbed in parcels up to 4095 bytes, which are then separated
 * into lines (\n), and those lines are sent to the output printer
 * Will shut the channel down upon error
 *
 * @param ioc the channel event source
 * @param cond flags indicating the condition(s) that have been satisfied
 * @param rt pointer to data struct for the command to which the channel applies
 * @param error TRUE if this call relates to the error channel
 * @return FALSE if the event source should be removed
 **/
static gboolean _e2_command_watch (GIOChannel *ioc,
	GIOCondition cond, E2_CommandRuntime *rt, gboolean error)
{
//	printd (DEBUG, "watch (ioc:,cond:%d,rt:)", cond);
	if (cond & (G_IO_IN | G_IO_PRI))
	{
		gchar buf[5000];	//space (rounded) to read 4096 and add trailing \0
		//buf[0] = buf[4096] = '\0';
		gsize len = 0;
		GIOStatus ret = g_io_channel_read_chars (ioc, buf, 4096 * sizeof(gchar), &len, NULL);
		if (ret == G_IO_STATUS_NORMAL && len > 0)	//len>0 is also a proxy for ret == G_IO_STATUS_EOF
		{
			if (rt->show || error)
			{
//				if (len < 4096)
					*(buf + len) = '\0';
				if (error)
					e2_output_print (rt->current_tab, buf, rt->pidstr,
						FALSE, E2_ERRORTAGS, NULL);
				else
					e2_output_print (rt->current_tab, buf, rt->pidstr,
						FALSE, NULL);
/* lin-by-line processing needed if we want to parse and handle CR's BS's etc
				{
					gchar save;	//ignore compiler warning
					gchar *sep;
					gchar *print = buf;
					gboolean more = TRUE;
					while (more)
					{
						//always ascii \n, don't need utf8 search
						if ((sep = strchr (print, '\n')) != NULL)
						{
							sep++;
							save = *sep;
							*sep = '\0';
						}
						else
							more = FALSE;

						//printd (DEBUG, "printing for tab %x", rt->current_tab);
						if (error)
							e2_output_print (rt->current_tab, print, rt->pidstr,
								FALSE, E2_ERRORTAGS, NULL);
						else
							e2_output_print (rt->current_tab, print, rt->pidstr,
								FALSE, NULL);
						if (more)
						{
							*sep = save;
							print = sep;
						}
					}
				}
*/
			}
			if (ret == G_IO_STATUS_NORMAL)
				return TRUE;
		}
	}
	if (rt->closed && (cond & (G_IO_ERR | G_IO_HUP | G_IO_NVAL)))
	{
		_e2_command_block ();
		if (rt->show)
		{
			gchar *message = g_strdup_printf (">%s (%d) %s '%d'\n",
				rt->command, rt->pid, _("returned"), WEXITSTATUS (rt->exit));
			e2_output_print (rt->current_tab, message, rt->pidstr, TRUE, "small",
				(WEXITSTATUS (rt->exit) == 0) ? "green" : "red", NULL);
			g_free (message);
			E2_CommandRuntime *rt2;
			//keep trying until fn not "busy"
			while ((rt2 = _e2_command_set_child_data
				(rt->pid, 2, NULL, 0, NULL)) == NULL);
		}
		g_free (rt->command);
		g_free (rt->pidstr);
		g_io_channel_shutdown (rt->ioin, TRUE, NULL);
		g_io_channel_unref (rt->ioin);
		g_free (rt);
		_e2_command_unblock ();
		return FALSE;
	}
	if (!rt->closed && (cond & (G_IO_ERR | G_IO_HUP | G_IO_NVAL)))
	{
		rt->closed = TRUE;
		g_io_channel_shutdown (ioc, TRUE, NULL);
	}
	return FALSE;
}
/**
 * @brief  set up channel for handling inter-process data
 * @param fd descriptor for the channel
 * @param cond flags setting the mode of channel operation
 * @param func the fuction to be called when @a cond is satisfied
 * @param rt pointer to data struct for the command to which the channel applies
 * @return the channel
 **/
static GIOChannel *_e2_command_set_up_io_channel 
	(gint fd, GIOCondition cond, gpointer func, E2_CommandRuntime *rt)
{
	GIOChannel *ioc = g_io_channel_unix_new (fd);

	// Set IOChannel encoding to none to make it fit for binary data
	//(which seems to mean that _any_ encoding will work)
	g_io_channel_set_encoding (ioc, NULL, NULL);
	g_io_channel_set_buffered (ioc, FALSE);
	g_io_channel_set_close_on_unref (ioc, TRUE);

	g_io_add_watch_full (ioc, e2_option_int_get ("command-watch-priority"),
		cond, (GIOFunc) func, rt, NULL);

	//undo reference added by g_io_add_watch ()
	g_io_channel_unref (ioc);
	return ioc;
}
/**
 * @brief
 * @param pid the id of the process we're looking for
 * @return the data struct for the command, or NULL if not found
 **/
static E2_CommandRuntime *_e2_command_find_child (guint pid)
{
	if (children != NULL)
	{
		GList *tmp = g_list_first (children);
		E2_CommandRuntime *rt;
		while ((tmp != NULL))
		{
			rt = tmp->data;
			if ((rt != NULL) && (rt->pid == pid))
				return rt;
			tmp = g_list_next (tmp);
		}
	}
	return NULL;
}
/**
 * @brief find the pid of the last-started child
 *
 * @return the PID or 0 if not found
 **/
guint e2_command_find_last_child (void)
{
	GList *end = g_list_last (children);
	if (end == NULL)
		return 0;
	E2_CommandRuntime *rt = end->data;
	if (rt == NULL || !rt->start)
		return 0;
	return rt->pid;
}
/**
 * @brief kill process with id @a pid
 * @param pid the id of the process we're looking for
 * @return TRUE if a child process with that pid was found
 **/
gboolean e2_command_kill_child (guint pid)
{
	E2_CommandRuntime *rt = _e2_command_find_child (pid);
	if (rt != NULL)
	{
		gchar *command = g_strdup_printf ("kill -9 %d", pid);
		e2_command_run (command, E2_COMMAND_RANGE_DEFAULT);
		g_free (command);
		children = g_list_remove (children, rt);
		return TRUE;
	}
	return FALSE;
}
/**
 * @brief check whether process with id @a pid is running
 * @param pid the id of the process we're looking for
 * @return TRUE if a process with that pid is running
 **/
gboolean e2_command_find_process (guint pid)
{
	pid_t result = waitpid (pid, NULL, WNOHANG);
	return (result == 0);	//will be -1 if finished before, result if finished now
}
/**
 * @brief
 * @param pid the id of the process we're looking for
 * @param mode 0, 1 or 2 if called from _e2_command_sigchld_handler(), _e2_command_run_async(), _e2_command_watch(), respectively
 * @param command  executed command string (@a mode = 1) or NULL (@a mode is 0 or 2)
 * @param exit integer exit code of executed command
 * @param in iochannel pointer (@a mode = 1) or NULL (@a mode is 0 or 2)
 * @return pointer to command data struct, or NULL if this fn is already being executed
 **/
static E2_CommandRuntime *_e2_command_set_child_data
	(gint pid, gint mode, gchar *command, gint exit, GIOChannel *in)
{
	static gboolean blocked = FALSE;
	if (blocked)
		return NULL;
	else
		blocked = TRUE;
	E2_CommandRuntime *rt = _e2_command_find_child (pid);
	if (rt != NULL)
	{
		if (mode == 1)
		{
			printd (DEBUG, "update child data (pid:%d,mode:%d,command:%s,exit:%d)", pid,
				mode, rt->command, exit);
			rt->start = TRUE;
			rt->command = g_strdup (command);
			rt->ioin = in;
		} else if (mode == 0)
		{
			printd (DEBUG, "update child data (pid:%d,mode:%d,command:N/A,exit:%d)", pid,
				mode, exit);
			rt->end = TRUE;
			rt->exit = exit;
		}
		if ((mode == 2) || ((mode == 0) && (!rt->show)))
		{
			if (g_list_length (children) < 2)
			{
				g_list_free (children);
				children = NULL;
			} else
				children = g_list_remove (children, rt);
		}
	}
	else	//rt == NULL, command not yet registered
	{
		printd (DEBUG, "register_child (pid:%d,mode:%d,command:%s,exit:%d)", pid,
		mode, command, exit);
		rt = g_malloc (sizeof (E2_CommandRuntime));
		if (mode == 1)
		{
			rt->start = TRUE;
			rt->end = FALSE;
			rt->command = g_strdup (command);
			rt->ioin = in;
		} else if (mode == 0)
		{
			rt->start = FALSE;
			rt->end = TRUE;
			rt->exit = exit;
		} else if (mode == 2)
		{
			printd (ERROR, "program bug in command handling, aborting");
			//quit_cb (NULL);
			e2_main_user_shutdown ();
		}
		rt->pid = pid;
		rt->pidstr = g_strdup_printf ("%d", rt->pid);
		
		children = g_list_append (children, rt);
	}
	blocked = FALSE;
	return rt;
}
/**
 * @brief update all child tab pointers after output pane tab change
 * @param currenttab
 * @param replacetab
 * @return
 **/
void e2_command_retab_children (E2_OutputTabRuntime *currenttab,
	E2_OutputTabRuntime *replacetab)
{
	_e2_command_block ();

	GList *tmp;
	for (tmp = children; tmp != NULL; tmp = tmp->next)
	{
		E2_CommandRuntime *rt = tmp->data;
		
		if (rt != NULL && rt->start)	//no need to check for end, that's toggled at the same time
		{
			if (rt->current_tab == currenttab)
			{
				rt->current_tab = rt->background_tab;
//	printd (DEBUG, "foreground tab reverted to background %x", rt->current_tab);
			}
			else if (rt->background_tab == replacetab)
			{
				rt->current_tab = currenttab;
//	printd (DEBUG, "foreground tab changed to %x", rt->current_tab);
			}
		}
	}

	_e2_command_unblock ();
}
/**
 * @brief
 * @param currenttab
 * @param replacetab
 * @return
 **/
void e2_command_retab2_children (E2_OutputTabRuntime *currenttab,
	E2_OutputTabRuntime *replacetab)
{
	_e2_command_block ();

//	printd (DEBUG, "looking for %x, replacement %x", currenttab, replacetab);
	GList *tmp;
	for (tmp = children; tmp != NULL; tmp = tmp->next)
	{
		E2_CommandRuntime *rt = tmp->data;
//	printd (DEBUG, "found %x for %s", rt->background_tab, rt->command);
		if (rt != NULL && rt->start	//no need to check for end, that's toggled at the same time
			&& rt->background_tab == currenttab)
		{
				rt->background_tab = replacetab;
//	printd (DEBUG, "replaced it");
//	printd (DEBUG, "foreground stays at %x", rt->current_tab);
			//CHECKME changing this too fixes crashes - WHY ??
				rt->current_tab = replacetab;
		}
//		else
//	printd (DEBUG, "NOT replaced");
	}

	_e2_command_unblock ();
}
/**
 * @brief count the children of the current process
 *
 * @return the number of current children
 **/
gint e2_command_count_children (void)
{
	if (children == NULL)
		return 0;
	else
		return (g_list_length (children));
}
/**
 * @brief list the children of the current process
 * This is an action, results shown in the current tab
 * @param data UNUSED action data specified when the action was registered
 * @return
 **/
void e2_command_list_children (gpointer data)
{
//	printd (DEBUG, "e2_command_list_children (data:)");
	if (children == NULL)
	{
		e2_output_print_default (_("no children\n"));
		return;
	}
	gchar *message = _("| id |  pid  | command\n|----|-------|-----------------\n"); //_I(
	e2_output_print_default (message);

	_e2_command_block ();

	gint i = 0;
	GList *tmp;
	for (tmp = children; tmp != NULL; tmp = tmp->next)
	{
		E2_CommandRuntime *rt = tmp->data;
		
		if (rt != NULL && rt->start)	//no need to check for end, that's toggled at the same time
		{
			i++;
			message = g_strdup_printf ("| %2d | %5d | %s\n",
				i, rt->pid, rt->command);
			e2_output_print_default (message);
			g_free (message);
		}
	}

	e2_output_print_default ("\n");

	_e2_command_unblock ();
}
/**
 * @brief pipe utf-8 string @a message to process @a pid
 * The specified process must be a child of the current e2 instance
 * @param pid the id of the destination process, or 0 for the last-created child
 * @param message utf string to send to process @a pid
 * @return TRUE if the string was sent without error
 **/
static gboolean _e2_command_send_to_pid (guint pid, gchar *message)
{
	gchar *errmsg;
	printd (DEBUG, "e2_command_send_to_pid (pid:%d,messge:%s)", pid, message);
	E2_CommandRuntime *rt;
	if (pid == 0)
	{
		GList *end = g_list_last (children);
		if (end == NULL)
			return FALSE;
		rt = end->data;
/*	} else if (pid == -1)
	{
 this is not correct logic, want to send to any child that expects something from stdin
	else if (range != E2_COMMAND_RANGE_FILE_OPERATION)
	{
		//as for a normal terminal, the "command" should go to the latest running child process, if any
		GList *end;
		for (end = g_list_last (children); end != NULL; end = g_list_previous (end))
		{
			E2_CommandRuntime *rt = end->data;
			if (e2_command_find_process (rt->pid))
			{
				printd (DEBUG, "sending '%s' to last active child process %d", command1, rt->pid);
				_e2_command_send_to_pid (rt->pid, command1);
				goto cleanup;
			}
		}
	}
*/
	} else
		rt = _e2_command_find_child (pid);
	if (rt == NULL || !rt->start)
	{
		errmsg = g_strdup_printf (_("The process with pid %d is not our child"), pid);
		e2_output_print_error (errmsg, TRUE);
		g_free (errmsg);
		return FALSE;
	}
	gsize size;
	GIOStatus stat;
	GError *error = NULL;
	//CHECKME what if the process doesn't understand utf-8 ?
	stat = g_io_channel_write_chars (rt->ioin, message, -1, &size, &error);
	if (error != NULL)
	{
		errmsg = g_strdup_printf (_("Failed writing to child: %s"), error->message);
		printd (WARN, errmsg);
		e2_output_print (rt->current_tab,
			errmsg, rt->pidstr, TRUE, E2_ERRORTAGS, NULL);
		g_error_free (error);
		g_free (errmsg);
		return FALSE;
	}
	error = NULL;
	stat = g_io_channel_write_chars (rt->ioin, "\n", 1, &size, &error);
	if (error != NULL)
	{
		errmsg = g_strdup_printf (_("Failed writing to child: %s"), error->message);
		printd (WARN, errmsg);
		e2_output_print (rt->current_tab,
			errmsg, rt->pidstr, TRUE, E2_ERRORTAGS, NULL);
		g_error_free (error);
		g_free (errmsg);
		return FALSE;
	}
	else if (rt->show)
	{	//signal what has been sent
		errmsg = g_strconcat (message, "\n", NULL);
		e2_output_print (rt->current_tab,
			errmsg, rt->pidstr, FALSE, NULL);
		g_free (errmsg);
	}
	return TRUE;
}
/**
 * @brief check whether @a command has any "modifiers"
 * @param command utf command string
 * @param macros pointer to flag, set FALSE if leading '!' found
 * @param shell pointer to flag, set TRUE if leading '>' found
 * @param wait pointer to flag, set TRUE if leading '|' found
 * @param show pointer to flag, set FALSE if trailing '&' found
 * @return integer, the byte-offset into @a command where the next char to be processed is
 **/
static gint _e2_command_run_checks (gchar *command, 
	gboolean *macros, gboolean *shell, gboolean *wait, gboolean *show)
{
	gint retval = 0;
	gboolean done = FALSE;
	//checks here are all for single-gchar characters, no need for utf functions
	while (!done)
	{
		switch (command[0])
		{
			case ' ':	//skip leading whitespace
			case '\t':
				break;
			case '!':
				*macros = FALSE;
				break;
			case '>':
				*shell = TRUE;
				break;
			case '|':
				*wait = TRUE;
				break;
			default:
				done = TRUE;
				break;
		}
		if (!done)
		{
			retval++;
			command++;
		}
	}
	if (g_str_has_suffix (command, "&"))
	{
		*show = FALSE;
		command[strlen (command) - sizeof (gchar)] = '\0';
	}
	return retval;
}
/**
 * @brief asynchronously run external command
 *
 * @param command the command string, used in messages, utf8
 * @param args pointer to command arguments array, each localised
 * @param show TRUE to show @a command and its pid in the output pane
 * @param prefix string to add to the start of @a command for its display
 *
 * @return the pid of process running the command, or 0 if there's an error
 **/
static gint _e2_command_run_async (gchar *command, gchar **args,
	gboolean show, gchar *prefix)
{
	printd (DEBUG, "asynchronously run command %s", command);
	_e2_command_block ();	//WHY block SIGCHLD signals here?

	GError *error = NULL;
	gint pid = 0, pin = 0, pout = 0, perr = 0;
	gchar *local = F_FILENAME_TO_LOCALE (curr_view->dir);
	g_spawn_async_with_pipes (local, args, NULL,
		G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_SEARCH_PATH,
		NULL, NULL, &pid, &pin, &pout, &perr, &error);
	F_FREE (local);

	if (error != NULL)
	{
		gchar *message = g_strconcat (prefix, command, " ", NULL);
		//always use the current tab at the start of the command
		e2_output_print (&app.tab, message, NULL, TRUE, "bold", "blue", NULL);
		g_free (message);
		message = g_strconcat (error->message, ".", NULL);
		e2_output_print_error (message, TRUE);
		g_error_free (error);
		_e2_command_unblock ();
		return 0;
	}

	GIOChannel *ioc = g_io_channel_unix_new (pin);
	g_io_channel_set_encoding (ioc, NULL, NULL);	//binary ok?
	g_io_channel_set_buffered (ioc, FALSE);
	g_io_channel_set_close_on_unref (ioc, TRUE);

	E2_CommandRuntime *rt = NULL;
	//keep trying this until data set fn not busy
	while ((rt = _e2_command_set_child_data
		(pid, 1, command, 0, ioc)) == NULL);
	rt->show = show;
	rt->closed = FALSE;
	//setup for foreground running
	rt->background_tab = curr_tab;
	rt->current_tab = &app.tab;
	
	if (show)
	{
		gchar *message = g_strconcat (prefix, command, NULL);
		e2_output_print (rt->current_tab,
			message, rt->pidstr, FALSE, "bold", "blue", NULL);
		g_free (message);
		message = g_strdup_printf ("  (%d)\n", pid);
		e2_output_print (rt->current_tab,
			message, rt->pidstr, FALSE, "small", "blue", NULL);
		g_free (message);
	}

	_e2_command_set_up_io_channel (pout,
		G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
		_e2_command_watch_std, rt);
	//error messages will come through, no matter what
	_e2_command_set_up_io_channel (perr,
		G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
		_e2_command_watch_err, rt);

	_e2_command_unblock ();

	return pid;
}
/**
 * @brief run command @a command synchronously (block until it's finished)
 *
 * @param command the command string, used in messages, utf8
 * @param args pointer to command arguments array, each localised
 * @param show TRUE to show @a command and its exit status in the output pane
 * @param prefix string to add to the start of @a command for its display
 *
 * @return the exit code of the command, 0 if error
 **/
static gint _e2_command_run_sync (gchar *command, gchar **args,
	gboolean show, gchar *prefix)
{
	printd (DEBUG, "synchronously run command %s", command);
	gchar *message;

	_e2_command_block ();	//CHECKME why block SIGCHLD signals ?

	//make sure we can handle tab swaps, by setting up a fake command rt
	E2_CommandRuntime rt;
	
	if (show)
	{
		rt.start = TRUE;
		rt.background_tab = curr_tab;
		rt.current_tab = &app.tab;
		children = g_list_append (children, &rt);
		message = g_strconcat (prefix, command, NULL);
		e2_output_print (rt.current_tab,
			message, NULL, TRUE, "bold", "blue", NULL);
		g_free (message);
		WAIT_FOR_EVENTS;
	}

	GError *error;
	gint exit;
	gchar *sout = NULL, *serr = NULL;
	gchar *local = F_FILENAME_FROM_LOCALE (curr_view->dir);
	gboolean success = g_spawn_sync (local, args, NULL,
		G_SPAWN_SEARCH_PATH, NULL, NULL, &sout, &serr, &exit, &error);
	F_FREE (local);

	if (success && show && sout != NULL)
		e2_output_print (rt.current_tab,
			sout, NULL, FALSE, NULL);
	if (success && serr != NULL)
		e2_output_print_error (serr, FALSE);	//CHECKME true ?

	if (success && show)
	{
		message = g_strdup_printf ("%s%s %s '%d'\n", prefix, command, _("returned"),
			WEXITSTATUS (exit));
		e2_output_print (rt.current_tab, message, NULL, TRUE, "small",
			(WEXITSTATUS (exit) == 0) ? "green" : "red", NULL);
		g_free (message);
	}
	if (!success)
	{
		message = g_strdup_printf (_("Strange error: could not run %s"), command);
		e2_output_print_error (message, TRUE);
		exit = 0;
	}
	if (show)
	{
		GList *dummy = g_list_find (children, &rt);
		children = g_list_delete_link (children, dummy);
	}

	_e2_command_unblock ();
	return exit;
}
/**
 * @brief run a single command @a raw, after interpreting its contents
 *
 * Command modifiers (!>&) are interpreted.
 * $... variables are interpreted
 * %... macros are expanded unless the "no-expand" modifier '!' is present at
 * the start of @a raw.
 * Any 'pid:' directive is interpreted. The 'pid' string must describe a base 10 integer.
 * Any "\" that doesn't follows another "\" is simply removed.
 * Any "~" followed by "/" is interpreted as $HOME, as is a "~" by itself.
 * If * or ? is in any argument, that will be expanded, if possible, into a
 * series of matching itemnames in the active directory.
 * "cd" commands that are not run in a separate shell just change the
 * dir shown in the active pane. In that case the new dir must be 
 * separated by one or more ' ' chars (ie no tabs), or "cd" alone is
 * interpreted as "cd $HOME".
 * Command arguments for must be separated by one or more ' ' chars (ie no tabs)
 * NOTE used to return pid of process running the command (if any)
 * @param raw utf command string
 * @param range flags for the scope of the command
 * 
 * @return 0 for success, 1 for error
 **/
//FIXME more consistent returned value
static gint _e2_command_run_single (gchar *raw, E2_CommandRange range)
{
	gchar *command1, *command2, *freeme;
	gboolean expand_macros = TRUE;
	gboolean run_in_shell = FALSE;
	gboolean wait = FALSE;
	gboolean show = TRUE;
	gint pid;

	//1st, find how the command is to be run
	command1 = raw + _e2_command_run_checks
		(raw, &expand_macros, &run_in_shell, &wait, &show);
	//(wait and show will be overridden, later, if range == E2_COMMAND_RANGE_FILE_ACTION)

	//replace any alias if so desired
	if (//CHECKME range != E2_COMMAND_RANGE_FILE_ACTION //this is not an action on selected items
		//&& expand_macros	//no instruction to not expand macros
		//&&
		e2_option_bool_get ("command-use-aliases")	//alias use is expected
	)
	{
		freeme = e2_alias_apply (command1, &app.aliases);
		//check command operation mode again, in case alias changed it
		command1 = freeme + _e2_command_run_checks (freeme,
			&expand_macros, &run_in_shell, &wait, &show);
	}
	else
		freeme = NULL;

	E2_Action *action = e2_action_check (command1);
	gboolean command_is_action = (action != NULL);
	if (command_is_action)
	{
		if (action->exclude & E2_ACTION_EXCLUDE_ACCEL)
		{
			gchar *msg = g_strdup_printf (_("Cannot run '%s'"), command1);
			e2_output_print_error (msg, TRUE);
			if (freeme != NULL)
				g_free (freeme);
			return 1;
		}
		//set the correct action-range according to whether the
		//name begins with "file" and has no argument
		if (g_str_has_prefix (command1, _A(5)))
		{
			command2 = e2_utils_find_whitepace (command1);
			if (command2 == NULL || e2_utils_pass_whitepace (command2) == NULL)
				range = E2_COMMAND_RANGE_FILE_ACTION;
		}
		else if (range == E2_COMMAND_RANGE_FILE_ACTION)
			range = E2_COMMAND_RANGE_DEFAULT;
	}

	//set some default operating modes for selected-items commands
	//(overrides show flag in raw or alias, if any)
	if (!command_is_action && range == E2_COMMAND_RANGE_FILE_ACTION)
	{
	//CHECKME is this of any real use ? it's intended mainly for actions
	//	wait = !e2_option_bool_get ("fileop-background");
		show = e2_option_bool_get ("fileop-show");
	}

	//if appropriate, arrange to append the selected item(s),
	//as the (or additional) command argument(s)
	if (range == E2_COMMAND_RANGE_FILE_ACTION	//this is an action on selected items
		&& expand_macros	//no instruction to not expand macros
		//FIXME support macros other than with english letters
		&& strstr (command1, "%f") == NULL
		&& strstr (command1, "%F") == NULL
		&& strstr (command1, "%p") == NULL
		&& strstr (command1, "%P") == NULL
	)
		//if there is no filelist macro already in the command, add one
		command2 = g_strconcat (command1, " %f", NULL);
	else
		command2 = g_strdup (command1);
	if (freeme != NULL)
		g_free (freeme);
	//newly-allocated command2 is now the focus ..

	//replace variables etc
	command1 = e2_utils_replace_vars (command2);
	g_free (command2);
	//replace wildcards
	if (expand_macros)
		command1 = e2_utils_replace_wildcards (command1);

/*	//as a convenience, eliminate escape chars from the command
	//(emelfm2 commands per se do not need anything escaped_
	if (strchr (command1, '\\') != NULL)
	{
		command2 = command1;
		while ((command2 = strchr (command2, '\\')) != NULL)
		{
			if (*(command2+1) != '\\')
				*command2 = '\1';
			command2++;
		}
		if (strchr (command1, '\1') != NULL)
		{
			gchar old[2] = {'\1', '\0'};
			command2 = command1;
			command1 = e2_utils_str_replace (command1, old, "");
			g_free (command2);
		}
	} */
	//now newly-allocated command1 has focus, command2 is clean

	//if command is: cd [%]%f, when updir '../' line was clicked, that
	//item would not be available as a selected item, so ...
	if (g_str_has_prefix (command1, "cd"))
	{
		//FIXME support macros other than with english letters
		if (strstr (command1, "%f") != NULL)
		{
			FileInfo *info = e2_fileview_get_selected_first_local (curr_view);
			if (info != NULL)
			{
				gchar *utf = F_FILENAME_FROM_LOCALE (info->filename);
				if (g_str_equal (utf, ".."))
				{
					g_free (command1);
					command1 = g_strdup ("cd ..");	//no translation
					expand_macros = FALSE;
				}
				F_FREE (utf);
			}
		}
	}
	//expand macros, if wanted
	if (expand_macros)
	{
		GString *str = e2_utils_expand_macros (command1, NULL);
		if (str == NULL)
		{
			e2_output_print_error (_("Failed to expand macros"), FALSE); //current tab
			pid = 1;
			goto cleanup;
		}
		else if (str == GINT_TO_POINTER(1))
		{	//the user cancelled in a prompt macro
			pid = 1;
			goto cleanup;
		}
		g_free (command1);
		command1 = str->str;
		g_string_free (str, FALSE);
	}

	//interpret any 'pid:' directive
	//the 'pid' string must describe a base 10 integer, no
	//bigger than the max integer for the current system

	//i.e. should be 5 bytes long (16385) for 32-bit systems,
	//10 bytes for 64-bit systems
//#if sizeof(gint) == 4
//#define E2_PID_SIZE 5
//#else
//#define E2_PID_SIZE 10
//#endif

	command2 = NULL;
	if (tochild)
	{
		pid = (_e2_command_send_to_pid (0, command1)) ? 0 : 1;
		goto cleanup;
	}
	if (//range != E2_COMMAND_RANGE_FILE_ACTION	//CHECKME why excluded ?
		//&& 
		(command2 = strchr (command1, ':')) != NULL	//if always ascii ':', don't need g_utf8_strchr()
	)
	{
		if (command2 > command1)	// there was something ahead of the ':'
		{
			*command2 = '\0';
			gchar *end;
			gint i = (gint) g_ascii_strtoull (command1, &end, 10);
			if (end == command2)
				//the whole string was used in conversion, so it was
				//a valid number and nothing else
			{
				printd (DEBUG, "sending '%s' to process '%s'", command2+1, command1);
				pid = (_e2_command_send_to_pid (i, command2+1)) ? 0 : 1;	//+1 to skip the '\0'
				goto cleanup;
			}
			*command2 = ':';
		}
	}

	//"cd" commands that are not run in a separate shell
	//just change the dir in the active pane
	if (!run_in_shell && g_str_has_prefix (command1, "cd"))	//no translation
	{	//note: tests here rely on ascii single-byte chars in command string
		gint len = strlen (command1);
		if ((len > 3) && (command1[2] == ' '))	//FIXME or \t ?
		{
			e2_pane_change_dir (NULL, command1+3);	//CHECKME command1 utf-8 or ascii
			pid = 0;
			goto cleanup;
		}
		if (len == 2)
		{
//			gchar *home = (gchar *) g_get_home_dir ();
//			e2_pane_change_dir (NULL, home);
			e2_pane_change_dir (NULL, "~");
			pid = 0;
			goto cleanup;
		}
	}
	else if /*(g_str_has_prefix (command1, "<"))
	{	//action
		if (strchr (command1, '.') == NULL)	//quick check
		{
			e2_output_print_error (_("You must specify an action to run, like '<file.copy'"), FALSE); //current tab
			pid = 1;
			goto cleanup;
		}
		if (strchr (command1, '.') != NULL)	//quick check */
		(command_is_action)
		{
			gchar *gap = e2_utils_find_whitepace (command1);
			if (gap != NULL)
			{
				*gap = '\0';
				gap = e2_utils_pass_whitepace (gap+1);
			}
			pid = (e2_action_run_simple (command1, gap)) ?	//now gap points to args, or NULL
				0 : 1;
			goto cleanup;
		}
//	}

	gboolean free;
	gchar **args;
	gint argc;
	if (run_in_shell)
	{	//run in new shell
		free = TRUE;	//free args, and separately free command1
		argc = 3;
		//construct freeable vector of args
		gchar *args_ = g_strconcat (
		G_DIR_SEPARATOR_S"bin"G_DIR_SEPARATOR_S"sh\1-c\1",
		command1, NULL);
		args = g_strsplit (args_, "\1", argc);
	}
	else
	{	//run in same shell
		//g_shell_parse_argv() leaves one of the args = command1
		//so don't trigger a double-free
		free = FALSE;	//free args, but not doubly-so for command1
		GError *error = NULL;
		//NB the glib 2.6 API doco is wrong about how this works !!
		g_shell_parse_argv (command1, &argc, &args, &error);
		if (error != NULL)
		{
			gchar *error_message = g_strdup_printf ("%s '%s' - ", _("Failed parsing command line"), command1);
			e2_output_print (curr_tab, error_message, NULL, FALSE, E2_ERRORTAGS, NULL);
			e2_output_print_errno ();
			g_free (error_message);
			g_error_free (error);
			pid = 1;
			goto cleanup;
		}
	}
#ifndef FILES_UTF8ONLY
	if (!app.utf8_filenames)
	{
		free = TRUE;	//free all args anyway
		//localise the actual command strings
		gchar *local;
		gchar **tmp = args;
		while (*tmp != NULL)
		{	//we always take a copy and strfreev later
			local = D_FILENAME_TO_LOCALE (*tmp);
			g_free (*tmp);
			*tmp = local;
			tmp++;
		}
	}
#endif
	gchar *prefix = run_in_shell ? "sh>" : ">";
	pid = (wait) ?
		_e2_command_run_sync (command1, args, show, prefix) :
		_e2_command_run_async (command1, args, show, prefix);

	if (free)
		g_free (command1);
	g_strfreev (args);

	return ((pid == 0) ? 1 : 0);

cleanup:
	g_free (command1);
	return pid;
}

/**
 * @brief send command to last child process, if any
 * Various checks are performed, to decide whether it's ok to send the command
 *
 * @param action_data UNUSED ptr to data assigned when action struct created at session start
 * @param rt_data UNUSED ptr to data sent to this command
 * @param widget the entry widget that was activated to trigger the action
 *
 * @return TRUE if task completed successfully, else FALSE
 */
static gboolean _e2_command_send_to_lastchild (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
	gchar *msg;
	gboolean retval;
	//get the command string
	const gchar *command_raw = gtk_entry_get_text (GTK_ENTRY (widget));
	//quick exit
	if (command_raw == NULL || *command_raw == '\0')
		return FALSE;
	gchar *command = g_strdup (command_raw);
	if (strchr (command, ';') != NULL)
		goto cleanup;
	if (e2_action_check (command) == NULL)
		goto cleanup;
	GList *end = g_list_last (children);
	if (end == NULL)
		goto cleanup;
	E2_CommandRuntime *rt = end->data;
	if (!rt->start)
		goto cleanup;
	tochild = TRUE;
	//this is an action, can only return T/F, so we lose the result of the command
	_e2_command_run_single (command, E2_COMMAND_RANGE_DEFAULT);
	gtk_entry_set_text (GTK_ENTRY (widget), "");	//we don't want any history, just clear it
	retval = TRUE;
exit:
	g_free (command);
	tochild = FALSE;
	return retval;
cleanup:
	msg = g_strdup_printf (_("Cannot send \"%s\" to a child process"), command);
	e2_output_print_error (msg, TRUE);
	retval = FALSE;
	goto exit;
}

  /*****************/
 /***** public *****/
/*****************/

/**
 * @brief run command @a command, after interpreting its contents
 *
 * @a command may be a single command or a series joined by ';'
 * (any ';' inside single- or double-parentheses is ignored)
 * NOTE used to return command's pid
 * @param command utf string containing the command(s)
 * @param range flags for the scope of the command (may be over-ridden, for actions)
 * @return integer return code from last command executed, 0 or 1
 **/
gint e2_command_run (gchar *command, E2_CommandRange range)
{
	if ((command == NULL) || (*command == '\0'))
		return 1;
	g_strstrip (command);
	if (*command == '\0')
		return 1;

	printd (DEBUG, "e2_command_run (command:%s,range:%d)", command, range);
	//parse & handle joined commands here
	//in particular, ignore separators inside parentheses
	if (strchr (command, ';') == NULL)	//if always ascii ;, don't need g_utf8_strchr()
		//no joined commands
		return _e2_command_run_single (command, range);
	gchar *s = command;
	gchar *p = s;
	gint result = 0;
	gint cnt1 = 0; //counter for ' restarted for each cmd
	gint cnt2 = 0; //counter for "
	while (*p != '\0')
	{
		if (*p == '\'')
		{
			if (p == command || *(p-1) != '\\')
				cnt1++;
		}
		else if (*p == '"')
		{
			if (p == command || *(p-1) != '\\')
				cnt2++;
		}
		else if (*p == ';')
		{	//check if separator seems to be outside parentheses
			if (cnt1%2 == 0 && cnt2%2 == 0)
			{ //process the command
				*p = '\0';
				result = _e2_command_run_single (s, range);
				s = ++p;	//resume from the next spot
				cnt1 = cnt2 = 0;
				continue;
			}
		}
		p++;
	}
	if (*s == '\0') //a separator at end of line
		return result;
	return (_e2_command_run_single (s, range));
}
/**
 * @brief display commands help document
 * The document name is hard-coded with the help-file name
 * defined at build-time
 * @return
 **/
void e2_command_output_help (void)
{	//check whether there's still an alias for the help command
	//(and BTW get a copy of the string, which needs to be freeable for e2_command_run)
	gchar *help_cmd = _("help");
	gchar *command = e2_alias_apply (help_cmd, &app.aliases);
	if (g_str_equal (help_cmd, command))
	{	//we can't find what the user has set for a help command
		//so use the default instead
		g_free (command);
		command = g_strconcat(_A(5),".",_A(88),
//			" "DOC_DIR, MAIN_HELP" [",_("commands"),"]",NULL);	//file.view_at
			" ", e2_option_str_get ("usage-help-doc"), " [",_("commands"),"]",NULL);	//file.view_at
	}
	e2_command_run (command, E2_COMMAND_RANGE_DEFAULT);
	g_free (command);
}
/**
 * @brief register command-related actions
 * @return
 **/
void e2_command_actions_register (void)
{
	_e2_command_set_sigchild_handler ();  //not an action, but this is a convenient spot for other init

	gchar *action_name = g_strconcat(_A(1),".",_A(50),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		e2_command_list_children, NULL, FALSE);
	action_name = g_strconcat(_A(1),".",_A(72),NULL);
	e2_action_register (action_name, E2_ACTION_TYPE_ITEM,
		_e2_command_send_to_lastchild, NULL, FALSE,
		E2_ACTION_EXCLUDE_TOOLBAR | E2_ACTION_EXCLUDE_MENU, NULL);
}
/**
 * @brief register command-related options
 * @return
 **/
void e2_command_options_register (void)
{
	//no screen rebuilds needed after any change to these options
	gchar *group_name = g_strconcat(_C(5),":",_C(24),NULL); //_("commands:misc"
	e2_option_bool_register ("use-external-viewer", group_name, _("use external file-viewer"),
		_("If activated, the command entered below will be run to view file content, instead of launching the internal viewer"),
		NULL, FALSE,
		E2_OPTION_FLAG_BASIC | E2_OPTION_FLAG_FREEGROUP);
	e2_option_str_register ("command-viewer", group_name, _("viewer command:"),
		_("This is a command to run an external application for viewing file content.\n"
		"The first selected item will be supplied as the first argument"),    //CHECKME = this is wrong ??
		"use-external-viewer", "gview",
		E2_OPTION_FLAG_BASIC);
	e2_option_bool_register ("use-external-editor", group_name, _("use external file-editor"),
		_("If activated, the command entered below will be run to edit file content, instead of launching the internal editor"),
		NULL, FALSE,
		E2_OPTION_FLAG_BASIC);
	e2_option_str_register ("command-editor", group_name, _("editor command:"),
		_("This is a command to run an external application for editing file content.\n"
		"The first selected item will be supplied as the first argument"),    //CHECKME = this is wrong ??
		"use-external-editor", "leafpad",
		E2_OPTION_FLAG_BASIC);
	e2_option_str_register ("command-xterm", group_name, _("x terminal emulator:"),
		_("This is the external command/application that will be be run when emelFM2 is asked to open a terminal"),
		NULL, "xterm",
		E2_OPTION_FLAG_BASIC);
	e2_option_bool_register ("command-use-aliases", group_name, _("use aliases"),
		_("This is a general switch to turn on/off alias handling for commands"),
		NULL, TRUE,
		E2_OPTION_FLAG_ADVANCED);
	e2_option_bool_register ("relative-symlinks", group_name, _("relative symlinks"),
		_("This gives each created symlink a relative path to its source, like '../../<otherbranch>/<file>', instead of a full path referenced to /"),
		NULL, TRUE,
		E2_OPTION_FLAG_ADVANCED);
	e2_option_bool_register ("transparent-dir-links", group_name, _("interpret 'relative' paths"),
		_("This enables correct interpretation of paths containing '..' etc. These might be typed in, or attached to a button, say"),
		NULL, TRUE,
		E2_OPTION_FLAG_ADVANCED);
	e2_option_int_register ("command-watch-priority", group_name, _("watch priority"),
		_("The watch priority of commands influences how fast program\n"
		"output is read from i/o channels. A too-high priority might\n"
		"decrease gui responsiveness and break automatic scrolling.\n"
		"(note: negative values mean high priority, positive values mean low priority)"),
		NULL, 300, -200, 400,
		E2_OPTION_FLAG_ADVANCED);

	group_name = g_strconcat(_C(5) ,":",_C(6),NULL); //_("commands:confirmation"
	e2_option_bool_register ("confirm-delete", group_name, _("confirm any delete"),
		_("If activated, you will be asked for confirmation before actually deleting anything"), NULL, TRUE,
		E2_OPTION_FLAG_BASIC | E2_OPTION_FLAG_FREEGROUP);
	e2_option_bool_register ("confirm-overwrite", group_name, _("confirm any overwrite"),
		_("If activated, you will be asked for confirmation before actually overwriting anything"), NULL, TRUE,
		E2_OPTION_FLAG_BASIC);

	group_name = g_strconcat(_C(5),":",_C(12),NULL);  //_("commands:file operations"
	e2_option_bool_register ("fileop-background", group_name, _("run in background"),
		_("This causes commands operating on the selected items to run in the background, not blocking the interface"), 
		NULL, FALSE,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP);
	e2_option_bool_register ("fileop-show", group_name, _("show commands"),
 		_("This displays, on the output pane, the commands that are run"), NULL, TRUE,
		E2_OPTION_FLAG_ADVANCED);
}
