/* $Id: e2_command_line.c 1861 2006-04-25 08:27:44Z tpgww $
 * 
 * Copyright (C) 2004-2005 tooar <tooar@gmx.net>
 * 
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file e2_command_line.c 
 * @brief command line functions
 *
 * This file contains functions for e2's command and dir lines.
 **/

#include <string.h>
#include "e2_command_line.h"
#include "e2_complete.h"
#include "e2_filelist.h"

static void _e2_command_line_default_cb (GtkWidget *entry, E2_CommandLineRuntime *rt);
//static GtkWidget *_e2_command_line_create (E2_CommandLineRuntime *rt);

//list of names of command and dir lines, used to ensure
//that each command- or dir-line has a unique name
static GList *line_names = NULL;

  /****************/
 /***** utils *****/
/***************/

/**
 * @brief sync command line tree model history with cachable glist history
 * This is called in the process of destroying a command-line (eg when
 * rebuilding toolbars) and when session is ending
 * @param rt  pointer to command line runtime data struct
 *
 * @return
 **/
static void _e2_command_line_sync_history (E2_CommandLineRuntime *rt)
{
	//free old history
	if (rt->history != NULL)
		e2_list_free_with_data (&rt->history);
	GtkTreeIter iter;
	//fill with new state
	//CHECKME how long does the model exist when the widget is being destroyed ?
	if (gtk_tree_model_get_iter_first (rt->model, &iter))
	{
		gint count = gtk_tree_model_iter_n_children (rt->model, NULL);
		gint max = (rt->original) ?
			e2_option_int_get ("command-line-history-max"):
			e2_option_int_get ("dir-line-history-max");
		gchar *value;
		do
		{
			if (count-- > max)
				continue;
			gtk_tree_model_get (rt->model, &iter, 0, &value, -1);
			rt->history = g_list_append (rt->history, value);
		} while (gtk_tree_model_iter_next (rt->model, &iter));
	}
}
/**
 * @brief check if strings @a arg1 and @a arg2 are different
 *
 * @param arg1 string to be compared
 * @param arg2 string to be compared
 *
 * @return TRUE if strings @a arg1 and @a arg2 are different
 **/
static gboolean _e2_command_line_match_list_str
	(gconstpointer arg1, gconstpointer arg2)
{
	return (!g_str_equal (arg1, arg2));
}
/**
 * @brief
 *
 * @param option
 * @param def
 *
 * @return
 **/
static gchar *_e2_command_line_find_name (gchar *option,
	gchar *def)
{
	if ((option == NULL) || (*option == '\0'))
		option = def;

	gchar *name = NULL;

	GList *tmp = g_list_find_custom (line_names, option, _e2_command_line_match_list_str );
	if (tmp == NULL)
		name = g_strdup (option);
	else
	{
		gint turn = 2;
		while (tmp != NULL)
		{
			g_free (name);
			name = g_strdup_printf ("%s %d", option, turn++);
			tmp = g_list_find_custom (line_names, name, _e2_command_line_match_list_str);
		}
	}
	return name;
}
/**
 * @brief
 *
 * @param rt  pointer to command line runtime data struct
 *
 * @return
 **/
static void _e2_command_line_update_entry (E2_CommandLineRuntime *rt)
{
	GtkWidget *entry = GTK_BIN (rt->cl)->child;
	if (e2_option_bool_get_direct (rt->opt_history_last))
	{
		gchar *text = e2_combobox_last_text (GTK_COMBO_BOX (rt->cl));
		if (text != NULL)
		{
			e2_combobox_set_active (rt->cl,
				e2_combobox_last_index (GTK_COMBO_BOX (rt->cl)));
//			gtk_entry_set_text (GTK_ENTRY (entry), text);
			gtk_editable_set_position (GTK_EDITABLE (entry), -1);
			g_free (text);
		}
		else
			gtk_entry_set_text (GTK_ENTRY (entry), "");
	}
	else
		gtk_entry_set_text (GTK_ENTRY (entry), "");
}
/**
 * @brief get pointer to rt data for first-recorded command line (as opposed to dir line)
 *
 * This does nothing if @a rt does not point to NULL
 *
 * @param rt pointer to location to store command line runtime data struct pointer
 *
 * @return
 **/
static void _e2_command_line_get_first (E2_CommandLineRuntime **rt)
{
	if (*rt == NULL)
	{
//		if (app.command_lines == NULL)
//			return;	//no command lines registered
//		else
//		{
			//find the first-registered command-line
			E2_CommandLineRuntime *cl_rt;
			GList *list;
			for (list = app.command_lines; list != NULL ; list = g_list_next (list))
			{
				cl_rt = (E2_CommandLineRuntime *) list->data;
				if (cl_rt->original)
				{
					*rt = cl_rt;
					break;
				}
			}
//		}
	}
}
/**
 * @brief  determine command line pointer(s)
 *
 * This function tries to find the corresponding entry widget if
 * @a rt != NULL
 * Or if @a rt is NULL, tries to find the @a rt corresponding to @a entry.
 * But then, if @a entry is not actually an entry widget, @a rt is just set
 * to the one for the first-registered command line (as opposed
 * to dir line) if any
 *
 * @param widget pointer to widget to check/set (may be NULL)
 * @param rt pointer to the command line runtime object to check/set, or to NULL
 *
 * @return  TRUE if @a entry and @a rt are set to valid values
 **/
static gboolean _e2_command_line_get (GtkWidget **widget,
	E2_CommandLineRuntime **rt)
{
	//the runtime object has precedence
	if (*rt == NULL)
	{
		//if the entry really is an entry, get the pointer to the
		//runtime object
		if ((*widget != NULL) && (GTK_IS_ENTRY (*widget)))
		{
			*rt = g_object_get_data (G_OBJECT (*widget), "command-line-runtime");
			if (*rt != NULL)
				return TRUE;
			else
			{
				E2_CommandLineRuntime *clrt;
				GList *line;
				for (line = app.command_lines; line != NULL ; line = g_list_next (line))
				{
					clrt = line->data;
					if (GTK_BIN (clrt->cl)->child == *widget)
						break;
				}
				if (line != NULL)
				{
					*rt = clrt;
					return TRUE;
				}
			}
		}
		else
		//can't find anything from widget, just try to find the first available command line
			_e2_command_line_get_first (rt);
		if (*rt == NULL)	//OOPS can't find anything !!
			return FALSE;
	}
	*widget = GTK_BIN ((*rt)->cl)->child;  //set the real entry widget
	return TRUE;
}

  /*********************/
 /***** callbacks *****/
/*********************/

/**
 * @brief change-dir hook function
 *
 * @param dir the dir that is now current
 * @param rt  pointer to command line runtime data struct
 *
 * @return TRUE always
 **/
static gboolean _e2_command_line_change_dir_cb
	(gchar *dir, E2_CommandLineRuntime *rt)
{
	GtkWidget *entry = GTK_BIN (rt->cl)->child;
	gtk_entry_set_text (GTK_ENTRY (entry), dir);
	e2_combobox_activated_cb (entry, GINT_TO_POINTER (e2_option_bool_get
		("dir-line-history-double")));
	if (!e2_option_bool_get_direct (rt->opt_history_last))
		gtk_entry_set_text (GTK_ENTRY (entry), "");
	if (e2_option_bool_get ("dir-line-pathname-hint"))
		gtk_tooltips_set_tip (app.tooltips, entry, dir, NULL);
	return TRUE;
}
/**
 * @brief mouse button-press callback
 * This triggers a directory selection dialog when the middle-button
 * is pressed while the mouse is on a directory-line
 * @param entry the entry widget where the button was pressed 
 * @param event pointer to gtk event data struct
 * @param rt  pointer to pane runtime data struct
 *
 * @return TRUE if the middle button was pressed
 **/
static gboolean _e2_command_line_button_press_cb (GtkWidget *entry,
	GdkEventButton *event, E2_PaneRuntime *rt)
{
//	curr_pane->focus_widget = entry;
	if (event->button == 2)
	{
		e2_pane_goto_choice (rt, NULL, NULL);
		return TRUE;
	}
	return FALSE;
}
/**
 * @brief This is the callback for dir-line keypress signals
 *
 * @param entry pointer to the widget that received the keypress
 * @param event event data struct
 * @param rt pointer to data struct for the dir-line
 *
 * @return TRUE if completion occurred
 **/
static gboolean _e2_command_line_key_press_cb (GtkWidget *entry,
	GdkEventKey *event, E2_CommandLineRuntime *rt)
{
	if ((event->state & (GDK_CONTROL_MASK | GDK_MOD1_MASK)) == 0 && event->keyval < 0xF000)
	{
		const gchar *text_now = gtk_entry_get_text (GTK_ENTRY (entry));
		gchar *text = g_strconcat (text_now, event->string, NULL);
		GList *found = NULL;
		//NOTE: pos is characters, not bytes
		gint pos = gtk_editable_get_position (GTK_EDITABLE (entry)) + 1;
		if (e2_complete_str (&text, &pos, &found, E2_COMPLETE_FLAG_DIRS, rt) > 0)
		{
			gtk_entry_set_text (GTK_ENTRY (entry), text);
			gtk_editable_set_position (GTK_EDITABLE (entry), pos);
			e2_list_free_with_data (&found);
			return TRUE;
		}
		g_free (text);
	}
	return FALSE;
}

  /****************************/
 /***** activate actions *****/
/****************************/

/**
 * @brief perform command entered in command line
 *
 * @param entry widget containing command to run
 * @param rt  pointer to command line runtime data struct
 *
 * @return
 **/
void _e2_command_line_default_cb (GtkWidget *entry, E2_CommandLineRuntime *rt)
{
	printd (DEBUG, "_e2_command_line_default_cb (entry:,rt:)");
	const gchar *command_raw = gtk_entry_get_text (GTK_ENTRY (entry));
	//quick exit
	if ((command_raw == NULL) || (*command_raw == '\0'))
		return;
	gchar *command = e2_utf8_unescape (command_raw, ' ');
	
	//actually run command
	e2_command_run (command, E2_COMMAND_RANGE_DEFAULT);

	g_free (command);

	_e2_command_line_update_entry (rt);
}
/**
 * @brief open directory entered in dir line
 *
 * @param entry widget containing directory string
 * @param rt  pointer to command line runtime data struct
 *
 * @return
 **/
void e2_command_line_directory_cb (GtkWidget *entry, E2_CommandLineRuntime *rt)
{
	printd (DEBUG, "e2_command_line_directory_cb (entry:,rt:)");
	const gchar *command_raw = gtk_entry_get_text (GTK_ENTRY (entry));
	//quick exit
	if ((command_raw == NULL) || (*command_raw == '\0'))
		return;

	gchar *command = e2_utf8_unescape (command_raw, ' ');

	E2_PaneRuntime *pane_rt = rt->pane;
	//actually run command
	e2_pane_change_dir (pane_rt, command);

	if (e2_option_bool_get ("dir-line-focus-after-activate"))
	{
		if (pane_rt == other_pane)
			e2_pane_activate_other ();
		else
//			gtk_widget_grab_focus (pane_rt->focus_widget);
			gtk_widget_grab_focus (pane_rt->view->treeview);
	}

	g_free (command);
}

  /*******************/
 /***** actions *****/
/*******************/

/**
 * @brief insert name of each active-pane selected item into @a entry
 *
 * @param data UNUSED pointer to action runtime data, NULL
 * @param arg action argument string
 * @param entry entry for command/dir line to be processed
 *
 * @return
 **/
static void _e2_command_line_insert_action
	(gpointer data, gchar *arg, GtkWidget *entry)
{
	//find out from the action argument if the user wants quotes/escaping
	g_strstrip (arg);
	gchar *str_down = g_utf8_strdown (arg, -1);
	gboolean escape = g_str_equal (arg, _A(92));	//_("escape"
	gboolean quote = (escape) ? FALSE : g_str_equal (arg, _A(97));	//_("quote"
	g_free (str_down);

#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, insert action");
#endif
	e2_filelist_disable_refresh ();

	gchar *utf;
	GString *cmd = g_string_sized_new (128);

	GList *tmp, *base;
	base = tmp = e2_fileview_get_selected_local (curr_view);
	for (; tmp != NULL; tmp=tmp->next)
	{
		utf = F_FILENAME_FROM_LOCALE (((FileInfo *) tmp->data)->filename);
		if (escape)
		{
			gchar *temp = e2_utf8_escape (utf, ' ');
			F_FREE (utf);
			utf = temp;
		}
		cmd = g_string_append_c (cmd, ' ');
		if (quote)
			g_string_append_printf (cmd, "\"%s\"", utf);
		else
			cmd = g_string_append (cmd, utf);
		if (escape)
			g_free (utf);
		else
			F_FREE (utf);
	}

	E2_CommandLineRuntime *rt = NULL;
	_e2_command_line_get (&entry, &rt);
	gint cursor = gtk_editable_get_position (GTK_EDITABLE (entry));
	gtk_editable_insert_text (GTK_EDITABLE (entry), cmd->str,
		cmd->len, &cursor);
	gtk_editable_set_position (GTK_EDITABLE (entry), cursor);

	g_string_free (cmd, TRUE);
	g_list_free (base);

#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, insert action");
#endif
	e2_filelist_enable_refresh();
}
/**
 * @brief complete the string in the command line associated with @a entry
 *
 * @param data UNUSED pointer to data from action data struct, NULL
 * @param arg action argument string, empty or name(s) of completion-type(s) 
 * @param entry the entry widget for the comand line where the action was inititated
 *
 * @return
 **/
static void _e2_command_line_complete_action
	(gpointer data, gchar *arg, GtkWidget *entry)
{
	printd (DEBUG, "complete_action (rt:_,arg:%s,entry:_", arg);
	//find which command line is active
	E2_CommandLineRuntime *rt = NULL;
	_e2_command_line_get (&entry, &rt);
	if (rt == NULL)
		return;	//can't find what to work on

	//find out what type of completion is wanted
	E2_CompleteFlags flags;
	if (!rt->original)
		flags = E2_COMPLETE_FLAG_DIRS;	//dir-lines just find dirs
	else
	{
		flags = 0;
		g_strstrip (arg);
//		gchar *str_down = g_utf8_strdown (arg, -1);
		if (strstr (arg, _A(91)) != NULL)
			flags |= E2_COMPLETE_FLAG_DIRS;
		if (strstr (arg, _A(94)) != NULL)
			flags |= E2_COMPLETE_FLAG_FILES;
		if (strstr (arg, _("mounts")) != NULL)	//_I("mounts")
			flags |= E2_COMPLETE_FLAG_MOUNT;
		if (strstr (arg, _("all")) != NULL)	//_I("all")
			flags |= E2_COMPLETE_FLAG_ALL;
		if (flags == 0)
			flags = E2_COMPLETE_FLAG_ALL;
//		g_free (str_down);
	}
	const gchar *_text = gtk_entry_get_text (GTK_ENTRY (entry));
	gchar *text = g_strdup (_text);	//could be "\0"
	GList *found = NULL;
	gint pos = gtk_editable_get_position (GTK_EDITABLE (entry));	//NOTE: characters, not bytes
	gint num = e2_complete_str (&text, &pos, &found, flags, rt);
//	printd (DEBUG, "found %d - %s - %d", num, text, pos);

	if (num > 0)
	{
		// if it isn't a dir: append a <space>
		if ((e2_option_bool_get ("command-line-complete-append-space")) &&
			(num == 1) && (g_utf8_get_char (g_utf8_offset_to_pointer (text, pos - 1)) != G_DIR_SEPARATOR))
		{
			gchar *part1 = e2_utf8_ndup (text, pos);
			gchar *part2 = g_utf8_offset_to_pointer (text, pos);
			gchar *text_spaced = g_strconcat (part1, " ", *part2 != '\0' ? part2 : "" , NULL);
			g_free (part1);
			gtk_entry_set_text (GTK_ENTRY (entry), text_spaced);
			gtk_editable_set_position (GTK_EDITABLE (entry), pos +1);
			g_free (text_spaced);
		} else 
		{
			gtk_entry_set_text (GTK_ENTRY (entry), text);
			gtk_editable_set_position (GTK_EDITABLE (entry), pos);
		}
		if (num > 1)
		{
			GList *tmp;
			for (tmp = found; tmp != NULL; tmp = g_list_next (tmp))
			{
				e2_output_print_default (tmp->data);
				e2_output_print_default ("\n");
			}
		}
	}
	e2_list_free_with_data (&found);
	g_free (text);
}
/**
 * @brief focus a command line
 *
 * @param rt  pointer to command line runtime data struct, or NULL to use the first-recorded line
 *
 * @return
 **/
static void _e2_command_line_focus_action (E2_CommandLineRuntime *rt)
{
	_e2_command_line_get_first (&rt);	//get the first line if rt == NULL
	if (rt == NULL)
		return;
	gtk_editable_set_position (GTK_EDITABLE (GTK_BIN (rt->cl)->child), -1);
	gtk_widget_grab_focus (GTK_BIN (rt->cl)->child);
//	curr_pane->focus_widget = GTK_BIN (rt->cl)->child;
//	gtk_widget_grab_focus (curr_pane->focus_widget);
}
/**
 * @brief toggle focus between command line and active pane
 *
 * @param rt pointer to command line runtime data struct, or NULL to use first-recorded line
 *
 * @return
 **/
static void _e2_command_line_focus_toggle_action (E2_CommandLineRuntime *rt)
{
	_e2_command_line_get_first (&rt); //gets the first line if rt == NULL
	if (rt == NULL)
		return;
	if (GTK_WIDGET_HAS_FOCUS (GTK_BIN (rt->cl)->child))
//		curr_pane->focus_widget = curr_pane->view->treeview;
		gtk_widget_grab_focus (curr_view->treeview);
	else
	{
		gtk_editable_set_position (GTK_EDITABLE (GTK_BIN (rt->cl)->child), -1);
//		curr_pane->focus_widget = GTK_BIN (rt->cl)->child;
		gtk_widget_grab_focus (GTK_BIN (rt->cl)->child);
	}
}
/**
 * @brief clear contents of @a entry
 *
 * @param data1 UNUSED pointer action data, NULL
 * @param data2 UNUSED pointer
 * @param widget used to initiate the action
 *
 * @return
 **/
static void _e2_command_line_clear_action (gpointer data1,
	gpointer data2, GtkWidget *widget)
{
	//widget may be entry, button, menuitem etc
	if (widget == NULL || ! GTK_IS_ENTRY (widget))
	{
		//FIXME this uses poor assumption of what the real entry will be
		E2_CommandLineRuntime *rt = NULL;
		_e2_command_line_get (&widget, &rt);
	}
	gtk_entry_set_text (GTK_ENTRY (widget), "");
}
/**
 * @brief clear command line history
 *
 * @param data1 UNUSED pointer to action data set when registered, NULL
 * @param data2 UNUSED pointer to runtime data sent to this action
 * @param entry the entry widget for the command line whose history is to be cleared
 *
 * @return
 **/
static void _e2_command_line_clear_history_action
	(gpointer data1, gpointer data2, GtkWidget *entry)
{
	//find the command line for the entry
	E2_CommandLineRuntime *rt = NULL;
	_e2_command_line_get (&entry, &rt);

	if (rt->history != NULL)
	{
		gint num = g_list_length (rt->history);
		for (; num > 0; num--)
			gtk_combo_box_remove_text (GTK_COMBO_BOX (rt->cl), 0);
		e2_list_free_with_data (&rt->history);
		g_object_unref (rt->model);
	}
}

  /******************/
 /***** public *****/
/******************/

/**
 * @brief  create a command line
 *
 * This creates a command line runtime struct and a command line
 * widget. The @a name_runtime parameter may be the same for
 * several command lines as a unique name is created at runtime
 * from it.
 *
 * @param name_runtime  the name base to determine different command lines
 * @param name_default  the default name in case the runtime name is NULL or empty
 * @param func the function to run on activation of the command line
 * @param flags commmand line flags
 * @param data pointer to pane rt struct for a dir line, or ?? for a command line
 *
 * @return pointer to the created runtime data struct
 **/
E2_CommandLineRuntime *e2_command_line_create (gchar *name_runtime,
	gchar *name_default, gpointer func, E2_CommandLineFlags flags, gpointer data)
{
	E2_CommandLineRuntime *rt = g_malloc (sizeof (E2_CommandLineRuntime));

	//get unique name for this line
	//BUT don't want 'public' name for cache entry - just use the default
	rt->name = _e2_command_line_find_name (name_default, "default cmd line");
//	rt->name = _e2_command_line_find_name (name_runtime, name_default);

	rt->func = func;
	rt->flags = flags;
	rt->original = (flags & E2_COMMAND_LINE_ORIGINAL);
	rt->pane = (E2_PaneRuntime *) data;	//CHECKME no usage for 'original' command line

	//init remaining pointers
	rt->cl = NULL;
	rt->history = NULL;
	rt->model = NULL;
	rt->opt_history_last = NULL;

	//append name to local name register to ensure uniqueness of each name
	//FIXME = use this register for all named runtime items
	line_names = g_list_append (line_names, rt->name);
	//also add line to command line register if it is not a dirline
//	if (rt->original) NOW ADD THEM ALL TO THE LIST
		app.command_lines = g_list_append (app.command_lines, rt);

	//now, create the widget after the runtime object has been initialized
//	_e2_command_line_create (rt);
	E2_ComboBoxFlags flags2 = E2_COMBOBOX_HAS_ENTRY;

	//sort out some options for command line or dir line;
	//the options have been registered in the general init function
	if (rt->original)
	{  //this is not a dir line
		if (e2_option_bool_get ("command-line-history-double"))
			flags2 |= E2_COMBOBOX_ALLOW_DOUBLE;
		if (e2_option_bool_get ("command-line-history-cycle"))
			flags2 |= E2_COMBOBOX_CYCLE_HISTORY;
		if (e2_option_bool_get ("command-line-menu-style"))
			flags2 |= E2_COMBOBOX_MENU_STYLE;
		flags2 |= E2_COMBOBOX_FOCUS_ON_CHANGE;
		//we need to access this often, thus we save the pointer to the
		//option set and will not have to look it up in the hashtable
		//each time
		rt->opt_history_last = e2_option_get ("command-line-history-last");
	}
	else
	{   //this is a dir line
		if (e2_option_bool_get ("dir-line-history-double"))
			flags2 |= E2_COMBOBOX_ALLOW_DOUBLE;
		if (e2_option_bool_get ("dir-line-history-cycle"))
			flags2 |= E2_COMBOBOX_CYCLE_HISTORY;
		if (e2_option_bool_get ("dir-line-menu-style"))
			flags2 |= E2_COMBOBOX_MENU_STYLE;
		flags2 |= E2_COMBOBOX_NO_AUTO_HISTORY;
		rt->opt_history_last = e2_option_get ("dir-line-history-last");
		//register to change_dir hook for dir-lines so we always know the
		//latest directory to add it to the history and show it if wanted
		E2_PaneRuntime *pane_rt = rt->pane;
		e2_hook_register (&pane_rt->hook_change_dir, _e2_command_line_change_dir_cb, rt);
	}

	//actually create the comboboxentry
	rt->cl = e2_combobox_get (rt->func, rt, flags2);
	g_object_set_data (G_OBJECT (GTK_BIN (rt->cl)->child), "command-line-runtime", rt);

	if (!rt->original)
	{  //this is a dir line, enable file-chooser
		g_signal_connect (G_OBJECT (GTK_BIN (rt->cl)->child), "button_press_event",
			G_CALLBACK (_e2_command_line_button_press_cb), rt->pane);
	}

	//name is 'internal' form
	E2_Cache *cache = e2_cache_list_register (rt->name, &rt->history);
	cache->sync_func = _e2_command_line_sync_history;
	cache->sync_data = rt;

	rt->model = gtk_combo_box_get_model (GTK_COMBO_BOX (rt->cl));
	g_object_ref (rt->model);

	//init history
	if (rt->history != NULL)  //if history cache was found
		e2_combobox_append_history (rt->cl, rt->history);

	//second part, connect focus callbacks and set up old pointer to keep
	//compatiblity
	gchar *bindingname;
	if (rt->original)
	{
		if (e2_option_bool_get ("command-line-show-output-on-focus-in"))
			g_signal_connect (G_OBJECT (GTK_BIN (rt->cl)->child),
				"focus-in-event", G_CALLBACK (e2_window_output_show), NULL);
		if (e2_option_bool_get ("command-line-hide-output-on-focus-out"))
			g_signal_connect (G_OBJECT (GTK_BIN (rt->cl)->child),
				"focus-out-event", G_CALLBACK (e2_window_output_hide), NULL);
		bindingname = g_strconcat (_C(16),".",_C(4),NULL);  //_(general.command-line"
	}
	else
	{
		g_signal_connect (G_OBJECT (GTK_BIN (rt->cl)->child),
			"key_press_event", G_CALLBACK (_e2_command_line_key_press_cb), rt);
		bindingname = g_strconcat (_C(16),".",_C(10),NULL);  //_(general.dir-line" 
	}

	e2_keybinding_register (bindingname, GTK_BIN (rt->cl)->child);
	g_free (bindingname);

	//select last history entry if necessary
	if ((!rt->original) || (e2_option_bool_get ("command-line-history-last")))
	{
		gint count = e2_combobox_last_index (GTK_COMBO_BOX (rt->cl));
		if (count > -1)
			gtk_combo_box_set_active (GTK_COMBO_BOX (rt->cl), count);
	}
	gtk_widget_show (GTK_WIDGET (rt->cl));

	return rt;
}
/**
 * @brief
 *
 * @param rt pointer to command line runtime data struct
 *
 * @return the commandline widget
 **/
//static GtkWidget *_e2_command_line_create (E2_CommandLineRuntime *rt)
//{
//	return GTK_WIDGET (rt->cl);
//}
/**
 * @brief clear runtime data for a command/dir line
 *
 * This is the 'destroy data' function supplied when command
 * lines are created.
 * It will sync the history and free all internal memory.
 *
 * @param rt  the runtime object to destroy
 *
 * @return 
 **/
void e2_command_line_destroy (E2_CommandLineRuntime *rt)
{
	line_names = g_list_remove (line_names, rt->name); //LEAK ??

/*	if (rt->original)
	{	//line is a command-line
		GList *list;
		for (list = app.command_lines; list != NULL ; list = g_list_next (list))
		{
			if (list->data == rt)
			{
				app.command_lines = g_list_delete_link (app.command_lines, list);
				break;
			}
		}
	}
	else */
	app.command_lines = g_list_remove (app.command_lines, rt);
	if (!rt->original)
	{	//line is a dir line
		E2_PaneRuntime *pane_rt = rt->pane;
		e2_hook_unregister (&pane_rt->hook_change_dir,
			_e2_command_line_change_dir_cb, rt, TRUE);
	}

	_e2_command_line_sync_history (rt);

	//now we can unregister
	e2_cache_unregister (rt->name);
	//and the combo box data model may be freed
	g_object_unref (rt->model);
	g_free (rt->name);
	g_free (rt);
}
/**
 * @brief initialize command line actions
 *
 * This function initializes the command line actions.
 * It is, and should only be, called once at startup.
 * 
 * @return 
 **/
void e2_command_line_actions_register (void)
{
	gchar *action_name = g_strconcat(_A(1),".",_A(23),NULL);
	e2_action_register (action_name, E2_ACTION_TYPE_COMMAND_LINE,
		_e2_command_line_default_cb, "command line", TRUE,  // internal default name not translated
		E2_ACTION_EXCLUDE_MENU | E2_ACTION_EXCLUDE_ACCEL, NULL);
	action_name = g_strconcat(_A(1),".",_A(39),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_command_line_focus_action, NULL, FALSE);
	action_name = g_strconcat(_A(1),".",_A(40),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_command_line_focus_toggle_action, NULL, FALSE);
	action_name = g_strconcat(_A(1),".",_A(27),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_command_line_clear_action, NULL, FALSE);
	action_name = g_strconcat(_A(1),".",_A(28),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_command_line_clear_history_action, NULL, FALSE);
	action_name = g_strconcat(_A(1),".",_A(29),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_command_line_complete_action, NULL, FALSE);
	action_name = g_strconcat(_A(1),".",_A(48),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_command_line_insert_action, NULL, FALSE);
}
/**
 * @brief initialize command line options
 *
 * This function initializes the options for the various command lines. 
 * It is, and should only be, called once at startup.
 * 
 * @return 
 **/
void e2_command_line_options_register (void)
{
	//none of these require a rebuild after change
	//command line
	gchar *group_name = g_strconcat(_C(5),".",_C(4),":",_C(24),NULL);  //_("commands.command line:misc"
	e2_option_bool_register ("command-line-history-last", group_name, _("show last"),
		_("If activated, the last-entered command will be displayed, instead of an empty line"),
		NULL, FALSE,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP);

	group_name = g_strconcat(_C(5),".",_C(4),":",_C(17),NULL);  //_("commands.command line:history"
	e2_option_int_register ("command-line-history-max", group_name, _("maximum number of history entries"),
		_("This is the largest number of command-line history entries that will be recorded"),
		NULL, 10, 0, 9999999,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP);
	e2_option_bool_register ("command-line-history-double", group_name, _("double entries"),
		_("This allows entries to be recorded more than once in the history list, so the last entry is always close to hand"),
		NULL, FALSE,
		E2_OPTION_FLAG_ADVANCED);
	e2_option_bool_register ("command-line-history-cycle", group_name, _("cyclic list"),
		_("When scanning the history list, cycle from either end around to the other end, instead of stopping"),
		NULL, TRUE,
		E2_OPTION_FLAG_ADVANCED);
	e2_option_bool_register ("command-line-menu-style",
		group_name, _("show as a menu"),
		_("If activated, the history entries will be presented as a menu. "
		"For most Gtk+2 themes, this will not be as attractive as the list view"),
		NULL, FALSE,
		E2_OPTION_FLAG_ADVANCED);

	group_name =g_strconcat(_C(5),".",_C(4),":",_C(38),NULL); //_("commands.command line:tab completion"
	e2_option_bool_register ("command-line-complete-append-space",
		group_name, _("append space after unique items"),
		_("This appends a 'space' character to the end of a unique successful match of a file"),
		NULL, TRUE,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP);

	//dir line
	group_name = g_strconcat(_C(32),".",_C(10),":",_C(24),NULL); //_("panes.directory line:misc"
	e2_option_bool_register ("dir-line-history-last",
		group_name, _("show last entry"),
		_("If activated, the last-entered directory will be displayed, instead of an empty line"),
		NULL, TRUE,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP);
	e2_option_bool_register ("dir-line-pathname-hint",
		group_name, _("show pathname as a tooltip"),
		_("If activated, the full directory pathname will display as a tooltip. "
		"This is useful when the path is too long for the normal display"),
		NULL, TRUE,
		E2_OPTION_FLAG_ADVANCED);

	group_name = g_strconcat(_C(32),".",_C(10),":",_C(17),NULL); //_("panes.directory line:history"
	e2_option_int_register ("dir-line-history-max", group_name, _("maximum number of history entries"),
		_("This is the largest number of directory-line history entries that will be retained"),
		NULL, 10, 0, 999999,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP);
	e2_option_bool_register ("dir-line-history-double",
		group_name, _("double entries"),
		_("This allows entries to be recorded more than once in the history list, so the last entry is always close to hand"),
		NULL, FALSE,
		E2_OPTION_FLAG_ADVANCED);
	e2_option_bool_register ("dir-line-history-cycle",
		group_name, _("cyclic list"),
		_("When scanning the history list, cycle from either end around to the other end, instead of stopping"),
		NULL, TRUE,
		E2_OPTION_FLAG_ADVANCED);
	e2_option_bool_register ("dir-line-menu-style",
		group_name, _("show as a menu"),
		_("If activated, the directory line history will be presented as a menu. "
		"For most for most Gtk+2 themes, this will not be as attractive as the list view"),
		NULL, FALSE,
		E2_OPTION_FLAG_ADVANCED);
}
