/* $Id: e2_cache.c 1861 2006-04-25 08:27:44Z tpgww $
 * 
 * Copyright (C) 2004 Florian Zaehringer (flo.zaehringer@web.de)
 * Copyright (C) 2004-2005 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file e2_cache.c
 * @brief cache system
 *
 * This file contains functions for e2's cache system: eg. to register
 * and unregister caches, and to write and read them to/from disk.
 **/

#include "e2_cache.h"
#include <string.h>
#include "e2_fs.h"
#include "e2_task.h"

static gchar *default_cache_file;
static GList *cache_list;
static GList *unknown_list;

  /*****************/
 /***** utils *****/
/*****************/

/**
 * @brief check whether the name member of @a cache_element matches @a name
 * @param cache_element struct with member to compare
 * @param name name string to compare
 * @return TRUE if the items do NOT match
 **/
static gboolean str_compare (E2_Cache *cache_element, gchar *name)
{
	return !g_str_equal (name, cache_element->name);
}

  /******************/
 /***** public *****/
/******************/

#ifdef E2_IMAGECACHE
/**
 * @brief cleanup helper for icon hash
 *
 * @param data pointer to E2_Image struct for an icon
 */
void _e2_cache_iconhash_remove (E2_Image *data)
{
	g_object_unref (G_OBJECT (data->pixbuf));
	g_free (data);
}
/**
 * @brief setup for icon caching
 * The destroy func assumes that pixbufs have refcount 1 then
 * when this is called
 */
void e2_cache_icons_init (void)
{
	app.icons = g_hash_table_new_full (g_str_hash, g_str_equal,
		g_free, (GDestroyNotify) _e2_cache_iconhash_remove);
	e2_utils_init_icon_sizes ();
}
/**
 * @brief helper func to clear cached icon pixbuf
 * @param key UNUSED key of hash table item being processed
 * @param value value of hash table item being processed
 * @param user_data UNUSED data specified when foreach was initiated
 * @return
 **/
/*static void _e2_cache_image_clear (gpointer key,
	E2_Image *value, gpointer user_data)
{
	value->refcount--;
//	g_object_unref (G_OBJECT (value->pixbuf));
} */
/**
 * @brief clear all cached icon pixbufs from X memory
 * @return
 **/
void e2_cache_image_clearall (void)
{
//	g_hash_table_foreach (app.icons, (GHFunc) _e2_cache_image_clear, NULL);
	g_hash_table_destroy (app.icons);
}
/**
 * @brief get cached image for icon named @a name with size @a size
 * If not already cached, the relevant image will be created and added to
 * the cache
 * @param name icon name string (NULL for missing image icon)
 * @param size icon size enumerator
 * @return pointer to E2_Image struct for the image
 **/
E2_Image *e2_cache_image_get (gchar *name, GtkIconSize size)
{
	if (name == NULL)
		name = GTK_STOCK_MISSING_IMAGE;	//revert to default icon image
	gint psize = e2_utils_get_icon_size (size);
	//make a unique cache key
	gchar *cachename = g_strdup_printf ("%s%d", name, psize);
	E2_Image *cached = g_hash_table_lookup (app.icons, cachename);
	if (cached == NULL)
	{	//need to cache this icon
		cached = g_malloc (sizeof (E2_Image));
		cached->size = psize;
//		cached->refcount = 1;
		if (e2_utils_check_stock_icon (name))
		{	//stock image
	//		cached->type = E2_IMAGE_BMP;
			cached->pixbuf = gtk_icon_set_render_icon (
				gtk_icon_factory_lookup_default (name),
				gtk_rc_get_style (app.main_window),
				gtk_widget_get_default_direction (),
				GTK_STATE_NORMAL, size, NULL, NULL);
		}
		else
		{	//try to sniff the icon file
			gchar *fullname;
			if (strchr (name, G_DIR_SEPARATOR) != NULL)
				fullname = name;	//icon name has a '/', assume it's a full path
			else
			{
				fullname = (e2_option_bool_get ("use-icon-dir")) ?
				g_strconcat (e2_option_str_get ("icon-dir"), G_DIR_SEPARATOR_S, name, NULL) :
				g_strconcat (ICON_DIR, name, NULL);
			}
	
			gint width, height;
			GdkPixbufFormat *pixinfo = gdk_pixbuf_get_file_info
				(fullname, &width, &height);
			if (pixinfo != NULL)
			{
	/*			gchar *fmt = gdk_pixbuf_format_get_name(pixinfo);
				if (g_str_equal (fmt, "svg"))	//CHECKME
					cached->type = E2_IMAGE_SVG;
				else
					cached->type = E2_IMAGE_BMP;
	*/
				cached->pixbuf = gdk_pixbuf_new_from_file_at_size
					(fullname, psize, psize, NULL);
				if (fullname != name)
					g_free (fullname);
			}
			else
			{	//can't do the custom image, revert to 'missing'
				g_free (cached);
				g_free (cachename);
				if (fullname != name)
					g_free (fullname);
				return e2_cache_image_get (NULL, size);	//recurse
			}
		}

		g_hash_table_insert (app.icons, cachename, cached);
	}
	else
	{
		g_free (cachename);
//		cached->refcount++;
//		g_object_ref (G_OBJECT (cached->pixbuf));
	}
	return cached;
}
#endif
/**
 * @brief set integer parameter from cache or default
 * @param name identifier string used in cache list
 * @param value pointer to store for cached value, if any, or @a def
 * @param def default value to use if none found in cache
 * @return 
 **/
void e2_cache_int_register (gchar *name, gint *value, gint def)
{
	E2_Cache *cache;
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found != NULL)
	{
		printd (DEBUG, "re-register int cache '%s'", name);
		return;
	}
	//first-time processing
	cache = g_malloc (sizeof (E2_Cache));
	cache->type = E2_CACHE_TYPE_INT;
	found = g_list_find_custom (unknown_list, name, 
		(GCompareFunc) str_compare);
	if (found == NULL)
	{ //not in the queue, either
		printd (DEBUG, "int cache '%s' not found, creating default with %d", name, def);
		cache->name = g_strdup (name);
		*value = def;
	} else
	{  //process the queued item
		printd (DEBUG, "cache '%s' found", name);
		E2_Cache *unknown = found->data;
		cache->name = unknown->name;
		gchar *end = NULL;
		gchar *str = (gchar *) unknown->data;
		*value = g_ascii_strtoull (str, &end, 10);
		if (end == str)
		{
			printd (WARN, "int cache '%s' data conversion failed from '%s", name, str);
			*value = def;
		} else
			g_free (str);
		unknown_list = g_list_remove_link (unknown_list, found);
		g_list_free (found);
		g_free (unknown);
	}
	cache->data = GINT_TO_POINTER (value);
	cache_list = g_list_append (cache_list, cache);
	return;
}
/**
 * @brief set long integer parameter from cache or default
 * This function needed as we cannot determine at compile-time
 * how big a time_t is
 * If there are ever any uses for a long int cache item, this could
 * be converted, and the uses for this changed to run-time checks
 * @param name identifier string used in cache list
 * @param value pointer to store for cached value, if any, or @a def
 * @param def default value to use if none found in cache
 * @return
 **/
void e2_cache_time_register (gchar *name, time_t *value, time_t def)
{
	E2_Cache *cache;
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found != NULL)
	{
		printd (DEBUG, "re-register long int cache '%s'", name);
		return;
	}
	//first-time processing
	cache = g_malloc (sizeof (E2_Cache));
	cache->type = E2_CACHE_TYPE_TIME;
	found = g_list_find_custom (unknown_list, name, 
		(GCompareFunc) str_compare);
	if (found == NULL)
	{ //not in the queue, either
		printd (DEBUG, "time cache '%s' not found, creating default with %d", name, def);
		cache->name = g_strdup (name);
		*value = def;
	} else
	{  //process the queued item
		printd (DEBUG, "cache '%s' found", name);
		E2_Cache *unknown = found->data;
		cache->name = unknown->name;
		gchar *end = NULL;
		gchar *str = (gchar *) unknown->data;
		*value = g_ascii_strtoull (str, &end, 10);
		if (end == str)
		{
			printd (WARN, "time cache '%s' data conversion failed from '%s", name, str);
			*value = def;
		} else
			g_free (str);
		unknown_list = g_list_remove_link (unknown_list, found);
		g_list_free (found);
		g_free (unknown);
	}
	cache->data = (gpointer) value;
	cache_list = g_list_append (cache_list, cache);
	return;
}
/**
 * @brief set double parameter from cache or default
 * @param name identifier string used in cache list
 * @param value pointer to store for cached value, if any, or @a def
 * @param def default value to use if none found in cache
 * @return 
 **/
void e2_cache_double_register (gchar *name, gdouble *value, gdouble def)
{
	E2_Cache *cache;
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found != NULL)
	{
		printd (DEBUG, "re-register double cache '%s'", name);
		return;
	}
	cache = g_malloc (sizeof (E2_Cache));
	cache->type = E2_CACHE_TYPE_DOUBLE;
	found = g_list_find_custom (unknown_list, name, 
		(GCompareFunc) str_compare);
	if (found == NULL)
	{  //not in queue either
		printd (DEBUG, "double cache '%s' not found, creating default with %f", name, def);
		cache->name = g_strdup (name);
		*value = def;
	} else
	{  //process queued item
		printd (DEBUG, "cache '%s' found", name);
		E2_Cache *unknown = found->data;
		cache->name = unknown->name;
		gchar *end = NULL;
		gchar *str = (gchar *) unknown->data;
		*value = g_ascii_strtod (str, &end);
		if (end == str)
		{
			printd (WARN, "double cache '%s' data conversion failed from '%s", name, str);
			*value = def;
		} else
			g_free (str);
		unknown_list = g_list_remove_link (unknown_list, found);
		g_list_free (found);
		g_free (unknown);
	}
	cache->data = (gpointer *) value;
	cache_list = g_list_append (cache_list, cache);
	return;
}
/**
 * @brief set string parameter from cache or default
 * @param name identifier string used in cache list
 * @param value pointer to store for cached value, if any, or @a def
 * @param def default value to use if none found in cache
 * @return 
 **/
void e2_cache_str_register (gchar *name, gchar **str, gchar *def)
{
	E2_Cache *cache;
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found != NULL)
	{
		printd (DEBUG, "re-register string cache '%s'", name);
		return;
	}
	cache = g_malloc (sizeof (E2_Cache));
	cache->type = E2_CACHE_TYPE_STR;
	found = g_list_find_custom (unknown_list, name, 
		(GCompareFunc) str_compare);
	if (found == NULL)
	{
		printd (DEBUG, "string cache '%s' not found, creating dedault with %s", name, def);
		cache->name = g_strdup (name);
		*str = strdup (def);
	} else
	{
		printd (DEBUG, "cache '%s' found", name);
		E2_Cache *unknown = found->data;
		cache->name = unknown->name;
		*str = (gchar *) unknown->data;
		unknown_list = g_list_remove_link (unknown_list, found);
		g_list_free (found);
		g_free (unknown);
	}
	cache->data = (gpointer *) str;
	cache_list = g_list_append (cache_list, cache);
	return;
}
/**
 * @brief get data-list element corresponding to @a name from the cache list
 * The element will be created if it does not exist
 * @param name identifier string used in cache file
 * @param list pointer to list to store the cached values
 * @return pointer to the cache data struct
 **/
E2_Cache *e2_cache_list_register (gchar *name, GList **list)
{
	E2_Cache *cache;
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found != NULL)
	{
		printd (DEBUG, "re-register list cache '%s'", name);
		cache = found->data;
		return cache;
	}
	cache = g_malloc (sizeof (E2_Cache));
	cache->type = E2_CACHE_TYPE_LIST;
	cache->sync_func = NULL;
	cache->sync_data = NULL;
	found = g_list_find_custom (unknown_list, name, 
		(GCompareFunc) str_compare);
	if (found == NULL)
	{
		printd (DEBUG, "list cache '%s' not found, creating NULL entry", name);
		cache->name = g_strdup (name);
		*list = NULL;
	} else
	{
		printd (DEBUG, "cache '%s' found", name);
		E2_Cache *unknown = found->data;
		cache->name = unknown->name;
		*list = (GList *) unknown->data;
		unknown_list = g_list_remove_link (unknown_list, found);
		g_list_free (found);
		g_free (unknown);
	}
	cache->data = (gpointer *) list;
	cache_list = g_list_append (cache_list, cache);
	return cache;
}
/**
 * @brief remove a data-list from the cache list
 * @param name identifier string used in cache list
 * @return 
 **/
void e2_cache_unregister (gchar *name)
{
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found == NULL)
	{
		printd (WARN, "trying to unregister cache '%s' which doesn't exist", name);
		return;
	}
	cache_list = g_list_remove_link (cache_list, found);
	E2_Cache *cache = found->data;
	E2_Cache *unknown = malloc (sizeof (E2_Cache));
	unknown->name = cache->name;
	unknown->type = cache->type;
	unknown->data = *cache->data;
	unknown_list = g_list_append (unknown_list, unknown);
	g_free (cache);
}
/**
 * @brief remove a data-list from the cache list
 * @param name identifier string used in cache list
 * @param size the number of elements in the array
 * @param value pointer to store for array of cached values, if any, or @a def
 * @param def array of default values to use if any are mission from cache
 * @return 
 **/
void e2_cache_array_register (gchar *name, guint size, gint *values, gint *defs)
{
	E2_Cache *cache;
	GList *found = g_list_find_custom (cache_list, name, 
		(GCompareFunc) str_compare);
	if (found != NULL)
	{
		printd (DEBUG, "re-register array cache '%s'", name);
		return;
	}
	//first-time processing
	cache = g_malloc (sizeof (E2_Cache));
	cache->type = E2_CACHE_TYPE_ARRAY;
	found = g_list_find_custom (unknown_list, name, 
		(GCompareFunc) str_compare);
	gint *stores = values;
	if (found == NULL)
	{ //not in the queue, either
		printd (DEBUG, "array cache '%s' not found, creating default", name);
		cache->name = g_strdup (name);
		guint i;
		for (i = 0 ; i < size ; i++)
			*stores++ = *defs++;
	} else
	{  //process the queued item
		printd (DEBUG, "cache '%s' found", name);
		E2_Cache *unknown = found->data;
		cache->name = unknown->name;
		gchar *end = NULL;
		gchar *str = (gchar *) unknown->data;
		gint i;
		gchar **split = g_strsplit (str, ",", -1);
		for (i = 0 ; split[i] != NULL ; i++)
		{
			*stores = g_ascii_strtoull (split[i], &end, 10);
			if (end == split[i])
			{
				printd (WARN, "array cache '%s' data conversion failed from '%s", name, split[i]);
				*stores = defs[i];
			}
			stores++;
		}
		if (i < size)
		{
			printd (WARN, "array cache '%s' missing value(s)", name);
			for (; i < size ; i++)
			*stores++ = defs[i];
		}
		g_strfreev (split);
		g_free (str);
		unknown_list = g_list_remove_link (unknown_list, found);
		g_list_free (found);
		g_free (unknown);
	}
	cache->data = (gpointer) values;
	cache->sync_data = GINT_TO_POINTER (size);
	cache_list = g_list_append (cache_list, cache);
	return;
}
/**
 * @brief write cache file to disk
 * Written file contents are encoded in utf-8, not localised
 * @return 
 **/
void e2_cache_file_write (void)
{
	if (cache_list == NULL)
	{
		printd (DEBUG, "no caches registered");
		return;
	}
	//these data need special 'syncing'
	e2_fileview_update_col_cachedata ();
	
	gchar *filename = g_build_filename
		(e2_cl_options.config_dir, default_cache_file, NULL);
	gchar *local = F_FILENAME_TO_LOCALE (filename);
	gchar *tempname = e2_utils_get_tempname (local);

	E2_FILE *f = e2_fs_open_file_to_write (tempname);

	if (f != NULL)
	{
		printd (DEBUG, "write cache file: %s", filename);

		if (e2_fs_file_write (f,
			//first line is language-independent, for version checking
			"# "PROGNAME" (v "VERSION")\n\n") == 0)
				goto error_handler;
		if (e2_fs_file_write (f,
			_("%sThis is the %s cache file.\n"
			"%sIt will be overwritten each time the program is shut down!\n\n"),
			 "# ", PROGNAME, "# ") == 0)
				goto error_handler;
		//no localisation of cache-file contents
		GList *list;
		for (list = cache_list; list != NULL; list = g_list_next (list))
		{
			E2_Cache *cache = list->data;
			switch (cache->type)
			{
				case E2_CACHE_TYPE_BOOL:
					if (e2_fs_file_write (f, "%s=%s\n", cache->name,
						*cache->data ? "true" : "false") == 0)
							goto error_handler;
					break;
				case E2_CACHE_TYPE_DOUBLE:
				{
					gchar doubl[G_ASCII_DTOSTR_BUF_SIZE];
					g_ascii_dtostr (doubl, G_ASCII_DTOSTR_BUF_SIZE, *((gdouble *) cache->data));
					if (e2_fs_file_write (f, "%s=%s\n", cache->name, doubl) == 0)
						goto error_handler;
				}
					break;
				case E2_CACHE_TYPE_INT:
					if (e2_fs_file_write (f, "%s=%d\n", cache->name, GPOINTER_TO_INT (*cache->data)) == 0)
						goto error_handler;
					break;
				case E2_CACHE_TYPE_TIME:
					//time_t may be int or long int
					if (e2_fs_file_write (f, "%s=%ld\n", cache->name, (glong) *cache->data) == 0)
						goto error_handler;
					break;
				case E2_CACHE_TYPE_STR:
					if (e2_fs_file_write (f, "%s=%s\n", cache->name, (gchar *) *cache->data) == 0)
						goto error_handler;
					break;
				case E2_CACHE_TYPE_LIST:
					if (cache->sync_func != NULL)
					{
						void (*fun) (gpointer) = cache->sync_func;
						fun (cache->sync_data);
					}
					if ((cache->data != NULL) && (*cache->data != NULL))
					{
						if (e2_fs_file_write (f, "<%s\n", cache->name) == 0)
							goto error_handler;
						GList *tmp;
						for (tmp = (GList *) *cache->data; tmp != NULL; tmp = g_list_next (tmp))
						{
							//escape anything that would seem like a tree-option start
							if (*((gchar *)tmp->data) == '<')	//ascii check is ok here
								e2_fs_file_write (f, "\\");
							if (e2_fs_file_write (f, "%s\n", (gchar *) tmp->data) == 0)
								goto error_handler;
						}
						if (e2_fs_file_write (f, ">\n") == 0)
							goto error_handler;
						//CHECKME = leaks when this is done before session-end ??
					}
					break;
				case E2_CACHE_TYPE_ARRAY:
					if (e2_fs_file_write (f, "%s=", cache->name) == 0)
						goto error_handler;
					gint *values = (gint *) cache->data;
					//j = array size - 1 - the last is handled separately
					gint i, j = GPOINTER_TO_INT (cache->sync_data) - 1;
					for (i = 0 ; i < j ; i++)
						if (e2_fs_file_write (f, "%d,", *values++) == 0)
							goto error_handler;
					if (e2_fs_file_write (f, "%d\n", *values) == 0)
						goto error_handler;
					break;
				default:
					printd (WARN, "don't know how to write '%s' to cache file",
						cache->name);
					break;
			}
		}
	}
	e2_fs_file_close (f);
	goto clean_exit;
error_handler:
	e2_fs_file_close (f);
	e2_task_backend_delete (tempname);
	gchar *msg = g_strdup_printf (_("Cannot write cache file: %s (%s)"), filename, g_strerror (errno));
	e2_output_print_error (msg, TRUE);
clean_exit:
	e2_task_backend_rename (tempname, local);
	F_FREE (local);
	g_free (tempname);
	g_free (filename);
}
/**
 * @brief interpret a line of the cache file
 * The line content is assumed to be encoded in utf-8
 * For each new cache-item, an E2_Cache struct is created,
 * and added to list unknown_list, ready for later use
 * @param f pointer to the start of the string
 * @return 
 **/
static void _e2_cache_read_from_string (gchar *f[])
{
	gboolean list_mode = FALSE;
	GList *list = NULL;
	E2_Cache *cache = NULL;
	
	gint i = 0;
	gchar *line;
	
	while ((line = f[i++]) != NULL)
	{
		g_strchomp (line);
		//ignore empty lines and comments
		if (*line == '\0')
			continue;
		if (line[0] == '#')
			continue;
		
		if (line[0] == '<')
		{
			list_mode = TRUE;
			list = NULL;
			cache = malloc (sizeof (E2_Cache));
			cache->name = g_strdup (line + 1);
			cache->type = E2_CACHE_TYPE_LIST;
			continue;
		}
		
		if (line[0] == '>')
		{
			list_mode = FALSE;
			if (list == NULL)
				continue;
			cache->data = (gpointer *) list;
			unknown_list = g_list_append (unknown_list, cache);
			continue;
		}
		
		if (list_mode)
		{
			//cleanup any escaped action-names
			if (g_str_has_prefix (line, "\\<"))
				line++;
			list = g_list_append (list, g_strdup (line));
		}
		else
		{
			cache = malloc (sizeof (E2_Cache));
			gchar **split = g_strsplit (line, "=", 2);
			cache->name = g_strdup (split[0]);
			cache->data = (gpointer *) g_strdup (split[1]);
			cache->type = E2_CACHE_TYPE_STR; //initially, all items assumed to be strings
			g_strfreev (split);
			unknown_list = g_list_append (unknown_list, cache);
		}
	} 
}
/**
 * @brief read cache file named @a fn from disk
 * @param fn name of cache file
 * @return 
 **/
static void _e2_cache_file_read (gchar *fn)
{
	//find absolute path to config file
	gchar *filepath = g_build_filename (e2_cl_options.config_dir, fn, NULL);
	gchar *cont;
	if (! e2_fs_get_file_contents (filepath, &cont))
	{
		printd (WARN, "could not open cache file '%s'", filepath);
	}
	else
	{
		printd (DEBUG, "cache read from file '%s'", filepath);
		gchar **split = g_strsplit (cont, "\n", -1);
		_e2_cache_read_from_string (split);
		g_strfreev (split);
		g_free (cont);
	}
	g_free (filepath);
}
/**
 * @brief initialize the cache
 *
 * This function will initilize the cache system of e2. It will automatically
 * be called at startup.
 *
 * @return 
 **/
void e2_cache_init (void)
{
	cache_list = NULL;
	unknown_list = NULL;
	default_cache_file = "cache"; //DO NOT TRANSLATE THIS
	_e2_cache_file_read (default_cache_file);
}
