/* $Id: e2_cl_option.c 1871 2006-04-27 09:31:07Z tpgww $
 *
 * Copyright (C) 2003-2006 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "emelfm2.h"
#include <getopt.h>
#include <string.h>
#include <glib/gprintf.h>
#include "e2_cl_option.h"

E2_CommandLineOptions e2_cl_options;

static void print_usage ()
{
	g_printf ("usage: %s [option]\n", BINNAME);
}

static void print_help ()
{
	print_usage ();
	g_printf
	(
		_("Program options:\n"
		"-1,--one=DIR           set 1st pane's start directory to DIR\n"
		"-2,--two=DIR           set 2nd pane's start directory to DIR\n"
		"-c,--config=DIR        set config directory to DIR (default: ~/.config/emelfm2)\n"
#ifdef DEBUG_MESSAGES
		"-d,--debug=[1-5]       set debug level from 1/error to 5/debug\n"
#endif
		"-e,--encoding=TYPE     set filesystem character encoding to TYPE\n"
		"-f,--fallback-encoding set fallback encoding (default: ISO-8859-1)\n"
		"-i,--ignore-problems   ignore encoding/locale problems (at your own risk!)\n"
		"-l,--leave-alone       don't suppress gtk output on console\n"
		"-n,--no-detach         don't detach from terminal\n"
		"-r,--run-at-start      command to run after session start\n"
		"-s,--set-option        set one-line gui options in config file format\n"
		"-t,--trash=DIR         set trash directory to DIR (default: ~/.Trash)\n"
		"-x,--verbose           display time/location info on debug output\n"
		"\n"
		"Help options:\n"
		"-h,--help              show this help message\n"
		"-u,--usage             display brief usage messsage\n"
		"-v,--version           display version and build info\n")
	);
}

static void print_version ()
{
	g_printf
	(
		_("%s v. %s\n"
		"Licensed under the GPL\n"
		"Copyright (C) %s\n"
		"Build date: %s\n"
		"Build platform: GTK+ %d.%d.%d %s \n"),
		PROGNAME, VERSION, COPYRIGHT, BUILDDATE,
		GTK_MAJOR_VERSION, GTK_MINOR_VERSION, GTK_MICRO_VERSION, BUILDINFO
	);
}

void e2_cl_option_process (int argc, char *argv[])
{
	e2_cl_options.pane1_path = NULL;
	e2_cl_options.pane2_path = NULL;
	gchar *freeme = g_filename_display_name (g_get_user_config_dir ());
	e2_cl_options.config_dir = g_build_filename (freeme, BINNAME, NULL);
	g_free (freeme);
	const gchar **encodings;
	g_get_filename_charsets (&encodings);
	if (encodings[0] != NULL && *encodings[0] != '\0')
		e2_cl_options.encoding = g_strdup (encodings[0]);
	else
		e2_cl_options.encoding = g_strdup ("ISO-8859-1");
	if (encodings[1] != NULL && *encodings[1] != '\0')
		e2_cl_options.fallback_encoding = g_strdup (encodings[1]);
	else
		e2_cl_options.fallback_encoding = g_strdup ("ISO-8859-1");
	e2_cl_options.overrides = NULL;
#ifdef DEBUG_MESSAGES
	e2_cl_options.debug_level = DEBUG_LEVEL;
#endif
	e2_cl_options.no_detach = (DEBUG_LEVEL > 4) ? TRUE : FALSE;
	e2_cl_options.verbose = FALSE;
	e2_cl_options.suppress_gtk_output = (DEBUG_LEVEL > 4) ? FALSE : TRUE;
	e2_cl_options.ignore_problems = FALSE;
	freeme = g_filename_display_name (g_get_user_data_dir ());
	e2_cl_options.trash_dir = g_build_filename (freeme, "Trash", "files", NULL);
	g_free (freeme);

	gint c, d = 0;
	while (1)
	{
		gint option_index = 0;
		static struct option long_options[] =
		{
			{"one", 1, 0, '1'},
			{"two", 1, 0, '2'},
			{"config", 1, 0, 'c'},
#ifdef DEBUG_MESSAGES
			{"debug", 1, 0, 'd'},
#endif
			{"ignore-problems", 0, 0, 'i'},
			{"encoding", 0, 0, 'e'},
			{"fallback-encoding", 0, 0, 'f'},
			{"leave-alone", 0, 0, 'l'},
			{"no-detach", 0, 0, 'n'},
			{"run-at-start", 1, 0, 'r'},
			{"set-option", 1, 0, 's'},
			{"verbose", 0, 0, 'x'},
			{"help", 0, 0, 'h'},
			{"trash", 1, 0, 't'},
			{"usage", 0, 0, 'u'},
			{"version", 0, 0, 'v'},
			{0, 0, 0, 0}
		};

#ifdef DEBUG_MESSAGES
		c = getopt_long (argc, argv, "1:2:c:d:e:f:t:ilnr:s:xhuv", long_options, &option_index);
#else
		c = getopt_long (argc, argv, "1:2:c:e:f:t:ilnr:s:xhuv", long_options, &option_index);
#endif
		if (c == -1) break;

		d++;

		if (optarg != NULL && *optarg != '\0')
		{
			switch (c)
			{
				case '1':
					printd (DEBUG, "setting pane 1 startup path '%s'", optarg);
					//no encoding conversion, the arrangements are not yet in place
					e2_cl_options.pane1_path = g_strdup (optarg);
					break;
				case '2':
					printd (DEBUG, "setting pane 2 startup path '%s'", optarg);
					//no encoding conversion, the arrangements are not yet in place
					e2_cl_options.pane2_path = g_strdup (optarg);
					break;
				case 'c':
					g_free (e2_cl_options.config_dir);
/*					gchar *freeme = (g_path_is_absolute (optarg)) ?
						g_strdup (optarg) :
						g_build_filename (g_get_home_dir (), optarg, NULL);
					//force utf-8
					e2_cl_options.config_dir = g_filename_display_name (freeme);
					g_free (freeme); */
					if (g_path_is_absolute (optarg))
						e2_cl_options.config_dir = g_filename_display_name (optarg);
					else
					{
						gchar *freeme = g_get_current_dir ();
						e2_cl_options.config_dir = g_build_filename (freeme, optarg, NULL);
						g_free (freeme);
						freeme = e2_cl_options.config_dir;
						e2_cl_options.config_dir = g_filename_display_name (freeme);
						g_free (freeme);
					}
					printd (DEBUG, "setting config directory '%s'", e2_cl_options.config_dir);
					break;
	#ifdef DEBUG_MESSAGES
				case 'd':
					printd (DEBUG, "setting debug level '%s'", optarg);
					{
						gchar *end;
						e2_cl_options.debug_level = g_ascii_strtoull (optarg, &end, 10);
						if (end == optarg)
						{
							e2_cl_options.debug_level = DEBUG_LEVEL;
							printd (WARN, "failed setting debug level '%s'", optarg);
						}
						if (e2_cl_options.debug_level > 4)
							e2_cl_options.suppress_gtk_output = FALSE;
					}
					break;
	#endif
				case 'e':
					g_free (e2_cl_options.encoding);
					e2_cl_options.encoding = g_strdup (optarg);
					break;
				case 'f':
					g_free (e2_cl_options.fallback_encoding);
					e2_cl_options.fallback_encoding = g_strdup (optarg);
					break;
				case 'r':
					e2_cl_options.startup_commands = g_list_append 
						(e2_cl_options.startup_commands, g_strdup (optarg));
					break;
				case 's':
					e2_cl_options.option_overrides = g_list_append
						(e2_cl_options.option_overrides, g_strdup (optarg));
					break;
				case 't':
					g_free (e2_cl_options.trash_dir);
					freeme = (g_path_is_absolute (optarg)) ?
						g_strdup (optarg) :
						g_build_filename (g_get_home_dir (), optarg, NULL);
					//force utf-8
					e2_cl_options.trash_dir = g_filename_display_name (freeme);
					g_free (freeme);
					printd (DEBUG, "setting trash directory '%s'", e2_cl_options.trash_dir);
					break;
				default:
					printd (ERROR, "unknown option with code: 0%o and value: %s", c, optarg);
					break;
			}
		}
		else	//no argument, or blank
		{
			switch (c)
			{
				case 'i':
					e2_cl_options.ignore_problems = TRUE;
					break;
				case 'l':
					e2_cl_options.suppress_gtk_output = FALSE;
				case 'n':
					e2_cl_options.no_detach = TRUE;
					break;
				case 'x':
					e2_cl_options.verbose = TRUE;
					break;
				case 'u':
					print_usage ();
					exit (0);
					break;
				case 'h':
					print_help ();
					exit (0);
					break;
				case 'v':
					print_version ();
					exit (0);
					break;
				case '?':
					exit (0);
					break;
				default:
					printd (ERROR, "unknown option with code: 0%o", c);
					break;
			}
		}
	}

	if (optind < argc)
	{
		g_printf ("You may only supply options starting with - or --.\n");
		exit (0);
	}
	printd (DEBUG, "parsed %d command line options", d);
}
