/* $Id: e2_dnd.c 1865 2006-04-25 23:23:01Z tpgww $
 *
 * Portions copyright (C) 2004-2005 tooar <tooar@gmx.net>.
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/e2_dnd.c 
 * @brief drag and drop functions
 *
 * This file contains functions related to selection and drag'ndrop.
 * Relevant button-events are handled in e2_fileview.c
 **/
 
#include "e2_dnd.h"
#include "e2_fileview.h"
#include "e2_filelist.h"
#include "e2_task.h"

GtkTargetEntry target_table[] = {
	{ "text/uri-list", 0, TARGET_URL },
	{ "text/plain",    0, TARGET_STRING }
};
guint n_targets = sizeof(target_table) / sizeof(target_table[0]);

//mouse pointer height, used for getting the path at the _top_ of the pointer
static gint pointer_height;
//default pointer height, used when we can't find the runtime value
//NOTE: setting the height too small eg 16 causes problems for
//un-lighting a row when the mouse moves slowly
#define POINTER_HEIGHT 26

extern gboolean btn2_released;

#ifdef E2_ALTLEFTMOUSE
/*this provides a quick check whether the left btn has been pressed
set and cleared in button-press and -release callbacks, respectively*/
extern gboolean left_pressed;
/*flag for whether selection-by-dragging is underway
set TRUE in drag-begin cb, cleared in drag-data-get cb*/
extern gboolean drag_sel_now;
#endif

static GtkTreeSelection *dragged_sel;
#ifdef E2_ALTLEFTMOUSE
static	GdkModifierType mask;
#endif

/**
 * @brief "drag_begin" signal callback
 *
 * For normal drag, sets dnd pointer height & multiple-item drag icon
 * For drag-selection, just sets item-counter=1
 *
 * @param treeview widget where drag started unused
 * @param dc dragcontext
 * @param view rt data for view where drag started unused
 * 
 * @return
**/
void e2_dnd_drag_begin_cb (GtkWidget *treeview,
	GdkDragContext *dc, FileView *view)
{
//	printd (DEBUG, "drag begin");
	GtkSettings *s = gtk_settings_get_default ();
	gint pointer_width;
	if (!gtk_icon_size_lookup_for_settings (s, GTK_ICON_SIZE_DND, //? is too big !
		&pointer_width, &pointer_height))
			pointer_height = POINTER_HEIGHT;	//can't find useful size, use the default
	pointer_height--;	//effectively make the 'y hotspot' 1-down from cursor top

	//remember where we began, for later drop-check
	dragged_sel = view->selection;
#ifdef E2_ALTLEFTMOUSE
	GdkDisplay *display =
	gdk_display_manager_get_default_display (gdk_display_manager_get());
	gdk_display_get_pointer (display, NULL, NULL, NULL, &mask);
	if (left_pressed && (mask & GDK_CONTROL_MASK) && !(mask & GDK_SHIFT_MASK))
	{
		drag_sel_now = TRUE;	//signal that a drag-sel has started
		//custom (blank) icon for drag-sel, if the relevant mod keys are ok
		//this special-purpose icon is always in the default icons dir
		E2_Image *cached = e2_cache_image_get (ICON_DIR"dragsel.png",
			GTK_ICON_SIZE_LARGE_TOOLBAR);
		gtk_drag_set_icon_pixbuf (dc, cached->pixbuf, 10, 0);	//x offset is ineffective ...
	}
	else
	{
#endif
		//setup dnd cursor icon
		guint count = gtk_tree_selection_count_selected_rows (dragged_sel);
		if (count > 1)
			gtk_drag_set_icon_stock (dc, GTK_STOCK_DND_MULTIPLE, 10, 0);	//y was 1, x makes no diff.
//		else //no point in checking for > 0  as default will be used, anyway
//			gtk_drag_set_icon_stock (dc, GTK_STOCK_DND, 10, 0);	//y was 1
#ifdef E2_ALTLEFTMOUSE
	}
#endif
	return;
}
/**
 * @brief "drag_motion" signal callback
 *
 * This is called when the user moves the cursor during a drag
 * For a normal drag it handles the highlighting of drop-target rows
 * For a drag-selection it selects item if needed and bumps counter
 *
 * @param treeview widget ptr
 * @param context drag context
 * @param x the left coordinate of the current cursor position
 * @param y the BOTTOM coordinate of the current cursor position
 * @param time the timestamp of the motion event
 * @param view rt data for the view where the drag [started?]
 * 
 * @return  TRUE unless drag-sel is happening (i.e. everywhere is considered a drop-zone)
**/
gboolean e2_dnd_drag_motion_cb (GtkWidget *treeview,
	GdkDragContext *context, gint x, gint y, guint time, FileView *view)
{
//	printd (DEBUG, "callback: drag motion");
	static FileView *last_view;
	GtkTreeIter iter;
	GtkTreePath *path;
	//CHECKME E2_NEWFILTER
//	GtkTreeModel *model = GTK_TREE_MODEL (view->store);
//	GtkTreeModel *model = view->model;
	GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));
	//get the path of the item under the cursor TOP ie y-pointer_height
	if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (treeview),
		x, y-pointer_height, &path, NULL, NULL, NULL))
	{
#ifdef E2_ALTLEFTMOUSE
		if (drag_sel_now	//a shift-left drag was started
//			&& (mask & GDK_CONTROL_MASK) && !(mask & GDK_SHIFT_MASK))//mod keys still correct
		)
		{
			GtkTreeSelection *sel = view->selection;
			if (!gtk_tree_selection_path_is_selected (sel, path)	//over an un-selected item
				&& (sel == dragged_sel))	//in the original treeview
			{
				//select it
				gtk_tree_selection_select_path (sel, path);
				gtk_tree_path_free (path);
				return FALSE;
			}
		}
#endif
		
		//list store path depth is 1, so row no. is 1st index
		gint row = *gtk_tree_path_get_indices (path);
		if (view == last_view)
		{
			if (row != view->drop_row //mouse has moved to another row
				&& view->lit
//				&& view->drop_row != -1
				&& gtk_tree_model_iter_nth_child (model, &iter, NULL, view->drop_row))
			{
//				printd (DEBUG, "restored color of row %d in same pane", view->drop_row);
				e2_fileview_clear_row_background (view, &iter);
			}
		}
		else
		{	//1st use, or changed pane
			// the drag-leave callback handles color-restoration  & drop_row = -1
			last_view = view;
		}

	//FIXME this does not highlight dirs when E2_NEWFILTER
		if (row != view->drop_row
			&& gtk_tree_model_get_iter (model, &iter, path))
		{
			view->drop_row = row;
			//only highlight dirs (either pane)
			FileInfo *info;
			gtk_tree_model_get (model, &iter, FINFO, &info, -1);
			if (e2_fs_is_dir (info))
			{
//				printd (DEBUG, "CHANGED color of row %d in same pane", row);
				e2_fileview_set_row_background (view, &iter, e2_option_color_get ("color-drag-highlight"));
			}
		}

		gtk_tree_path_free (path);
	}
	else //there is no row under the mouse-top
	{
//		printd (DEBUG, "No row under mouse now");
		if (view == last_view)	//still in same pane
		{
			if (view->lit
//				&& view->drop_row != -1
				&& gtk_tree_model_iter_nth_child (model, &iter, NULL, view->drop_row))
			{
//				printd (DEBUG, "restored color of row %d in same pane", view->drop_row);
				//restore the color of the last row that was lit
				e2_fileview_clear_row_background (view, &iter);
				view->drop_row = -1;
			}
		}
		else	//now in the other file-pane
		{
			// the drag-leave callback andles color restoration
			last_view = view;
		}
	}

	return TRUE;  
}
/**
 * @brief revert a drop target directory to default background
 *
 * This is the callback for "drag-leave" signals
 * It works when changing file panes, and for exits from the
 * app window
 *
 * @param treeview widget which was departed
 * @param context  dragcontext unused
 * @param time unused
 * @param view rt data for the view where the drop ocurred
 * 
 * @return  TRUE if the conversion completed successfully, else FALSE
**/
gboolean e2_dnd_drag_leave_cb (GtkWidget *treeview, 
	GdkDragContext *context, guint time, FileView *view)
{
//	printd (DEBUG, "callback: drag leave");
	gboolean retval;
	GtkTreeIter iter;
	//CHECKME E2_NEWFILTER
//	GtkTreeModel *model = GTK_TREE_MODEL (view->store);
//	GtkTreeModel *model = view->model;
	GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));
	if (!view->lit
		|| view->drop_row == -1
		|| ! gtk_tree_model_iter_nth_child (model, &iter, NULL, view->drop_row))
	{
		retval = FALSE;
	}
	else
	{
		e2_fileview_clear_row_background (view, &iter);
		retval = TRUE;
	}
	view->drop_row = -1;	//make sure any re-entry will trigger a highlight
	return retval;
}
/**
 * @brief UNUSED "drag-data-delete" signal callback
 *
 * This is called when a drag of type GDK_ACTION_MOVE 
 * was successfully completed.
 *
 * @param treeview widget where drag started unused
 * @param dc dragcontext
 * @param view rt data for view where drag started
 * 
 * @return
**/
/*void e2_dnd_drag_delete_cb (GtkWidget *treeview,
	GdkDragContext *dc, FileView *view)
{
	printd (DEBUG, "drag data delete");
	e2_task_delete (NULL, NULL, NULL);	//delete active-pane selected items
	return;
} */
/**
 * @brief callback for "drag-data-get" signal
 *
 * This is the emitted on the drag source
 * when the drop site requests the data which is dragged.
 * For a drag-selection it just cleans up some related flags
 * For a normal drag it builds a URI list of the selected filenames
 * to set as the drag data
 * Any directory item will not have a trailing separator, as the
 * names are freshly converted from the relevant FileInfo field
 *
 * @param widget UNUSED the treeview which received the signal
 * @param context UNUSED the drag context
 * @param data the GtkSelectionData to be filled with the dragged data
 * @param info_arg the info that has been registered with the target in the GtkTargetList.
 * @param time the timestamp at which the data was requested
 * @param view data structure for the view 
 * 
 * @return 
**/
void e2_dnd_drag_data_get_cb (
		GtkWidget *treeview,
		GdkDragContext *context,
		GtkSelectionData *data,
		guint info_arg,
		guint time,
		FileView *view)
{
//	printd (DEBUG, "callback: drag data get");
#ifdef E2_ALTLEFTMOUSE
	if (drag_sel_now)
	{
//		GtkTreeSelection *sel = view->selection;
//		if (view->selection == dragged_sel)
//		}
			drag_sel_now = FALSE;
			return;
//		}
	}
#endif

#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, drag data get");
#endif
	e2_filelist_disable_refresh();

	GList *base, *tmp;
	base = tmp = e2_fileview_get_selected_local (view);
	if (base != NULL)
	{	//there is actually a selection to process ...
		FileInfo *info;
		gchar *utf;
		GString *uri_list = g_string_sized_new (PATH_MAX);
	
	/* this is the emerging standard for uir list entries ?
	<?xml version="1.0" encoding="UTF-8"?>
   <resource-lists xmlns="urn:ietf:params:xml:ns:resource-lists" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">\r\n
	 <list>\r\n
	   <entry uri="sip:bill@example.com" />\r\n
	   <entry uri="sip:joe@example.org" />\r\n
	   <entry uri="sip:ted@example.net" />\r\n
	 </list>\r\n
   </resource-lists>\r\n
for now, at least, in accord with old w3c standard (which is currently under review)
/foo/bar => file:///foo/bar
OR
/foo/bar => file://localhost/foo/bar
NOTE some apps omit the leading "//" from the first example, resulting in file:/foo/bar
*/
		for (; tmp != NULL; tmp = tmp->next)
		{
			info = tmp->data;
			utf = F_FILENAME_FROM_LOCALE (info->filename);
			g_string_append_printf (uri_list, "file://%s%s\r\n", view->dir, utf);
			F_FREE (utf);
		}

		gtk_selection_data_set (data, data->target, 8, (guchar *)uri_list->str, uri_list->len);
		g_string_free (uri_list, TRUE);
		g_list_free (base);
	}
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, drag data get");
#endif
	e2_filelist_enable_refresh ();
}
/**
 * @brief get drag op menu item
 *
 * @param menuitem the clicked menu item widget
 * @param chosen store for the chosen item
 *
 * @return
**/
static void _e2_dnd_drag_op_selected_cb (GtkWidget *menuitem, GtkWidget **chosen)
{
	*chosen = menuitem;
	gtk_main_quit ();
}
/**
 * @brief execute drop
 *
 * This is the callback for "drag_data_received" signals, emitted
 * on @a treeview when dragged data has been received
 * It initiates action for the dropped items
 *
 * @param treeview widget for pane where the drop is to occur
 * @param context the dragcontext
 * @param x where the drop happened, mouse-cursor-left
 * @param y where the drop happened, mouse-cursor-BOTTOM
 * @param data pointer to selected items data struct
 * @param drag_info the info that has been registered with the target in the GtkTargetList
 * @param time the timestamp at which the data was received
 * @param view pointer to FileView data struct for the pane shown in @a treeview
 * 
 * @return 
**/
void e2_dnd_drag_data_received_cb (
		GtkWidget *treeview,
		GdkDragContext *context,
		gint x,
		gint y,
		GtkSelectionData *data,
		guint drag_info,
		guint time,
		FileView *view)
{
//	printd (DEBUG, "callback: drag data received");
	if (*(data->data) == '\0')
	{
		gtk_drag_finish (context, FALSE, FALSE, time);
		return;
	}

	//get the sender's or user's choice on what to do with the dragged items
	gboolean menu_wanted;
	if (btn2_released)
		menu_wanted = TRUE;
	else
	{
		//need to get fresh mask here, the one set in drag-begin doesn't work
		GdkDisplay *display =
		gdk_display_manager_get_default_display (gdk_display_manager_get());
		GdkModifierType mask;
		gdk_display_get_pointer (display, NULL, NULL, NULL, &mask);
//		guint modifiers = gtk_accelerator_get_default_mod_mask ();
//		menu = ((mask & modifiers) == GDK_MOD1_MASK); //alt key pressed
		menu_wanted = (mask & GDK_MOD1_MASK); //alt key pressed
	}

	GdkDragAction action = context->actions;
	if (menu_wanted && (action & GDK_ACTION_ASK))
	{
		GtkWidget *drag_op_menu = gtk_menu_new();
		GtkWidget *item1, *item2, *item3, *item4, *chosen;

		if (action & GDK_ACTION_COPY)
			item1 = e2_menu_add (drag_op_menu, _("_Copy"), NULL, NULL,
				_e2_dnd_drag_op_selected_cb, &chosen);
		else
			item1 = NULL;

		if (action & GDK_ACTION_MOVE)
			item2 = e2_menu_add (drag_op_menu, _("_Move"), NULL, NULL,
				_e2_dnd_drag_op_selected_cb, &chosen);
		else
			item2 = NULL;

		if (action & GDK_ACTION_LINK)
			item3 = e2_menu_add (drag_op_menu, _("_Link"), NULL, NULL,
				_e2_dnd_drag_op_selected_cb, &chosen);
		else
			item3 = NULL;

		item4 = e2_menu_add (drag_op_menu, _("C_ancel"), NULL, NULL,
			_e2_dnd_drag_op_selected_cb, &chosen);

		chosen = item4;
		gtk_menu_popup (GTK_MENU (drag_op_menu), NULL, NULL, NULL, NULL, 0, (guint32) time);
		gtk_main ();
		gtk_widget_destroy (drag_op_menu);
		
		if (chosen == item1)
			action = GDK_ACTION_COPY;
		else if (chosen == item2)
			action = GDK_ACTION_MOVE;
		else if (chosen == item3)
			action = GDK_ACTION_LINK;
		else
		{	//the user cancelled
			gtk_drag_finish (context, FALSE, FALSE, time);
			return;
		}
	}
	else	//don't ask
		action = context->suggested_action;
	
	//create relevant drop-directory path string
	//default is to just drop to the pane dir
	gchar *dest_dir = view->dir; //utf-8 string with trailing /
	GtkTreePath *path;
	if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (treeview),
		x, y-pointer_height, &path, NULL, NULL, NULL))
	{
		GtkTreeIter iter;
		GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));
		if (gtk_tree_model_get_iter (model, &iter, path))
		{
			gchar *filename;
			FileInfo *info;
			gtk_tree_model_get (model, &iter, FILENAME, &filename, FINFO, &info, -1);
			//NICE to add check for executable, here, and do a different drop menu
			//check if dropping onto a dir
			if (e2_fs_is_dir (info))
			{
				//now that we're finishing, un-light the row
				e2_fileview_clear_row_background (view, &iter);
				//prevent dropping into a dir that's part of the selection
				GtkTreeSelection *sel = view->selection;
				if ((sel == dragged_sel)	//we're in the same selection as we started with
					&& gtk_tree_selection_path_is_selected (sel, path))
				{
					g_free (filename);
					gtk_tree_path_free (path);
					return;
				}
				//append target dir's name to the pane dir path
				//view->dir, filename both utf, both have trailing /
				dest_dir = g_strconcat (view->dir, filename, NULL);
			}
			g_free (filename);
		}
		gtk_tree_path_free (path);
	}

	//do it
	E2_TaskType droptask;
	if (action & GDK_ACTION_COPY)
		droptask = E2_TASK_COPY;
	else if (action & GDK_ACTION_MOVE)
		droptask = E2_TASK_MOVE;
	else if (action & GDK_ACTION_LINK)
		droptask = E2_TASK_LINK;
	else
		droptask = -1;
	//URI list at data->data is a utf8 string
	gboolean result = (droptask != -1) ?
		e2_task_drop (droptask, data->data, dest_dir): FALSE;
	context->action = action;
	gtk_drag_finish (context, result, action & GDK_ACTION_MOVE, time);

	if (dest_dir != view->dir)
		g_free (dest_dir);
	
//CHECKME E2_NEWFILTER
/* Must override the default 'drag_data_received' handler on GtkTreeView when
using models that don't support the GtkTreeDragDest interface and enabling
drag-and-drop. The simplest way to do this is to connect to 'drag_data_received'
and call g_signal_stop_emission_by_name() in your signal handler to prevent the
default handler from running. Look at the source code for the default handler in
gtktreeview.c to get an idea what your handler should do.
OR ?? appropriate usage of
if (GTK_IS_TREE_MODEL_FILTER (model))
{
	model = gtk_tree_model_filter_get_model
	(GTK_TREE_MODEL_FILTER (model));
} */
}
