/* $Id: e2_filelist.c 1867 2006-04-26 23:19:19Z tpgww $
 *
 * Portions copyright (C) 2004-2006 tooar <tooar@gmx.net>.
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/e2_filelist.c 
 * @brief directory content liststore functions
 *
 * This file contains functions related to creation, filling and emptying a
 * liststore of directory content data, and fns for the associated model.
 * Also filtering of store content, and other related utilties.
 **/

#include "e2_filelist.h"
#include <string.h>
#include <sys/time.h>
#include <langinfo.h>
#include <pwd.h>
#include <time.h>
#include <grp.h>
#include "e2_option.h"
#include "e2_task.h"
#include "e2_dialog.h"

#ifdef E2_SELTXT_RECOLOR
extern GdkColor selectedtext;
#endif

#ifdef E2_FAM_DNOTIFY
#include <signal.h>
extern sigset_t dnotify_signal_set;
#endif

//disable-refresh counter
gint refresh_ref;
//static gboolean case_sensitive; //local copy of option value, for faster access

#ifndef E2_NEWFILTER
/**
 * @brief decide whether an item should be included in the filelist, on the basis of its name
 *
 * Item name is compared with current filters
 * Item name string is localised
 * though this fn assumes ascii = FIXME
 *
 * This function is also "borrowed" to identify selected items which match a filter
 *
 * @param view data structure for view being processed
 * @param info data structure for item being checked
 * @param pattern string with pattern(s) to match, comma-separated if > 1
 *
 * @return TRUE if item 'passes' (i.e. to be included in file list)
 */
static gboolean _e2_filelist_match_name_filter (FileView *view,
	FileInfo *info, gchar *pattern)
{
	gchar *s, *p, *utf, *freeme;
	gchar save;
	gboolean negated, matched, positive_check = FALSE, result = FALSE;

	p = pattern;
	utf = F_FILENAME_FROM_LOCALE (info->filename);
	//maybe several patterns, separated by commas
	while ((s = strchr (p, ',')) != NULL)	//if always ascii ',', don't need g_utf8_strchr()
	{	//check each pattern that is followed by a ','
		save = *s;
		*s = '\0';
		while (p[0] == ' ')
			p++;
		if (p[0] == '!')
		{
			negated = !view->name_filter.invert_mask;
			p++;
		}
		else
		{
			negated = view->name_filter.invert_mask;
			if (p[0] == '\\' && p[1] == '!')
				p++;
		}
		if (!positive_check)
			positive_check = !negated;

		if (!view->name_filter.case_sensitive)
		{
		  freeme = g_utf8_strdown (utf, -1);
		  matched = g_pattern_match_simple (p, freeme);
		  g_free (freeme);
		}
		else
		  matched = g_pattern_match_simple (p, utf);

		*s = save;
		if (matched && negated)
		{
		  F_FREE (utf);
		  return FALSE;
		}
		if (matched && !negated)
		  result = TRUE;	//but keep looking for any later exclude
//		else if (!matched && negated && !positive_check)
//			result = TRUE;
		//if neither negated nor matched, we don't change result

		NCHR(s);	//in case the matched char is bigger than 1 byte
		p = s;
	}
	//check the last (or only) pattern
	while (p[0] == ' ')
		p++;
	if (p[0] == '\0')
		return result;
	if (p[0] == '!')
	{
		negated = !view->name_filter.invert_mask;
		p++;
	}
	else
	{
		negated = view->name_filter.invert_mask;
		if (p[0] == '\\' && p[1] == '!')
			p++;
	}
	if (!positive_check)
		positive_check = !negated;

	if (!view->name_filter.case_sensitive)
	{
		freeme = g_utf8_strdown (utf, -1);
		matched = g_pattern_match_simple (p, freeme);
		g_free (freeme);
	}
	else
		matched = g_pattern_match_simple (p, utf);
	if (matched)
		result = !negated;
	//extra check for unmatched final check
	else if (negated && !positive_check)
		result = TRUE;

	F_FREE (utf);
	return result;
}
/**
 * @brief decide whether an item should be included in the filelist, on the basis of its size
 *
 * Item statbuf.st_size is compared with curent filters
 *
 * @param view data structure for view being processed
 * @param info data structure for item being checked
 *
 * @return TRUE if item does NOT need to be filtered out
 */
static gboolean _e2_filelist_match_size_filter (FileView *view, FileInfo *info)
{
	switch (view->size_filter.op)
	{
		case GT:
		  return (info->statbuf.st_size > view->size_filter.size);
		  break;
		case LT:
		  return (info->statbuf.st_size < view->size_filter.size);
		  break;
		case EQ:
		  return (info->statbuf.st_size == view->size_filter.size);
		  break;
		default:
		  return TRUE;
	}
	// this path can never happen
	return TRUE;
}
/**
 * @brief decide whether an item should be included in the filelist, on the basis of its access time
 *
 * Item statbuf.st_atime is compared with curent filters
 *
 * @param view data structure for view being processed
 * @param info data structure for item being checked
 *
 * @return TRUE if item does NOT need to be filtered out
 */
static gboolean _e2_filelist_match_date_filter (FileView *view, FileInfo *info)
{
	switch (view->date_filter.time_type)
	{
		case MTIME:
		  if (view->date_filter.op == GT)
		    return (difftime(view->date_filter.time, info->statbuf.st_mtime) <= 0);
		  else
		    return (difftime(view->date_filter.time, info->statbuf.st_mtime) > 0);
		  break;
		case ATIME:
		  if (view->date_filter.op == GT)
		    return (difftime(view->date_filter.time, info->statbuf.st_atime) <= 0);
		  else
		    return (difftime(view->date_filter.time, info->statbuf.st_atime) > 0);
		  break;
		case CTIME:
		  if (view->date_filter.op == GT)
		    return (difftime(view->date_filter.time, info->statbuf.st_ctime) <= 0);
		  else
		    return (difftime(view->date_filter.time, info->statbuf.st_ctime) > 0);
		  break;
		default:
		  return TRUE;
	}
	// this path can't happen
	return TRUE;
}
#endif //ndef E2_FILTER
/* These functions must encapsulate any instructions that access the FileInfo
 * structs produced when a selection is interrogated.
 * This is because a refresh in the middle of such an operation would free
 * the FileInfo structs as a side-effect.
 */
/**
 * @brief  disable refreshing of config data and pane filelists
 *
 * Increase ref count, if == 1, stop the timer process which calls the refresh fns
 * Why is config data check bundled here ?
 *
 * @return
 */
void e2_filelist_disable_refresh (void)
{
	refresh_ref++;
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, refresh ref now = %d", refresh_ref);
#endif
	if (refresh_ref == 1)
	{
		if (e2_option_bool_get ("auto-refresh"))
		{
//don't cancel timer, to reduce risk that dirty-reports queue may over-fill
//			g_source_remove (app.filelist_timer_id);
			curr_view->blocked = TRUE;
			other_view->blocked = TRUE;
		}
		if (e2_option_bool_get ("auto-refresh-config"))
			//CHECKME just block this too?
			//(can't reasonably be many changes to the config file)
			g_source_remove (app.config_timer_id);
		e2_window_disable_status_update ();
	}
}
/**
 * @brief  cause polling of config data and pane filelists at specified intervals
 *
 * Polling (re)starts after the interval 
 * reference count is back to 0
 *
 * @param interval  milliseconds between polls
 *
 * @return
 */
/*void e2_filelist_set_refresh (guint interval)
{
	e2_filelist_disable_refresh ();
	e2_filelist_enable_refresh_custom (interval);
}

void e2_filelist_set_refresh_default (void)
{
	e2_filelist_disable_refresh ();
	e2_filelist_enable_refresh_custom (E2_FILESCHECK_INTERVAL);
} */
/**
 * @brief  re-enable polling of config data and pane filelists
 *
 * Polling (re)starts after the default 1-second interval, if the 
 * reference count is back to 0
 *
 * @return
 */
void e2_filelist_enable_refresh (void)
{
	e2_filelist_enable_refresh_custom (E2_FILESCHECK_INTERVAL);
}
/**
 * @brief  re-enable polling of config data and pane filelists, with a specified interval
 *
 * Decrease ref count, if == 0, restart the timers which call the refresh fns
 *  Config timer is bundled here because the fn refreshes the panes
 *
 * @param interval milliseconds between polls
 *
 * @return
 */
void e2_filelist_enable_refresh_custom (guint interval)
{
	refresh_ref--;
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, refresh ref now = %d", refresh_ref);
#endif
	if (refresh_ref < 0)
	{
		printd (WARN, "The auto-refresh refresh count is < 0");
		e2_window_output_show (NULL, NULL);
		e2_output_print_error (
		_("Something is wrong with the auto-refresh.\nPlease submit a bug report"), FALSE);
	}

	if (refresh_ref == 0)
	{
		if (e2_option_bool_get ("auto-refresh"))
		{
//			app.filelist_timer_id = g_timeout_add 
//				(interval, (GSourceFunc) e2_filelist_check_dirty, NULL);
			curr_view->blocked = FALSE;
			other_view->blocked = FALSE;
		}
		//this timer is used, even with fam backends
		if (e2_option_bool_get ("auto-refresh-config"))
			app.config_timer_id = g_timeout_add
				(E2_CONFIGCHECK_INTERVAL, (GSourceFunc) e2_option_check_config_files, NULL);
		e2_window_enable_status_update (-1);
	}
}
/**
 * @brief refresh filepane contents (and with FAM, config too) if the corresponding changed is detected
 *
 * Among other uses, this is the timeout function for filelists auto refresh
 * When FAM/gamin in use, it also checks for and flags (not processes)
 * config file updates, as that data comes from the same data stream.
 *
 * @param userdata data specified when the timer was initialised (NULL), or when called directly, non-NULL
 *
 * @return TRUE so the timer keeps working
 */
gboolean e2_filelist_check_dirty (gpointer userdata)
{
	static gboolean p1deferred = FALSE;
	static gboolean p2deferred = FALSE;
#ifdef E2_FAM
	static gboolean cfgdeferred = FALSE;
#endif

	gboolean p1dirty, p2dirty;
#ifdef E2_FAM
	gboolean localcfgdirty;	//use a local copy so that the extern value doesn't get cleared before it's noticed
	extern gboolean cfgdirty;
#endif

#ifdef E2_FAM
	if (userdata != NULL
		 && app.monitor_type != E2_MONITOR_DEFAULT)
			g_usleep (E2_FAMWAIT);	//wait for things to be noticed
#endif
	//this reports 'gone' as 'dirty', and then the list-refesh
	//function handles choosing a replacement dir
	e2_fs_FAM_check_dirty (&p1dirty, &p2dirty
#ifdef E2_FAM
	, &localcfgdirty
#endif
	);
	
	//CHECKME process active pane first, more likely to be attending to that ?

	if (p1dirty || p1deferred)
	{
		if (!app.pane1_view.completed	//not finished from last time
			|| (app.pane1_view.blocked && userdata == NULL))	//blocked auto-refresh
		{
			p1deferred = TRUE;
			printd (DEBUG, "pane 1 refresh deferred");
		}
		else
		{
			p1deferred = FALSE;
			e2_fileview_refresh_list (&app.pane1_view);
			printd (DEBUG, "pane 1 refreshed");
		}
	}

	if (p2dirty || p2deferred)
	{
		if (!app.pane2_view.completed
			|| (app.pane2_view.blocked && userdata == NULL))
		{
			p2deferred = TRUE;
			printd (DEBUG, "pane 2 refresh deferred");
		}
		else
		{
			p2deferred = FALSE;
			e2_fileview_refresh_list (&app.pane2_view);
			printd (DEBUG, "pane 2 refreshed");
		}
	}

#ifdef E2_FAM
	if (localcfgdirty || cfgdeferred)
	{
		if (!app.rebuild_enabled)
		{
			cfgdeferred = TRUE;
			printd (DEBUG, "config refresh deferred");
		}
		else
		{
			cfgdeferred = FALSE;
			if (localcfgdirty)
			{
				cfgdirty = TRUE;	//set, but don't clear, the main flag
				printd (DEBUG, "config refresh flag set");
			}
		}
	}
#endif
	return TRUE;
}
#ifdef E2_NEWSTORES
/**
 * @brief idle function to clear old liststores
 *
 * @param user_data UNUSED data specified when the idle was setup
 *
 * @return FALSE, to stop the callbacks
 */
gboolean e2_filelist_clear_old_stores (gpointer user_data)
{
	GSList *tmp;
#ifdef DEBUG_MESSAGES
	gint debug = g_slist_length (app.used_stores);
#endif
	for (tmp = app.used_stores; tmp != NULL; tmp = tmp->next)
	{
		GtkListStore *store = tmp->data;
		GtkTreeModel *mdl = GTK_TREE_MODEL (store);
		GtkTreeIter iter;
		if (gtk_tree_model_get_iter_first (mdl, &iter))
		{	//it's not empty already
			//clear file info data structs referred to in the store
			//CHECKME need to clear anything else?
			FileInfo *info;
			do
			{
				gtk_tree_model_get (mdl, &iter, FINFO, &info, -1);
#if defined(E2_CURRENTGTK_VERSION) && GLIB_CHECK_VERSION (2,10,0)
				g_slice_free1 (sizeof(FileInfo), info);
#else
				g_free (info);
#endif
			} while (gtk_tree_model_iter_next (mdl, &iter));
		}
#ifdef E2_NEWFILTER
//CHECKME clear any filtermodel ?
#endif
		g_object_unref (G_OBJECT (store));
	}
	g_slist_free (app.used_stores);
	app.used_stores = NULL;
	printd (DEBUG, "%d old liststore(s) cleared", debug);
	return FALSE;
}
#endif
/**
 * @brief create and populate filelist-compatible list store with rows for each item in @a entries
 * This also sets 
 * @param entries list of FileInfo's to be processed
 * @param view pointer to view data struct
 * @param omitted store for count of filtered items in @a entries
 *
 * @return pointer to the liststore, or NULL if there's a problem
 */
GtkListStore *e2_filelist_fill_store (GList *entries
#ifndef E2_NEWFILTER
	, FileView *view, guint *omitted
#endif
)
{
//	printd (DEBUG, "start store fill");
	GtkTreeIter iter;
	GtkListStore *store = e2_filelist_make_store ();
	if (store == NULL)
		return NULL;

	GList *tmp;
	FileInfo *infoptr;
	struct tm *tm_ptr;
	struct passwd *pwd_buf;
	struct group *grp_buf;
	gchar size_buf[20];	//enough for 999 Tb
	gchar modified_buf[25];
	gchar accessed_buf[25];
	gchar changed_buf[25];
	gchar perm_buf[11];
	gchar uid_buf[20];
	gchar gid_buf[20];
	gchar *buf[NAMEKEY+1];	//pointers to NAMEKEY strings
	gchar *freeme;

	//get format parameters

	//format for dates display
	gchar *strf_string;
	switch	(e2_option_int_get ("date-string"))
	{
		case 1:
			strf_string = "%d/%m/%y %H:%M";	//standard
			break;
		case 2:
			strf_string = "%m/%d/%y %H:%M";	//american
			break;
		case 3:
/*			{
				strf_string = nl_langinfo (D_T_FMT);
				if (strf_string != NULL && *strf_string != '\0')
					break;
			} */
			strf_string = "%x %X";	//localised
			break;
		default:
			strf_string = "%b %d %H:%M";
			break;
	}
	//get format for size display
	gchar *comma;
	switch	(e2_option_int_get ("size-string"))	
	{
		case 1:
		{
			comma = nl_langinfo (THOUSEP);
			if (comma == NULL || *comma == '\0')
				comma = ",";
			break;
		}
		default:
			comma = NULL;	//signal to use the condensed version
			break;
	}

	gboolean caseignore = ! e2_option_bool_get ("namesort-case-sensitive");
	GtkStyle *style = gtk_rc_get_style (curr_view->treeview);
	GdkColor *default_color = &style->text[GTK_STATE_NORMAL];

	GdkColor *link_color = e2_option_color_get ("color-ft-link");
	GdkColor *dir_color = e2_option_color_get ("color-ft-dir");
	GdkColor *dev_color = e2_option_color_get ("color-ft-dev");
	GdkColor *sock_color = e2_option_color_get ("color-ft-socket");
	GdkColor *exec_color = e2_option_color_get ("color-ft-exec");

#ifndef E2_NEWFILTER
	gboolean test1 = !e2_option_bool_get ("show-updir-entry")
			|| g_str_equal (view->dir, G_DIR_SEPARATOR_S);  //at root directory now?
	//convert this once
	gchar *pattern;	//ignore compiler warning about unitialized usage
	//FIXME faster to compile the pattern(s) once, and match against that
	if (view->name_filter.active)
	{
		if (!view->name_filter.case_sensitive)
			pattern = g_utf8_strdown (view->name_filter.patternptr, -1);
		else pattern = g_strdup (view->name_filter.patternptr);
	}
	guint totalskips = 0;	//item counter to assist status-line counts
#endif //ndef E2_NEWFILTER

	//iterate through the listed FileInfo's
	for (tmp = entries; tmp != NULL; tmp = tmp->next)
	{
		infoptr = (FileInfo *) tmp->data;
#ifndef E2_NEWFILTER
		//some items we're not interested in
		if (infoptr->filename[0] == '.'
			&& (
				   (!view->show_hidden && (infoptr->filename[1] != '.' || infoptr->filename[2] != '\0'))
				|| (infoptr->filename[1] == '\0')
				|| (test1 && infoptr->filename[1] == '.' && infoptr->filename[2] == '\0')
				)
			)
		{
			totalskips++;
			continue;
		}
		if (view->filter_directories || !e2_fs_is_dir (infoptr))
		{
			if (
				(view->name_filter.active &&
					!_e2_filelist_match_name_filter (view, infoptr, pattern))
			||	(view->size_filter.active &&
					!_e2_filelist_match_size_filter (view, infoptr))
			||	(view->date_filter.active &&
					!_e2_filelist_match_date_filter (view, infoptr))
			)
			{
				totalskips++;
				continue;
			}
		}
		//this one is to keep
#endif //ndef E2_NEWFILTER

		//convert to utf if appropriate
		buf[FILENAME] = F_FILENAME_FROM_LOCALE (infoptr->filename);
		//directories (and links to dirs) get a trailing separator
		if (e2_fs_is_dir (infoptr))
		{
			freeme = buf[FILENAME];
			buf[FILENAME] = e2_utils_strcat (freeme, "/");
			if (freeme != infoptr->filename)
				g_free (freeme);
		}
		else if (buf[FILENAME] == infoptr->filename)
			buf[FILENAME] = g_strdup (buf[FILENAME]);

		if (comma == NULL)
		{	//use condensed size format
			if (infoptr->statbuf.st_size < 10240) // less than 10k
			{
			  g_snprintf(size_buf, sizeof(size_buf), "%llu",
					infoptr->statbuf.st_size);
			}
			else if (infoptr->statbuf.st_size < 1048576) // less than a meg
			{
			  g_snprintf(size_buf, sizeof(size_buf), "%.1f%s",
				(gfloat) infoptr->statbuf.st_size / 1024.0, _("k"));
			}
			else  // more than a meg
			{
			  g_snprintf(size_buf, sizeof(size_buf), "%.1f%s",
				(gfloat) infoptr->statbuf.st_size / 1048576.0, _("M"));
			}
		}
		else
		{	//use actual size, with commas
			g_snprintf(size_buf, sizeof(size_buf), "%llu",
				infoptr->statbuf.st_size);

			guint len = strlen (size_buf);
			guint ths = len-1;  //0-based index
			guint i;
			while (ths > 2 && len < sizeof(size_buf))
			{
				for (i = len-1; i > ths-3; i--)
					size_buf[i+1] = size_buf[i];

				size_buf[i+1] = *comma;
				size_buf[++len] = '\0';
				ths = i;
			}
		}
		buf[SIZE] = g_strdup (size_buf);	//content is ascii, no need to convert to utf

		e2_fs_get_perm_string (perm_buf, sizeof(perm_buf), infoptr->statbuf.st_mode);
//		buf[PERM] = e2_utf8_from_locale (perm_buf);
		buf[PERM] = g_strdup (perm_buf); //content is ascii, no need to convert to utf

		if ((pwd_buf = getpwuid (infoptr->statbuf.st_uid)) == NULL)
		{
		  g_snprintf (uid_buf, sizeof(uid_buf), "%d",
		     (guint) infoptr->statbuf.st_uid);
		  buf[OWNER] = g_strdup (uid_buf); //content is ascii number, no need to convert to utf
		}
		else
		{
		  buf[OWNER] = e2_utf8_from_locale (pwd_buf->pw_name);
		}

		if ((grp_buf = getgrgid (infoptr->statbuf.st_gid)) == NULL)
		{
		  g_snprintf(gid_buf, sizeof(gid_buf), "%d",
		       (guint) infoptr->statbuf.st_gid);
		  buf[GROUP] = g_strdup (gid_buf); //content is ascii number, no need to convert to utf
		}
		else
		{
		  buf[GROUP] = e2_utf8_from_locale (grp_buf->gr_name);
		}
		tm_ptr = localtime (&(infoptr->statbuf.st_mtime));
		strftime (modified_buf, sizeof(modified_buf), strf_string, tm_ptr);
		buf[MODIFIED] = e2_utf8_from_locale (modified_buf);

		tm_ptr = localtime (&(infoptr->statbuf.st_atime));
		strftime (accessed_buf, sizeof(accessed_buf), strf_string, tm_ptr);
		buf[ACCESSED] = e2_utf8_from_locale (accessed_buf);

		tm_ptr = localtime(&(infoptr->statbuf.st_ctime));
		strftime (changed_buf, sizeof(changed_buf), strf_string, tm_ptr);
		buf[CHANGED] = e2_utf8_from_locale (changed_buf);

		if (caseignore)
		{
			freeme = g_utf8_casefold (buf[FILENAME], -1);
#if defined (E2_CURRENTGTK_VERSION) && GTK_CHECK_VERSION (2, 8, 0)
			buf[NAMEKEY] = g_utf8_collate_key_for_filename (freeme, -1);
#else
			buf[NAMEKEY] = g_utf8_collate_key (freeme, -1);
#endif
			g_free (freeme);
		}
		else
#if defined (E2_CURRENTGTK_VERSION) && GTK_CHECK_VERSION (2, 8, 0)
			buf[NAMEKEY] = g_utf8_collate_key_for_filename (buf[FILENAME], -1);
#else
			buf[NAMEKEY] = g_utf8_collate_key (buf[FILENAME], -1);
#endif
		GdkColor *foreground;
		switch (infoptr->statbuf.st_mode & S_IFMT)
		{
		  case S_IFLNK:
		    foreground = link_color;
		    break;
		  case S_IFDIR:
		    foreground = dir_color;
		    break;
		  case S_IFCHR:
		  case S_IFBLK:
		    foreground = dev_color;
		    break;
		  case S_IFSOCK:
		    foreground = sock_color;
		    break;
		  case S_IFREG:
			  //show as executable if _anyone_ has that permission
		    if ((S_IXUSR & infoptr->statbuf.st_mode)
		  		|| (S_IXGRP & infoptr->statbuf.st_mode)
		  		|| (S_IXOTH & infoptr->statbuf.st_mode))
			  foreground = exec_color;
			else
			{
#ifdef E2_RAINBOW
				//find extension of current item
				//localised text, but assumes . is ascii
				//hashed extension-strings used for comparison have been localised
				foreground = NULL;	//in case there's no '.' at all
				gchar *ext = infoptr->filename + 1;	//+1 in case item is hidden
				while ((ext = strchr (ext,'.')) != NULL)
				{
					ext++;	//pass the '.'
					//use its color, if any
					foreground = g_hash_table_lookup (app.colors, ext);
					if (foreground != NULL)
						break;
				}
				if (foreground == NULL)
					foreground = default_color;
#else
				foreground = default_color;
#endif
			}
		    break;
		  default:
			foreground = default_color;
		    break;
		}

		gtk_list_store_insert_with_values (store, &iter, -1,
			FILENAME, buf[FILENAME],
			SIZE, buf[SIZE],
			PERM, buf[PERM],
			OWNER, buf[OWNER],
			GROUP, buf[GROUP],
			MODIFIED, buf[MODIFIED],
			ACCESSED, buf[ACCESSED],
			CHANGED, buf[CHANGED],
			NAMEKEY, buf[NAMEKEY],
			FINFO, infoptr,
			FORECOLOR, foreground,
//			VISIBLE, TRUE,
			-1);

		//cleanup
		gint i;
		for (i = 0; i <= NAMEKEY; i++)
			g_free (buf[i]);

/*
//		gettimeofday (&strt, NULL); //FOR DEBUGGING, check time this load takes
//		gettimeofday (&nd, NULL);
		if (nd.tv_usec < strt.tv_usec)
		{
		    int nsec = (strt.tv_usec - nd.tv_usec) / 1000000 + 1;
    		strt.tv_usec -= 1000000 * nsec;
    		strt.tv_sec += nsec;
		}
		if (nd.tv_usec - strt.tv_usec > 1000000)
		{
			int nsec = (strt.tv_usec - nd.tv_usec) / 1000000;
			strt.tv_usec += 1000000 * nsec;
			strt.tv_sec -= nsec;
		}
		result.tv_sec = result.tv_sec + nd.tv_sec - strt.tv_sec;
		result.tv_usec = result.tv_usec + nd.tv_usec - strt.tv_usec;
		if (result.tv_usec > 1000000)
		{
			int nsec = result.tv_usec / 1000000 + 1;
    		result.tv_usec -= 1000000 * nsec;
    		result.tv_sec += nsec;
		}
	
//	printd (DEBUG, "populating rows took %f seconds", result.tv_sec + result.tv_usec / 1000000.0 );
*/
	}
#ifndef E2_NEWFILTER
	if (view->name_filter.active)
		g_free (pattern);
	//record count of items in the list, adjusted for . and ..
	*omitted = totalskips;
#endif	//ndef E2_NEWFILTER

	return store;
}
/**
 * @brief create list store structure
 *
 * Creates basic structure of list store, for panes filelists
 * No rows or data are added.
 *
 * @return the new liststore
 */
GtkListStore *e2_filelist_make_store (void)
{
	GtkListStore *store = gtk_list_store_new (MODEL_COLUMNS,
		G_TYPE_STRING,  //FILENAME
		G_TYPE_STRING,  //SIZE
		G_TYPE_STRING,  //PERM
		G_TYPE_STRING,  //OWNER
		G_TYPE_STRING,  //GROUP
		G_TYPE_STRING,  //MODIFIED
		G_TYPE_STRING,  //ACCESSED
		G_TYPE_STRING,  //CHANGED
		// the rest are not displayed
		G_TYPE_STRING,  //NAMEKEY for i18n name sorts
		G_TYPE_POINTER,  //FINFO pr to FileInfo for the item
		GDK_TYPE_COLOR,  //FORECOLOR line colour
		GDK_TYPE_COLOR  //BACKCOLOR line colour
	//	G_TYPE_BOOLEAN	//VISIBLE whether filtered or not
		);
	return store;
}
