/* $Id: e2_fileview.c 1861 2006-04-25 08:27:44Z tpgww $
 *
 * Portions copyright (C) 2004-2006 tooar <tooar@gmx.net>.
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/e2_fileview.c 
 * @brief directory content view functions
 *
 * This file contains functions related to creation of treeviews to display
 * directory content, plus related selections and sorting.
 **/

#include "emelfm2.h"
//#include <unistd.h>
#include <string.h>
#ifdef E2_NEWFILTER
#include <time.h>
#endif
#include "e2_dnd.h"
#include "e2_filelist.h"
#include "e2_context_menu.h"
#include "e2_option.h"
#include "e2_dialog.h"
#include "e2_task.h"

typedef struct _DirHistoryEntry
{
	gchar path[PATH_MAX];
	gint row;
	gboolean selected;  //TRUE if row was selected
} DirHistoryEntry;

extern GtkTargetEntry target_table[];
extern guint n_targets;

//this holds, in fixed column order (used in model & config, cache data),
// the displayed column widths, a row for each pane
gint col_width_store[2][MAX_COLUMNS];
//this holds, in fixed column order (used in model & config, cache data),
// the displayed column order, a row for each pane
gint stored_col_order[2][MAX_COLUMNS];
//this holds column data from the above arrays, for cache reading/writing
//GList *cols_data;
//this holds, in displayed column order
// the fixed column order, a row for each pane
gint displayed_col_order[2][MAX_COLUMNS];

// Colors
GdkColor LIST_COLOR;
#ifdef E2_SELTXT_RECOLOR
GdkColor selectedtext;
#endif

gboolean btn2_released;	//flag used to decide id middle-btn DnD menu is wanted

//local copy of flag
static gboolean button2updir;
//number of last mouse button press, for distinguishing drags
static gboolean pane1_all = FALSE, pane2_all=FALSE;
//these are related to column dragging
static gint header_button;
static gboolean block;
#ifdef E2_ALTLEFTMOUSE
/*this provides a quick check whether the left btn has been pressed
set and cleared in button-press and -release callbacks, respectively*/
gboolean left_pressed;
/*flag for whether selection-by-dragging is underway
set TRUE in drag-begin cb, cleared in drag-data-get cb*/
gboolean drag_sel_now;
//max msec between clicks when checking for a doubleclick
//extern guint click_interval;
#endif

//functions for sorting of treeviews

static gint _e2_fileview_name_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);
static gint _e2_fileview_size_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);
static gint _e2_fileview_perm_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);
static gint _e2_fileview_user_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);
static gint _e2_fileview_group_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);
static gint _e2_fileview_mdate_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);
static gint _e2_fileview_adate_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);
static gint _e2_fileview_cdate_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction);

E2_Column e2_all_columns[MAX_COLUMNS] = {
	{N_("Filename"), 120, _e2_fileview_name_sort},
	{N_("Size"), 80, _e2_fileview_size_sort},
	{N_("Permissions"), 80, _e2_fileview_perm_sort},
	{N_("Owner"), 60, _e2_fileview_user_sort},
	{N_("Group"), 60, _e2_fileview_group_sort},
	{N_("Modified"), 120, _e2_fileview_mdate_sort},
	{N_("Accessed"), 120, _e2_fileview_adate_sort},
	{N_("Changed"), 120, _e2_fileview_cdate_sort} };

/**
 * @brief item-name sort-order comparison function
 *
 * This compares 'sort-key' row data, which is utf-8 compatible
 * If the case-insensitive sorting option was in force when the data 
 * was loaded into the model, the keys are 'case-folded'
 * ie this is not so dynamic as we might want
 * Directories are placed before other things.
 * NB utf8 collate keys for ascii text are case insensitive !!
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_fileview_name_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	gchar *keya; // = data;
	gchar *keyb; // = data + E2_UTFNAME_MAX;  //this is the size set in e2_fileview_column_header_clicked_cb() etc
	FileInfo *infoa, *infob;
	gint result;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	//locks in this fn mist be local, to avoid blocking
	gtk_tree_model_get (model, a, NAMEKEY, &keya, FINFO, &infoa, -1);
	gtk_tree_model_get (model, b, NAMEKEY, &keyb, FINFO, &infob, -1);

	struct stat statfast;

	if (g_str_equal (infoa->filename, ".."))
		result = (order) ? -1 : 1;
	else
		if (g_str_equal (infob->filename, ".."))
			result = (order) ? 1 : -1;
	else
		if (e2_fs_is_dir_fast (infoa, &statfast))
		{
			if (e2_fs_is_dir_fast (infob, &statfast))
				result = strcmp (keya, keyb);
			else
		  		result = (order) ? -1 : 1;
		}
	else
		if (e2_fs_is_dir_fast (infob, &statfast))
			result = (order) ? 1 : -1;
	else
		result = strcmp (keya, keyb);

	g_free (keya);
	g_free (keyb);

	return result;
}
/**
 * @brief item-size sort-order comparison function
 *
 * Directories are placed before other things.
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_fileview_size_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	FileInfo *infoa, *infob;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	gtk_tree_model_get (model, a, FINFO, &infoa, -1);
	if (g_str_equal (infoa->filename, ".."))
		return (order) ? -1 : 1;

	gtk_tree_model_get (model, b, FINFO, &infob, -1);
	if (g_str_equal (infob->filename, ".."))
		return (order) ? 1 : -1;

	struct stat statfast;

	gboolean bf = e2_fs_is_dir_fast (infob, &statfast);
	
	if (e2_fs_is_dir_fast (infoa, &statfast))
	{
		if (bf)
		  return (infoa->statbuf.st_size - infob->statbuf.st_size);
		else
		  return (order) ? -1 : 1;
	}

	if (bf)
		return (order) ? 1 : -1;
	
	return (infoa->statbuf.st_size - infob->statbuf.st_size);
}
/**
 * @brief item-permission sort-order comparison function
 *
 * Directories are placed before other things.
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_fileview_perm_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	FileInfo *infoa, *infob;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	gtk_tree_model_get (model, a, FINFO, &infoa, -1);
	if (g_str_equal (infoa->filename, ".."))
		return (order) ? -1 : 1;
	
	gtk_tree_model_get (model, b, FINFO, &infob, -1);
	if (g_str_equal (infob->filename, ".."))
		return (order) ? 1 : -1;

	struct stat statfast;

	gboolean bf = e2_fs_is_dir_fast (infob, &statfast);

	if (e2_fs_is_dir_fast (infoa, &statfast))
	{
		if (bf)
		  return (infoa->statbuf.st_mode - infob->statbuf.st_mode);
		else
		  return (order) ? -1 : 1;
	}

	if (bf)
		return (order) ? 1 : -1;
	
	return (infoa->statbuf.st_mode - infob->statbuf.st_mode);
}
/**
 * @brief item-owner sort-order comparison function
 *
 * Directories are placed before other things.
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_fileview_user_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	FileInfo *infoa, *infob;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	gtk_tree_model_get (model, a, FINFO, &infoa, -1);
	if (g_str_equal (infoa->filename, ".."))
		return (order) ? -1 : 1;

	gtk_tree_model_get (model, b, FINFO, &infob, -1);
	if (g_str_equal (infob->filename, ".."))
		return (order) ? 1 : -1;

	struct stat statfast;

	gboolean bf = e2_fs_is_dir_fast (infob, &statfast);
	
	if (e2_fs_is_dir_fast (infoa, &statfast))
	{
		if (bf)
		  return (infoa->statbuf.st_uid - infob->statbuf.st_uid);
		else
		  return (order) ? -1 : 1;
	}

	if (bf)
		return (order) ? 1 : -1;

	return (infoa->statbuf.st_uid - infob->statbuf.st_uid);
}
/**
 * @brief item-group sort-order comparison function
 *
 * Directories are placed before other things.
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_fileview_group_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	FileInfo *infoa, *infob;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	gtk_tree_model_get (model, a, FINFO, &infoa, -1);
	if (g_str_equal (infoa->filename, ".."))
		return (order) ? -1 : 1;

	gtk_tree_model_get (model, b, FINFO, &infob, -1);
	if (g_str_equal (infob->filename, ".."))
		return (order) ? 1 : -1;

	struct stat statfast;

	gboolean bf = e2_fs_is_dir_fast (infob, &statfast);
	
	if (e2_fs_is_dir_fast (infoa, &statfast))
	{
		if (bf)
		  return (infoa->statbuf.st_gid - infob->statbuf.st_gid);
		else
		  return (order) ? -1 : 1;
	}

	if (bf)
		return (order) ? 1 : -1;

	return (infoa->statbuf.st_gid - infob->statbuf.st_gid);
}
/**
 * @brief item-acess-date sort-order comparison function
 *
 * Directories are placed before other things.
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_fileview_adate_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	FileInfo *infoa, *infob;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	gtk_tree_model_get (model, a, FINFO, &infoa, -1);
	if (g_str_equal (infoa->filename, ".."))
		return (order) ? -1 : 1;

	gtk_tree_model_get (model, b, FINFO, &infob, -1);
	if (g_str_equal (infob->filename, ".."))
		return (order) ? 1 : -1;

	struct stat statfast;

	gboolean bf = e2_fs_is_dir_fast (infob, &statfast);
	
	if (e2_fs_is_dir_fast (infoa, &statfast))
	{
		if (bf)
		  return (infoa->statbuf.st_atime - infob->statbuf.st_atime);
		else
		  return (order) ? -1 : 1;
	}

	if (bf)
		return (order) ? 1 : -1;

	return (infoa->statbuf.st_atime - infob->statbuf.st_atime);
}
/**
 * @brief item-modification-date sort-order comparison function
 *
 * Directories are placed before other things.
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_fileview_mdate_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	FileInfo *infoa, *infob;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	gtk_tree_model_get (model, a, FINFO, &infoa, -1);
	if (g_str_equal (infoa->filename, ".."))
		return (order) ? -1 : 1;

	gtk_tree_model_get (model, b, FINFO, &infob, -1);
	if (g_str_equal (infob->filename, ".."))
		return (order) ? 1 : -1;

	struct stat statfast;

	gboolean bf = e2_fs_is_dir_fast (infob, &statfast);
	
	if (e2_fs_is_dir_fast (infoa, &statfast))
	{
		if (bf)
		  return (infoa->statbuf.st_mtime - infob->statbuf.st_mtime);
		else
		  return (order) ? -1 : 1;
	}

	if (bf)
		return (order) ? 1 : -1;

	return (infoa->statbuf.st_mtime - infob->statbuf.st_mtime);
}

/**
 * @brief item-status-change-date sort-order comparison function
 *
 * Directories are placed before other things.
 *
 * @param model the data model to be interrogated
 * @param a pointer to model iter for the first row to be compared
 * @param b pointer to model iter for the second row to be compared
 * @param direction pointer to the views's sort direction
 *
 * @return  integer <0, 0 or >0 according to whether a belongs before, = or after b
 */
static gint _e2_fileview_cdate_sort
	(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, GtkSortType *direction)
{
	FileInfo *infoa, *infob;
	gboolean order = (*direction == GTK_SORT_ASCENDING);  //TRUE for ascending, FALSE for descending
	gtk_tree_model_get (model, a, FINFO, &infoa, -1);
	if (g_str_equal (infoa->filename, ".."))
		return (order) ? -1 : 1;

	gtk_tree_model_get (model, b, FINFO, &infob, -1);
	if (g_str_equal (infob->filename, ".."))
		return (order) ? 1 : -1;

	struct stat statfast;

	gboolean bf = e2_fs_is_dir_fast (infob, &statfast);
	
	if (e2_fs_is_dir_fast (infoa, &statfast))
	{
		if (bf)
		  return (infoa->statbuf.st_ctime - infob->statbuf.st_ctime);
		else
		  return (order) ? -1 : 1;
	}

	if (bf)
		return (order) ? 1 : -1;

	return (infoa->statbuf.st_ctime - infob->statbuf.st_ctime);
}

// Callbacks

/**
 * @brief cursor change detection
 *
 * This records the row to which the cursor is moved
 *
 * @param treeview the widget where the cursor moved
 * @param view data structure for the pane
 *
 * @return FALSE, so related signals are not blocked
 */
static gboolean _e2_fileview_cursor_change_cb
	(GtkTreeView *treeview, FileView *view)
{
	printd (DEBUG, "callback: cursor changed");
	GtkTreePath *path;
	//can this be called during a filelist refresh ?
	if (gtk_tree_selection_count_selected_rows (view->selection) == 0)
	{	//there is deliberately no selection, or
		//the selection is gone (after deletion), but we want
		//to stay in the same vicinity
		if (view->row > 0)
		{
			path = gtk_tree_path_new_from_indices (view->row, -1);
			while (--view->row > 0 && !gtk_tree_path_prev (path)) {}
			gtk_tree_view_set_cursor (treeview, path, NULL, FALSE);
			gtk_tree_selection_unselect_path (view->selection, path);
			gtk_tree_path_free (path);
		}
	}
	else
	{
		gtk_tree_view_get_cursor (treeview, &path, NULL);
//		if (path != NULL) 	path can't be invalid for a cursor move, even if item at prior path is now deleted
//		{
			view->row = *gtk_tree_path_get_indices (path);
			gtk_tree_path_free (path);
//			printd (DEBUG, "focus row is %d", view->row);
//		}
	}
	return FALSE;
}
/**
 * @brief column-header button-press detection
 *
 * This fn records which button was clicked, so that when a "release"
 * callback occurs (which may be when a drag starts, not an actual
 * release) we can check whether it is still pressed
 * This is a workaround for gtk's (2.4 at least) behaviour
 * It relies on detecting clicks on the undocumented column->button widget
 *
 * @param header UNUSED the col button header widget which was clicked
 * @param event pointer to event data struct
 * @param view UNUSED data structure for the view to which the file list belongs
 *
 * @return FALSE, so related signals are not blocked
 */
static gboolean _e2_fileview_column_header_buttonpress_cb
	(GtkWidget *header, GdkEventButton *event, FileView *view)
{
	printd (DEBUG, "callback: column header button press");
	header_button = event->button;
	return FALSE;
}
/**
 * @brief column-header button-release detection
 *
 * This checks whether a release event is real. In gtk 2.4 
 * at least, such an event may be when a column-drag starts,
 * not an actual release.
 * It relies on detecting clicks on the undocumented 
 * column->button widget
 *
 * @param UNUSED header the col button widget which was clicked
 * @param event pointed to event data struct
 * @param view data structure for the view to which the file list belongs
 *
 * @return TRUE, if this is a 'bogus' release, else FALSE
 */
static gboolean _e2_fileview_column_header_buttonrel_cb
	(GtkWidget *header, GdkEventButton *event, FileView *view)
{
	printd (DEBUG, "callback: column header button release");

	if (view != curr_view)
		e2_pane_activate_other ();

	//check for whether button is really still pressed
	//gtk 2.4 emits a bogus extra release signal when column-drag occurs
	gboolean pressed = FALSE;
	if (!block && event->send_event)
	{
		pressed = TRUE;
		block = TRUE;
	}
	else if (block)
		block = FALSE;  //on the second pass, just unblock
//	else
//		pressed = FALSE;

	if (!pressed)
		header_button = -1;  //signal to "clicked" callback not to do anything
	return pressed;	//stop further action on this button if not released yet
}
/**
 * @brief column-click callback
 *
 * This fn updates the view sorting, by applying an ascending sort
 * to a column that is not sorted, or by toggling the sort direction 
 * of a column that is already sorted.
 * Sort indicators are updated.
 * This is called when a column header is clicked, before any
 * associated "column-change" signal.
 *
 * @param col the view column which was clicked
 * @param rt data structure for the pane to which the file list belongs
 *
 * @return
 */
static void _e2_fileview_column_header_clicked_cb 
	(GtkTreeViewColumn *col, FileView *view)
{
	if (header_button != -1)
		return;		//the button is still down, this is drag action
	
	printd (DEBUG, "callback: column header click");

	//get which column was cliicked
	GList *cols = gtk_tree_view_get_columns (GTK_TREE_VIEW (view->treeview));
	gint clicked_col = g_list_index (cols, (gpointer) col);
	g_list_free (cols);

	//get colum-order array for the pane
	gint *order_array = (view == app.pane1.view) ? displayed_col_order[0] : displayed_col_order[1];
	//get clicked-model-column number
	gint i = order_array[clicked_col];
/*	if (i == 0)
	{ //filename column  DO WE WANT TO MAKE THIS DYNAMIC BUT SLOWER ??
		//get a local copy of the case-sensitive flag
		E2_OptionSet *set = e2_option_get_simple ("namesort-case-sensitive");
		case_sensitive = _e2_option_bool_get (set);
	} */
	//get current-sorted-column details
	gint m;
	GtkSortType old_order;
	GtkTreeSortable *sortable = GTK_TREE_SORTABLE (view->store);
	gtk_tree_sortable_get_sort_column_id (sortable, &m, &old_order);
	
	if (m == FILENAME)
	//revert any sort-function change by the extension-sort plugin
		gtk_tree_sortable_set_sort_func (sortable, FILENAME, 
		e2_all_columns[FILENAME].sort_func, &view->sort_order, NULL);

	GtkSortType sort_order;
	if (m == i)
	{	//column is already sorted, toggle it
		view->sort_order = sort_order = (old_order == GTK_SORT_ASCENDING) ?
			GTK_SORT_DESCENDING : GTK_SORT_ASCENDING;
		//set flag for row-comparison fn
//		order = (sort_order == GTK_SORT_ASCENDING);
	}
	else
	{  //this is a new sort column
		//replace sort indicator
		gtk_widget_hide (view->sort_arrows[m]);
		gtk_widget_show (view->sort_arrows[i]);
		//1st click makes it ascending
		view->sort_order = sort_order = GTK_SORT_ASCENDING;
//		order = TRUE;
	}
	//set appropriate arrow type
	gtk_arrow_set (GTK_ARROW (view->sort_arrows[i]),
		(sort_order == GTK_SORT_ASCENDING) ? GTK_ARROW_DOWN : GTK_ARROW_UP,
			GTK_SHADOW_IN);
/* no use, gtk always returns gchar data to heap space
	//setup stack variable space for speedier name sorts
	if (i == FILENAME)
	{
		gchar namespace [2][E2_UTFNAME_MAX];
		gtk_tree_sortable_set_sort_func (sortable, FILENAME,
			e2_all_columns[FILENAME].sort_func,
            (gpointer *) namespace, NULL);
	} */
	//remember the column, for cacheing
	view->sort_column = i;
	//do the sort
	gtk_tree_sortable_set_sort_column_id (sortable, i, sort_order);

	gtk_widget_grab_focus (view->treeview);
	return;
}
/**
 * @brief column-change callback
 *
 * This fn refreshes the columns-order array for the pane, by
 * interating over pane columns glist, grabbing each title
 * like "4", converting it to corresponding integer.
 * It is called when a column is dropped to a different position
 * and when the window is closed
 * (also, multiple times when the window is re-created - CHECKME why)
 *
 * @param treeview the widget where the action occurred
 * @param view data structure for the view
 *
 * @return
 */
static void _e2_fileview_col_change_cb (GtkTreeView *treeview,
	FileView *view)
{
	printd (DEBUG, "callback: column change");
	gint *order_array = (view == &app.pane1_view) ?
		displayed_col_order[0] : displayed_col_order[1];
	GList *cols = gtk_tree_view_get_columns (treeview);
	GList *tmp;
	const gchar *title;
	gint colnum = 0;
	for (tmp = cols; tmp != NULL; tmp = tmp->next)
	{
		title = gtk_tree_view_column_get_title (tmp->data);
		order_array[colnum++] = atoi (title);
	}
	g_list_free (cols);

	//update order of "slave" pane if appropriate
	gint *slave_array;
	FileView *slave_view;
	if (e2_option_bool_get ("pane2-uses-other")
		&& view == &app.pane1_view)
	{
		slave_array = displayed_col_order[1];
		slave_view = &app.pane2_view;
	}
	else
	if (e2_option_bool_get ("pane1-uses-other")
		&& view == &app.pane2_view)
	{
		slave_array = displayed_col_order[0];
		slave_view = &app.pane1_view;
	}
	else
		return;
	
	printd (DEBUG, "slave-view column change");
	g_signal_handlers_block_by_func (G_OBJECT (slave_view->treeview),
		_e2_fileview_col_change_cb, slave_view);
	cols = gtk_tree_view_get_columns (GTK_TREE_VIEW (slave_view->treeview));
	for (colnum = 0; colnum < MAX_COLUMNS; colnum++)
	{
		gint current_col = order_array[colnum];
		if (slave_array[colnum] != current_col)
		{
			gint j;
			GtkTreeViewColumn *prior_col, *moved_col;
			if (colnum == 0)
				prior_col = NULL;
			else
				prior_col = g_list_nth_data (cols, colnum-1);
			for (j = colnum+1; j < MAX_COLUMNS; j++)
			{
				if (slave_array[j] == current_col)
				{
					moved_col = g_list_nth_data (cols, j);
					gtk_tree_view_move_column_after (
						GTK_TREE_VIEW (slave_view->treeview),
							moved_col, prior_col);
					//this approach seems too complex, but several simpler tries failed !!
					g_list_free (cols);
					cols = gtk_tree_view_get_columns (GTK_TREE_VIEW (slave_view->treeview));
					j = colnum;
					for (tmp = g_list_nth (cols, j); tmp != NULL; tmp=tmp->next)
					{
						const gchar *title = gtk_tree_view_column_get_title (tmp->data);
						slave_array[j] = atoi(title);
						j++;
					}
					break;
				}
			}
		}
	}
	g_list_free (cols);
	g_signal_handlers_unblock_by_func (G_OBJECT (slave_view->treeview),
		_e2_fileview_col_change_cb, slave_view);
}
/**
 * @brief mouse button double-click callback
 *
 * This "executes" the clicked row item
 * view->row will have been set by the button-release callback
 *
 * @param treeview the widget where the button was pressed
 * @param path model path to the clicked row
 * @param col UNUSED clicked view column
 * @param view UNUSED rt data for the view to be worked on
 *
 * @return
**/
static void _e2_fileview_row_activated_cb (
		GtkTreeView        *treeview,
		GtkTreePath        *path,
		GtkTreeViewColumn  *col,
		FileView              *view)
{
	printd (DEBUG, "callback: double-click");
	GtkTreeIter iter;
	GtkTreeModel *model = gtk_tree_view_get_model (treeview);
    if (gtk_tree_model_get_iter (model, &iter, path))
	{
		FileInfo *info;
		gtk_tree_model_get (model, &iter, FINFO, &info, -1);
		e2_task_backend_open (info);
	}
}
/**
 * @brief mouse button press callback
 *
 * This makes the clicked view active, if it wasn't already.
 * A button-1 press when there is nothing else selected logs
 * the clicked model row, if any, in case it's a drag start. 
 * (For anything other than a drag, the row is logged whem the
 * cursor moves.) The gtk handler will select the row.
 * Double-clicks are detected (based on cb sequence) here, to
 * work around gtk's behaviour when its handler is aborted
 * after a single click
 * A button-2 press does an updir, if that option is in force,
 * or else logs the clicked model row, if any.
 * A button-3 press sets up the relevant context menu, and may
 * select the line or unselect everything, depending on options in force
 *
 * @param treeview the widget where the button was pressed
 * @param event gdk event data
 * @param view rt data for the view to be worked on
 *
 * @return TRUE (stop other handlers) for btn 2 (updir) or btn 3 press, else FALSE (allow other handlers)
**/
static gboolean _e2_fileview_button_press_cb (GtkWidget *treeview, 
	GdkEventButton *event, FileView *view)
{
	printd (DEBUG, "callback: mouse button press");
	//check that we're in the actual treeview (not its header) by making
	//sure that the window is correct
	GdkWindow *window = gtk_tree_view_get_bin_window (GTK_TREE_VIEW (treeview));
	if (window != event->window)
		return FALSE;
	//make the clicked pane active, if it wasn't already
	if (view != curr_view)  // && !panechange_started)
		e2_pane_activate_other ();

	GtkTreePath *path;
	GtkTreeSelection *sel;
#ifdef E2_ALTLEFTMOUSE
	gboolean oldleft = left_pressed;
	//set flag to assist processing any drag
	left_pressed = (event->button == 1); //&& event->type == GDK_BUTTON_PRESS);
	if (left_pressed
	//we don't handle any 'modified' button-press (so that range-selection works as normal)
	//or a click within the specified double-click interval from the prior click
		&& !(event->state & (GDK_SHIFT_MASK | GDK_CONTROL_MASK | GDK_MOD1_MASK)) 
		)
	{
		//event-time for detecting double-clicks
/* because we block returns for selected items, double-clicks have to be detected
	and processed here, too
	instead of click-interval, test now relies on the normal cb sequence for doubles
	(pr, rl, pr, pr, rl) i.e. a press without an intervening release is a double (or triple!)
		static guint32 last_event_time = 0;
		guint32 interval = event->time - last_event_time;
		last_event_time = event->time;
		if (interval >= click_interval)	//time-based check for double-click (the 3rd pr will pass this test too)
		{ */
			/*if we've passed all the tests, and a selected item is clicked,
			we block gtk's deselection of other items until the button is
			released. Then, we'll know how to treat it*/
			GtkTreePath *path;
			if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (treeview),
					event->x, event->y, &path, NULL, NULL, NULL))
			{
//				sel = view->selection;
				if (gtk_tree_selection_path_is_selected (view->selection, path))
				{
					//sometimes the focus doesn't get transferred properly by gtk, so ...
					gtk_widget_grab_focus (view->treeview);

					if (oldleft)	//this is a fake press, generated as part of a double-click
						//process it as a double-click
						_e2_fileview_row_activated_cb (GTK_TREE_VIEW (treeview), path, NULL, view);

					gtk_tree_path_free (path);
					//note that returning TRUE blocks some normal treeview behaviour
					//like refocussing a selected item and repeated double-clicks
					return TRUE;	//wait until release, to see what to do with the selected row
				}
//				else
					//this is a new selection, log the clicked row, and allow it to be selected as normal
//					view->row = *gtk_tree_path_get_indices (path); NOW IN CURSOR MOVE CB
				gtk_tree_path_free (path);
			}
//		}
	}
	else
#endif
/*		if (event->button == 1 &&
			(event->type == GDK_2BUTTON_PRESS || event->type == GDK_3BUTTON_PRESS))
			)
		{	//double or triple left-click
		}
	else if (event->button == 2
			return TRUE;	//wait until release, to see whether to select the row
	else */
		if (event->button == 3)
	{
//		printd (DEBUG, "callback: mouse button-3 press");
		if (e2_option_bool_get ("windows-right-click")
//			&& (view->tagged == NULL)
		)
		{
			//if windows right click is enabled, we want to also select the item,
			//if not already done, and clear any other selected item(s) unless this
			//one is selected already
			sel = view->selection;
			if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (treeview),
				event->x, event->y, &path, NULL, NULL, NULL))
			{
				if (!gtk_tree_selection_path_is_selected (sel, path))
					gtk_tree_selection_unselect_all (sel);
				gtk_tree_selection_select_path (sel, path);
				//do this now because menu created before cursor-move cb
				view->row = *gtk_tree_path_get_indices (path);
				gtk_tree_path_free (path);
			}
			else if (! e2_option_bool_get ("windows-right-click-extra"))
				gtk_tree_selection_unselect_all (sel);

		}
		//context menu
		gint type = 0;
		guint modifiers = gtk_accelerator_get_default_mod_mask ();
		if ((event->state & modifiers) == GDK_SHIFT_MASK) type = 1;
		else if ((event->state & modifiers) == GDK_CONTROL_MASK) type = 2;
		e2_context_menu_show (event->button, event->time, type);
		return TRUE;
	}
	return FALSE;
}
/**
 * @brief  mouse button release callback
 *
 * This checks if the click was a left click. If so, and it's not the
 * end of a drag, or an 'un-select' click, the clicked row is logged.
 * Detection of the end of a drag relies on an undocumented 
 * check whether the release happened in the same pane as
 * was clicked.
 *
 * @param treeview UNUSED where the button was _pressed_
 * @param event
 * @param view rt data for the view where the button was _pressed_
 * 
 * @return TRUE (stop other handlers) if updir initiated, else FALSE (allow other handlers)
**/
static gboolean _e2_fileview_button_release_cb (GObject *treeview, 
	GdkEventButton *event, FileView *view)
{
	printd (DEBUG, "callback: button release");
	//check that we're in the actual treeview (not its header) by making
	//sure that the window is correct
	GdkWindow *window = gtk_tree_view_get_bin_window (GTK_TREE_VIEW (treeview));
	if (window != event->window)
		return FALSE;
	//for releases, there is no distinction between 1- 2- or 3- buttons
#ifdef E2_ALTLEFTMOUSE
	if (event->button == 1)
	{
		left_pressed = FALSE;
		//we ignore releases when a modifier is pressed,
		//and normal-drag ends,
		//  i.e. event->send_event seems to be FALSE (ie the event was not sent explicitly)  
		//and drag-select ends
		if (!(event->state & (GDK_SHIFT_MASK | GDK_CONTROL_MASK | GDK_MOD1_MASK))
			&& event->send_event == 0
			&& !drag_sel_now)
		{	/*there was no drag-selection, or the drag did not cover
			anything but the clicked item
			we want an outcome same as from a normal button-press
			i.e. select just the clicked row and log it, if any */
			GtkTreePath *path;
			if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (view->treeview),
				event->x, event->y, &path, NULL, NULL, NULL))
			{	//we're releasing over a treeview-item (which must have been selected)
				GtkTreeSelection *sel = view->selection;
				if (gtk_tree_selection_count_selected_rows (sel) > 1)
				{
					gtk_tree_selection_unselect_all (sel);
					gtk_tree_selection_select_path (sel, path);
					//record the row as the last-selected one
//					view->row = *gtk_tree_path_get_indices (path);NOW IN CURSOR MOVE CB
				}
				gtk_tree_path_free (path);
			}
		}
	}
	else
#endif
		if (event->button == 2)
	{
//		printd (DEBUG, "callback: mouse button-2 release");
		//correctly distinguish drag releases, and normal clicks
		//event->send_event seems to be FALSE (ie the event was not sent explicitly)  
		//for releases in the same pane as clicked??
		if (event->send_event == 0) 
		{	//not a drag release
			if (button2updir)
			{
				printd (DEBUG, "up");
				e2_pane_change_dir (NULL, "..");
				return TRUE;
			}
			//otherwise, maybe log the row
			//don't bother with modifers-check in this context
//			guint modifiers = gtk_accelerator_get_default_mod_mask ();
//			if ((event->state & modifiers) != GDK_CONTROL_MASK)
			if (!(event->state & GDK_CONTROL_MASK))
			{	//not an "unselect" click
				GtkTreePath *path;
				if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (view->treeview),
					event->x, event->y, &path, NULL, NULL, NULL))
				{
					// row was NOT selected when the mouse was clicked
//					GtkTreeSelection *sel = view->selection;
					gtk_tree_selection_select_path (view->selection, path);
					//record the row as the last-selected one
//					view->row = *gtk_tree_path_get_indices (path);NOW IN CURSOR MOVE CB
					gtk_tree_path_free (path);
				}
			}
		}
		else
			//set signal for DnD handler
			btn2_released = TRUE;
	}
	else
		btn2_released = FALSE;

	return FALSE;
}
/**
 * @brief set popup menu position
 *
 * This function is supplied when calling gtk_menu_popup(), to position
 * the displayed menu.
 * set @a push_in to TRUE for menu completely inside the screen,
 * FALSE for menu clamped to screen size
 *
 * @param menu UNUSED the GtkMenu to be positioned
 * @param x	place to store gint representing the menu left
 * @param y  place to store gint representing the menu top
 * @param push_in place to store pushin flag
 * @param treeview in focus when the menu key was pressed
 *
 * @return
 */
void e2_fileview_set_menu_position (GtkMenu *menu,
	gint *x, gint *y, gboolean *push_in, GtkWidget *treeview)
{
	gint left, top;
	gtk_window_get_position (GTK_WINDOW (app.main_window),
		&left, &top);
	GtkAllocation alloc = treeview->allocation;
	*x = left + alloc.x + alloc.width/2;
	*y = top + alloc.y +alloc.height/2 - 30;
	*push_in = TRUE;
}
/**
 * @brief menu button callback (does not apply for 'modified' menu button)
 *
 * Sets up the relevant context menu
 *
 * @param treeview UNUSED the widget in focus when the action was initiated
 * @param view UNUSED rt data for the view to be worked on
 *
 * @return TRUE always
**/
/*static gboolean _e2_fileview_popup_cb (GtkWidget *treeview, 
	FileView *view)
{
	gint menu_type = 0;	//default to 'normal' context menu
	//FIXME this check irrelevant, as cb never used for menu btn + Ctrl/Shift
/ *	GdkModifierType state;
	if (gtk_get_current_event_state (&state))
	{
		if (state & GDK_SHIFT_MASK) menu_type = 1;
		else
			if (state & GDK_CONTROL_MASK) menu_type = 2;
	} * /
	gint event_time = gtk_get_current_event_time ();
	//button '0' signals this is not a mouse-button click
	e2_context_menu_show (0, event_time, menu_type);
	return TRUE;
} */
/**
 * @brief treeview key-press callback
 *
 * @param widget UNUSED the focused treeview widget when the key was pressed
 * @param event pointer to event data struct
 * @param view rt data for the view to be worked on
 * 
 * @return TRUE when we handle a 'vi key'
**/
static gboolean _e2_fileview_key_press_cb (GtkWidget *widget, 
	GdkEventKey *event, FileView *view)
{
	guint modifiers = gtk_accelerator_get_default_mod_mask ();
	modifiers = event->state & modifiers;
	if (event->keyval == GDK_Menu)
	{
		gint menu_type = 0;	//default to 'normal' context menu
		if (modifiers == GDK_SHIFT_MASK) menu_type = 1;
		else if (modifiers == GDK_CONTROL_MASK) menu_type = 2;
		//button '0' signals this is not a mouse-button click
		e2_context_menu_show (0, event->time, menu_type);
		return TRUE;
	}
/*		{
  no need for this: the treeview handles it itself
			//select row beginning with letter entered, if we can
			gint row = e2_fileview_find_filename_begining_with (event->keyval);
			if (row > -1)
				e2_fileview_focus_row (view, row, TRUE, TRUE, TRUE, TRUE);
			return TRUE;
		} */
	return FALSE;
}

// some filter menu callbacks (others are in filter-dialog files)

void e2_fileview_filter_dirs_cb (GtkWidget *widget, FileView *view)
{  //toggle cached flag to match the menu check button
//	view->filter_directories = gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (widget));
	view->filter_directories = !view->filter_directories;
#ifdef E2_NEWFILTER
	e2_fileview_refilter_list (view);
#else
	e2_fileview_refresh_list (view);
#endif
}

void e2_fileview_remove_filters_cb (GtkWidget *widget, FileView *view)
{
	// turn off local flags
	view->name_filter.active = FALSE;
	view->size_filter.active = FALSE;
	view->date_filter.active = FALSE;
	view->filter_directories = FALSE;
	//show the results
	e2_toolbar_toggle_filter_button (view);
#ifdef E2_NEWFILTER
	e2_fileview_refilter_list (view);
#else
	e2_fileview_refresh_list (view);
#endif
}
/**
 * @brief initialise file filter flags to show everything
 *
 * @param view data structure for view being processed
 *
 * @return
 */
/*void e2_fileview_initialize_filters (FileView *view)
{
	view->filter_directories = FALSE;

	view->name_filter.active = FALSE;
	g_strlcpy (view->name_filter.pattern, "*",
		      sizeof(view->name_filter.pattern));
	view->name_filter.case_sensitive = TRUE;

	view->size_filter.active = FALSE;
	view->size_filter.size = 0;
	view->size_filter.op = GT;

	view->date_filter.active = FALSE;
	view->date_filter.time = time(NULL);
	view->date_filter.time_type = ATIME;
	view->date_filter.op = GT;
} */

#ifdef E2_NEWFILTER
static void _e2_fileview_clear_filter_patterns (FileView *view)
{
	E2_SelectPattern *patterninfo;
	GSList *member;
	if (view->name_filter.compiled_patterns == NULL)
		return;
	for (member = view->name_filter.compiled_patterns; member != NULL; member = member->next)
	{
		patterninfo = (E2_SelectPattern *)member->data;
		g_pattern_spec_free (patterninfo->pspec);
		g_free (patterninfo);
	}
	g_slist_free (view->name_filter.compiled_patterns);
	view->name_filter.compiled_patterns = NULL;
}

gboolean _e2_fileview_compile_filter_patterns (FileView *view)
{
	if (view->name_filter.compiled_patterns != NULL)
		_e2_fileview_clear_filter_patterns (view);
	
	E2_SelectPattern *patterninfo;
	GSList *members = NULL;
	gchar *s, *p, *freeme;
	gchar save;

	p = view->name_filter.patternptr;
	//maybe several patterns, separated by commas
	while ((s = strchr (p, ',')) != NULL)	//if always ascii ',', don't need g_utf8_strchr()
	{	//check each pattern that is followed by a ','
		patterninfo = (E2_SelectPattern *) g_try_malloc (sizeof (E2_SelectPattern));
		if (patterninfo == NULL)
		{
			g_slist_free (members); //improper cleanup, too bad
			return FALSE;
		}
		members = g_slist_append (members, patterninfo);
		save = *s;
		*s = '\0';
		while (p[0] == ' ')
			p++;
		if (p[0] == '!')
		{
			patterninfo->negated = !view->name_filter.invert_mask;
			p++;
		}
		else
		{
			patterninfo->negated = view->name_filter.invert_mask;
			if (p[0] == '\\' && p[1] == '!')
				p++;
		}
		if (!view->name_filter.case_sensitive)
		{
			freeme = g_utf8_strdown (p, -1);
			patterninfo->pspec = g_pattern_spec_new (freeme);
			g_free (freeme);
		}
		else
			patterninfo->pspec = g_pattern_spec_new (p);

		*s = save;
		NCHR(s);
		p = s;
	}
	//now the last (or only) pattern
	while (p[0] == ' ')
		p++;
	if (p[0] == '\0')
	{
		g_slist_free (members);
		return FALSE;
	}
	patterninfo = (E2_SelectPattern *) g_try_malloc (sizeof (E2_SelectPattern));
	if (patterninfo == NULL)
	{
		g_slist_free (members); //improper cleanup, too bad
		return FALSE;
	}
	members = g_slist_append (members, patterninfo);

	if (p[0] == '!')
	{
		patterninfo->negated = !view->name_filter.invert_mask;
		p++;
	}
	else
	{
		patterninfo->negated = view->name_filter.invert_mask;
		if (p[0] == '\\' && p[1] == '!')
			p++;
	}
	if (!view->name_filter.case_sensitive)
	{
		freeme = g_utf8_strdown (p, -1);
		patterninfo->pspec = g_pattern_spec_new (freeme);
		g_free (freeme);
	}
	else
		patterninfo->pspec = g_pattern_spec_new (p);

	view->name_filter.compiled_patterns = members;
	return TRUE;
}
static gchar *name_pattern = NULL;
/**
 * @brief treeview row visibility function
 *
 * @param model the child model of the GtkTreeModelFilter
 * @param iter pointer to a GtkTreeIter for the row in @a model whose visibility is to be determined
 * @param view rt data for the view, assigned when function was initiated
 *
 * @return TRUE if the row is visible
**/
static gboolean _e2_fileview_filter_check (GtkTreeModel *model,
	GtkTreeIter *iter, FileView *view)
{
	FileInfo *info;
	gtk_tree_model_get (model, iter, FINFO, &info, -1);
	gchar *name = info->filename;
	if (name[0] == '.'
		&& (
			//name[1] == '\0' || filtered out when dir read
			   (!view->show_hidden && name[1] != '.' && name[2] != '\0')
			|| (name[1] == '.' && name[2] == '\0' &&
				(!e2_option_bool_get ("show-updir-entry")
				|| (view->dir[0] == G_DIR_SEPARATOR && view->dir[1] == '\0') //at root directory now?
				)
			   )
			)
		)
			return FALSE;

	if (!view->filter_directories && e2_fs_is_dir (info))
		return TRUE;

	if (view->name_filter.active)
	{
#if 0
		gchar *s, *p, *utf, *freeme;
		gchar save;
		gboolean negated, matched, positive_check = FALSE, result = FALSE;

		p = name_pattern;
		utf = F_FILENAME_FROM_LOCALE (info->filename);
		//maybe several patterns, separated by commas
		while ((s = strchr (p, ',')) != NULL)	//if always ascii ',', don't need g_utf8_strchr()
		{	//check each pattern that is followed by a ','
			save = *s;
			*s = '\0';
			while (p[0] == ' ')
				p++;
			if (p[0] == '!')
			{
				negated = !view->name_filter.invert_mask;
				p++;
			}
			else
			{
				negated = view->name_filter.invert_mask;
				if (p[0] == '\\' && p[1] == '!')
					p++;
			}
			if (!positive_check)
				positive_check = !negated;

			if (!view->name_filter.case_sensitive)
			{
			  freeme = g_utf8_strdown (utf, -1);
			  matched = g_pattern_match_simple (p, freeme);
			  g_free (freeme);
			}
			else
			  matched = g_pattern_match_simple (p, utf);

			*s = save;
			if (matched && negated)
			{
			  F_FREE (utf);
			  return FALSE;
			}
			if (matched && !negated)
			  result = TRUE;	//but keep looking for any later exclude
			//if neither negated nor matched, we don't change result

			NCHR(s);	//in case the matched char is bigger than 1 byte
			p = s;
		}
		//check the last (or only) pattern
		while (p[0] == ' ')
			p++;
		if (p[0] != '\0')
		{
			if (p[0] == '!')
			{
				negated = !view->name_filter.invert_mask;
				p++;
			}
			else
			{
				negated = view->name_filter.invert_mask;
				if (p[0] == '\\' && p[1] == '!')
					p++;
			}
			if (!positive_check)
				positive_check = !negated;

			if (!view->name_filter.case_sensitive)
			{
				freeme = g_utf8_strdown (utf, -1);
				matched = g_pattern_match_simple (p, freeme);
				g_free (freeme);
			}
			else
				matched = g_pattern_match_simple (p, utf);
			if (matched)
				result = !negated;
			//extra check for unmatched final check
			else if (negated && !positive_check)
				result = TRUE;
		}
		F_FREE (utf);
		if (!result)
			return FALSE;
#else
		GSList *member;
		E2_SelectPattern *patterninfo;
		gboolean negated, matched, positive_check = FALSE, result = FALSE;
		gchar *freeme, *utf = F_FILENAME_FROM_LOCALE (info->filename);

		for (member = view->name_filter.compiled_patterns; member != NULL;
			member= member->next)
		{
			patterninfo = (E2_SelectPattern *)member->data;
			negated = patterninfo->negated;
			if (!positive_check)
				positive_check = !negated;

			if (!view->name_filter.case_sensitive)
			{
			  freeme = g_utf8_strdown (utf, -1);
			  matched = g_pattern_match_string (patterninfo->pspec, freeme);
			  g_free (freeme);
			}
			else
			  matched = g_pattern_match_string (patterninfo->pspec, utf);

			if (matched && negated)
			{
			  F_FREE (utf);
			  return FALSE;
			}
			if (matched && !negated)
			  result = TRUE;	//but keep looking for any later exclude
			//if neither negated nor matched, we don't change result
		}
		F_FREE (utf);
		//extra check for unmatched final check
		if (!matched && negated && !positive_check)
			result = TRUE;
		if (!result)
			return FALSE;
#endif //def 0
	}

	if (view->size_filter.active)
	{
		switch (view->size_filter.op)
		{
			case LT:
			  if (info->statbuf.st_size >= view->size_filter.size)
				  return FALSE;
			  break;
			case EQ:
			  if (info->statbuf.st_size != view->size_filter.size)
				  return FALSE;
			  break;
			default:
//			case GT:
			  if (info->statbuf.st_size <= view->size_filter.size)
				  return FALSE;
			  break;
		}
	}

	if (view->date_filter.active)
	{
		switch (view->date_filter.time_type)
		{
			case ATIME:
			  if (view->date_filter.op == GT)
			  {
				if (difftime (view->date_filter.time, info->statbuf.st_atime) > 0)
					return FALSE;
			  }
			  else
				if (difftime (view->date_filter.time, info->statbuf.st_atime) <= 0)
					return FALSE;
			  break;
			case CTIME:
			  if (view->date_filter.op == GT)
			  {
				if (difftime (view->date_filter.time, info->statbuf.st_ctime) > 0)
					return FALSE;
			  }
			  else
				if (difftime (view->date_filter.time, info->statbuf.st_ctime) <= 0)
					return FALSE;
			  break;
			default:
//			case MTIME:
			  if (view->date_filter.op == GT)
			  {
				if (difftime (view->date_filter.time, info->statbuf.st_mtime) > 0)
					return FALSE;
			  }
			  else
				if (difftime (view->date_filter.time, info->statbuf.st_mtime) <= 0)
					return FALSE;
			  break;
		}
	}
	return TRUE;
}
/**
 * @brief treeview row visibility function for unfiltered view
 *
 * @param model the child model of the GtkTreeModelFilter
 * @param iter pointer to a GtkTreeIter for the row in @a model whose visibility is to be determined
 * @param data data assigned when function was initiated
 *
 * @return TRUE always //unless the row is "."
**/
static gboolean _e2_fileview_filter_none (GtkTreeModel *model,
	GtkTreeIter *iter, gpointer data)
{
	/* filtered out when dir read
	FileInfo *info;
	gtk_tree_model_get (model, iter, FINFO, &info, -1);
	return (info->filename[0] != '.' || info->filename[1] != '\0'); */
	return TRUE;
}
/**
 * @brief update filters for treeview
 * This changes model, and attaches view to new model, as appropriate
 * Assumes view attached to a model (store ?)
 * When new store is created, view->filtered_before will be false
 * @param view rt data for the view
 *
 * @return
**/
void e2_fileview_refilter_list (FileView *view)
{
	gboolean filtered_now = (
		 view->name_filter.active ||
		 view->size_filter.active ||
		 view->date_filter.active ||
		!view->show_hidden ||
		!e2_option_bool_get ("show-updir-entry") );

	if (filtered_now)
	{
		//setup for filter func
		if (view->name_filter.active)
		{
			_e2_fileview_compile_filter_patterns (view);
			if (!view->name_filter.case_sensitive)
				name_pattern = g_utf8_strdown (view->name_filter.patternptr, -1);
			else
				name_pattern = g_strdup (view->name_filter.patternptr);
		}

		if (!view->filtered_before)
		{
			//gtk won't accept a changed visibility function
			g_object_ref (G_OBJECT (view->store)); //preserve the store
			g_object_unref (G_OBJECT (view->model)); //kill the current model
			view->model = gtk_tree_model_filter_new (GTK_TREE_MODEL (view->store), NULL);
			g_object_unref (G_OBJECT (view->store)); //kill the ref just added
			gtk_tree_model_filter_set_visible_func (GTK_TREE_MODEL_FILTER (view->model),
				(GtkTreeModelFilterVisibleFunc) _e2_fileview_filter_check, view, NULL);
		}

		gtk_tree_view_set_model (GTK_TREE_VIEW (view->treeview), view->model);
		gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER (view->model));

		if (view->name_filter.active)
		{
			_e2_fileview_clear_filter_patterns (view);
			g_free (name_pattern);
		}
	}
	else if (view->filtered_before)
	{	//de-filter
		//gtk won't accept a changed, or NULL, visibility function
		{
			g_object_ref (G_OBJECT (view->store)); //preserve the store
			g_object_unref (G_OBJECT (view->model)); //kill the current model
			view->model = gtk_tree_model_filter_new (GTK_TREE_MODEL (view->store), NULL);
			g_object_unref (G_OBJECT (view->store)); //kill the ref just added
			gtk_tree_model_filter_set_visible_func (GTK_TREE_MODEL_FILTER (view->model),
				(GtkTreeModelFilterVisibleFunc) _e2_fileview_filter_none, NULL, NULL);
		}
		gtk_tree_view_set_model (GTK_TREE_VIEW (view->treeview), view->model);
		gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER (view->model));
	}
	view->filtered_before = filtered_now;
}
#endif //ndef E2_NEWFILTER

/** @brief decide whether type-ahead search matches a row
 *
 * @param model the treemodel being searched
 * @param column UNUSED the search column (= FILENAME)
 * @param key the aggregate string of keys sofar pressed
 * @param iter an iter pointing the row of @a model that should be compared with @a key
 * @param view pointer to data struct for the view
 * 
 * @return FALSE when a match is found
**/
static gboolean _e2_fileview_match_filename (GtkTreeModel *model, gint column,
	const gchar *key, GtkTreeIter *iter, FileView *view)
{
	gchar *itemname;
	gtk_tree_model_get (model, iter, FILENAME, &itemname, -1);
	gboolean match = g_str_has_prefix (itemname, key);
	if (match)
	{	//we have a match
		GtkTreePath *path = gtk_tree_model_get_path (model, iter);		
		view->row = *gtk_tree_path_get_indices (path);
		if (view->row > 0) view->row++;	//CHECKME why is this needed .. 
		e2_fileview_focus_row (view, view->row,
			FALSE, FALSE, TRUE, TRUE);
		gtk_tree_path_free (path);
	}
	g_free (itemname);
	return !match;
}
/** @brief  UNUSED find filename begining with specified char, in current view
 *
 *
 * @param ch character at the start of the name
 * 
 * @return the row number where the matching name was found, or -1 if not found
**/
/*static gint e2_fileview_find_filename_begining_with (gchar ch)
{
	gint i;
	//number of rows in the model
	gint n = gtk_tree_model_iter_n_children
		(GTK_TREE_MODEL (curr_view->store), NULL);
	GtkTreeIter iter;
	GtkTreeModel *model = GTK_TREE_MODEL (curr_view->store);
	gchar *name;
	// search from after the current row to the end of the store
	for (i = curr_view->row+1; i < n; i++)
	{
		if (! gtk_tree_model_iter_nth_child
			(model, &iter, NULL, i))
				continue;  //the specified row doesn't exist any more
		gtk_tree_model_get (model, &iter, FILENAME, &name, -1);
		
		if (name[0] == ch)
		{
			g_free (name);
			return i;
		}
		else
			g_free (name);
	}
	// search from the first row to the current row
	for (i = 0; i < curr_view->row; i++)
	{
		if (! gtk_tree_model_iter_nth_child
			(model, &iter, NULL, i))
				continue;  //the specified row doesn't exist any more
		gtk_tree_model_get (model, &iter, FILENAME, &name, -1);
		
		if (name[0] == ch)
		{
			g_free (name);
			return i;
		}
		else
			g_free (name);
	}
	return -1;
}
*/

// Public functions

/**
 * @brief display (NOT select) a numbered row, optionally show its middle column center-pane
 *
 *
 * @param view rt data for ther view to be worked on
 * @param row number of the row that is to be focused
 * @param clear_selection TRUE if whole selection for the view is to be un-selected first
 * @param center TRUE if cursor focus should be moved to the center column of the view
 * @param grab_focus TRUE if the treeview is to be given the focus
* 
 * @return
**/
//when be called during a refresh, lock applied externally
void e2_fileview_focus_row (FileView *view,
		      gint row,
		      gboolean select_row,
		      gboolean clear_selection,
		      gboolean center,
		      gboolean grab_focus)
{
	if (view != NULL)
	{
		GtkTreeIter iter;
		GtkTreeModel *model = //(GTK_TREE_MODEL (view->store);
			gtk_tree_view_get_model (GTK_TREE_VIEW (view->treeview));

		//FIXME filter & child
		if (! gtk_tree_model_iter_nth_child (model, &iter, NULL, row))
		{
			return;  //the specified row doesn't exist
		}
		GtkTreePath* path = gtk_tree_model_get_path (model, &iter);
		GtkTreeView *tvw = GTK_TREE_VIEW (view->treeview);
		gtk_tree_view_scroll_to_cell (tvw, path, NULL, center, 0.382, 0.0);
		GtkTreeSelection *sel = view->selection;
		if (clear_selection)
			gtk_tree_selection_unselect_all (sel);
		if (select_row)
			gtk_tree_selection_select_path (sel, path);
		if (grab_focus)
			gtk_widget_grab_focus (view->treeview);
		gtk_tree_path_free (path);
	}
}
/**
 * @brief UNUSED select 1st row, in specified view, where the filename matches a specified name
 *
 * model content update is suspended while this is performed
 *
 * @param view rt data for ther view to be worked on
 * @param filename string with name of file to be used for selection
 * 
 * @return
**/
/*void e2_fileview_select_row_by_filename (FileView *view, gchar *filename)
{
	gint i;
	//number of rows in the model
	gint n = gtk_tree_model_iter_n_children
		(GTK_TREE_MODEL (view->store), NULL);
	GtkTreeIter iter;
	GtkTreeModel *model = GTK_TREE_MODEL (view->store);
	gchar *thisname;
	e2_filelist_disable_refresh();
	// search from after the current row to the end of the store
	for (i = 0; i < n; i++)
	{
		if (! gtk_tree_model_iter_nth_child
			(model, &iter, NULL, i))
				continue;  //the specified row doesn't exist any more
		gtk_tree_model_get (model, &iter, FILENAME, &thisname, -1);

		if (g_str_equal (thisname, filename))
		{
			e2_fileview_focus_row (view, i, TRUE, FALSE, TRUE, TRUE);
			g_free (thisname);
			break;
		}
		else
			g_free (thisname);
	}
	e2_filelist_enable_refresh();
} */
/**
 * @brief get file info for 1st selected row
 *
 *
 * @param view rt data for ther view to be worked on
 * 
 * @return file info structure for the 1st valid selected row, or if none, then NULL
**/
FileInfo *e2_fileview_get_selected_first_local (FileView *view)
{
	FileInfo *info = NULL;
//	GtkTreeSelection *sel = view->selection;
	GtkTreeModel *model;
	GList *base = gtk_tree_selection_get_selected_rows (view->selection, &model);
	GList *rowpaths;
	GtkTreeIter iter;
	gchar *itemname;
	for (rowpaths=base; rowpaths!=NULL; rowpaths=rowpaths->next)
	{
		GtkTreePath *first_sel = (GtkTreePath *) rowpaths->data;
		if (gtk_tree_model_get_iter (model, &iter, first_sel))
		{
			gtk_tree_model_get (model, &iter, FILENAME, &itemname, FINFO, &info, -1);
			//ignore possible selection of ".."
			if (!g_str_equal (itemname, "../"))
			{
				g_free (itemname);
				break;
			}
			g_free (itemname);
		}
	}
	g_list_foreach (base, (GFunc) gtk_tree_path_free, NULL);
	g_list_free (base);

	return info;
}
/**
 * @brief get list of selected rows' info items
 *
 * This is usable only when the data is used before any chance
 * of filelist refresh, which would invalidate the listed data pointers
 *
 * @param view rt data for the view to be worked on
 * 
 * @return glist with the selected rows' info items, excluding ".."
**/
GList *e2_fileview_get_selected_local (FileView *view)
{
//	g_signal_emit_by_name (G_OBJECT (view->treeview), "end-selection");

//	if (view->tagged != NULL)
//		return g_list_copy (view->tagged);

	GList *selectedrow_data = NULL;
	FileInfo *info;
	GtkTreePath *path;
	GtkTreeIter iter;
//	GtkTreeSelection *sel = view->selection;
	GtkTreeModel *model;
	GList *base = gtk_tree_selection_get_selected_rows (view->selection, &model);
	GList *rowpaths;

	for (rowpaths=base; rowpaths!=NULL; rowpaths=rowpaths->next)
	{
		path = (GtkTreePath *) rowpaths->data;
		if (gtk_tree_model_get_iter (model, &iter, path))
		{
			gtk_tree_model_get (model, &iter, FINFO, &info, -1);
			if (!g_str_equal (info->filename, ".."))	//CHECKME ok, even with localised string?
				selectedrow_data = g_list_append (selectedrow_data, info);
		}
	}
	g_list_foreach (base, (GFunc) gtk_tree_path_free, NULL);
	g_list_free (base);
	return selectedrow_data;
}
/**
 * @brief  toggle selection of all items in a pane
 *
 * This function selects or unselects all** items in a pane, depending on
 * the current state of the corresponding flag, pane1_all or pane2_all
 *  ** except '..' entries, (if displayed)
 * The pane will be activated, if it wasn't already.
 * The function handles clicks of the select-all toggle buttons 
 *
 * @param blah unused widget ptr (in case this is a callback)
 * @param view fileview for pane that is to be (un)selected
 *
 * @return 
 **/
void e2_fileview_select_all (GtkWidget *blah, FileView *view)
{
	if (curr_view != view)
		e2_pane_activate_other ();
	GtkTreeSelection *sel = curr_view->selection;
	gboolean *allnow = (view == app.pane1.view) ? &pane1_all : &pane2_all ;
	if (*allnow)
	{
		gtk_tree_selection_unselect_all (sel);
		*allnow = FALSE;
	}
	else //select
	{
		gtk_tree_selection_select_all (sel);
		if (e2_option_bool_get ("show-updir-entry")
			&& !g_str_equal (curr_view->dir, G_DIR_SEPARATOR_S))  //not at root dir now
		{
//			gtk_clist_unselect_row (GTK_CLIST (curr_view->clist), 0, 0);
//			GtkTreeIter iter = <iter for .. row>;  //not necessarily child 0 ?
//			gtk_tree_selection_unselect_iter (selection, &iter);
		}
		*allnow = TRUE;
	}
}
/**
 * @brief cleanup content of selected file info struct
 *
 * foreach function for e2_fileview_clean_selected()
 *
 * @param seldata ptr to the struct to be processed
 * @param user_data user data passed to g_ptr_array_foreach() UNUSED
 * 
 * @return
**/
void e2_fileview_selected1_clean (E2_SelectedItemInfo *seldata,
	gpointer user_data)
{
	g_free (seldata->filename);
#ifdef E2_INCLIST
	gtk_tree_row_reference_free (seldata->ref);
#else
	gtk_tree_path_free (seldata->path);
#endif
}
/**
 * @brief cleanup pointer array of selected file info structs
 *
 *
 * @param selected ptr to the array to be freed
 * 
 * @return
**/
void e2_fileview_clean_selected (GPtrArray *selected)
{
	g_ptr_array_foreach (selected, (GFunc) e2_fileview_selected1_clean, NULL);
	g_ptr_array_free (selected, TRUE);
}
/**
 * @brief create pointer array of selected file info structs
 *
 * This fn creates a pointer array, with pointers to structs containing
 * selected items' names and other info.
 * If found, ".." is filtered out.
 * If found, any trailing / is removed.
 * Item name strings are utf-8
 * Data is duplicated, so the selection is refresh-proof.
 * Treerow refs support incremental store changes as the selection is processed
 * The returned array needs to be freed with its data.
 *
 * @param view is data structure for the view to be interrogated
 * 
 * @return pointer to GPtrArray structure, or NULL if nothing selected
**/
GPtrArray *e2_fileview_get_selected (FileView *view)
{
//	if (view == curr_view)
//		g_signal_emit_by_name (G_OBJECT(view->treeview), "end-selection");
	GtkTreeSelection *sel = view->selection;
	gint count = gtk_tree_selection_count_selected_rows (sel);
	if (count > 0)
	{
		GPtrArray *array = g_ptr_array_sized_new (count);
		GtkTreeModel *model;
		GList *base = gtk_tree_selection_get_selected_rows (sel, &model);
		GList *paths;
		GtkTreeIter iter;
		gchar *filename, *localfilename;
//		FileInfo *info;
		E2_SelectedItemInfo *fileinfo;
		for (paths = base; paths != NULL; paths = paths->next)
		{
			if (!gtk_tree_model_get_iter (model, &iter, paths->data))
				continue;  //the row has been altered since it was selected
//			gtk_tree_model_get (model, &iter, FILENAME, &filename, FINFO, &info, -1);
			gtk_tree_model_get (model, &iter, FILENAME, &filename, -1);
			if (!g_str_equal (filename, ".."))
			{
				fileinfo = g_malloc (sizeof (E2_SelectedItemInfo));
				//always replicate filename
				localfilename = g_strdup (filename);

				//strip any trailer (can't have it on links, at least)
				gchar *s = localfilename + strlen (localfilename) - 1;
				if (*s == G_DIR_SEPARATOR)
					*s = '\0';
				fileinfo->filename = localfilename;  //this is utf8
//				fileinfo->statbuf = info->statbuf; NO LONGER USED
#ifdef E2_INCLIST
				//get a reference in case the model changes during the operation on the selection
				fileinfo->ref = gtk_tree_row_reference_new (model, paths->data);
#else
				//this is only relevant if store data not changed on the fly
				fileinfo->path = gtk_tree_path_copy (paths->data);
#endif
				g_ptr_array_add (array, fileinfo);
			}
// always free it now			else
			g_free (filename);
		}
		g_list_foreach (base, (GFunc) gtk_tree_path_free, NULL);
		g_list_free (base);
		return array;
	}
	return NULL;
}
/**
 * @brief get itemname for a specified view and row
 *
 *
 * @param view is data structure for the view to be interrogated
 * @param row number of the row to be interrogated REAL or FILTERED ?
 * 
 * @return pointer to duplicated string, or NULL if there's nothing found
**/
gchar *e2_fileview_get_row_name (FileView *view, gint row)
{
	GtkTreeIter iter;
//	GtkTreeModel *model = GTK_TREE_MODEL (view->store); maybe filter
	GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW (view->treeview));

	if (! gtk_tree_model_iter_nth_child (model, &iter, NULL, row))
		return NULL;
	gchar *name;
	gtk_tree_model_get (model, &iter, FILENAME, &name, -1);
	return name;
}
/**
 * @brief helper to set file-list font
 * @param view pointer to data struct for the view to be changed
 * @param fontstr name of font to apply to the filelist associated with @a view
 *
 * @return
**/
static void _e2_fileview_set_font (FileView *view, gchar *fontstr)	
{
	GList* columns, *renderers, *base1, *base2;
	columns = base1 = gtk_tree_view_get_columns (GTK_TREE_VIEW (view->treeview));
	for (; columns != NULL ; columns = columns->next)
	{
		GtkTreeViewColumn *column = columns->data;
		renderers = base2 = gtk_tree_view_column_get_cell_renderers (column);
		for (; renderers != NULL ; renderers = renderers->next)
			g_object_set (renderers->data, "font", fontstr, NULL);	
		g_list_free (base2);
	}
	g_list_free (base1);
}
/**
 * @brief set file-lists font
 * This is needed to change font after a config dialog that does
 * not rebuild the whole window
 * @return
**/
void e2_fileview_set_font (void)
{
	gchar *fontstr = (e2_option_bool_get ("custom-list-font")) ?
		e2_option_str_get ("list-font") : NULL;	//NULL forces default
	_e2_fileview_set_font (curr_view, fontstr);
	_e2_fileview_set_font (other_view, fontstr);
}
/**
 * @brief set background color of treeview line
 *
 *
 * @param view data struct for the view to process
 * @param iter pointer to iter for the row to be changed
 * @param color pointer to struct with data for the new color
 *
 * @return
**/
void e2_fileview_set_row_background (FileView *view,
	GtkTreeIter *iter, GdkColor *color)
{
	if (gtk_tree_selection_iter_is_selected (view->selection, iter))
		return;	//no point in changing selected rows, gtk color prevails

	//set the attribute column for this row
	gtk_list_store_set (GTK_LIST_STORE (view->store), iter, BACKCOLOR, color, -1);
/*	we always set the renderers to use the background column,
	as we cannot viably remove the renderer property on the fly
	GList* columns, *renderers, *base1, *base2;
	columns = base1 = gtk_tree_view_get_columns (GTK_TREE_VIEW (view->treeview));
	for (; columns != NULL ; columns = columns->next)
	{
		GtkTreeViewColumn *column = columns->data;
		{
			renderers = base2 = gtk_tree_view_column_get_cell_renderers (column);
			for (; renderers != NULL ; renderers = renderers->next)
				gtk_tree_view_column_add_attribute (column,
					(GtkCellRenderer *) renderers->data,
					"cell-background-gdk", BACKCOLOR);
			g_list_free (base2);
		}
	}
	g_list_free (base1); */
	view->lit = TRUE;
}
/**
 * @brief revert background color of treeview line
 *
 *
 * @param view data struct for the view to process
 * @param iter pointer to iter for the row to be changed
 *
 * @return
**/
void e2_fileview_clear_row_background (FileView *view, GtkTreeIter *iter)
{
	if (gtk_tree_selection_iter_is_selected (view->selection, iter))
		return;	//no need to revert selected rows, they're not re-colored

	//clear the attribute column for this row
	gtk_list_store_set (GTK_LIST_STORE (view->store), iter, BACKCOLOR, NULL, -1);
/* this does not work, it stuffs up clearing of background color
	so we just use the backround column attribute all the time
	//unfortunately, no way to remove a single attribute, so set
	//all attributes the way they were when the treeview was created
	GList* columns, *renderers, *base1, *base2;
	columns = base1 = gtk_tree_view_get_columns (GTK_TREE_VIEW (view->treeview));
	for (; columns != NULL ; columns = columns->next)
	{
		GtkTreeViewColumn *column = columns->data;
		//the column number is 'textified' as the title
		const gchar *title = gtk_tree_view_column_get_title (column);
		gint m = title[0] - '0';
		renderers = base2 = gtk_tree_view_column_get_cell_renderers (column);
		for (; renderers != NULL ; renderers = renderers->next)
			gtk_tree_view_column_set_attributes (column,
					(GtkCellRenderer *) renderers->data,
					"text", m,
					"foreground-gdk", FORECOLOR,
					NULL);
			g_list_free (base2);
	}
	g_list_free (base1); */
	view->lit = FALSE;
}
/**
 * @brief translate array into another form sorted in index order
 *
 * This is needed to convert the 'fixed' column order used in data models
 * and config/cache data, to variable order in accord with the displayed view columns
 * and vice versa
 *
 * @param src_array pointer to array of ints that is to be translated
 * @param dest_array pointer to array of ints that stores the translated order
 * @param size the number of elements to be translated
 *
 * @return
 */
void e2_fileview_translate_cols_array (gint *src_array, gint *dest_array,
 gint size)
{
	gint i;
	for (i=0; i< size; i++)
		dest_array[src_array[i]] = i;
}
/**
 * @brief update columns data, that is not automatically updated, for caching
 *
 * Columns order and widths are udated and stored in a list
 * List format is: pane1 {order, width, ...}, pane2 {order, width, ...}
 *
 * @return
**/
void e2_fileview_update_col_cachedata (void)
{
	//get current column widths
	GtkTreeViewColumn *col;
	G_CONST_RETURN gchar *title;
	gint i;
	FileView *view;
	view = app.pane1.view;
	GList *columns = 	gtk_tree_view_get_columns (GTK_TREE_VIEW (view->treeview));
	GList *base = columns;
	for (; columns != NULL; columns=columns->next)
	{
		col = columns->data;
		title = gtk_tree_view_column_get_title (col);
		//title is "0" ... "7", for cols 0 ... 7
		i = atoi (title);
		col_width_store [0][i] = gtk_tree_view_column_get_width (col);
	}
	g_list_free (base);

	view = app.pane2.view;
	columns = 	gtk_tree_view_get_columns (GTK_TREE_VIEW (view->treeview));
	base = columns;
	for (; columns != NULL; columns=columns->next)
	{
		col = columns->data;
		title = gtk_tree_view_column_get_title (col);
		i = atoi (title);
		col_width_store [1][i] = gtk_tree_view_column_get_width (col);
	}
	g_list_free (base);

	//get columns displayed order
	e2_fileview_translate_cols_array (displayed_col_order[0], stored_col_order[0], MAX_COLUMNS);
	e2_fileview_translate_cols_array (displayed_col_order[1], stored_col_order[1], MAX_COLUMNS);

/*SUPERSEDED BY ARRAY CACHEING
	//setup for list-cache writing
	//start fresh list
	e2_list_free_with_data (&cols_data);
	gchar *buffer;
	gint j;
	for (j = 0; j < 2; j++)
	{ for (i = 0; i < MAX_COLUMNS; i++) {
		//data need to be strings, for the cache writer
		buffer = g_strdup_printf ("%d", stored_col_order[j][i]);
		cols_data = g_list_append (cols_data, buffer);
		buffer = g_strdup_printf ("%d", col_width_store[j][i]);
		cols_data = g_list_append (cols_data, buffer);
	}}
*/
}
/**
 * @brief UNUSED find which view column shows filenames
 *
 * Columns may be in any order
 * Scans column titles, looking for "0"
 *
 * @param view data structure for the view to which the file list belongs
 *
 * @return integer column number, 0 to MAX_COLUMNS-1
 */
/* gint e2_fileview_find_name_col (FileView *view)
{
	GList *cols = gtk_tree_view_get_columns (GTK_TREE_VIEW (view->treeview));
	GList *base = cols;
	G_CONST_RETURN gchar *title;
	gint result = 0;
	for ( ; cols != NULL; cols = cols->next )
	{
		title = gtk_tree_view_column_get_title (cols->data);
		if (g_str_equal (title, "0"))  break;
		result++;
	}
	g_list_free (base);
//	return (result < MAX_COLUMNS) ? result : -1;
	return result;
} */
/**
 * @brief exchange active and inactive panes
 *
 * This function is called from e2_pane_activate_other (), i.e.
 * whenever we need to exchange curr_view and other_view.
 * It toggles pointers and changes colours
 * The latter is done by interrogating the selections in both file views,
 * and so this fn should be blocked if either view is still being
 * populated at the time this is called ?? CHECKME
 *
 * @return
 */
void e2_fileview_switch_views (void)
{
	FileView *temp;

	if (! e2_fs_chdir (other_view->dir))
		return;

	temp = other_view;
	other_view = curr_view;
	curr_view = temp;

	e2_pane_flag_active ();	//change the status-indicator for the panes
}
/**
 * @brief UNUSED swap view pointers 
 *
 * Exchange curr_view and other_view, 
 * without doing anything else
 *
 * @return 
 **/
/*void e2_fileview_switch_views_simple (void)
{
	FileView *temp = other_view;
	other_view = curr_view;
	curr_view = temp;
}*/
/**
 * @brief get indices of top line and left column shown in a filelist treeview
 * This gets scroll info for treeview rebuilds
 * @param view data structure for the view to be changed
 * @param leftcol pointer to store for the left column index
 * @param toprow pointer to store for the top row index
 *
 * @return
 */
void e2_fileview_get_scroll_data (FileView *view, gint *leftcol, gint *toprow)
{
	GtkTreePath *startpath;
	GtkTreeViewColumn *startcol;
//	gint cell_x, cell_y;
	if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (view->treeview),
		0, 0, &startpath, &startcol, NULL, NULL))	//&cell_x, &cell_y))
	{
		*toprow = *gtk_tree_path_get_indices (startpath);
		printd (DEBUG, "path found, row is %d", *toprow);
		gtk_tree_path_free (startpath);
	}
	else	//liststore is empty
		*toprow = 0;
//gtk bug, startcol always is the one at index 0 (maybe fixed after 2.8.6)
	*leftcol = 0;
}
/**
 * @brief scroll a filelist treeview to top line @a toprow and left column @a leftcol
 * This is used in treeview rebuilds
 * It expects that the treeview exists - i.e. be careful when threaded
 * @param view data structure for the view to be changed
 * @param leftcol left column index
 * @param toprow top row index
 *
 * @return
 */
void e2_fileview_scroll_to_position (FileView *view, gint leftcol, gint toprow)
{
	if (toprow > 0 || leftcol > 0)
	{
		GtkTreeView *tvw = GTK_TREE_VIEW (view->treeview);	
		GtkTreeModel *model = gtk_tree_view_get_model (tvw);
		GtkTreeViewColumn* col = NULL;
		GtkTreePath* path = NULL;
		if (toprow > 0)
		{
			GtkTreeIter iter;
			if (gtk_tree_model_iter_nth_child (model, &iter, NULL, toprow))
				path = gtk_tree_model_get_path (model, &iter);
		}
		else
			path = gtk_tree_path_new_first ();

		gtk_tree_view_scroll_to_cell (tvw, path, col, TRUE, 0, 0);
		gtk_tree_path_free (path);
	}
}
/**
 * @brief helper function to clean up redundant tree hash tables
 *  when there's nothing better to do
 * @param data pointer to table to be destroyed
 * @return FALSE so that this is not called again
 */
gboolean e2_fileview_treehash_free (GHashTable *data)
{
	g_hash_table_destroy (data);
	return FALSE;
}
/**
 * @brief get hash of selected-item names for pane in @a view
 *
 * Logged names are in utf-8 encoding, dirs have trailing /
 * It is used as part of the view refresh or recreate processes
 * (even with FAM, needed for refresh action, (un)hide, (un)filter,
 *  window re-create)
 *
 * @param view data structure for view being processed
 *
 * @return hash table of names, NULL if nothing selected
 */
//CHECKME filter & child
GHashTable *e2_fileview_log_selected_names (FileView *view)
{
	GtkTreeModel *model;
	GList *base = gtk_tree_selection_get_selected_rows (view->selection, &model);
	if (g_list_length (base) == 0)
		return NULL;

	//build names-hash
	GHashTable *treehash = g_hash_table_new_full
			(g_str_hash, g_str_equal, g_free, NULL);
	GtkTreeIter iter;
	gchar *itemname;
	GList *selected_paths;

	for (selected_paths=base; selected_paths!=NULL; selected_paths=selected_paths->next)
	{
		if (gtk_tree_model_get_iter (model, &iter, selected_paths->data))
  		{
			gtk_tree_model_get (model, &iter, FILENAME, &itemname, -1);
			g_hash_table_insert (treehash, itemname, NULL);
		}
		gtk_tree_path_free (selected_paths->data);
	}
	g_list_free (base);
	return treehash;
}
/**
 * @brief re-select items in list store for specified view
 *
 * @a view ->selected_names is freed here
 *
 * @param view data structure for view being processed
 *
 * @return
 */
void e2_fileview_reselect_names (FileView *view)
{
//	GtkTreeModel *model = GTK_TREE_MODEL (view->store);
	GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW (view->treeview));
	GtkTreeIter iter;
	gboolean notempty = gtk_tree_model_get_iter_first (model, &iter);
	GHashTable *oldselected = view->selected_names;
	if (oldselected != NULL && notempty)
	{
		guint more_hits = g_hash_table_size (oldselected);
		GtkTreeSelection *sel = view->selection;
		gchar *itemname;
		gpointer key;
		gpointer value;
		do
		{
			gtk_tree_model_get (model, &iter, FILENAME, &itemname, -1);
			if (g_hash_table_lookup_extended (oldselected, itemname,
					&key, &value))
			{
				gtk_tree_selection_select_iter (sel, &iter);
				more_hits--;
			}
			g_free (itemname);
		} while (more_hits > 0 && gtk_tree_model_iter_next (model, &iter));
	}

	if (oldselected != NULL)
	{
		//setup to get rid of hash when there's more time
		g_idle_add_full (G_PRIORITY_LOW,
			(GSourceFunc) e2_fileview_treehash_free, oldselected, NULL);
		view->selected_names = NULL;
	}
}
/**
 * @brief find matching path entry in history list
 *
 * @param list Glist of past path strings
 * @param path path string to be matched
 *
 * @return list entry whose data matches path, or else NULL
 */
static GList *_e2_fileview_find_history (GList *list, gchar *path)
{
	GList *tmp;
	DirHistoryEntry *entry;

	for (tmp = list; tmp != NULL; tmp = tmp->next)
	{
		entry = tmp->data;
		if (g_str_equal (entry->path, path))
		  return tmp;
	}
	return NULL;
}
/**
 * @brief update all items' info in the treeview associated with @a view
 *
 * After updating the underlying liststore contents, item(s) are
 * (re) selected as appropriate, and the view is scrolled to the
 * relevant position
 * This is used for refreshes and cd's
 * It can be a thread func - though not yet thread-safe
 *
 * @param view runtime data for the view to be displayed
 *
 * Directory whose content we want is named in view->dir (utf-8)
 * Current store content is cleared after successful preliminary things,
 * but before new data is added, of course
 * Active filter(s) are applied, but no sorting.
 * If relevant, the text-color for selected items is updated here, rather 
 * than in the selection function (too many un-necessary repeats, there)
 * This function is not necessarily thread-safe
 * NOTE sets view->total_items
 *
 * @param reload TRUE to get a fresh lot of data from the source directory
 * @param view data structure for the view to which the file list belongs
 *
 * @return
 **/
void e2_fileview_prepare_list (FileView *view)
{
	//prevent re-entrant refreshes
	view->completed = FALSE;  //set blocked signal
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, prepare list");
#endif
	e2_filelist_disable_refresh ();

	e2_widget_set_cursor (GDK_WATCH);

	GtkTreeModel *mdl;
	GtkTreeSortable *sortable;
	GList *entries;
	if (view->refreshtype == E2_REFRESH || view->refreshtype == E2_CHANGE)
	{
		mdl = GTK_TREE_MODEL (view->store);
		sortable = GTK_TREE_SORTABLE (mdl);
		//remember the current sorting arrangements
		gtk_tree_sortable_get_sort_column_id (sortable, &view->sort_column,
			&view->sort_order);
		//turn off model sorting before any rows are added
		//CHECKME do we really benefit from turning off sorting when the
		//view is detached from the model?
		//GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID (-2) appeared in gtk 2.6
		gtk_tree_sortable_set_sort_column_id (sortable,
			GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID, GTK_SORT_ASCENDING);
#ifndef E2_NEWFILTER
		//make sure the store stays alive after the treeview unrefs the model
		g_object_ref (G_OBJECT (view->store));
#endif
	/*	for speed, and no-blanking in the case of repeated
		refreshes, detach model from view
		it doesn't matter how far ahead of (or behind) the
		store-clear that we do this */
		gtk_tree_view_set_model (GTK_TREE_VIEW (view->treeview), NULL);

	//construct relevant items data for store-filler
	//FIXME if we can, just get relevant items for a refresh
	//get a fresh set of FileInfo data from source
		entries = e2_fs_read_dir (view);
		if (entries == GINT_TO_POINTER (-1))
			return;
#if defined (E2_FAM_INOTIFY) || defined (E2_FAM_DNOTIFY) || defined (E2_FAM_KQUEUE)
		if (view->refreshtype == E2_REFRESH)
			e2_fs_FAM_clean_reports (view->dir);
	/*	prior reports about the dir are now gone.
		Henceforth any report about the dir or an item in it
		will trigger a new refresh, even if the change that
		caused the report is shown in this refresh.
		The alternative is to do this at the end, and lose any
		change after an item is processed in this refresh */
#endif	//which fam
	}
	else	//doing a window recreation
	{	//use current data for the view
		entries = NULL;
		FileInfo *infoptr, *newinfo;
		GtkTreeIter iter;
		mdl = GTK_TREE_MODEL (view->store);
		if (gtk_tree_model_get_iter_first (mdl, &iter))
		{
			do
			{
				gtk_tree_model_get (mdl, &iter, FINFO, &infoptr, -1);
#if defined(E2_CURRENTGTK_VERSION) && GLIB_CHECK_VERSION (2,10,0)
				newinfo = g_slice_alloc (sizeof(FileInfo));
				g_memmove (newinfo, infoptr, sizeof(FileInfo));
#else
				newinfo = g_memdup (infoptr, sizeof(FileInfo));
#endif
				entries = g_list_append (entries, newinfo);
			} while (gtk_tree_model_iter_next (mdl, &iter));
		}
	}

	//if we can, get new directory data into a store
#ifndef E2_NEWFILTER
	guint omitted = 0;
#endif
	GtkListStore *store = e2_filelist_fill_store (entries
#ifndef E2_NEWFILTER
	, view, &omitted
#endif
	);
	//FIXME entries may be just a few changes
#ifdef E2_NEWFILTER
	//entries count adjusted for .. (. filtered from list)
	view->total_items = g_list_length (entries) - 1;
#else
	//entries count adjusted for .. and .
	view->total_items = g_list_length (entries) - 2;
#endif

	if (store != NULL)
	{	//there was a valid store created
#ifdef E2_NEWSTORES
# ifdef E2_NEWFILTER
		//can't re-assign the filter model to new store, nave to replace model
		//make sure the store stays alive after the filtermodel goes
		g_object_ref (G_OBJECT (view->store));
		g_object_unref (G_OBJECT (view->model));	//kill the old filtermodel
		view->model = gtk_tree_model_filter_new (
			GTK_TREE_MODEL (store), NULL);
		g_object_unref (G_OBJECT (store));	//kill the ref from model creation
		view->filtered_before = FALSE;	//trigger re-attachment of filter func
# endif
		//arrange to cleanup the old store at an idle time
		gboolean newtimer = (app.used_stores == NULL);
		app.used_stores = g_slist_append (app.used_stores, view->store);
		if (newtimer)
		{
			printd (DEBUG, "setup to clear stores later");
			g_idle_add (e2_filelist_clear_old_stores, NULL);
		}
#else
		//clear the current store now
		//(use the "base" model for direct store manipulation)
		GtkTreeModel *mdl = GTK_TREE_MODEL (view->store);
		GtkTreeIter iter;
		if (gtk_tree_model_get_iter_first (mdl, &iter))
		{	//it's not empty already
			//clear file info data referred to in the store
			FileInfo *currentinfo;
			do
			{
				gtk_tree_model_get (mdl, &iter, FINFO, &currentinfo, -1);
# if defined(E2_CURRENTGTK_VERSION) && GLIB_CHECK_VERSION (2,10,0)
				g_slice_free1 (currentinfo, sizeof(FileInfo));
# else
				g_free (currentinfo);
# endif
			} while (gtk_tree_model_iter_next (mdl, &iter));

			gtk_list_store_clear (view->store);
		}
		g_object_unref (G_OBJECT (view->store));
#endif	//def E2_NEWSTORES

		view->store = store;
		sortable = GTK_TREE_SORTABLE (GTK_TREE_MODEL (store));
		gtk_tree_sortable_set_default_sort_func (sortable, NULL, NULL, NULL);
		gint array_row = (view == app.pane1.view) ? 0 : 1;
		gint *order_array = displayed_col_order[array_row];
		gint i, m;
		for (i = 0; i < MAX_COLUMNS; i++)
		{
			m = order_array[i];  //get the model/'external' index
			gtk_tree_sortable_set_sort_func (sortable, m,
				e2_all_columns[m].sort_func, &view->sort_order, NULL);
		}

		//sort the view using the current sort column & direction
		gtk_tree_sortable_set_sort_column_id (sortable, view->sort_column,
			view->sort_order);

#ifdef E2_NEWFILTER
		e2_fileview_refilter_list (view);
#else	//ndef E2_NEWFILTER
# ifdef E2_NEWSTORES
		mdl = GTK_TREE_MODEL (view->store);	//the new store has a different model
# endif
		//re-attach model to view (before scrolling to old positions)
		gtk_tree_view_set_model (GTK_TREE_VIEW (view->treeview), mdl);
		//kill ref just added, to destroy model with the view
		g_object_unref (G_OBJECT (view->store));
#endif	//def E2_NEWFILTER
	}
	//signal to the world ...
	view->completed = TRUE;

	e2_widget_set_cursor (GDK_LEFT_PTR);

#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, prepare list");
#endif
	e2_filelist_enable_refresh ();
//	printd (DEBUG, "end view update for %s", view->dir);
}
/**
 * @brief get new directory contents for a view
 * This may be called manually (=in callback, inside gtk lock)
 * or in a timer func (explicit locking needed)
 * @param view data structure for the view to be changed
 * @param oldpath NULL, or utf8 string with old path, needed by _e2_fileview_threadedprepare_list() when changing dir
 * @param newpath utf8 string with new path
 *
 * @return
 */
void e2_fileview_change_dir (FileView *view, gchar *newpath)
{
	printd (DEBUG, "change dir: %s", newpath);
	DirHistoryEntry *entry;
	//update row history for current dir
	//at session-start, view->dir = "", hence no match found in list
	GList *outdir = _e2_fileview_find_history (view->dir_history, view->dir);
	if (outdir != NULL)
	{
		entry = outdir->data;
		/*even though the sort-order and/or dir contents
		  might change before we re-visit here,
		  if there is a selection now, remember its start,
		  or else remember the start of the current display*/
		
		/*this func may be called during a window-rebuild, in which
		   case the old dir will the same as the new one, and the
		   selection and visible rect data will be meaningless*/

		GtkTreeSelection *sel = view->selection;
		if (gtk_tree_selection_count_selected_rows (sel) > 0)
		{
			entry->selected = TRUE;
			GtkTreeModel *model;
			GList *selpaths = gtk_tree_selection_get_selected_rows
				(sel, &model);
			//get the 1st selected item's row
			entry->row = *gtk_tree_path_get_indices (selpaths->data);
			g_list_foreach (selpaths, (GFunc) gtk_tree_path_free, NULL);
			g_list_free (selpaths);
		}
		else
		{
			entry->selected = FALSE;
			GtkTreePath *startpath;
			if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (view->treeview),
				0, 0, &startpath, NULL, NULL, NULL))
			{
				entry->row = *gtk_tree_path_get_indices (startpath);
				gtk_tree_path_free (startpath);
			}
			else	//dir is empty
				entry->row = 0 ;
		}
	}

	//now update the local path store
	g_strlcpy (view->dir, newpath, sizeof (view->dir));

	if (view->selected_names != NULL)
	{
		//setup to get rid of any old hash
		g_idle_add_full (G_PRIORITY_LOW,
			(GSourceFunc) e2_fileview_treehash_free, view->selected_names, NULL);
		view->selected_names = NULL;	//signal no need to reselect anything
	}

	//get the replacement filelist
	view->refreshtype = E2_CHANGE;
	e2_fileview_prepare_list (view);

	//goto where we were last time, if any
	GtkTreeModel *mdl = gtk_tree_view_get_model (GTK_TREE_VIEW (view->treeview));
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter_first (mdl, &iter))
	{
		GList *indir = _e2_fileview_find_history (view->dir_history, view->dir);
		if (indir != NULL)
		{  //new dir is in history list
			DirHistoryEntry *entry = indir->data;
			if (entry->selected)
			{
				//goto & select the 1st row we selected last time
				//(tuff luck if there's been content change or sort-order change)
				e2_fileview_focus_row (view, entry->row, TRUE, FALSE, FALSE, FALSE);
			}
			else if (entry->row)
			{	//goto the top of the visible list as it was last time
				if (gtk_tree_model_iter_nth_child (mdl, &iter, NULL, entry->row))
				{
					GtkTreePath *path = gtk_tree_model_get_path (mdl, &iter);
					gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (view->treeview), path, NULL, TRUE, 0, 0);
					view->row = entry->row;	//remember the focus row
				}
				else
					view->row = 0;
			}
			else
				view->row = 0;
		}
		else //the new dir has never been opened in this session
		{
			if (e2_option_bool_get ("select-first-item"))
			{
				e2_fileview_focus_row (view, 0, TRUE, FALSE, FALSE, FALSE);
			}
			view->row = 0;
		}
	}
	
	if (_e2_fileview_find_history (view->dir_history, view->dir) == NULL)
	{  //need a new entry for the history list
		entry = g_new0 (DirHistoryEntry, 1);
		//just set the path, for now - other data set when dir is departed
		g_strlcpy (entry->path, view->dir, sizeof(entry->path));
		view->dir_history = g_list_append (view->dir_history, entry);
	}

	if (view == curr_view)
	{
		e2_hook_list_run (&app.hook_pane_focus_changed, curr_pane);
		e2_window_update_status_bar (NULL);
	}
}
/**
 * @brief reload list store for @a view and re-select relevant items 
 *
 * This records the names of currently-selected items in the view,
 * reloads the directory contents (which clears the old selection),
 * finds the rows in the new store that match the selected names,
 * and selects them.
 * If the directory associated with @a view is no longer accessible,
 * a suitable alternative is found and displayed.
 * This may be called manually (=in callback, inside gtk lock)
 * or in a timer func (explicit locking needed)
 *
 * @param view data structure for view being processed
 *
 * @return
 */
//FIXME make this do incremental updates of current liststore,
//then abandon e2_fileview_prepare_list()
void e2_fileview_refresh_list (FileView *view)
{
	g_return_if_fail (view != NULL);

	if (!view->completed)
		return;	//refresh is blocked, or the view is still being processed from last time

	gchar *utf = g_strdup (view->dir);	//copy, so that view->dir not clobbered
	if (e2_fs_get_valid_path (&utf, TRUE))
		g_free (utf);
	else
	{	//if desired path not accessible, go to somewhere that is ..
		gchar *msg = g_strdup_printf ("Cannot access %s, going to %s instead", view->dir, utf);
		e2_output_print_error (msg, TRUE);

		E2_PaneRuntime *rt = (view == curr_view ) ? curr_pane : other_pane;
		e2_pane_change_dir (rt, utf);
		g_free (utf);
		return;
	}

	if (view->selected_names != NULL)
	{
		//setup to get rid of any old hash
		g_idle_add_full (G_PRIORITY_LOW,
			(GSourceFunc) e2_fileview_treehash_free, view->selected_names, NULL);
	}
	//log the names of selected items in this view
	view->selected_names = e2_fileview_log_selected_names (view);

	//clear and refill the store
	//FIXME support incremental updates
	view->refreshtype = E2_REFRESH;
	e2_fileview_prepare_list (view);

	GtkTreeModel *mdl = gtk_tree_view_get_model (GTK_TREE_VIEW (view->treeview));
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter_first (mdl, &iter))
	{
		//restore scroll positions
		/*Not clear why this works, but ....
		the adjustments' values are unchanged, but the sw 
		doesn't know about them unless signals are issued */
		GtkAdjustment *adj;
		adj = gtk_tree_view_get_vadjustment (GTK_TREE_VIEW (view->treeview));
		gtk_adjustment_value_changed (adj);
//		adj = gtk_tree_view_get_hadjustment (GTK_TREE_VIEW (view->treeview));
//		gtk_adjustment_value_changed (adj);
	}
	if (view->selected_names != NULL)
//	{
//		clear out any old selection
//		gtk_tree_selection_unselect_all (view->selection);
		//try to reselect items, then at least clean the hash
		e2_fileview_reselect_names (view);
//	}
//	printd (DEBUG, "finished new selection for %s", view->dir);
}
/**
 * @brief create filelist for the pane related to @a rt
 *
 * This creates the backend list store, and related scrolled window
 * containing a treeview
 *
 * @param view runtime data for view being processed
 *
 * @return a scrolled window widget
 **/
GtkWidget *e2_fileview_create_list (FileView *view)
{
	//create liststore framework for the pane
	view->store = e2_filelist_make_store ();
	view->completed = TRUE;	//start without any blocking
	//setup column-order array
	gint array_row = (view == app.pane1.view) ? 0 : 1;
	e2_fileview_translate_cols_array (stored_col_order[array_row],
		displayed_col_order[array_row], MAX_COLUMNS);
	gint *width_array = col_width_store[array_row];
	gint *order_array = displayed_col_order[array_row];
	//element of cache-name strings
	gchar *panename = (view == app.pane1.view) ?
		app.pane1.name : app.pane2.name;

	/*now create tree view
	column order is determined by array displayed_col_order[].
	columns are named "0" to "7" for later searching */
	gint i, m;
	gchar *option_name;
	E2_OptionSet *set;
	gboolean show_this;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	GtkTreeSelection *sel;
	gchar *fontstr = (e2_option_bool_get ("custom-list-font")) ?
		e2_option_str_get ("list-font") : NULL;	//NULL will cause default font
	static gchar id[] = {'0', '\0'};  //for column header tiles, used to check which column

#ifdef E2_NEWFILTER
	//unfiltered model needed for getting the sortable
	GtkTreeModel *mdl = GTK_TREE_MODEL (view->store);
	view->model = gtk_tree_model_filter_new (mdl, NULL);
	g_object_unref (G_OBJECT (view->store));	//kill the ref from model creation
	view->filtered_before = FALSE;	//start with fresh filters
	GtkWidget *tvw = gtk_tree_view_new_with_model (view->model);
#else
	GtkTreeModel *mdl = GTK_TREE_MODEL (view->store);
	GtkWidget *tvw = gtk_tree_view_new_with_model (mdl);
	g_object_unref (G_OBJECT (view->store));  //kill the ref from the view attachment
#endif
	//save it for others to use
	view->treeview = tvw;

	//setup for non-sorted dir loads
	GtkTreeSortable *sortable = GTK_TREE_SORTABLE (mdl);
	gtk_tree_sortable_set_default_sort_func (sortable, NULL, NULL, NULL);

	//set general treeview properties
	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (tvw), TRUE);
	gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (tvw),	//useless with managed backcolors 
		e2_option_bool_get ("panes-hinted"));
	//cache the view's selection pointer
	view->selection = sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (tvw));
	gtk_tree_selection_set_mode (sel, GTK_SELECTION_MULTIPLE);
#ifdef E2_SELTXT_RECOLOR
	if (e2_option_bool_get ("selectedtext-colorchange"))
		gtk_tree_selection_set_select_function (sel,
			(void *) _e2_fileview_selection_func, view, NULL);
#endif
/*	//when both panes are "same", make the slave-pane columns un-draggable
	gboolean fixed_order = (
		(e2_option_bool_get ("pane1-uses-other") && view == app.pane1.view)
		||
		(e2_option_bool_get ("pane2-uses-other") && view == app.pane2.view) );
*/
	//set columns' properties
	for (i = 0; i < MAX_COLUMNS; i++)
	{
		m = order_array[i];  //get the model/'external' index
		gtk_tree_sortable_set_sort_func (sortable, m,
			e2_all_columns[m].sort_func, &view->sort_order, NULL);

		//get the column visibility
		option_name = g_strdup_printf ("%s-show-column%d", panename, m);
		// e2_option_bool_get  defaults to FALSE
		set = e2_option_get (option_name);
		show_this = (set != NULL) ? e2_option_bool_get_direct (set) : TRUE;
		g_free (option_name);

		//create header widget, to support styling and avoid gtk bug with arrows
		GtkWidget *headerbox = gtk_hbox_new (FALSE, 1);
		GtkWidget *label = gtk_label_new (gettext(e2_all_columns[m].title)); //same title for each pane
		//expand = TRUE prevents col width < title width + sort indicator (if any)
		//spacing 0 when label expand = FALSE (or ese it crashes when col is too narrow to show all)
		gtk_box_pack_start (GTK_BOX(headerbox), label, FALSE, TRUE, 0);
		gtk_widget_show (label);
		if (m == FILENAME)
			//remember the Filename column label in case we need to use it for active-pane flagging
			view->name_label = GTK_LABEL (label);
		//set the current sort arrow
		GtkArrowType arrow;
		if (m == view->sort_column
				&& view->sort_order == GTK_SORT_DESCENDING)
			arrow = GTK_ARROW_UP;
		else
			arrow = GTK_ARROW_DOWN;
		view->sort_arrows[m] = gtk_arrow_new (arrow, GTK_SHADOW_IN);
		//this doesn't show the indicator at the end of the header ! (and different exp, fill don't help)
		gtk_box_pack_end (GTK_BOX(headerbox), view->sort_arrows[m], FALSE, FALSE, 0);
		gtk_widget_show (headerbox);

		renderer = gtk_cell_renderer_text_new ();
		g_object_set (renderer, 
			"yalign", 0.0,
			"ypad", 0, //supposedly default, but this makes the rows closer
			"font", fontstr,	//this property not stored in model, as it doesn't change on the fly ...
			NULL);
//		gtk_cell_renderer_text_set_fixed_height_from_font
//			(GTK_CELL_RENDERER_TEXT (renderer), 1);
		//right-align the size column only
		if (m == SIZE)
			g_object_set (renderer, 
			"xalign", 1.0,
			"xpad", 0,
			NULL);
		
  		id[0] = m+'0';
		column = gtk_tree_view_column_new_with_attributes (
			id, //this is for seaching & sorting, displayed title is in widget label
			renderer,
			"text", m,
			"foreground-gdk", FORECOLOR,
			"cell-background-gdk", BACKCOLOR,	//this column is generally NULLed, except for drop-target highlighting
			NULL);

		gint thiswidth = width_array[m];
		//doesn't like 0 widths set by gtk when a col is hidden
		if (thiswidth == 0)
			thiswidth = e2_all_columns[m].size;

		g_object_set (column,
			"sizing", GTK_TREE_VIEW_COLUMN_FIXED,
			"resizable", TRUE,
			"fixed-width", thiswidth,   //width_array[m],
			"visible", show_this,
			"clickable", TRUE,
//			"reorderable", !fixed_order,
			"reorderable", TRUE,
			"widget", headerbox,
			NULL);

		gtk_tree_view_append_column (GTK_TREE_VIEW (tvw), column);

		g_signal_connect (column, "clicked",
		 G_CALLBACK (_e2_fileview_column_header_clicked_cb), view);
		 
		//NB undocumented ->button
		//workaround for detecting header clicks to distinguish from drags
		g_signal_connect (column->button, "button_press_event",
		 G_CALLBACK (_e2_fileview_column_header_buttonpress_cb), view);
 		g_signal_connect (column->button, "button_release_event",
		 G_CALLBACK (_e2_fileview_column_header_buttonrel_cb), view); 
	}
	//setup inital sort parameters, ascending name col
//	view->sort_order = GTK_SORT_ASCENDING;
//	gtk_widget_show (view->sort_arrows[0]);
	//inital sort parameters set from cache
	gtk_tree_sortable_set_sort_column_id (sortable,
		view->sort_column, view->sort_order);
	gtk_widget_show (view->sort_arrows[view->sort_column]);

	//FIXME do this once, not for both panes
	button2updir = e2_option_bool_get ("button2-updir");

	g_signal_connect (tvw, "columns-changed",
	 G_CALLBACK(_e2_fileview_col_change_cb), view);

	//for E2_ALTLEFTMOUSE, double-clicks are handled in the click callback
	//BUT gtk issues this signal when Enter is pressed during a type-ahead
	//search, which we want
	g_signal_connect (tvw, "row-activated",
	 G_CALLBACK (_e2_fileview_row_activated_cb), view);

	//by default, type-ahead searching is enabled on column 0
	gtk_tree_view_set_search_equal_func	(GTK_TREE_VIEW (tvw),
		(GtkTreeViewSearchEqualFunc)_e2_fileview_match_filename, view, NULL);
	//DnD connections
//	gtk_drag_source_set (tvw, GDK_BUTTON1_MASK, target_table, n_targets,
//		GDK_ACTION_COPY); //can't use 2 of this fn, it seems
	gtk_drag_source_set (tvw, GDK_BUTTON1_MASK | GDK_BUTTON2_MASK, target_table, n_targets,
		GDK_ACTION_COPY | GDK_ACTION_MOVE | GDK_ACTION_LINK | GDK_ACTION_ASK);
	gtk_drag_dest_set (tvw, GTK_DEST_DEFAULT_MOTION |
		GTK_DEST_DEFAULT_DROP, target_table, n_targets,
//		, target_table, n_targets,
		GDK_ACTION_COPY | GDK_ACTION_MOVE | GDK_ACTION_LINK | GDK_ACTION_ASK);

	g_signal_connect (G_OBJECT (tvw), "drag_begin",
		G_CALLBACK (e2_dnd_drag_begin_cb), view);
	g_signal_connect (G_OBJECT (tvw), "drag_data_get",
		G_CALLBACK (e2_dnd_drag_data_get_cb), view);
	g_signal_connect (G_OBJECT (tvw), "drag_motion",
		G_CALLBACK (e2_dnd_drag_motion_cb), view);
	g_signal_connect (G_OBJECT (tvw), "drag_leave",
		G_CALLBACK (e2_dnd_drag_leave_cb), view);
//needed if GTK_DEST_DEFAULT_DROP not set, above
//	g_signal_connect (G_OBJECT (tvw), "drag_drop",
//		G_CALLBACK (e2_dnd_drag_drop_cb), view);
	g_signal_connect (G_OBJECT (tvw), "drag_data_received",
		 G_CALLBACK(e2_dnd_drag_data_received_cb), view);
//	g_signal_connect (G_OBJECT (tvw), "drag-data-delete",
//		G_CALLBACK (e2_dnd_drag_delete_cb), view);
	g_signal_connect (G_OBJECT (tvw), "button_press_event",
		G_CALLBACK (_e2_fileview_button_press_cb), view);
	g_signal_connect (G_OBJECT (tvw), "button_release_event",
		G_CALLBACK (_e2_fileview_button_release_cb), view);

	//pick up any key-bindings before the general key-press
	//BUT CHECKME order probably does not matter
	gchar *category = g_strconcat(_C(16),".",_C(32),NULL);
	e2_keybinding_register (category, tvw);  //_("general.panes
	g_free (category);

//	g_signal_connect (G_OBJECT (tvw), "popup-menu", handled by keypress cb now
//		G_CALLBACK (_e2_fileview_popup_cb), view);
	g_signal_connect (G_OBJECT (tvw), "key_press_event",
		G_CALLBACK (_e2_fileview_key_press_cb), view);

	g_signal_connect (G_OBJECT (tvw), "cursor-changed",
		G_CALLBACK (_e2_fileview_cursor_change_cb), view);

	GtkWidget *sw = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW(sw),
		GTK_SHADOW_ETCHED_IN);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_placement(GTK_SCROLLED_WINDOW(sw),
		e2_option_int_get ("scrollbar-position"));
	gtk_container_add (GTK_CONTAINER (sw), tvw);

	gtk_widget_show (tvw);
	gtk_widget_show (sw);

	return sw; 
}
