/* $Id: e2_output.c 529 2005-03-25 12:44:57Z tpgww $
 *
 * Copyright (C) 2003-2006 tooar <tooar@gmx.net>
 * Portions copyright (C) 2004 Florian Zähringer <flo.zaehringer@web.de>
 * Portions copyright (C) 1999 Michael Clark.
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/output/e2_output.c 
 * @brief output pane creation and action functions
 *
 * includes actions related to output contents, but not 
 * to the output visibility
 *
 **/

#include "emelfm2.h"
#include <string.h>
#include "e2_output.h"
#include "e2_dialog.h"
//for pane-text activation
#include "e2_task.h"
#include "e2_filetype.h"
//for selection-save
#include "e2_view_dialog.h"

#define E2_PANED_TOLERANCE 10

static void _e2_output_tabchange_cb (GtkNotebook *notebook,
	GtkNotebookPage *page, guint page_num, gpointer user_data);

//static gboolean output_activated = FALSE;

  /*****************/
 /***** utils *****/
/*****************/
/**
 * @brief determine whether a position in the output pane text buffer is presently visible
 *
 * @param text_view the output pane textview widget
 * @param iter pointer to data for the position in @a text_view that is to be checked
 *
 * @return TRUE if @a iter relates to a point in @a text_view that is NOT presently visible
 **/
static inline gboolean _e2_output_iter_offscreen 
	(GtkTextView *text_view, GtkTextIter *iter)
{
	GdkRectangle visible_rect, iter_rect;
	gtk_text_view_get_visible_rect (text_view, &visible_rect);
	gtk_text_view_get_iter_location (text_view, iter, &iter_rect);
	return
	//this makes the pane scroll 1 line too many, but is faster
	 ((iter_rect.y + iter_rect.height) > (visible_rect.y + visible_rect.height)
	//this causes jiggling, due to trailing blank line (\n at line-ends)
	//being re-scrolled off-screeen-bottom
//	 ((iter_rect.y > (visible_rect.y + visible_rect.height)
	|| iter_rect.y < visible_rect.y);
}
/**
 * @brief move the 'page' displayed in the output pane, relative to its text content
 *
 * @param down TRUE to move down, FALSE to move up
 * @param arg string containing a number, the no. of 'moves'
 * @param page TRUE move @a arg 'pages', FALSE move @a arg 'steps'
 *
 * @return
 **/
static void _e2_output_scroll_helper (gboolean down, gchar *arg, gboolean page)
{
	printd (DEBUG, "scroll_helper (down:%d,arg:%s,page:%d)", down, arg, page);
	E2_OutputTabRuntime *rt = &app.tab;
	g_return_if_fail (rt->scroll != NULL);
	gchar *end = NULL;
	gdouble times = 1.0;
	if (arg != NULL)
		times = g_ascii_strtod (arg, &end);
	if (end == arg)
		times = 1.0;
	GtkAdjustment *vadj = gtk_scrolled_window_get_vadjustment 
		(GTK_SCROLLED_WINDOW (rt->scroll));
	gdouble value = gtk_adjustment_get_value (vadj);
	gint inc = page ? vadj->page_increment : vadj->step_increment;
	if (down)
	{
		value += (inc * times);
		if (value > (vadj->upper - vadj->page_size))
			value = vadj->upper - vadj->page_size;
	}
	else
		value -= (inc * times);

	gtk_adjustment_set_value (vadj, value);
}
/**
 * @brief set popup menu position
 *
 * This function is supplied when calling gtk_menu_popup(), to position
 * the displayed menu.
 * set @a push_in to TRUE for menu completely inside the screen,
 * FALSE for menu clamped to screen size
 *
 * @param menu the GtkMenu to be positioned
 * @param x	place to store gint representing the menu left
 * @param y  place to store gint representing the menu top
 * @param push_in place to store pushin flag
 * @param textview output pane widget in focus when the menu key was pressed
 *
 * @return
 **/
static void _e2_output_set_menu_position (GtkWidget *menu,
	gint *x, gint *y, gboolean *push_in, GtkWidget *textview)
{
	gint left, top;
	gtk_window_get_position (GTK_WINDOW (app.main_window),
		&left, &top);
	GtkAllocation alloc = textview->allocation;
	*x = left + alloc.x + alloc.width/2;
	*y = top + alloc.y + alloc.height/2;
	*push_in = TRUE;
}

  /*******************/
 /***** actions *****/
/*******************/
/**
 * @brief print to output pane, with default settings
 *
 * Any escaped newlines are converted to real ones.
 *
 * @param data UNUSED data from the action system
 * @param message string to be printed
 *
 * @return
 **/
static void _e2_output_print_action (gpointer data, gchar *message)
{
	gchar *real_message = e2_utils_str_replace (message, "\\n", "\n");
	e2_output_print (&app.tab, real_message, NULL, FALSE, NULL);
	g_free (real_message);
}
/**
 * @brief print help message to output pane
 *
 * @param data UNUSED data from the action system
 * @param arg string indicating which message is to be displayed
 *
 * @return
 **/
static void _e2_output_help_action (gpointer data, gchar *arg)
{
	if ((arg == NULL) || (*arg == '\0') || g_str_has_prefix (arg, _("commands")))
		e2_command_output_help ();
	else if (g_str_has_prefix (arg, _("keys")))	//this must be translated same as in default aliases
		e2_keybinding_output_help (arg);
}
/**
 * @brief move the displayed text window by @a arg 'steps'
 *
 * @param down TRUE to move down, FALSE to move up
 * @param arg string containing a number, the no. of 'steps' to move
 *
 * @return
 **/
static void _e2_output_scroll_step (gboolean down, gchar *arg)
{
	_e2_output_scroll_helper (down, arg, FALSE);
}
/**
 * @brief move the displayed text window by @a arg 'pages'
 *
 * @param down TRUE to move down, FALSE to move up
 * @param arg string containing a number, the no. of 'pages' to move
 *
 * @return
 **/
static void _e2_output_scroll_page (gboolean down, gchar *arg)
{
	_e2_output_scroll_helper (down, arg, TRUE);
}
/**
 * @brief move the displayed text window to start or end of its content
 *
 * @param down TRUE to move to end, FALSE to move to start
 * @param arg string UNUSED
 *
 * @return
 **/
static void _e2_output_scroll_all (gboolean down, gchar *arg)
{
	E2_OutputTabRuntime *rt = &app.tab;
	g_return_if_fail (rt->scroll != NULL);
	GtkAdjustment *vadj = gtk_scrolled_window_get_vadjustment 
		(GTK_SCROLLED_WINDOW (rt->scroll));
	if (down)
		gtk_adjustment_set_value (vadj, vadj->upper - vadj->page_size);
	else
		gtk_adjustment_set_value (vadj, 0.0);
}
/**
 * @brief clear all content of the output pane
 *
 * @return
 **/
static void _e2_output_clear (void)
{
	//order of things here is to handle clearing when the text buffer is being added to
	GtkTextTagTable *table = gtk_text_buffer_get_tag_table (app.tab.buffer);
	GtkTextBuffer *buffer = gtk_text_buffer_new (table);
	GtkTextIter start;
	gtk_text_buffer_get_start_iter (buffer, &start);
	GtkTextMark *mark = gtk_text_buffer_create_mark (buffer, "internal-end-mark", &start, FALSE);
	WAIT_FOR_EVENTS
	gtk_text_view_set_buffer (app.tab.text, buffer);
	app.tab.buffer = buffer;
	app.tab.mark = mark;
	app.tab.onscreen = TRUE;
	*(curr_tab) = app.tab; //listed tab data needs the updated values too
	g_object_unref (G_OBJECT (buffer));
}
/**
 * @brief update gtk's flag which sets output pane text wrapping
 *
 * This is a hook fn
 *
 * @param pvalue pointerised version of the T/F value to be set
 * @param rt pointer to data struct for output pane
 *
 * @return
 **/
static void _e2_output_set_op_wrap (gpointer pvalue, E2_OutputTabRuntime *rt)
{
//	printd (DEBUG, "_e2_output_set_op_wrap (pvalue:_,rt:_)");
	gint value = GPOINTER_TO_INT (pvalue);
	gint cur = gtk_text_view_get_wrap_mode (rt->text);
	if (cur != value)
		gtk_text_view_set_wrap_mode (rt->text, value);
}
  /**************/
 /**** menu ****/
/**************/
/**
 * @brief save selected output-pane text
 *
 * @param menuitem UNUSED the selected widget, or NULL
 * @param rt runtime struct to work on
 *
 * @return
 **/
/*static void _e2_output_savesel_cb (GtkWidget *menuitem,	E2_OutputTabRuntime *rt)
{
	e2_edit_dialog_save_selected (rt->buffer, app.main_window);
} */
/**
 * @brief save selected output-pane text
 *
 * @param menuitem UNUSED the selected widget, or NULL
 * @param rt UNUSED runtime struct to work on
 *
 * @return
 **/
static void _e2_output_edit_cb (GtkWidget *menuitem, E2_OutputTabRuntime *rt)
{
	e2_edit_dialog_create (NULL);
}
/**
 * @brief construct and show output-pane context menu
 *
 * @param textview the textview widget where the click happened
 * @param event_button which mouse button was clicked (0 for a menu key)
 * @param event_time time that the event happened (0 for a menu key)
 * @param rt runtime struct to work on
 *
 * @return
 **/
static void _e2_output_show_context_menu (GtkWidget *textview,
	guint event_button, gint event_time, E2_OutputTabRuntime *rt)
{
	gchar *item_name;
	GtkWidget *menu = gtk_menu_new ();
	item_name = g_strconcat (_A(8),".",_A(26),NULL);
	e2_menu_add_action (menu, _("_Hide"), "output_hide_"E2IP".png",
		_("Do not show the output pane"),
		item_name, "1", NULL);  //no string translation
	g_free(item_name);
	item_name = g_strconcat (_A(8),".",_A(26),NULL);
	e2_menu_add_action (menu, _("_Toggle full"), NULL,
		_("Toggle output pane size to/from the full window size"),
		item_name, "*,0", NULL);  //no string translation
	g_free(item_name);
	item_name = g_strconcat (_A(8),".",_A(27),NULL);
	e2_menu_add_action (menu, _("_Clear"), GTK_STOCK_CLEAR,
		_("Clear output pane"),
		item_name, NULL, NULL);
	g_free(item_name);
	GtkWidget *item;	// =
/*	e2_menu_add (menu, _("Save as.."), "save_selection_"E2IP".png",	//no suitable mnemonic
		_("Save the selected text"), _e2_output_savesel_cb, rt);
	if (!gtk_text_buffer_get_selection_bounds (rt->buffer, NULL, NULL))
		gtk_widget_set_sensitive (item, FALSE); */
	e2_menu_add (menu, _("_Edit"), "edit_"E2IP".png",
		_("Edit the pane contents"), _e2_output_edit_cb, rt);
	item_name = g_strconcat (_A(8),".",_A(25),NULL);
	e2_menu_add_action (menu, _("_New tab"), GTK_STOCK_ADD,
		_("Add another tab for the output pane"),
		item_name, NULL, NULL);
	g_free(item_name);
	item_name = g_strconcat (_A(8),".",_A(35),NULL);
	item =
	e2_menu_add_action (menu, _("_Remove tab"), GTK_STOCK_REMOVE,
		_("Close this this tab"),
		item_name, NULL, NULL);
	g_free(item_name);
	if (app.tabcount == 1)
		gtk_widget_set_sensitive (item, FALSE);
/*	item_name = g_strconcat (_A(8),".",_A(46),NULL);
	e2_menu_add_action (menu, _("C_ommand help"), GTK_STOCK_HELP,
		_("Show information about using the command line"),
		item_name, "", NULL);
	g_free(item_name);
*/	GtkWidget *submenu = e2_menu_add_submenu (menu, _("_Settings"));
	e2_menu_create_options_menu (GTK_WIDGET (rt->text), submenu,
		rt->opt_wrap, NULL, NULL,
		app.output.opt_show_on_new, NULL, NULL,
		app.output.opt_show_on_focus_in, NULL, NULL,
		app.output.opt_hide_on_focus_out, NULL, NULL,
		app.output.opt_jump, NULL, NULL,
		app.output.opt_jump_follow, NULL, NULL,
		app.output.opt_jump_end, NULL, NULL,
		NULL);
	e2_menu_add_separator (submenu);
	item_name = g_strconcat (_A(2),".",_A(30),NULL);
	e2_menu_add_action (submenu, _("_Other"), GTK_STOCK_PREFERENCES,
		_("Open the configuration dialog at the output options page"),
		item_name, _C(27), //_("output")
		NULL);
	g_free(item_name);
	g_signal_connect (G_OBJECT (menu), "selection-done",
		G_CALLBACK (e2_menu_destroy_cb), NULL);
	if (event_button == 0)
		gtk_menu_popup (GTK_MENU (menu), NULL, NULL,
			(GtkMenuPositionFunc) _e2_output_set_menu_position,
			textview, event_button, event_time);
	else
		//this was a button-3 click
		gtk_menu_popup (GTK_MENU (menu), NULL, NULL,
			NULL, NULL, event_button, event_time);
}

  /*********************/
 /***** callbacks *****/
/*********************/

/**
 * @brief process an 'intercepted' double-click on the output pane
 * This handles 'activated' text in the output pane.
 *
 * @param rt runtime struct for the pane
 *
 * @return TRUE if the press was handled
 **/
static gboolean _e2_output_activated_cb (E2_OutputTabRuntime *rt)
{
	printd (DEBUG, "output activated cb");
	GtkTextIter start, end;
	if (gtk_text_buffer_get_selection_bounds (rt->buffer, &start, &end))
	{
		gchar *seltext = gtk_text_buffer_get_text (rt->buffer, &start, &end, FALSE);
		printd (DEBUG, "output text is %s", seltext);
		FileInfo info;
		if (!e2_fs_stat (seltext, &info.statbuf))
		{
			gchar *name = g_path_get_basename (seltext);
			gchar *local = F_FILENAME_TO_LOCALE (name);
			g_strlcpy (info.filename, local, sizeof (info.filename));
			F_FREE (local);
			//can't use task backend function here, as that relies on item(s)
			//selected in active filepane
			gchar *ext, *freeme, *action;
			if (e2_fs_is_dir (&info))
				freeme = ext = g_strconcat (".", _("<directory>"), NULL);
			else if (e2_fs_is_executable (&info))
				freeme = ext = g_strconcat (".", _("<executable>"), NULL);
			else
			{
				freeme = NULL; //for this extension, no free needed
				ext = strchr (name, '.');	//assumes '.' is ascii
			}
			if (ext == NULL) //no extension
			{
				if (e2_fs_is_executable (&info))
				{
					gchar *path = g_path_get_dirname (seltext);
					if (g_str_equal (path, ".")
						&& ! g_str_has_prefix (name, E2_COMMAND_PREFIX))
							action = g_strconcat (E2_COMMAND_PREFIX, name, NULL);
					else
						action = name;
					e2_command_run (action, E2_COMMAND_RANGE_DEFAULT);
					if (action != name)
						g_free (action);
					g_free (path);
					ext = (gchar *)0x2;	//prevent dialog
				}
				else if (e2_fs_is_text (info.filename))
				{
					e2_task_backend_view (seltext);
					ext = (gchar *) 0x2;
				}
				else
					ext = (gchar *)0x1;	//force other dialog
			}
			else if (ext == name) //it's a dot file
			{
				if (e2_fs_is_text (info.filename))
					e2_task_backend_view (seltext);
				else
					ext = (gchar *)0x1;	//force other dialog
			}
			else
			{ // the file has an extension
				do
				{
					NCHR(ext); //skip the . prefix
					action = e2_filetype_get_default_action (ext);
					if (action != NULL)
					{
						gchar *command;
						if ((command = strchr (action, '@')) != NULL)	//if always ascii @, don't need g_utf8_strchr()
							command++;	//get the real action
						else
							command = action;
						//leading ! blocks wildcards or %f etc in command interpreter
						command = g_strconcat ("!", command, " ", seltext, NULL);
						e2_command_run (command, E2_COMMAND_RANGE_DEFAULT);
						g_free (command);
						g_free (action);
						break;
					}
				} while ((ext = strchr (ext, '.')) != NULL);	//if always ascii '.', don't need g_utf8_strchr()

				if (freeme != NULL)
					g_free (freeme);
			}

			if (ext == NULL) //no matching extension
				//CHECKME what if we add a filetype when seltext is a full path ?
				e2_filetype_dialog_create (seltext, TRUE); //enter filetype, command or view
			else if (ext == GINT_TO_POINTER (1)) //no extension or hidden file
				e2_filetype_dialog_create (seltext, FALSE); //enter command or view

			g_free (name);
		}
		g_free (seltext);
//BAD - this locks text DnD on
//		output_activated = TRUE;	//prevent normal button-release handling
		return TRUE;	//block internal click handling
	}
	printd (DEBUG, "nothing selected in output");
//	output_activated = FALSE;	//allow normal button-release handling
	return FALSE;
}
/**
 * @brief process mouse button-click in the output pane
 *
 * 1/L = focus, 2/M = hide, 3/R = context menu
 * This also detects left-button double-clicks, for which there is no API
 *
 * @param textview the widget where the click happened
 * @param event gdk event data struct
 * @param rt runtime struct to work on
 *
 * @return TRUE if the signal has been handled (button 2 or 3)
 **/
static gboolean _e2_output_button_press_cb (GtkWidget *textview,
	GdkEventButton *event, E2_OutputTabRuntime *rt)
{
/*	GtkTextWindowType wtype = gtk_text_view_get_window_type (
		GTK_TEXT_VIEW (textview), event->window);
	if (wtype != GTK_TEXT_WINDOW_TEXT)
		return FALSE;
*/
	printd (DEBUG, "output button press cb");
	/*for double-clicks, the callback sequence is:
	1, then another < click interval, then 1 more with 0 click interval
	for triple-clicks, the callback sequence is:
	1, then 2 * last 2 of double click sequence = 5 total
	The 0-click interval clicks are of type GDK_2BUTTON_PRESS or
	GDK_3BUTTON_PRESS as appropriate
	*/
	if (event->button == 1 && event->type == GDK_BUTTON_PRESS)
	{
		static GtkTextIter start, end;
		extern guint click_interval;
		static guint32 last_event_time = 0;
		guint32 interval = event->time - last_event_time;
		last_event_time = event->time;
//		printd (DEBUG, "output button press interval %d", interval);
		//we don't want to repeat the action, if there was a double/triple click
		if (interval >= click_interval)
		{

//			output_activated = FALSE;	//reinstate normal release handling
			//because a selection will be cleared by thge normal release
			//callback, note what is selected in case it's a double-click
			gtk_text_buffer_get_selection_bounds (rt->buffer, &start, &end);

/*			//CHECKME is this sensible ?? it does not focus the commandline
			gchar *action_name = g_strconcat (_A(1),".",_A(39),NULL);  //_("command.focus")	
			e2_action_run_simple (action_name, NULL);
			g_free (action_name);
*/
			gtk_widget_grab_focus (textview);
		}
		else
		{
			//reselect what was de-selected by the intervening release callback,
			//ready for the next press callback (type GDK_2BUTTON_PRESS etc)
			gtk_text_buffer_select_range (rt->buffer, &start, &end);
		}
	}
	else if (event->button == 1 &&
	//		(
			event->type == GDK_2BUTTON_PRESS //|| event->type == GDK_3BUTTON_PRESS)
			)
	{
		printd (DEBUG, "output button press event type %d", event->type);
		return (_e2_output_activated_cb (rt));
		//FIXME stop the selected text from being de-selected in the
		//button-release callback - but we can't just block that !!
	}
	else if (event->button == 2)
	{
		e2_window_output_hide (NULL, NULL);
		return TRUE;
	}
	else if (event->button == 3)
	{
		_e2_output_show_context_menu (textview, event->button, event->time, rt);
		return TRUE;
	}
	return FALSE;
}
/**
 * @brief process mouse button-release in the output pane
 *
 * This is essentially to re-select text which is de-selected by
 * the standard release callback
 *
 * @param textview the widget where the release happened
 * @param event gdk event data struct
 * @param rt runtime struct to work on
 *
 * @return TRUE if the signal has been handled
 **/
/*static gboolean _e2_output_button_release_cb (GtkWidget *textview,
	GdkEventButton *event, E2_OutputTabRuntime *rt)
{
	printd (DEBUG, "output button release cb");
	return FALSE;
	gtk_text_buffer_select_range (rt->buffer, &start, &end);
	return TRUE;
} */
/**
 * @brief process menu button press in the output pane
 *
 * @param widget the textview widget where the press happened
 * @param rt output runtime struct to work on
 *
 * @return TRUE if the signal has been handled
 **/
static gboolean _e2_output_popup_menu_cb (GtkWidget *widget, E2_OutputTabRuntime *rt)
{
	gint event_time = gtk_get_current_event_time ();
	_e2_output_show_context_menu (widget, 0, event_time, rt);
	return TRUE;
}
/**
 * @brief Determine whether the buffer end-mark is on-screen, after manual change to the vertical adjustment 
 *
 * This is a callback for manual changes to the output textview vertical adjustment
 *
 * @param adjust UNUSED the vertical adjustment for the textview
 * @param rt data struct for the output pane
 *
 * @return
 **/
static void _e2_output_scrolled_cb (GtkAdjustment *adjust, E2_OutputTabRuntime *rt)
{
	//FIXME make this faster - local copy of the option value ?
	GtkTextIter iter;
	if (e2_option_bool_get_direct (app.output.opt_jump_end))
		//scrolling only when new output belongs to the last context in the textbuffer
		gtk_text_buffer_get_end_iter (rt->buffer, &iter);
	else
		gtk_text_buffer_get_iter_at_mark (rt->buffer, &iter, rt->mark);

	if (_e2_output_iter_offscreen (rt->text, &iter))
	{
		if (rt->onscreen)
			printd (NOTICE, "not following anymore");
		rt->onscreen = FALSE;
	}
	else
	{
		if (!rt->onscreen)
			printd (NOTICE, "following again");
		rt->onscreen = TRUE;
	}
}
/*
static gboolean test_cb (GtkWidget *widget, GdkEventMotion *event)
{
	int x, y;
	GdkModifierType state;
	if (event->is_hint)
	{
	        gdk_window_get_pointer(event->window, &x, &y, &state);
	        printf("x = %d y = %d\n", x, y);
	}
	else 
	{
	        printf("x: %lf y: %lf\n", event->x, event->y);
	}
	return FALSE;  
} 
*/
  /******************/
 /***** public *****/
/******************/


/**
 * @brief show error message @a msg in output pane, with beep
 *
 * @param msg message
 *
 * @return 
 **/
void e2_output_print_error (gchar *msg, gboolean freemsg)
{
	e2_output_print (curr_tab, msg, NULL, TRUE, E2_ERRORTAGS, NULL);
	if (freemsg)
		g_free (msg);
	GdkDisplay *display =
		gdk_display_manager_get_default_display (gdk_display_manager_get());
	gdk_display_beep (display);
}
/**
 * @brief show system error message in output pane, with beep
 *
 * @return 
 **/
void e2_output_print_errno (void)
{
	e2_output_print (curr_tab, (gchar *) g_strerror (errno),
		NULL, TRUE, E2_ERRORTAGS, NULL);
	GdkDisplay *display =
		gdk_display_manager_get_default_display (gdk_display_manager_get());
	gdk_display_beep (display);
}
/**
 * @brief show @a msg in output pane
 *
 * @a msg is converted to utf-8 if it's not that form already
 * Provides special handling of any '\b', '\r' char(s) in @a msg
 * for an error message, style parameters are ignored
 *
 * @param msg actual message
 * @param origin context to which the message belongs or NULL for default
 * @param error TRUE if @a msg is an error message
 * @param debug TRUE if @a msg is a debug message
 * @param newline TRUE if @a msg is to be printed with pre- and post- newline
 * @param first_tag one or more tags to apply, terminated by NULL
 *
 * @return 
 **/
void e2_output_print (E2_OutputTabRuntime *tab,
	gchar *msg, gchar *origin, //gboolean error, gboolean debug,
		gboolean newline, const gchar *first_tag, ...)
{
	//early exit
	if (msg == NULL) return;

//	printd (DEBUG, "e2_ouput_print (msg:,origin:%s,error:%d,debug:%d,newline:%d,first_tag:)",
//		origin, error, debug, newline);

	//if this is not a debug message
	//show the output pane if the user wishes that
	if (//!debug &&
		!app.output.visible && e2_option_bool_get ("show-output-window-on-output"))
	{
//		gdk_threads_enter ();
		e2_window_output_show (NULL, NULL);
//		gdk_threads_leave ();
	}

	//parse the msg string to handle special characters

	gint len = strlen (msg);
	//process leading backspaces
	//these delete characters from the last message of this context
	gint back = 0;
	while (((gint) msg[0]) == '\b')
	{
		back++;
		msg++;
		len--;
	}

	//FIXME: handle \r correctly
	//flag whether there's a carriage return '\r' in the message
	gboolean line_back = FALSE;
	//flag for backspaces not at the beginning of the message
	gboolean backspaces_left = FALSE;
	//flag whether message variable should be freed
	gboolean free_message = FALSE;
	gint i;

	//FIXME make this faster !!
	for (i = 0; i < len; i++)
	{
//		printd (DEBUG, "%c - %d", msg[i], (gint) msg[i]);
/* no point in handling just this escape code ...
		//escape
		if (((gint) msg[i]) == 27)
		{
			//clear sequence: ESC[2J
			if (((i + 3) < len) && (msg[i + 1] == '[') &&
				(msg[i + 2] == '2') && (msg[i + 3] == 'J'))
			{
				_e2_output_clear ();
				return;
			}
			
		}
		else */
		//carriage return
			if (((gint) msg[i]) == '\r')
			line_back = TRUE;
		//backspace
		else if (((gint) msg[i]) == '\b')
		{
			//yes, there are backspaces in the middle
			backspaces_left = TRUE;
			gint j = 1;
			//find a character back in the string that is not
			//already a backspace
			//FIXME handle multi-byte chars
			while (j <= i)
			{
				if (msg[i - j] == '\b')
					j++;
				else
				{
					//overwrite it with backspace (i.e. later replaced)
					msg[i - j] = '\b';
					break;
				}
			}
			//if there is no prior non-backspace character,
			//setup to delete one from the previous message
			if (j > i)
				back++;
		}
	}

	//eliminate any backspace char(s)
	//FIXME handle multi-byte chars
	if (backspaces_left)
	{
		gchar *s1 = msg, *s2;
		while (*s1 != '\0')
		{
			if (*s1 == '\b')
			{
				s2 = s1+1;
				while (*s2 == '\b') s2++;
				gint slide = (msg+len+1-s2);
				memcpy (s1, s2, slide);
				len -= (s2-s1);
			}
			else
				s1++;
		}
	}

	gchar *utf;
	//check if message is already utf8
	if (g_utf8_validate (msg, -1, NULL))
		utf = msg;
	else
	//	utf = e2_utf8_from_locale (msg); 
	{
		//convert to utf before inserting
		GError *error = NULL;
		utf = g_locale_to_utf8 (msg, -1, NULL, NULL, &error);
		if (error != NULL)
		{
			printd (WARN, "locale string to UTF8 conversion failed: %s", error->message);
			g_error_free (error);
			utf = e2_utf8_from_locale_fallback (msg);
		}
	}
	
	//ensure that there's an origin set
	if (origin == NULL)
		origin = "default";	//no translate
	gboolean is_default_origin = g_str_equal (origin, "default");
	//flag whether this origin is new
	gboolean is_new;
	//get the text buffer for the output pane
	GtkTextBuffer *buffer = tab->buffer;
	GtkTextIter start, end;

	//try to get insert mark
	//it exists if there has been output in this origin before
	GtkTextMark *origin_mark = gtk_text_buffer_get_mark (buffer, origin);
	//if not, create it at the end
	if (origin_mark == NULL)
	{
		is_new = TRUE;
		gtk_text_buffer_get_end_iter (buffer, &end);
		origin_mark = gtk_text_buffer_create_mark (buffer, origin, &end, TRUE);
	}
	else
	{
		is_new = FALSE;
		//the default context always outputs to the end of the textview
		//(ie the output is not "glued together"
		if (is_default_origin)
		{
			gtk_text_buffer_get_end_iter (buffer, &end);
			gtk_text_buffer_move_mark (buffer, origin_mark, &end);
		}
		else
			gtk_text_buffer_get_iter_at_mark (buffer, &end, origin_mark);
	}
	//move the scroll helper mark
	//(it is used by the scroll callback to find out if the user is
	//following output)
	gtk_text_buffer_move_mark (buffer, tab->mark, &end);
	gtk_text_buffer_place_cursor (buffer, &end);

	//do we have to overwrite a previous message's characters in front of us?
	//(if a previous message had a carriage return '\r' in it)
	gchar *mark_del_name = g_strconcat (origin, "-del", NULL);	//no translate
	GtkTextMark *mark_del = gtk_text_buffer_get_mark (buffer, mark_del_name);
	if (mark_del != NULL)
	{
		GtkTextIter iter_del;
		gtk_text_buffer_get_iter_at_mark (buffer, &iter_del, mark_del);
		gtk_text_buffer_delete (buffer, &end, &iter_del);
	}

	if (back > 0)
	{
		gint _offset = gtk_text_iter_get_offset (&end);
		GtkTextIter backi;
		gtk_text_buffer_get_iter_at_offset (buffer, &backi, _offset - back);
		gtk_text_buffer_delete (buffer, &backi, &end);
		gtk_text_buffer_get_iter_at_offset (buffer, &end, _offset - back);
	}
	//delete the delete mark, it's only used one time
	if (mark_del != NULL)
		gtk_text_buffer_delete_mark (buffer, mark_del);

	gint lfcount = 0;

	//find out if it is necessary to insert a newline character before
	//the message because it is has a different context from the last one
	//and the last one hasn't printed one yet;
	static gchar *origin_buffer = NULL;
	if (origin_buffer != origin)
	{
		//save the origin of the last message
		origin_buffer = origin;
		if ((!gtk_text_iter_starts_line (&end))
			&& (is_new || is_default_origin))
			lfcount = 1;
	}

	if (newline && (!gtk_text_iter_starts_line (&end)))
	{
		printd (DEBUG, "inserted additional newline because the message requested it");
		lfcount++;
	}

	GtkTextMark *mark_end;
	if (lfcount > 0)
	{
		gtk_text_buffer_insert (buffer, &end, "\n", lfcount);
		mark_end = gtk_text_buffer_get_insert (buffer);
		gtk_text_buffer_get_iter_at_mark (buffer, &end, mark_end);
	}
	//AT LAST, WE PUT IT IN !
	gtk_text_buffer_insert (buffer, &end, utf, -1);
	//get start and end of text just inserted
	gtk_text_buffer_get_iter_at_mark (buffer, &start, origin_mark);
	mark_end = gtk_text_buffer_get_insert (buffer);
	gtk_text_buffer_get_iter_at_mark (buffer, &end, mark_end);
	if (first_tag != NULL)
	{
		//apply style-tags (if any) for added text
		va_list args;
		va_start (args, first_tag);
		const gchar *tag_name = first_tag;
		while (tag_name != NULL)
		{
			GtkTextTag *tag = gtk_text_tag_table_lookup (buffer->tag_table, tag_name);
/* tags are all pre-configured
			if (tag == NULL)
				printd (WARN, "%s: no tag with name '%s'!", G_STRLOC, tag_name);
			else */
				gtk_text_buffer_apply_tag (buffer, tag, &start, &end);
			tag_name = va_arg (args, const gchar*);
		}
		va_end (args);
	}
/*	when there is a context after the current one, the font weight of
	the later one (eg bold) is sometimes (eg current has error then normal)
	used instead of the proper font (normal)
	EVEN IF the inserted text is explicitly set to normal style
	probably a gtk bug as the font color is correct
	the following is a workaround */
	else //if (!gtk_text_iter_is_end (&end))
	{
		//gtk_text_buffer_remove_all_tags (buffer, &start, &end);
		GSList *tags = gtk_text_iter_get_tags (&start);
		if (tags != NULL)
		{
			gtk_text_buffer_remove_all_tags (buffer, &start, &end);
			g_slist_free (tags);
		}
	}

	if (newline)
	{	//trailing \n
		gtk_text_buffer_insert (buffer, &end, "\n", 1);
		mark_end = gtk_text_buffer_get_insert (buffer);
		gtk_text_buffer_get_iter_at_mark (buffer, &end, mark_end);
	}

	//scroll down if necessary
	if (//scrolling to new output is in effect
		e2_option_bool_get_direct (app.output.opt_jump)
		&& (//manual scrolling has not moved the insert mark off-screen
			tab->onscreen
			//scroll regardless of whether manually scrolled away
			|| !e2_option_bool_get_direct (app.output.opt_jump_follow))
		&& (
			//added text is at the end of the buffer
			gtk_text_iter_is_end (&end)
			//don't care whether added text is at the end of the buffer
			|| !e2_option_bool_get_direct (app.output.opt_jump_end))
			//the end of the added text is not visible already
		&& _e2_output_iter_offscreen (tab->text, &end))
			gtk_text_view_scroll_to_mark (tab->text, mark_end, 0.0, TRUE, 0.1, 1.0);

	//if there was a carriage return, save a delete mark for the next message
	if ((line_back) && (!is_new))
		//CHECKME origin_mark moved in this case too ?
		gtk_text_buffer_create_mark (buffer, mark_del_name, &end, TRUE);
	else
		gtk_text_buffer_move_mark (buffer, origin_mark, &end);

	if (utf != msg)
		g_free (utf);
	if (free_message)
		g_free (msg);
	g_free (mark_del_name);

//	gdk_threads_leave ();
}
/**
 * @brief update several output pane textview settings to conform to current config parameters
 *
 * @param rt runtime struct to work on
 *
 * @return
 **/
void e2_output_update_style (void)
{
	const gchar *fntname = e2_utils_get_output_font ();
	PangoFontDescription *font_desc = pango_font_description_from_string
			(fntname);
	app.output.font_size = pango_font_description_get_size (font_desc);

	GList *member;
	for (member = app.tabslist; member != NULL; member = member->next)
	{
		GtkTextView *tvw = ((E2_OutputTabRuntime *)member->data)->text;
		gtk_text_view_set_wrap_mode (tvw,
			e2_option_int_get ("output-wrap-mode")); 
		gtk_text_view_set_left_margin (tvw,
			e2_option_int_get ("output-left-margin"));
		gtk_text_view_set_right_margin (tvw,
			e2_option_int_get ("output-right-margin"));
		gtk_widget_modify_font (GTK_WIDGET (tvw), font_desc);

		GtkTextBuffer *buf = ((E2_OutputTabRuntime *)member->data)->buffer;
		GtkTextTagTable *table = gtk_text_buffer_get_tag_table (buf);
		GtkTextTag *tag = gtk_text_tag_table_lookup (table, "green");
		g_object_set (G_OBJECT (tag), "foreground",
			e2_option_str_get ("color-positive"), NULL);
		tag = gtk_text_tag_table_lookup (table, "red");
		g_object_set (G_OBJECT (tag), "foreground",
			e2_option_str_get ("color-negative"), NULL);
		tag = gtk_text_tag_table_lookup (table, "grey");
		g_object_set (G_OBJECT (tag), "foreground",
			e2_option_str_get ("color-unimportant"), NULL);
		tag = gtk_text_tag_table_lookup (table, "small");
		g_object_set (G_OBJECT (tag), "size",
			(gint) (PANGO_SCALE_SMALL * app.output.font_size), NULL);
	}
	pango_font_description_free (font_desc);
}
/**
 * @brief create a new textview and buffer and some basic tags
 *
 * @param rt runtime struct to work on
 *
 * @return scrolled window containing a textview
 **/
static GtkWidget *_e2_output_create_view (E2_OutputTabRuntime *rt)
{
	//visible flag is set elswhere, depending on cache data
	//init some vars
	//at the outset, assume that the current content is on-screen
	rt->onscreen = TRUE;
	rt->mark = NULL;
	//create scrolled window
	rt->scroll = e2_widget_get_sw
		(GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (rt->scroll),
		GTK_SHADOW_OUT);

	//create text view
	rt->text = GTK_TEXT_VIEW (gtk_text_view_new ());
	gtk_container_add (GTK_CONTAINER (rt->scroll), GTK_WIDGET (rt->text));
	gtk_text_view_set_editable (rt->text, FALSE);
	gtk_text_view_set_cursor_visible (rt->text, FALSE);
//	allow focus so popup menu signal can happen, & can select text in the pane
//	GTK_WIDGET_UNSET_FLAGS (rt->text, GTK_CAN_FOCUS);

	gtk_text_view_set_wrap_mode (rt->text,
		e2_option_int_get ("output-wrap-mode")); 
	gtk_text_view_set_left_margin (rt->text,
		e2_option_int_get ("output-left-margin"));
	gtk_text_view_set_right_margin (rt->text,
		e2_option_int_get ("output-right-margin"));
	const gchar *fntname = e2_utils_get_output_font ();
	PangoFontDescription *font_desc = pango_font_description_from_string
			(fntname);
	gtk_widget_modify_font (GTK_WIDGET (rt->text), font_desc);
	pango_font_description_free (font_desc);

	//signal used for "links" in the output pane
//	g_signal_connect (G_OBJECT (rt->text), "motion-notify-event",
//		G_CALLBACK (test_cb), NULL);
	gtk_widget_set_events (GTK_WIDGET (rt->text),
		gtk_widget_get_events (GTK_WIDGET (rt->text))
		| GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK);
	g_signal_connect (G_OBJECT (rt->text), "popup_menu",
		G_CALLBACK (_e2_output_popup_menu_cb), rt);
	g_signal_connect (G_OBJECT (rt->text), "button_press_event",
		G_CALLBACK (_e2_output_button_press_cb), rt);
//	g_signal_connect_after (G_OBJECT (rt->text), "button_release_event",
//		G_CALLBACK (_e2_output_button_release_cb), rt);
//YUK no "activate" available, nor so for any ancestor
	g_signal_connect_after (G_OBJECT (rt->text->vadjustment), "value-changed",
		G_CALLBACK (_e2_output_scrolled_cb), rt);
	gtk_widget_show (GTK_WIDGET (rt->text));

	rt->buffer = gtk_text_view_get_buffer (rt->text);

	gtk_text_buffer_create_tag (rt->buffer, "bold", "weight", PANGO_WEIGHT_BOLD, NULL);
	gtk_text_buffer_create_tag (rt->buffer, "italic", "style", PANGO_STYLE_ITALIC, NULL);
	gtk_text_buffer_create_tag (rt->buffer, "uline", "underline", PANGO_UNDERLINE_SINGLE, NULL);
	gtk_text_buffer_create_tag (rt->buffer, "small", "size",
			(gint) (PANGO_SCALE_SMALL * app.output.font_size), NULL);
	gtk_text_buffer_create_tag (rt->buffer, "blue", "foreground", "blue", NULL);
	gtk_text_buffer_create_tag (rt->buffer, "green", "foreground",
		e2_option_str_get ("color-positive"), NULL);
	gtk_text_buffer_create_tag (rt->buffer, "red", "foreground",
		e2_option_str_get ("color-negative"), NULL);
	gtk_text_buffer_create_tag (rt->buffer, "grey", "foreground", 
		e2_option_str_get ("color-unimportant"), NULL);
/* these not needed unless output links are parsed
	gtk_text_buffer_create_tag (rt->tab.buffer, "link", "foreground", "blue",
		"underline", PANGO_UNDERLINE_NONE, NULL);
	gtk_text_buffer_create_tag (rt->tab.buffer, "link-active", "foreground", "blue",
		"underline", PANGO_UNDERLINE_SINGLE, NULL); */

	GtkTextIter end;
	gtk_text_buffer_get_end_iter (rt->buffer, &end);
	rt->mark = gtk_text_buffer_create_mark (rt->buffer,
		"internal-end-mark", &end, FALSE);

	//attach options
	rt->opt_wrap = e2_option_attach_value_changed ("output-wrap-mode",
		GTK_WIDGET (rt->text), _e2_output_set_op_wrap, rt);

	gtk_widget_show (rt->scroll);
	return rt->scroll;
}
/**
 * @brief create an output pane
 *
 * Create initial output pane data - with initial no. of panes as per cache
 *
 * @return tabbed notebook, each page containing an output pane
 **/
GtkWidget *e2_output_tabbed_initialise (void)
{
	//create notebook for tab(s), no callback until all pages created
	app.outbook = gtk_notebook_new ();
	GtkNotebook *book = GTK_NOTEBOOK (app.outbook);
	gtk_notebook_set_show_tabs (book, (app.tabcount > 1));
	gtk_notebook_popup_enable (book);
	gtk_notebook_set_tab_pos (book, GTK_POS_LEFT);
	gtk_notebook_set_scrollable (book, TRUE);
	gtk_notebook_set_show_border (book, FALSE);

	const gchar *fntname = e2_utils_get_output_font ();
	PangoFontDescription *font_desc = pango_font_description_from_string
			(fntname);
	app.output.font_size = pango_font_description_get_size (font_desc);
	pango_font_description_free (font_desc);

	printd (DEBUG, "stacked tab is at %x", &app.tab);
	//we're always going to have at least 1 tab, which gets the initial focus
	gint i;
	E2_OutputTabRuntime *tab; //ignore compiler warning about unitialized usage
	//iterate backward so that we end with the first (default) tab
	for (i = app.tabcount ; i > 0 ; i--)
	{
		GtkWidget *sw = _e2_output_create_view (&app.tab);
		tab = (E2_OutputTabRuntime *) g_malloc (sizeof (E2_OutputTabRuntime));
		printd (DEBUG, "created tab at %x", tab);
		*tab = app.tab;	//copy across for the save
		app.tabslist = g_list_prepend (app.tabslist, tab);
		gchar *txt = g_strdup_printf ("%d", i);	//no translation
		GtkWidget *label = gtk_label_new (txt);
		gtk_notebook_prepend_page (book, sw, label);
		g_free (txt);
	}
	g_signal_connect (G_OBJECT(app.outbook), "switch-page",
		G_CALLBACK (_e2_output_tabchange_cb), NULL);	//CHECKME data ?

	//last-added one becomes current
	//its page is set when the window is shown
	curr_tab = tab;

	//setup data common to all tabs
	//repetitive, but option data structs may move after options hash recreation,
	//and possibly without re-running the options init function
	app.output.opt_show_on_new = e2_option_get ("show-output-window-on-output");
	app.output.opt_show_on_focus_in = e2_option_get ("command-line-show-output-on-focus-in");
	app.output.opt_hide_on_focus_out = e2_option_get ("command-line-hide-output-on-focus-out");
	app.output.opt_jump = e2_option_get ("output-jump-new");
	app.output.opt_jump_follow = e2_option_get ("output-jump-new-following");
	app.output.opt_jump_end = e2_option_get ("output-jump-new-end");

	return app.outbook;
}
/**
 * @brief create another output pane tab, and go there
 *
 * @return
 **/
static void _e2_output_tab_add (void)
{
	E2_OutputTabRuntime savetab = app.tab;
	GtkWidget *sw = _e2_output_create_view (&app.tab);
	E2_OutputTabRuntime *tab = (E2_OutputTabRuntime *) g_malloc0 (sizeof (E2_OutputTabRuntime));
	*tab = app.tab;
	app.tab = savetab;	//reinstate current, ready for switch
	app.tabslist = g_list_append (app.tabslist, tab);
	//there may be gaps in the tab labels, so find the last one and bump its label
	GtkWidget *wid = gtk_notebook_get_nth_page (GTK_NOTEBOOK (app.outbook),
		app.tabcount-1);
	wid = gtk_notebook_get_tab_label (GTK_NOTEBOOK (app.outbook), wid);
	gchar *txt = (gchar *) gtk_label_get_text (GTK_LABEL (wid));
	gint new = atoi (txt) + 1;
	txt = g_strdup_printf ("%d", new);	//no translation
	wid = gtk_label_new (txt);
	g_free (txt);
	gtk_notebook_append_page (GTK_NOTEBOOK (app.outbook), sw, wid);
	if (++app.tabcount == 2)	//log, for cache
		gtk_notebook_set_show_tabs (GTK_NOTEBOOK (app.outbook), TRUE);
	//focus the new tab
	gtk_notebook_set_current_page (GTK_NOTEBOOK (app.outbook), app.tabcount-1);
}
/**
 * @brief remove currently focused output pane tab
 * This may be called from a keybinding or context menu item
 * Essentially, tab-specific data are swpped between stack and heap space,
 * and pointers for affected running commands are adjusted accordingly
 *
 * @return
 **/
static void _e2_output_tab_remove (void)
{
	//ignore removal of only tab
	if (app.tabcount == 1) return;
	//cleanup the notebook
	gint index = gtk_notebook_get_current_page (GTK_NOTEBOOK (app.outbook));
	gtk_notebook_remove_page (GTK_NOTEBOOK (app.outbook), index);
	if (--app.tabcount == 1)
		gtk_notebook_set_show_tabs (GTK_NOTEBOOK (app.outbook), FALSE);

	GList *thismember = g_list_nth (app.tabslist, index);
	GList *newmember;
	if (index == 0)
		newmember = thismember->next;
	else
		newmember = thismember->prev;
	E2_OutputTabRuntime *newtab = newmember->data;
//	printd (DEBUG, "curr_tab is %x", curr_tab);
//	printd (DEBUG, "replacement tab at %x", newtab);
	//quickly install replacement data, ready for use by any printing children
	app.tab = *newtab;
	//anthing running in the focused tab stays there,
	//so no need to update foreground pointers
	//but do update background-tab ptrs of children using the tab
	e2_command_retab2_children (thismember->data, newtab); //!= curr_tab?
	//now its ok to update the last pointer
	curr_tab = newtab;
	//clean the old tab's data FIXME any leak ? text buffer ?
	GHashTable *hash = ((E2_OutputTabRuntime *)thismember->data)->origins;
	if (hash != NULL)
		g_hash_table_destroy (hash);
//	printd (DEBUG, "removing tab at %x", thismember->data);
	g_free (thismember->data);
	app.tabslist = g_list_delete_link (app.tabslist, thismember);
}
/**
 * @brief handle an output-pane tab change
 * This is callback for the notebook's "switch-page" signal
 * Essentially, tab-specific data are swpped between stack and heap space,
 * and pointers for affected running commands are adjusted accordingly
 * @param notebook UNUSED the notebook widget
 * @param page UNUSED the notebook page which is now focused
 * @param page_num the index of the new page
 * @param user_data UNUSED pointer to data specified when notebook was created
 *
 * @return
 **/
static void _e2_output_tabchange_cb (GtkNotebook *notebook,
	GtkNotebookPage *page, guint page_num, gpointer user_data)
{
//	printd (DEBUG, "output tab change cb, NEW page %d", page_num);
//	printd (DEBUG, "curr_tab is %x", curr_tab);
	//backup current tab's data from stack to heap
	*curr_tab = app.tab;
	//get the replacement stuff into stackspace
	curr_tab = g_list_nth_data (app.tabslist, page_num);
	app.tab = *curr_tab;
//	printd (DEBUG, "curr_tab NOW is %x", curr_tab);
	//adjust all relevant child foreground-tab pointers to/from the stacked tab data
	e2_command_retab_children (&app.tab, curr_tab);
}
/**
 * @brief register actions related to output pane
 *
 * @return
 **/
void e2_output_actions_register (void)
{
	gchar *action_name = g_strconcat(_A(8),".",_A(27),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_output_clear, NULL, FALSE);
	action_name = g_strconcat(_A(8),".",_A(63),NULL);
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_output_print_action, NULL, TRUE);
	action_name = g_strconcat(_A(8),".",_A(46),NULL);  //output.help
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_output_help_action, NULL, TRUE);
	action_name = g_strconcat(_A(8),".",_A(69),NULL);  //output.scrolldown
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_output_scroll_step, GINT_TO_POINTER (TRUE), TRUE);
	action_name = g_strconcat(_A(8),".",_A(70),NULL);  //output.scrollup
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_output_scroll_step, GINT_TO_POINTER (FALSE), TRUE);
	action_name = g_strconcat(_A(8),".",_A(60),NULL); //output.pagedown
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_output_scroll_page, GINT_TO_POINTER (TRUE), TRUE);
	action_name = g_strconcat(_A(8),".",_A(61),NULL); //output.pageup
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,	
		_e2_output_scroll_page, GINT_TO_POINTER (FALSE), TRUE);
	action_name = g_strconcat(_A(8),".",_A(44),NULL); //output.goto bottom
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_output_scroll_all, GINT_TO_POINTER (TRUE), TRUE);
	action_name = g_strconcat(_A(8),".",_A(45),NULL); //output.goto top
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_output_scroll_all, GINT_TO_POINTER (FALSE), TRUE);
	action_name = g_strconcat (_A(8),".",_A(25),NULL); //output.add
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_output_tab_add, NULL, FALSE);
	action_name = g_strconcat (_A(8),".",_A(35),NULL); //output.delete
	e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM,
		_e2_output_tab_remove, NULL, FALSE);	//FIXME data
}
/**
 * @brief register options related to output pane
 *
 * @return
 **/
void e2_output_options_register (void)
{
	gchar *group_name = g_strconcat(_C(5),".",_C(27),":",_C(24),NULL); //_("commands.output:miscellaneous"
	e2_option_bool_register ("show-output-window-on-output", group_name,
		_("show output pane when a new message appears"),
		_("This will ensure you don't miss any messages"), NULL, FALSE,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP);
	e2_option_bool_register ("command-line-show-output-on-focus-in", group_name,
		_("show output pane if the command line is focused"),
		_("This causes the output pane to be opened when you are about enter a command"),
		NULL, FALSE,
		E2_OPTION_FLAG_ADVANCED);
	e2_option_bool_register ("command-line-hide-output-on-focus-out", group_name,
		_("hide output pane if the command line is unfocused"), 
		_("This causes the output pane to be closed when you move focus away from the command line"),
		NULL, FALSE,
		E2_OPTION_FLAG_ADVANCED);
	e2_option_bool_register ("output-jump-new", group_name, _("scroll to new output"),
		_("This will automatically scroll the output pane content, to show new message(s)"), NULL, TRUE,
		E2_OPTION_FLAG_ADVANCED);
	e2_option_bool_register ("output-jump-new-following", group_name, _("only scroll when really following"),
		_("This stops automatic pane scrolling to new output if you've manually scrolled away"),
		"output-jump-new", TRUE,
		E2_OPTION_FLAG_ADVANCED);
	e2_option_bool_register ("output-jump-new-end", group_name, _("only scroll when new output is at the end"),
		_("This stops pane scrolling if a different process has displayed text after the current insert position"),
  		"output-jump-new", FALSE,
		E2_OPTION_FLAG_ADVANCED);
	const gchar *opt_wrap_mode[] = {_("none"), _("everywhere"), _("words"), NULL};
	group_name = g_strconcat(_C(5),".",_C(27),":",_C(37),NULL);	//_("commands.output:style"
	e2_option_sel_register ("output-wrap-mode", group_name, _("line wrap mode"),
		_("If mode is 'none', a horizontal scrollbar will be available. Mode 'words' will only break the line between words"),
		NULL, 2, opt_wrap_mode,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_FREEGROUP | E2_OPTION_FLAG_BUILDALL);
	e2_option_int_register ("output-left-margin", group_name, _("left margin (pixels)"),
		_("This is the left margin between the output-pane edge and the text in it"),
		NULL, 6, 0, 1000,
		E2_OPTION_FLAG_ADVANCED);
	e2_option_int_register ("output-right-margin", group_name, _("right margin (pixels)"),
		_("This is the right margin between the output-pane edge and the text in it"),
		 NULL, 2, 0, 1000,
		E2_OPTION_FLAG_ADVANCED);
	e2_option_bool_register ("custom-output-font",
		group_name, _("use custom font"),
		_("If activated, the font specified below will be used, instead of the theme default"),
		NULL, FALSE,
		E2_OPTION_FLAG_BASIC);  //no rebuild
	e2_option_font_register ("output-font", group_name, _("custom output font"),
		_("This is the font used for text in the output pane"), "custom-output-font", "Sans 10", 	//_I( font name
		E2_OPTION_FLAG_BASIC);  //no rebuild	
}
