/* $Id: e2_plugins.c 1861 2006-04-25 08:27:44Z tpgww $
 *
 * Copyright (C) 2003-2006 tooar <tooar@gmx.net>
 * Portions copyright (C) 1999 Michael Clark.
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "e2_plugins.h"
#include <string.h>
#include "e2_action.h"
#include "e2_output.h"
#include "e2_dialog.h"
#include "e2_filelist.h"

typedef struct
{
	gchar *name;
	guint refcount;
} E2P_Ref;

static GList *action_refs = NULL;
static GList *option_refs = NULL;

/**
 * @brief find member of @a list which matches @a name
 *
 * @param list list of E2P_Ref's to scan
 * @param name name of option or action to match
 *
 * @return the matching list member, or NULL if no match
 **/
static GList *_e2_plugins_find_ref (GList *list, gchar *name)
{
	GList *member;
	for (member = list; member != NULL; member = member->next)
	{
		E2P_Ref *data = member->data;
		if (g_str_equal (data->name, name))
			break;
	}
	return member;
}
/**
 * @brief increase refcount for action/item named @a name
 * A list member is added when needed
 * @param list address of list of E2P_Ref's to scan
 * @param name name of option or action to ref
 *
 * @return the new refcount, or 0 in case of error
 **/
static guint _e2_plugins_ref (GList **list, gchar *name)
{
	guint newrefcount;
	E2P_Ref *data;
	GList *member = _e2_plugins_find_ref (*list, name);
	if (member == NULL)
	{
		data = (E2P_Ref *) g_malloc (sizeof (E2P_Ref));
		if (data != NULL)
		{
			*list = g_list_append (*list, data);
			data->name = g_strdup (name);
			data->refcount = 1;
			return 1;
		}
		return 0;
	}
	else
	{
		data = (E2P_Ref *)member->data;
		newrefcount = data->refcount + 1;
		data->refcount = newrefcount;
		return newrefcount;
	}
}
/**
 * @brief decrease refcount for action/item named @a name
 * The list member is cleared when the new count reaches 0
 * @param list address of list of E2P_Ref's to scan
 * @param name name of option or action to ref
 *
 * @return the new refcount, or -1 in case of error
 **/
static gint _e2_plugins_unref (GList **list, gchar *name)
{
	guint newrefcount;
	E2P_Ref *data;
	GList *member = _e2_plugins_find_ref (*list, name);
	if (member == NULL)
		return -1;
	else
	{
		data = (E2P_Ref *)member->data;
		newrefcount = data->refcount - 1;
		if (newrefcount == 0)
		{
			g_free (data->name);
			*list = g_list_delete_link (*list, member);
		}
		else
			data->refcount = newrefcount;
		return (gint)newrefcount;
	}
}
/**
 * @brief get function address
 *
 * This enables a plugin to get the address of a main-program
 * function that the plugin wishes to use, which allows
 * plugins to be version-independent
 * BUT coverage is too limited
 *
 * @param type integer defining which function to get
 *
 * @return ptr to function sought, or NULL if not found
 **/
/*void *e2_plugins_api_lookup (gint type)
{
	if (type >= E2API_POINTER_COUNT)
		return NULL;
	//these are in the same order as the enerator
	//FIXME add other relevant fns
	gpointer interface [E2API_POINTER_COUNT] = {
		e2_action_register_simple,
		e2_action_unregister,
		e2_action_get,
		e2_option_get,
		e2_option_register,
		e2_filelist_disable_refresh,
		e2_filelist_enable_refresh
	};
	return interface [type];
}
*/
/**
 * @brief obliterate specified plugin
 *
 * The plugin's action is unregistered, any 'unload' function in the plugin
 * is called, then the module is dumped and memory freed
 *
 * @param p plugin data struct
 * @param force TRUE to force removal even if not unloaded
 *
 * @return tRUE if the plugin was unloaded
 **/
gboolean e2_plugins_unload1 (Plugin *p, gboolean force)
{
	printd (DEBUG, "unload plugin: %s", p->menu_name);
	gboolean (*clean)(Plugin *);
	if (g_module_symbol (p->module, "clean_plugin", (void *) &clean))
	{
		if (clean (p) || force)
		{
			g_module_close (p->module);
			g_free (p);
			return TRUE;
		}
		gchar *msg = g_strdup_printf ("%s \"%s\"",
			_("Unable to unload plugin"), p->menu_name);
		e2_output_print_error (msg, TRUE);
	}
	return FALSE;
}
/**
 * @brief unload all plugins
 *
 * Each plugin listed in app.plugins is removed, and the list itself is cleared
 *
 * @return 
 **/
void e2_plugins_unload_all (void)
{
	GList *tmp;

	for (tmp = app.plugins; tmp != NULL; tmp = tmp->next)
	{
		if (tmp->data != NULL)
		{
			Plugin *p = tmp->data;
		//FIXME handle plugins that are not allowed to be unloaded
			e2_plugins_unload1 (p, TRUE);
		}
	}

	g_list_free (app.plugins);
	app.plugins = NULL;
}
/**
 * @brief load a specified plugin, if possible
 *
 * @param filepath path+name string for the plugin
 *
 * @return plugin data struct for the loaded plugin, or NULL if plugin not found or has no initialize fn
 **/
Plugin *e2_plugins_load1 (gchar *filepath)
{
	printd (DEBUG, "load plugin: %s", filepath);

	gboolean (*init)(Plugin *);
	GModule *module = g_module_open (filepath, 0);
	if (module == NULL)
	{
		printd (DEBUG, "failed to load module: %s", g_module_error());
		return NULL;
	}

	if (!g_module_symbol (module, "init_plugin", (void *)&init))
	{
		printd (DEBUG, "couldn't find init_plugin in module: %s", g_module_error());
		return NULL;
	}

	Plugin *p = g_new0 (Plugin, 1);
	p->module = module;
	p->plugin_init = init;

	return p;
}
/**
 * @brief add plugin
 *
 * If the specified plugin file is found (at custom or default path),
 * it is loaded, relevant data is recorded, then the plugin
 * is discarded if it is not wanted.
 *
 * @param model treemodel for the plugins config data
 * @param iter ptr to iter referring to model/store row being processed
 *
 * @return 
 **/
static void _e2_plugins_add1 (GtkTreeModel *model, GtkTreeIter *iter)
{
	void (*unload)(Plugin *);
	gboolean load_this, in_menu;
	gchar *icon, *label, *tip;	//these are for checking if currently in config
	gchar *file_name, *file_path;

	gtk_tree_model_get (model, iter, 0, &load_this, 1, &in_menu,
		2, &label, 3, &icon, 4, &tip,
		5, &file_name, 6, &file_path, -1);

	if (*file_path == '\0')
	//no path specified, so use the default plugins path
		file_path = g_strdup (PLUGINS_DIR);
	//CEHECKME do we ever need to NULL the stored path?
	gchar *local = F_FILENAME_TO_LOCALE (file_path);
	if (e2_fs_is_dir3 (local))
	{
		gchar *freeme = file_path;
		file_path = g_build_filename (file_path, file_name, NULL);
		g_free (freeme);

		if (load_this || in_menu //this is one we want
			|| *label == '\0')	//no data logged CHECKME is this test ok ?
		{
			Plugin *p;
			if ((p = e2_plugins_load1 (file_path)) != NULL)
			{
				if (load_this || in_menu)
				{
	//				p->api_lookup = e2_plugins_api_lookup;
					if (p->plugin_init (p))
					{
						//update the menu & config dialog data if need be
						if (*label == '\0' && p->menu_name != NULL)
							gtk_tree_store_set (GTK_TREE_STORE (model), iter,
								2, g_strdup(p->menu_name), -1);
						else
							p->menu_name = g_strdup (label);
						if (*icon == '\0' && p->icon != NULL)
							gtk_tree_store_set (GTK_TREE_STORE (model), iter,
								3, g_strdup(p->icon), -1);
						else
							p->icon = g_strdup (icon);
						if (*tip == '\0' && p->description != NULL)
							gtk_tree_store_set (GTK_TREE_STORE (model), iter,
								4, g_strdup(p->description), -1);
						else
							p->description = g_strdup (tip);

						p->show_in_menu = in_menu;
						app.plugins = g_list_append (app.plugins, p);
					}
					else
					{
						printd (WARN, "can't initialize plugin: %s", file_name);
						gtk_tree_store_remove (GTK_TREE_STORE (model), iter);
						if (g_module_symbol (p->module, "clean_plugin", (void *) &unload))
						//CHECKME handle case where want to retain ??
							unload(p);
						g_module_close (p->module);
						g_free (p);
					}
				}
				else
				{	//this is one we don't want but we do want its dialog data
					p->plugin_cb = GINT_TO_POINTER (1); //set to non-NULL to just get config data
					//if (
						p->plugin_init (p);
					//)
					//{
						if (p->menu_name != NULL)
							gtk_tree_store_set (GTK_TREE_STORE (model), iter,
								2, g_strdup(p->menu_name), -1);
						if (*icon == '\0' && p->icon != NULL)
							gtk_tree_store_set (GTK_TREE_STORE (model), iter,
								3, g_strdup(p->icon), -1);
						if (*tip == '\0')
						{
							gchar *faketip = (p->description != NULL) ?
								g_strdup(p->description) :
								g_strdup(_("Plugin not loaded"));	//some extra user info
							gtk_tree_store_set (GTK_TREE_STORE (model), iter,
								4, faketip, -1);
						}
						//junk it again, now, if not otherwise wanted
						//e2_plugins_unload1 (p, FALSE);
						g_module_close (p->module);
						g_free (p);
/*					}
					else
					{	//junk unitialised module
						g_module_close (p->module);
						g_free (p);
					} */
				}
			}
			else	//plugin file not loaded
			{
				printd (WARN, "plugin doesn't exist: %s", file_name);
				gtk_tree_store_remove (GTK_TREE_STORE (model), iter);
			}
		}
	}
	else //no valid dir for the plugin
	{
		printd (WARN, "plugin dir doesn't exist: %s", file_path);
		gtk_tree_store_remove (GTK_TREE_STORE (model), iter);
	}
	g_free (icon);
	g_free (label);
	g_free (tip);
	g_free (file_name);
	g_free (file_path);
	F_FREE (local);
}
/**
 * @brief load all known plugins
 *
 * Register the plugins that are marked for loading
 * For the rest, just get data for the config dialog, then discard
 *
 * @return 
 **/
void e2_plugins_add_all (void)
{
	E2_OptionSet *set = e2_option_get_simple ("plugins");
	GtkTreeIter iter;

	if ( (set != NULL) &&
		gtk_tree_model_get_iter_first (set->ex.tree.model, &iter) )
	{
		do
		{
			_e2_plugins_add1 (set->ex.tree.model, &iter);
		} while (gtk_tree_model_iter_next (set->ex.tree.model, &iter));

	} else
		printd (WARN, "plugins config doesn't exist");
}

/**
 * @brief update loaded plugins
 *
 * This is the 'apply' fn for the plugins dialog, also used in general config dialog
 * It performs a differential load/unload relative to the status quo
 * Status quo is determined by checking menu name, which is in both Plugin and model
 * ( so if a user edits the menu name, the plugin will be improperly, but not fatally, reloaded )
 *
 * @return
 **/
void e2_plugins_apply_new (void)
{
	gchar *icon;
	gchar *name;
	gboolean load, menu;
	GList *tmp;
	E2_OptionSet *set = e2_option_get_simple ("plugins");
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter_first (set->ex.tree.model, &iter))
	{
		do
		{
			gtk_tree_model_get (set->ex.tree.model, &iter,
				0, &load, 1, &menu, 2, &name, 3, &icon, -1);
			if (menu && !load) //fix any mistake by user
			{
				gtk_tree_store_set (set->ex.tree.model, &iter,
					0, TRUE, -1);
				load = TRUE;
			}
			//check if this one is already in play
			if (name == NULL)
				tmp = NULL;
			else
			{
				for (tmp = app.plugins; tmp != NULL; tmp = tmp->next)
				{
					if (g_str_equal (name, ((Plugin *)tmp->data)->menu_name))
						break;
				}
			}
			if (load && tmp != NULL)
			{
				//update on-menu status
				((Plugin *)tmp->data)->show_in_menu = menu;
				//update icon
				((Plugin *)tmp->data)->icon = icon;
			}
			else
				if (load)
				//we need to load this one
					_e2_plugins_add1 (set->ex.tree.model, &iter);
			else
				if (!load && (tmp != NULL))
				{	//we need to unload this one if we can
					if (e2_plugins_unload1 (tmp->data, FALSE))
						app.plugins = g_list_delete_link (app.plugins, tmp);
					else
						gtk_tree_store_set (set->ex.tree.model, &iter, 0, TRUE,-1);
				}

			g_free (name);
		} while (gtk_tree_model_iter_next (set->ex.tree.model, &iter));
	}
}
/**
 * @brief get the list of installed plugins
 *
 * This is intended to be used by plugins themselves, in case they
 * want to interact
 *
 * @return pointer to glist, in which each ->data is a Plugin* for a loaded plugin
 **/
GList *e2_plugins_get_list (void)
{
	return app.plugins;
}
/**
* @brief check whether plugin with specified @a signature is loaded
 *
 * This is intended to be used by plugins, if they want to interact
 *
 * @param signature unique name string for the desired plugin
 *
 * @return pointer to data struct for plugin which has the specified ID, else NULL
 **/
Plugin *e2_plugins_check_installed (const gchar *signature)
{
	GList *tmp;
	Plugin *p;
	for (tmp = app.plugins; tmp != NULL; tmp = tmp->next)
	{
		p = tmp->data;
		if (g_str_equal (p->signature, signature))
			return p;
	}
	return NULL;
}
/**
 * @brief find address of a function in another plugin
 *
 * This is intended to be used by plugins, if they want to interact
 *
 * @param signature unique name string for the desired plugin
 * @param func_name the name of the function to find
 * @param address store for the located address
 *
 * @return TRUE if the address is found
 **/
gboolean e2_plugins_find_function (const gchar *signature, const gchar *func_name,
	gpointer *address)
{
	Plugin *p = e2_plugins_check_installed (signature);
	if (p == NULL)
	{
		//find and load the plugin
		E2_OptionSet *set = e2_option_get_simple ("plugins");
		GtkTreeIter iter;
		if ((set != NULL) &&
			gtk_tree_model_get_iter_first (set->ex.tree.model, &iter))
		{
			gboolean loaded;
			do
			{
				gtk_tree_model_get (set->ex.tree.model, &iter, 0, &loaded, -1);
				if (!loaded)
				{
					gtk_tree_store_set (set->ex.tree.model, &iter, 0, TRUE, -1);
					_e2_plugins_add1 (set->ex.tree.model, &iter);
					GList *member = g_list_last (app.plugins);
					p = member->data;
					if (g_str_equal (p->signature, signature))
						break;
					else
					{
						if (e2_plugins_unload1 (p, FALSE))
							gtk_tree_store_set (set->ex.tree.model, &iter, 0, FALSE, -1);
						p = NULL;
					}
				}
			} while (gtk_tree_model_iter_next (set->ex.tree.model, &iter));
			if (p == NULL)
			{
				printd (WARN, "module with %s not found", signature);
				return FALSE;
			}
		}
		else
		{
			printd (WARN, "plugins config doesn't exist");
			return FALSE;
		}
	}
	if (!g_module_symbol (p->module, func_name, address))
	{
		printd (WARN, "couldn't find %s in module %s: %s",
			func_name, signature, g_module_error());
		return FALSE;
	}
	return TRUE;
}
/**
 * @brief run a dialog showing plugins config data
 *
 * @return 
 **/
void e2_plugins_configure (void)
{
	e2_config_dialog_single ("plugins", e2_plugins_apply_new, TRUE); //internal name, no translation
}
/**
 * @brief register plugin action
 *
 * This is a wrapper for the standard action registration function
 * It also manages a refcount for the action
 *
 * @param name freeable string with possible '.' separating 'parent/ and 'child' parts of the action
 * @param type flags for one of: item (?) submenu, file_actions, custom_command, plugins, separator,
	tear_off_menu, toggle, command_line, bookmarks, dummy
 * @param func the function to be called when the action is run
 * @param data pointer to data to be supplied to @a func, or NULL
 * @param has_arg boolean flag signalling whether ??WHAT has argument(s)
 * @param exclude flags indicating whether to exclude this action from toolbars, menus &| keybindings
 * @param data2 additional data needed by the action, or NULL
 *
 * @return registered action E2_Action
 **/
E2_Action *e2_plugins_action_register (gchar *name, E2_ACTION_TYPE type,
	void *func, gpointer data, gboolean has_arg, E2_ACTION_EXCLUDE exclude,
	gpointer data2)
{
	E2_Action *action;
	guint refcount = _e2_plugins_ref (&action_refs, name);
	if (refcount == 1)
		action = e2_action_register (name, type, func, data, has_arg,
			exclude, data2);
	else
		action = e2_action_get (name);
	return action;
}
/**
 * @brief unregister plugin action named @a name
 * This is a wrapper for the standard action deregistration function
 * It also manages a refcount for the action
 * @param name action name string
 *
 * @return TRUE if the action was actually unregistered
 **/
gboolean e2_plugins_action_unregister (gchar *name)
{
	guint refcount = _e2_plugins_unref (&action_refs, name);
	if (refcount == 0)
		return (e2_action_unregister (name));
	else
		return FALSE;
}
/**
 * @brief perform action for a specified plugin
 *
 * @param widget unused
 * @param p plugin data 
 *
 * @return 
 **/
void e2_plugins_do_action (GtkWidget *widget, Plugin *p)
{
	if (p == NULL)
	{
		e2_output_print_error (_("Plugin not found!"), FALSE);
		return;
	}
//#ifdef E2_REFRESH_DEBUG
//	printd (DEBUG, "enable refresh, plugin do action");
//#endif
	// fix refresh ref. count = 1 from context menu setup
//	e2_filelist_enable_refresh ();
	p->plugin_cb(NULL, NULL, NULL);
	//revert the counter, to avoid cleanup problems
//#ifdef E2_REFRESH_DEBUG
//	printd (DEBUG, "disable refresh, plugin do action");
//#endif
//	e2_filelist_disable_refresh ();
}
/**
 * @brief register a plugin config option
 * This is a wrapper for the standard option registration function
 * It also manages a refcount for the option
 *
 * @param type flag for the type of set that it is
 * @param name name of the option, a constant string
 * @param group group the option belongs to, used in config dialog, a r-t string  FREEME
 * @param desc textual description of the option used in config dialog, a r-t _() string FREEME ?
 * @param tip tooltip used when displaying the config dialog, a _() string, or NULL
 * @param depends name of another option this one depends on, or NULL
 *
 * @return the option data struct
 **/
E2_OptionSet *e2_plugins_option_register (E2_OptionType type, gchar *name,
	gchar *group, gchar *desc, gchar *tip, gchar *depends, E2_OptionFlags flags)
{
	E2_OptionSet *set;
	guint refcount = _e2_plugins_ref (&option_refs, name);
	if (refcount == 1)
		set = e2_option_register (type, name, group, desc, tip, depends, flags);
	else
		set = e2_option_get (name);
	return set;
	//NOTE caller needs special care for tree-options
}
/**
 * @brief unregister plugin option named @a name
 * This is a wrapper for the standard option deregistration function
 * It also manages a refcount for the option
 * @param name option name string
 *
 * @return TRUE if the option was actually unregistered
 **/
gboolean e2_plugins_option_unregister (gchar *name)
{
	guint refcount = _e2_plugins_unref (&option_refs, name);
	if (refcount == 0)
		return (e2_option_unregister (name));
	else
		return FALSE;
}
/**
 * @brief install default tree options for plugins
 * This function is called only if the default is missing from the config file
 * This is essentially a list of all the expected plugin filenames
 * the menu name and description at collected from the plugin file
 * the 'path' field at the end is empty if the plugins default dir is to be used
 * @param set pointer to set data
 * @return
 **/
static void _e2_plugins_tree_defaults (E2_OptionSet *set)
{
	e2_option_tree_setup_defaults (set,
	g_strdup("plugins=<"), //internal name
	g_strdup ("true|true||||e2p_glob.so|"),
	g_strdup ("true|true||||e2p_rename_ext.so|"),
	g_strdup ("true|true||||e2p_for_each.so|"),
	g_strdup ("true|true||||e2p_pack.so|"),
	g_strdup ("true|true||||e2p_names_clip.so|"),
	g_strdup ("true|true||||e2p_sort_by_ext.so|"),
	g_strdup ("true|true||||e2p_du.so|"),
	g_strdup ("true|false||||e2p_unpack.so|"),
	g_strdup ("true|true||||e2p_find.so|"),
	g_strdup ("false|false||||e2p_cpbar.so|"),
	g_strdup ("false|false||||e2p_mvbar.so|"),
	g_strdup ("false|false||||e2p_times.so|"),
	g_strdup ("false|false||||e2p_dircmp.so|"),
	g_strdup ("false|false||||e2p_config.so|"),
	g_strdup ("false|false||||e2p_clone.so|"),
	g_strdup ("false|false||||e2p_view.so|"),
	g_strdup(">"),
	NULL);
}
/**
 * @brief register plugins actions
 *
 * @return 
 **/
void e2_plugins_options_register (void)
{
	//no screen rebuilds needed after any change to this option
	gchar *group_name = _C(33); //_("plugins")
	E2_OptionSet *set = e2_option_tree_register ("plugins", group_name,
		group_name, NULL, NULL, NULL,
		E2_OPTION_TREE_LIST | E2_OPTION_TREE_UP_DOWN | E2_OPTION_TREE_ADD_DEL,
		E2_OPTION_FLAG_ADVANCED | E2_OPTION_FLAG_BUILDPLUGS);
	e2_option_tree_add_column (set, _("Loaded"), E2_OPTION_TREE_TYPE_BOOL, TRUE, "true",
		0, NULL, NULL);
	e2_option_tree_add_column (set, _("Menu"), E2_OPTION_TREE_TYPE_BOOL, TRUE, "true",
		0, NULL, NULL);
	e2_option_tree_add_column (set, _("Label"), E2_OPTION_TREE_TYPE_STR, 0, "",
		0, NULL, NULL);
	e2_option_tree_add_column (set, _("Icon"), E2_OPTION_TREE_TYPE_ICON, 0, "",
		0, NULL, NULL);
	e2_option_tree_add_column (set, _("Tooltip"), E2_OPTION_TREE_TYPE_STR, 0, "",
		0, NULL, NULL);
	e2_option_tree_add_column (set, _("Filename"), E2_OPTION_TREE_TYPE_STR, 0, "?.so",
		0, NULL, NULL);
	e2_option_tree_add_column (set, _("Path"), E2_OPTION_TREE_TYPE_STR, 0, "",
		0, NULL, NULL);
	e2_option_tree_create_store (set);

	e2_option_tree_prepare_defaults (set, _e2_plugins_tree_defaults);
}
