/* $Id: e2_task.c 1866 2006-04-26 12:02:29Z tpgww $
 *
 * Copyright (C) 2003-2006 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/* TODO 25/10/04
some functions do not need TreeRowReferences
 (as of now, none do - no on-the-fly store content changes)
when stability confirmed, move surrounding dis/enable refresh to e2_dialog_ow_check() itself ?
*/

#include "e2_task.h"
#include <string.h>
#include "e2_dialog.h"
#include "e2_option.h"
#include "e2_ownership_dialog.h"
#include "e2_filelist.h"

gboolean _e2_task_move (gpointer action_data, gpointer rt_data,
	GtkWidget *widget, gchar *trashpath);

typedef struct _E2_TaskData
{
	E2_TaskType task_func;
	gchar *slocal;
	gchar *dlocal;
	gboolean sametime;
#ifdef E2_INCLIST
	GtkTreeRowReference *ref; //used for post-move deletion
#else
	GtkTreePath *path;	//used for post-move un-selection
#endif
} E2_TaskData;

GMutex *task_mutex = NULL;

//only 1 thread in pool, as no real merit in parallel disk accesses
#define TPOOL_SIZE 1

#ifdef E2_INCLIST
/**
 * @brief 
 *
 * @param ref pointer to reference in current view
 *
 * @return
 */
static void _e2_task_treeview_line_remove (GtkTreeRowReference *ref)
{
	GtkTreeModel *model = GTK_TREE_MODEL (curr_view->store);
	GtkTreeIter iter;
	GtkTreePath *path = gtk_tree_row_reference_get_path (ref);
	if (path != NULL && gtk_tree_model_get_iter (model, &iter, path))
		gtk_list_store_remove (curr_view->store, &iter);
}
/**
 * @brief 
 *
 * @param ref pointer to reference in current view
 * @param newname utf-8 string with replacement name for an item
 *
 * @return
 */
static void _e2_task_treeview_line_rename (GtkTreeRowReference *ref, gchar *newname)
{
	gchar *savename;
	gchar *local = F_FILENAME_TO_LOCALE (newname);
	if (e2_fs_is_dir3 (local))
		savename = g_strconcat (newname, G_DIR_SEPARATOR_S, NULL);		
	else
		savename = newname;

	GtkTreeModel *model = GTK_TREE_MODEL (curr_view->store);
	GtkTreeIter iter;
	GtkTreePath *path = gtk_tree_row_reference_get_path (ref);
	if (path != NULL && gtk_tree_model_get_iter (model, &iter, path))
		gtk_list_store_set (curr_view->store, &iter, FILENAME, savename, -1);

	F_FREE (local);
	if (savename != newname)
		g_free (savename);
}
//FIXME add fn to change permissions field in a row
#endif	//def E2_INCLIST
/**
 * @brief thread-pool function to perform an async move or copy task
 *
 *	As it's async, this must handle any consequences of task 'success'
 * Cleans up @a tdata before finishing
 *
 * @param tdata pointer to task data struct
 * @param setup_data UNUSED pointer to data specified when pool created
 *
 * @return
 */
void _e2_task_threadfunc (E2_TaskData *tdata, gpointer setup_data)
{
	if (tdata->task_func == E2_TASK_COPY)
		e2_task_backend_copy (tdata->slocal, tdata->dlocal, tdata->sametime);
	else if (tdata->task_func == E2_TASK_MOVE)
	{
		if (e2_task_backend_move (tdata->slocal, tdata->dlocal))
#ifdef E2_INCLIST
			_e2_task_treeview_line_remove (tdata->ref);
		gtk_tree_row_reference_free (tdata->ref);
#else
			gtk_tree_selection_unselect_path (curr_view->selection, tdata->path);
		gtk_tree_path_free (tdata->path);
#endif
	}
	g_free (tdata->slocal);
	g_free (tdata->dlocal);
	g_free (tdata);
}
/**
 * @brief depending on @a func, copy or move item @a slocal to @a dlocal
 * @param func code for the function to be used
 * @param slocal localised path string of the item to be processed
 * @param dlocal localised path string of the item's destination
 * @param preserve TRUE to preserve times of copied files
 * @param path treeview path of item to be moved, or NULL if it's to be copied
 *
 * @return FALSE for async tasks, TRUE if sync task competed successfully
 */
gboolean e2_task_do_task (E2_TaskType func, gchar *slocal, gchar *dlocal,
	gboolean preserve,
#ifdef E2_INCLIST
	GtkTreeRowReference *ref)
#else
	GtkTreePath *path)
#endif
{
	if (e2_option_bool_get ("fileop-background"))
	{
		static GThreadPool *tpool = NULL;
		if (tpool == NULL)
			tpool = g_thread_pool_new (
				(GFunc) _e2_task_threadfunc,
				NULL,//pointer, 2nd parameter supplied to GFunc
				TPOOL_SIZE,
				TRUE,	//exclusive
				NULL	//no error reports
			);

		if (task_mutex == NULL)
			task_mutex = g_mutex_new ();	//sometimes a lock is needed
		E2_TaskData *data = g_malloc (sizeof (E2_TaskData));
		data->task_func = func;
		data->slocal = g_strdup (slocal);
		data->dlocal = g_strdup (dlocal);
		data->sametime = preserve;
#ifdef E2_INCLIST
		if (ref != NULL)
			data->ref = gtk_tree_row_reference_copy (ref);
#else
		if (path != NULL)
			data->path = gtk_tree_path_copy (path);
#endif
		g_thread_pool_push (tpool, data, NULL);	//this will handle successful completion
		return FALSE;	//can't yet be sure of success !!
	}
	else	//foreground task
	{
		gboolean success = FALSE;
		if (func == E2_TASK_COPY)
			success = e2_task_backend_copy (slocal, dlocal, preserve);
		else if (func == E2_TASK_MOVE)
		{
			success = e2_task_backend_move (slocal, dlocal);
			if (success)
#ifdef E2_INCLIST
			_e2_task_treeview_line_remove (ref);
#else
			gtk_tree_selection_unselect_path (curr_view->selection, path);
#endif
		}
		return success;
	}
}
/**
 * @brief display error message @a msg, with current system error description
 *
 * @param msg message string to display
 *
 * @return FALSE (==error signal) always
 */
gboolean e2_task_error (gchar *msg)
{
	gchar *long_msg = g_strconcat (msg, " - ",
		(gchar *) g_strerror (errno), NULL);
	e2_output_print_error(long_msg, TRUE);
	return FALSE;
}
/**
 * @brief construct and display error message, with current system error description
 * @param format format string (with a "%s") for the message to display
 * @param local localised string to be inserted in @a format
 * @return FALSE (==error signal) always
 */
gboolean e2_task_error_local (const gchar *format, gchar *local)
{
	gchar *utf = F_FILENAME_FROM_LOCALE (local);
	gchar *msg = g_strdup_printf (format, utf);
	e2_task_error (msg);
	F_FREE (utf);
	g_free (msg);
	return FALSE;
}
#ifdef LATE_DELETE
/**
 * @brief get temp item name if needed
 * @param path localised path of item to check
 * @return pointer to name string, = @a path, or newly-allocated replacement
 */
gchar *e2_task_tempname (gchar *path)
{
	struct stat dest_sb;
	if (e2_fs_lstat (path, &dest_sb))
		return path;
	else
		return (e2_utils_get_tempname (path));
}
#endif
/**
 * @brief copy selected items from active to inactive pane
 *
 * Overwrite checking is performed, if that option is enabled
 * The actual operation is performed by a separate back-end function
 *
 * @param preserve pointerised TRUE or FALSE - TRUE to preserve mtime, atinme of copied items
 * @param rt_data ptr to data sent to this command, "" to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_copy (gpointer preserve, gpointer rt_data, GtkWidget *widget)
{
//	printd (DEBUG, "task: copy");
	if (g_str_equal (curr_view->dir, other_view->dir))
	{
		//display some message
		return FALSE;
	}
	//handle case of specific data instead of selection??
	GPtrArray *names = e2_fileview_get_selected (curr_view);
	if (names == NULL) return FALSE;	//nothing selected
	GString *src = g_string_sized_new (PATH_MAX+NAME_MAX);
	GString *dest = g_string_sized_new (PATH_MAX+NAME_MAX);
	guint count;
	E2_SelectedItemInfo **iterator = (E2_SelectedItemInfo **) names->pdata;
	gboolean check = e2_option_bool_get ("confirm-overwrite");
	//setup for tailoring over-write dialog
	gboolean multisrc =  (check) ? names->len > 1 : FALSE;
	OW_ButtonFlags extras = (multisrc)? BOTHALL : NONE;

	//no window desensitize ...
	e2_widget_set_cursor (GDK_WATCH);
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, copy task");
#endif
	e2_filelist_disable_refresh ();  //prevent pauses in the copy process
	gchar *itemname, *slocal, *dlocal;
	for (count=0; count < names->len; count++, iterator++)
	{
//		if (g_str_equal((*iterator)->filename, "..")) continue;
		itemname = (*iterator)->filename;
		//curr_view->dir, other_view->dir are utf-8
		g_string_printf(src, "%s%s", curr_view->dir, itemname);  //separator comes with dir
		g_string_printf(dest, "%s%s", other_view->dir, itemname);
		slocal = F_FILENAME_TO_LOCALE (src->str);
		dlocal = F_FILENAME_TO_LOCALE (dest->str);

		if (check && e2_fs_access2 (dlocal) == 0)
		{
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, copy dialog");
#endif
			e2_filelist_enable_refresh ();  //allow updates while we wait
			DialogButtons result = e2_dialog_ow_check (slocal, dlocal, extras);
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, copy dialog");
#endif
			e2_filelist_disable_refresh ();
			switch (result)
			{
				case YES_TO_ALL:
					check = FALSE;
				case OK:
					e2_task_do_task (E2_TASK_COPY, slocal, dlocal,
						GPOINTER_TO_INT (preserve), NULL);
				case CANCEL:
					break;
				default:
					result = NO_TO_ALL;
					break;
			}
			if (result == NO_TO_ALL)
			{
				F_FREE (slocal);
				F_FREE (dlocal);
				break;
			}
		}
		else  //file doesn't exist, or don't care
		{
			e2_task_do_task (E2_TASK_COPY, slocal, dlocal,
				GPOINTER_TO_INT (preserve), NULL);
		}
		F_FREE (slocal);
		F_FREE (dlocal);
	}
	e2_fileview_clean_selected (names);
	g_string_free (src, TRUE);
	g_string_free (dest, TRUE);

	//FIXME E2_INCLIST - do other view
	e2_filelist_check_dirty (GINT_TO_POINTER(1));

	e2_widget_set_cursor (GDK_LEFT_PTR);
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, copy task");
#endif
	e2_filelist_enable_refresh ();
//	printd (DEBUG, "task: copy FINISHED");
	return TRUE;
}
/**
 * @brief copy selected items from active to inactive pane, with renaming
 *
 * Overwrite checking is performed, if that option is enabled
 * The actual operation is performed by a separate back-end function
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_copy_as (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
	//handle case of specific data instead of selection??
	GPtrArray *names = e2_fileview_get_selected (curr_view);
	if (names == NULL) return FALSE;	//nothing selected
	gchar *new_name, *slocal, *dlocal;
	GString *prompt = g_string_sized_new (NAME_MAX+64);
	GString *src = g_string_sized_new (PATH_MAX+NAME_MAX);
	GString *dest = g_string_sized_new (PATH_MAX+NAME_MAX);
	gboolean check = e2_option_bool_get ("confirm-overwrite");
	guint count;
	E2_SelectedItemInfo **iterator = (E2_SelectedItemInfo **) names->pdata;
	//setup for tailoring over-write dialog
	gboolean multisrc =  (check) ? names->len > 1 : FALSE;
	// setup for showing stop button in rename dialog
	OW_ButtonFlags extras = (multisrc)? NOALL : NONE;

//	prepare_dialog();
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, copy as task");
#endif
	e2_filelist_disable_refresh ();
	e2_widget_set_cursor (GDK_WATCH);

	for (count=0; count < names->len; count++, iterator++)
	{
//		if (g_str_equal((*iterator)->filename, "..")) continue;
		g_string_assign (src, (*iterator)->filename);
		g_string_printf (prompt, "%s: <b>%s</b>", _("Enter new name for"), src->str);
		DialogButtons result = 0;
	//FIXME = SPEED UP DIALOG CREATION OUTSIDE LOOP ?
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, copy as dialog");
#endif
		e2_filelist_enable_refresh ();  //allow updates while we wait
		DialogButtons result2 = e2_dialog_line_input (_("copy"), prompt->str, src->str, extras, FALSE, &new_name);
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, copy as dialog");
#endif
		e2_filelist_disable_refresh ();

		if (result2 == OK)
		{
			g_string_printf (src, "%s%s", curr_view->dir, (*iterator)->filename);  //separator comes with dir
			g_string_printf (dest, "%s%s", other_view->dir, new_name);
			g_free (new_name);
			if (strcmp(src->str, dest->str) == 0) continue;
			slocal = F_FILENAME_TO_LOCALE (src->str);
			dlocal = F_FILENAME_TO_LOCALE (dest->str);

			if (check && e2_fs_access2 (dlocal) == 0)
			{
#ifdef E2_REFRESH_DEBUG
				printd (DEBUG, "enable refresh, copy as dialog");
#endif
				e2_filelist_enable_refresh ();  //allow updates while we wait
				result = e2_dialog_ow_check (NULL, dlocal, extras);
				e2_filelist_disable_refresh ();
				if (result == OK)
				{
					e2_task_do_task (E2_TASK_COPY, slocal, dlocal, FALSE, NULL);
				}
/*				else if (result == YES_TO_ALL)
				{
					do something smart about multiple-renames
				} */
				else if (result == NO_TO_ALL)
				{
					F_FREE (slocal);
					F_FREE (dlocal);
					break;
				}
			}
			else
			{
				e2_task_do_task (E2_TASK_COPY, slocal, dlocal, FALSE, NULL);
			}
			F_FREE (slocal);
			F_FREE (dlocal);
		}
		else if (result2 == NO_TO_ALL)
		{
//			result = NO_TO_ALL;
			break;
		}
	}
	e2_fileview_clean_selected (names);
	g_string_free (prompt,TRUE);
	g_string_free (src,TRUE);
	g_string_free (dest,TRUE);

	//refresh other view & current if same
	e2_filelist_check_dirty (GINT_TO_POINTER(1));

	gtk_widget_set_sensitive(app.main_window, TRUE);
	e2_widget_set_cursor (GDK_LEFT_PTR);
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, copy as task");
#endif
	e2_filelist_enable_refresh ();
	return TRUE;
}
/**
 * @brief move selected items from active pane to trash location
 * THIS IS OUT-OF-DATE
 * Selected item(s) are moved to the last-used trash dir, which will be:
 * if it exists, folder named '.Trash' in current dir, or
 * if it exists, ~/.Trash, or
 * if it exists, config dir/Trash, or
 * not moved, and error msg is printed
 * No overwrite checking is performed.
 * The actual operation is performed by a separate back-end function
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, "" to act on selection
 * @param widget 
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_trashit (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
//	printd (DEBUG, "task: trash");
	//save pointer for other fn to use
	const gchar *trashpath = e2_utils_get_trash_path ();
	if (trashpath != NULL)
	{
		if (!g_str_equal (trashpath, curr_view->dir))  //do nothing if we're in the trash place already
			return (_e2_task_move (action_data, rt_data, widget, (gchar *) trashpath));
	}
	//error message displayed by the utils fn
	return FALSE;
}
/**
 * @brief move selected items from active to inactive pane
 *
 * Overwrite checking is performed, if that option is enabled
 * The actual operation is performed by a separate back-end function
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_move (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
//	printd (DEBUG, "task: move");
	if (g_str_equal (curr_view->dir, other_view->dir))
	{
		//display some message
		return FALSE;
	}
	return (_e2_task_move (action_data, rt_data, widget, NULL));
}
/**
 * @brief move selected items
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 * @param trashpath string with path to trash dir, or NULL if moving to inactive pane
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean _e2_task_move (gpointer action_data, gpointer rt_data,
	GtkWidget *widget, gchar *trashpath)
{
	GPtrArray *names = e2_fileview_get_selected (curr_view);
	if (names == NULL) return FALSE;	//nothing selected
	gchar *slocal, *dlocal;
	GString *src = g_string_sized_new (PATH_MAX+NAME_MAX);
	GString *dest = g_string_sized_new (PATH_MAX+NAME_MAX);
	guint count;
	E2_SelectedItemInfo **iterator = (E2_SelectedItemInfo **) names->pdata;
	gboolean trash = (trashpath != NULL);
	gboolean check = (trash) ? FALSE : e2_option_bool_get ("confirm-overwrite");
	//setup for tailoring over-write dialog
	gboolean multisrc =  (check) ? names->len > 1 : FALSE;
	OW_ButtonFlags extras = (multisrc) ? BOTHALL : NONE;

	//no window de-sensitize
	e2_widget_set_cursor (GDK_WATCH);
	e2_filelist_disable_refresh ();  //avoid pauses in the move process
	gchar *itemname;

	for (count=0; count < names->len; count++, iterator++)
	{
//		if (g_str_equal((*iterator)->filename, "..")) continue;
		itemname = (*iterator)->filename;
		g_string_printf (src, "%s%s", curr_view->dir, itemname);  //separator comes with dir
		gchar *p = (trash) ? trashpath : other_view->dir ;	//CHECKME trashpath = utf-8, not localised
		g_string_printf (dest, "%s%s", p, itemname);
		slocal = F_FILENAME_TO_LOCALE (src->str);
		dlocal = F_FILENAME_TO_LOCALE (dest->str);
		
		if (trash && e2_fs_access2 (dlocal) == 0)
		{
			//trash without destroying existing file there
			gchar *tdlocal = e2_utils_get_tempname (dlocal);
			e2_task_do_task (E2_TASK_MOVE, slocal, tdlocal, FALSE,
#ifdef E2_INCLIST
			(*iterator)->ref);
#else
			(*iterator)->path);
#endif
			g_free (tdlocal);
		}
		else if (check && e2_fs_access2 (dlocal) == 0)
		{
			e2_filelist_enable_refresh ();  //allow updates while we wait
			DialogButtons result = e2_dialog_ow_check (slocal, dlocal, extras);
			e2_filelist_disable_refresh ();
			switch (result)
			{
				case YES_TO_ALL:
					check = FALSE;
				case OK:
					e2_task_do_task (E2_TASK_MOVE, slocal, dlocal, FALSE,
#ifdef E2_INCLIST
					(*iterator)->ref);
#else
					(*iterator)->path);
#endif
				case CANCEL:
					break;
				default:
					result = NO_TO_ALL;
					break;
			}
			if (result == NO_TO_ALL)
			{
				F_FREE (slocal);
				F_FREE (dlocal);
				break;
			}
		}
		else  //file doesn't exist, or don't care
		{
			e2_task_do_task (E2_TASK_MOVE, slocal, dlocal, FALSE,
#ifdef E2_INCLIST
			(*iterator)->ref);
#else
			(*iterator)->path);
#endif
		}
		F_FREE (slocal);
		F_FREE (dlocal);
	}
	e2_fileview_clean_selected (names);
	g_string_free (src, TRUE);
	g_string_free (dest, TRUE);

	//FIXME E2_INCLIST - just do other view
	e2_filelist_check_dirty (GINT_TO_POINTER(1));
//	e2_fileview_clean_selected (names);

	e2_widget_set_cursor (GDK_LEFT_PTR);
	e2_filelist_enable_refresh ();

	return TRUE;
}
/**
 * @brief move selected items from active to inactive pane, with renaming
 *
 * Overwrite checking is performed, if that option is enabled
 * The actual operation is performed by a separate back-end function
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_move_as (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
	GPtrArray *names = e2_fileview_get_selected (curr_view);
	if (names == NULL) return FALSE;	//nothing selected
	gchar *new_name, *slocal, *dlocal;
	GString *prompt = g_string_sized_new (NAME_MAX+64);
	GString *src = g_string_sized_new (PATH_MAX+NAME_MAX);
	GString *dest = g_string_sized_new (PATH_MAX+NAME_MAX);
	gboolean check = e2_option_bool_get ("confirm-overwrite");
	guint count;
	E2_SelectedItemInfo **iterator = (E2_SelectedItemInfo **) names->pdata;
	gboolean multisrc =  (check) ? names->len > 1 : FALSE;
	OW_ButtonFlags extras = (multisrc)? NOALL : NONE;

//	prepare_dialog();  //disable refresh needed ??
	e2_widget_set_cursor (GDK_WATCH);
	e2_filelist_disable_refresh ();

	for (count=0; count < names->len; count++, iterator++)
	{
//		if (g_str_equal((*iterator)->filename, "..")) continue;
		g_string_assign (src, (*iterator)->filename);
		g_string_printf (prompt, "%s: <b>%s</b>", _("Enter new name for"), src->str);
		DialogButtons  result = 0;
		e2_filelist_enable_refresh ();  //allow updates while we wait
		DialogButtons result2 = e2_dialog_line_input (_("move"), prompt->str, src->str, extras, FALSE, &new_name);
		e2_filelist_disable_refresh ();

		if (result2 == OK)
		{
			g_string_printf(src, "%s%s", curr_view->dir, (*iterator)->filename);  //separator comes with dir
			g_string_printf(dest, "%s%s", other_view->dir, new_name);
			g_free (new_name);
			if (strcmp (src->str, dest->str) == 0) continue;
			slocal = F_FILENAME_TO_LOCALE (src->str);
			dlocal = F_FILENAME_TO_LOCALE (dest->str);

			if (check && e2_fs_access2 (dlocal) == 0)
			{
				e2_filelist_enable_refresh ();  //allow updates while we wait
				result = e2_dialog_ow_check (NULL, dlocal, extras);
				e2_filelist_enable_refresh ();
				if (result == OK)
				{
					e2_task_do_task (E2_TASK_MOVE, slocal, dlocal, FALSE,
#ifdef E2_INCLIST
					(*iterator)->ref);
#else
					(*iterator)->path);
#endif
				}
/*				else if (result == YES_TO_ALL)
				{
					do something smart about multiple-renames ??
				} */
				else if (result == NO_TO_ALL)
				{
					F_FREE (slocal);
					F_FREE (dlocal);
					break;
				}
			}
			else
			{
				e2_task_do_task (E2_TASK_MOVE, slocal, dlocal, FALSE,
#ifdef E2_INCLIST
				(*iterator)->ref);
#else
				(*iterator)->path);
#endif
			}
			F_FREE (slocal);
			F_FREE (dlocal);
		}
		else if (result2 == NO_TO_ALL)
		{
//			result = NO_TO_ALL;
			break;
		}
	}
	e2_fileview_clean_selected (names);
	g_string_free (prompt,TRUE);
	g_string_free (src,TRUE);
	g_string_free (dest,TRUE);

	//refresh other view & current if same
	e2_filelist_check_dirty (GINT_TO_POINTER(1));

	gtk_widget_set_sensitive(app.main_window, TRUE);
	e2_widget_set_cursor (GDK_LEFT_PTR);
	e2_filelist_enable_refresh ();

	return TRUE;
}

//FIXME where to put this ??
static void _e2_task_response_decode_cb (GtkWidget *button,
	gint response, gint *dialog_result)
{
	*dialog_result = response;
	gtk_main_quit ();
	return;
}
/**
 * @brief deleted selected items in the active pane
 *
 * Confirmation is sought, if that option is enabled
 * The actual operation is performed by a separate back-end function
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_delete (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
//	printd (DEBUG, "task: delete");
	GPtrArray *names = e2_fileview_get_selected (curr_view);
	if (names == NULL) return FALSE;	//nothing selected
	gchar *slocal;
#ifndef E2_INCLIST
	//handle case of specific data instead of selection??
	GtkTreeSelection *selection = curr_view->selection;
#endif
	GString *src = g_string_sized_new (NAME_MAX);
	GString *prompt = g_string_sized_new (NAME_MAX +64);
	gboolean check = e2_option_bool_get ("confirm-delete");
	guint count;
	E2_SelectedItemInfo **iterator = (E2_SelectedItemInfo **) names->pdata;
	gboolean multisrc =  (check) ? names->len > 1 : FALSE;
	gint result = 0;

//	prepare_dialog ();
	e2_filelist_disable_refresh ();
	e2_widget_set_cursor  (GDK_WATCH);

	for (count=0; count < names->len; count++, iterator++)
	{
//		if (g_str_equal((*iterator)->filename, ".."))continue;
		gchar *filename = g_build_filename (curr_view->dir, (*iterator)->filename, NULL);
		slocal = F_FILENAME_TO_LOCALE (filename);

		if (check)
		{
			g_string_assign (src, (*iterator)->filename);
			g_string_printf (prompt, _("Are you sure you want to delete <b>%s</b>?"), src->str);
			GtkWidget *dialog = e2_dialog_create (GTK_STOCK_DIALOG_QUESTION,
				prompt->str, _("confirm"), _e2_task_response_decode_cb, &result);	
//FIXME make this support single-setup when multisrc is FALSE
			//setup buttons states
			if (!multisrc)
			{
				E2_BUTTON_NOTOALL.showflags |= E2_BTN_GREY;
				E2_BUTTON_YESTOALL.showflags |= E2_BTN_GREY;
			}
			E2_BUTTON_CANCEL.showflags |= E2_BTN_DEFAULT;
			E2_BUTTON_OK.showflags &= !E2_BTN_DEFAULT;
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, delete dialog");
#endif
			e2_filelist_enable_refresh ();  //allow updates while we wait
			e2_dialog_show (dialog, app.main_window,
				0,
				&E2_BUTTON_NOTOALL, &E2_BUTTON_YESTOALL,
				&E2_BUTTON_CANCEL, &E2_BUTTON_OK, NULL);
			gtk_main ();
			gtk_widget_destroy (dialog);
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, delete dialog");
#endif
			e2_filelist_disable_refresh ();

			switch (result)
			{
				case E2_RESPONSE_YESTOALL:
					check = FALSE;
				case GTK_RESPONSE_OK:
					if (e2_task_backend_delete (slocal))
#ifdef E2_INCLIST
						_e2_task_treeview_line_remove ((*iterator)->ref);
#else
						gtk_tree_selection_unselect_path (selection, (*iterator)->path);
#endif
				case GTK_RESPONSE_CANCEL:
					break;
//				case E2_RESPONSE_NOTOALL:
				default:
					result = E2_RESPONSE_NOTOALL;
					break;
			}
		}
		else  //no checking
		{
			if (e2_task_backend_delete (slocal))
#ifdef E2_INCLIST
				_e2_task_treeview_line_remove ((*iterator)->ref);
#else
				gtk_tree_selection_unselect_path (selection, (*iterator)->path);
#endif
		}
		g_free (filename);
		F_FREE (slocal);
		if (result == E2_RESPONSE_NOTOALL)
			break;
	}
	e2_fileview_clean_selected (names);
	g_string_free (src, TRUE);
	g_string_free (prompt, TRUE);

	//FIXME E2_INCLIST - just do other view if it's same
	e2_filelist_check_dirty (GINT_TO_POINTER(1));

//	gtk_widget_set_sensitive(app.main_window, TRUE);
	e2_widget_set_cursor (GDK_LEFT_PTR);
	e2_filelist_enable_refresh ();

	return TRUE;
}
/**
 * @brief delete all items from trash location
 *
 * The last-used trash dir, which will be:
 * if it exists, folder named '.Trash' in current dir, or
 * if it exists, ~/.Trash, or
 * if it exists, config dir/Trash, 
 * will be opened, and all item(s) there will be selected and deleted
 * The actual operation is performed by a separate back-end function
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, "" to act on selection
 * @param widget 
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_trashempty (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
//	printd (DEBUG, "task: empty trash");
	const gchar *trashpath = e2_utils_get_trash_path ();
	if (trashpath != NULL)
	{
		gchar *pathsave = NULL;
		//make sure we're in the trash place
		if (!g_str_equal (trashpath, curr_view->dir))
		{
			pathsave = g_strdup (curr_view->dir);
			e2_pane_change_dir (NULL, (gchar *)trashpath);
		}
		
		e2_fileview_select_all (NULL, curr_view);
		gboolean retval = e2_task_delete (action_data, rt_data, widget);

		if (pathsave != NULL)
		{
			e2_pane_change_dir (NULL, pathsave);
			g_free (pathsave);
		}
		return retval;
	}
	//error message displayed by the utils fn
	return FALSE;
}
/**
 * @brief symlink selected items in active pane to the inactive pane
 *
 * Overwrite checking is performed, if that option is enabled
 * The actual operation is performed by a separate back-end function
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_symlink (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
	if (g_str_equal (curr_view->dir, other_view->dir))
	{
		//display some message
		return FALSE;
	}
	//handle case of specific data instead of selection??
	GPtrArray *names = e2_fileview_get_selected (curr_view);
	if (names == NULL) return FALSE;	//nothing selected
	GString *src = g_string_sized_new (PATH_MAX);	//used for target path
	GString *dest = g_string_sized_new (PATH_MAX);	//used for link path
	gboolean use_rel_links = e2_option_bool_get ("relative-symlinks");
	gboolean check = e2_option_bool_get ("confirm-overwrite");
	guint count;
	E2_SelectedItemInfo **iterator = (E2_SelectedItemInfo **) names->pdata;
	gboolean multisrc =  (check) ? names->len > 1 : FALSE;
	OW_ButtonFlags extras = (multisrc)? BOTHALL : NONE;

	e2_widget_set_cursor (GDK_WATCH);
	e2_filelist_disable_refresh ();  //avoid any interruption to linking
	gchar *itemname, *slocal, *dlocal;

	for (count=0; count < names->len; count++, iterator++)
	{
//		if (g_str_equal((*iterator)->filename, "..")) continue;
		itemname = (*iterator)->filename;
		g_string_printf (src, "%s%s", curr_view->dir, itemname);  //separator comes with dir
		g_string_printf (dest, "%s%s", other_view->dir, itemname);
		if (use_rel_links)
			e2_utils_create_relative_path (src, dest);
		slocal = F_FILENAME_TO_LOCALE (src->str);	//target, possibly relative
		dlocal = F_FILENAME_TO_LOCALE (dest->str);	//link name

		if (check && e2_fs_access2 (dlocal) == 0)
		{
			e2_filelist_enable_refresh ();  //allow updates while we wait
			DialogButtons result = e2_dialog_ow_check (slocal, dlocal, extras);
			e2_filelist_disable_refresh ();
			switch (result)
			{
				case YES_TO_ALL:
					check = FALSE;
				case OK:
					e2_task_backend_link (slocal, dlocal);
				case CANCEL:
					break;
				default:
					result = NO_TO_ALL;
					break;
			}
			if (result == NO_TO_ALL) break;
		}
		else  //file not present or don't care
			e2_task_backend_link (slocal, dlocal);

		F_FREE (slocal);
		F_FREE (dlocal);
	}
	e2_fileview_clean_selected (names);
	g_string_free (src, TRUE);
	g_string_free (dest, TRUE);

	//FIXME E2_INCLIST - just do other view
	e2_filelist_check_dirty (GINT_TO_POINTER(1));

	e2_widget_set_cursor (GDK_LEFT_PTR);
	e2_filelist_enable_refresh ();

	return TRUE;
}
/**
 * @brief symlink selected items in active pane to the inactive pane, with renaming
 *
 * Overwrite checking is performed, if that option is enabled
 * The actual operation is performed by a separate back-end function
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_symlink_as (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
	//handle case of specific data instead of selection??
	GPtrArray *names = e2_fileview_get_selected (curr_view);
	if (names == NULL) return FALSE;	//nothing selected
	gchar *new_name, *slocal, *dlocal;
	GString *prompt = g_string_sized_new (NAME_MAX+64);
	GString *src = g_string_sized_new (PATH_MAX);	//used for target path
	GString *dest = g_string_sized_new (PATH_MAX);	//used for link path
	gboolean check = e2_option_bool_get ("confirm-overwrite");
	gboolean use_rel_links = e2_option_bool_get ("relative-symlinks");
	guint count;
	E2_SelectedItemInfo **iterator = (E2_SelectedItemInfo **) names->pdata;
	gboolean multisrc =  (check) ? names->len > 1 : FALSE;
	OW_ButtonFlags extras = (multisrc) ? NOALL : NONE;
	
//	prepare_dialog();
	e2_widget_set_cursor (GDK_WATCH);
	e2_filelist_disable_refresh ();

	for (count=0; count < names->len; count++, iterator++)
	{
//		if (g_str_equal((*iterator)->filename, "..")) continue;
		g_string_assign (src, (*iterator)->filename);
		g_string_printf (prompt, "%s: <b>%s</b>", _("Enter new name for"), src->str);
		DialogButtons result = 0;
		e2_filelist_enable_refresh ();  //allow updates while we wait
		DialogButtons result2 = e2_dialog_line_input (_("symlink"), 
			prompt->str, src->str, extras, FALSE, &new_name);
		e2_filelist_disable_refresh ();

		if (result2 == OK)
		{
			g_string_printf(src, "%s%s", curr_view->dir, (*iterator)->filename);  //separator comes with dir
			g_string_printf(dest, "%s%s", other_view->dir, new_name);
			g_free (new_name);
			if (strcmp (src->str, dest->str) == 0) continue;
			if (use_rel_links)
				e2_utils_create_relative_path (src, dest);
			slocal = F_FILENAME_TO_LOCALE (src->str);	//target, possibly relative
			dlocal = F_FILENAME_TO_LOCALE (dest->str);	//link name

			if (check && (e2_fs_access2 (dlocal) == 0))
			{
				e2_filelist_enable_refresh ();  //allow updates while we wait
				result = e2_dialog_ow_check (NULL, dlocal, extras);
				e2_filelist_disable_refresh ();
					if (result == OK)
					e2_task_backend_link (slocal, dlocal);
/*				if (result == YES_TO_ALL)
				{
					do something smart about multiple-renames ??
				} */
				else if (result == NO_TO_ALL)
					break;
			}
			else
				e2_task_backend_link (slocal, dlocal);
			F_FREE (slocal);
			F_FREE (dlocal);
		}
		else if (result2 == NO_TO_ALL)
		{
//			result = NO_TO_ALL;
			break;
		}
	}
	e2_fileview_clean_selected (names);
	g_string_free (prompt,TRUE);
	g_string_free (src,TRUE);
	g_string_free (dest,TRUE);

	//FIXME E2_INCLIST - just do other view
	e2_filelist_check_dirty (GINT_TO_POINTER(1));

//	gtk_widget_set_sensitive(app.main_window, TRUE);
	e2_widget_set_cursor (GDK_LEFT_PTR);
	e2_filelist_enable_refresh ();
	return TRUE;
}
/**
 * @brief show/change permission of selected items in active pane
 *
 * The actual operation is performed by a separate back-end function
 * Directories are treated separately, to support recursive changes to
 * directories only
 * View is not updated immediately as the change does not trigger an auto-refresh  FIXME
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_permissions (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
//	printd (DEBUG, "task: permissions");
	//handle case of specific data instead of selection??
	GPtrArray *names = e2_fileview_get_selected (curr_view);
	if (names == NULL) return FALSE;	//nothing selected
	GString *path = g_string_sized_new (PATH_MAX+NAME_MAX);
	GString *mode = g_string_sized_new (16);
	GtkWidget *dialog=NULL;
	guint count;
	gint result;
	gint myuid = -1;
	E2_SelectedItemInfo **iterator = (E2_SelectedItemInfo **) names->pdata;
	gboolean multisrc = names->len > 1;
	gboolean all = FALSE;

/* out-of-loop setup = FIXME
	GtkWidget *dialog;
	dialog = e2_permissions_dialog_setup (info, &recurse);
	e2_dialog_add_buttons_simple (dialog, buttons ..., NULL);
*/	
//	prepare_dialog();
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, permissions task");
#endif
	e2_filelist_disable_refresh ();
	e2_widget_set_cursor (GDK_WATCH);

	for (count=0; count < names->len; count++, iterator++)
	{
//		if (g_str_equal((*iterator)->filename, ".."))continue;
		E2_RecurseType recurse;
		gboolean permission;
//FIXME for single-setup: instead of the following, adjust file details in dialog, reset default button

		if (all)
		{
			//check if we have permission to change this item
			//NB this tests _all_ w permissions, chmod is not so sophisticated ...
//			gchar *itempath = g_strconcat (curr_view->dir, (*iterator)->filename, NULL);
// 			permission = e2_utils_check_write_permission (myuid, itempath);
//			g_free (itempath);
			//path is to current dir
 			permission = e2_fs_check_write_permission (myuid, (*iterator)->filename);
			if (permission)
				result = GTK_RESPONSE_OK;
			else
			{
				result = GTK_RESPONSE_CANCEL;
				gchar *msg = g_strdup_printf ("%s %s", 
					_("You do not have authority to change permission(s) for"),
					(*iterator)->filename);
				e2_output_print_error(msg, TRUE);
			}
		}
		else
		{
			recurse=E2_RECURSE_NONE;
			permission=TRUE;
			g_string_assign (mode, "");
			dialog = e2_permissions_dialog_setup (*iterator, &recurse, 
				&permission, &mode, multisrc);
			g_signal_connect (G_OBJECT (dialog), "response",
				G_CALLBACK (_e2_task_response_decode_cb), &result);
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, permissions dialog");
#endif
			e2_filelist_enable_refresh ();  //allow updates while we wait
			_e2_dialog_show (dialog, app.main_window, 0);
			gtk_main ();
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, permissions dialog");
#endif
			e2_filelist_disable_refresh ();
		}

		switch (result)
		{
		  case E2_RESPONSE_APPLYTOALL:
  			gtk_widget_destroy (dialog);
			all = TRUE;
			myuid = getuid ();  //do this once, for speed
			result = GTK_RESPONSE_OK;
		  case GTK_RESPONSE_OK:
			if (permission)
			{
				g_string_printf (path, "%s%s", curr_view->dir, (*iterator)->filename); //separator comes with dir
				gchar *slocal = F_FILENAME_TO_LOCALE (path->str);
#ifdef E2_INCLIST
				if (e2_task_backend_chmod (slocal, mode->str, recurse))
				{
					//FIXME update line in treeview
				}
#else
#ifdef E2_FAM
				e2_task_backend_chmod (slocal, mode->str, recurse);
#else
				if (e2_task_backend_chmod (slocal, mode->str, recurse))
					//make the file-list refresher notice successful change
					e2_fs_touchnow (curr_view->dir);
#endif

#endif
				F_FREE (slocal);
			}
		  case GTK_RESPONSE_CANCEL:
			break;
		  default:
			result = E2_RESPONSE_NOTOALL;  // break flag;
			break;
		}
		if (result == E2_RESPONSE_NOTOALL)
			break;
		if (!all)	// && (count < names->len - 1))
			//destroy dialog, ready for next one
			gtk_widget_destroy (dialog);
	}
	e2_fileview_clean_selected (names);
	g_string_free (path, TRUE);
	g_string_free (mode, TRUE);

	//FIXME E2_INCLIST - just do other view if it's same
	e2_filelist_check_dirty (GINT_TO_POINTER(1));

	if ((count == names->len) || (result == E2_RESPONSE_NOTOALL))
		gtk_widget_destroy (dialog);
//	gtk_widget_set_sensitive (app.main_window, TRUE);
	e2_widget_set_cursor (GDK_LEFT_PTR);
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, permissions task");
#endif
	e2_filelist_enable_refresh ();
	gtk_widget_grab_focus (curr_view->treeview);

	return TRUE;
}
/**
 * @brief show/change user and/or group of selected items in active pane
 *
 * The actual operation is performed by a separate back-end function
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_ownership (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
	//handle case of specific data instead of selection??
	GPtrArray *names = e2_fileview_get_selected (curr_view);
	if (names == NULL) return FALSE;	//nothing selected
	GtkWidget *dialog=NULL;
	GString *path = g_string_sized_new (PATH_MAX+NAME_MAX);
	guint count;
	gint result;
	gint myuid = -1;
	E2_SelectedItemInfo **iterator = (E2_SelectedItemInfo **) names->pdata;
	gboolean multisrc = names->len > 1;
	gboolean all = FALSE;

//	prepare_dialog();
	e2_filelist_disable_refresh ();
	e2_widget_set_cursor (GDK_WATCH);

	for (count=0; count < names->len; count++, iterator++)
	{
		gboolean permission;
		uid_t owner_id;
		gid_t group_id;
		gboolean recurse;
//		if (g_str_equal((*iterator)->filename, ".."))continue;

		if (all)
		{
			//check if we have permission to change this item
//			gchar *itempath = g_strconcat (curr_view->dir, (*iterator)->filename, NULL);
// 			permission = e2_utils_check_write_permission (myuid, itempath);
//			g_free (itempath);
			//path is to current dir
 			permission = e2_fs_check_write_permission (myuid, (*iterator)->filename);
			if (permission)
				result = GTK_RESPONSE_OK;
			else
			{
				result = GTK_RESPONSE_CANCEL;
				gchar *msg = g_strdup_printf ("%s  %s", 
					_("You do not have authority to change owner(s) for"),
					(*iterator)->filename);
				e2_output_print_error (msg, TRUE);
			}
		}
		else
		{
			owner_id=-1;  //CHECKME = should these be set outside loop ?
			group_id=-1;
			permission=TRUE;
			recurse=FALSE;
			dialog = e2_ownership_dialog_setup (*iterator, &owner_id, 
				&group_id, &recurse, &permission, multisrc);
			//leave dialog open as long as possible to prevent 'shadow' prior to screen refresh
			g_signal_connect (G_OBJECT (dialog), "response",
				G_CALLBACK (_e2_task_response_decode_cb), &result);
			e2_filelist_enable_refresh ();  //allow updates while we wait
			_e2_dialog_show (dialog, app.main_window, 0);
			gtk_main ();
			e2_filelist_disable_refresh ();
		}

		switch (result)
		{
		  case E2_RESPONSE_APPLYTOALL:
			gtk_widget_destroy (dialog);
			all = TRUE;
			myuid = getuid ();
			result = GTK_RESPONSE_OK;
		  case GTK_RESPONSE_OK:
			if (permission)
			{
//				printd (DEBUG, "ownership task: ok");
//				printd (DEBUG, "new owner id is: %d",  owner_id);
				g_string_printf (path, "%s%s", curr_view->dir, (*iterator)->filename); //separator comes with dir
				gchar *slocal = F_FILENAME_TO_LOCALE (path->str);
#ifdef E2_INCLIST
				if (e2_task_backend_chown (slocal, owner_id, group_id, recurse))
				{
					//FIXME update line in treeview
				}
#else
#ifdef E2_FAM
				e2_task_backend_chown (slocal, owner_id, group_id, recurse);
#else
				if (e2_task_backend_chown (slocal, owner_id, group_id, recurse))
					//make the file-list refresher notice successful change
					e2_fs_touchnow (curr_view->dir);
#endif
#endif
				F_FREE (slocal);
			}
		  case GTK_RESPONSE_CANCEL:
			break;
		  default:
			result = E2_RESPONSE_NOTOALL;  // break flag;
			break;
		}
		if (result == E2_RESPONSE_NOTOALL)
			break;
		if (!all)	// && (count < names->len - 1))
			//destroy dialog, ready for next one
			gtk_widget_destroy (dialog);
	}
	e2_fileview_clean_selected (names);
	g_string_free (path,TRUE);

	//FIXME E2_INCLIST - just do other view if it's same
	e2_filelist_check_dirty (GINT_TO_POINTER(1));

	if ((count == names->len) || (result == E2_RESPONSE_NOTOALL))
		gtk_widget_destroy (dialog);
//	gtk_widget_set_sensitive(app.main_window, TRUE);
	e2_widget_set_cursor (GDK_LEFT_PTR);
	e2_filelist_enable_refresh ();
	gtk_widget_grab_focus (curr_view->treeview);

	return TRUE;
}
/**
 * @brief show info about each select item in active pane
 *
 * NB the active pane directory is not necessarily the current file system
 * dir, so the path is added to the filename
 * FIXME on this ?
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_file_info (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
	//handle case of specific data instead of selection??
	GPtrArray *names = e2_fileview_get_selected (curr_view);
	if (names == NULL) return FALSE;	//nothing selected
	GtkWidget *dialog;
	guint count;
	E2_SelectedItemInfo **iterator = (E2_SelectedItemInfo **) names->pdata;

//	prepare_dialog ();
	e2_widget_set_cursor (GDK_WATCH);
//	gtk_widget_set_sensitive (app.main_window, FALSE);
	e2_filelist_disable_refresh ();	//prevent access-related refresh

	for (count=0; count < names->len; count++, iterator++)
	{
//		if (g_str_equal((*iterator)->filename, "..")) continue;
		gchar *local = F_FILENAME_TO_LOCALE ((*iterator)->filename);
		if (!e2_fs_access2 (local))
		{
			dialog = e2_file_info_dialog_setup (*iterator);
			_e2_dialog_show (dialog, app.main_window, 0);
			gtk_main ();
		}
		F_FREE (local);
	}
	e2_fileview_clean_selected (names);

//	postpare_dialog (FALSE, NULL);
	e2_widget_set_cursor (GDK_LEFT_PTR);
//	gtk_widget_set_sensitive (app.main_window, TRUE);
	e2_filelist_enable_refresh ();

	//get focus back to where we started
	gtk_widget_grab_focus (curr_view->treeview);
//	e2_fileview_focus_row (curr_view, curr_view->row, FALSE, FALSE, FALSE, TRUE);
	return TRUE;
}
/**
 * @brief rename selected items in active pane to the inactive pane
 *
 * Overwrite checking is performed, if that option is enabled
 * The actual operation is performed by a separate back-end function
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_rename (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
	GPtrArray *names = e2_fileview_get_selected (curr_view);
	if (names == NULL) return FALSE;	//nothing selected
	gchar *new_name, *slocal, *dlocal;
#ifndef E2_INCLIST
	//handle case of specific data instead of selection??
	GtkTreeSelection *selection = curr_view->selection;
#endif
	GString *prompt = g_string_sized_new (NAME_MAX+64);
	GString *src = g_string_sized_new (PATH_MAX+NAME_MAX);
	GString *dest = g_string_sized_new (PATH_MAX+NAME_MAX);
	gboolean check = e2_option_bool_get ("confirm-overwrite");
	guint count;
	E2_SelectedItemInfo **iterator = (E2_SelectedItemInfo **) names->pdata;
	gboolean multisrc =  (check) ? names->len > 1 : FALSE;
	OW_ButtonFlags extras = (multisrc) ? NOALL : NONE;

//	prepare_dialog();
	e2_filelist_disable_refresh ();
	e2_widget_set_cursor (GDK_WATCH);

	for (count=0; count < names->len; count++, iterator++)
	{
//		if (g_str_equal((*iterator)->filename, ".."))continue;
		g_string_assign (src, (*iterator)->filename);
		g_string_printf (prompt, "%s: <b>%s</b>", _("Enter new name for"), src->str);
		DialogButtons result = 0;
		e2_filelist_enable_refresh ();  //allow updates while we wait
		DialogButtons result2 = e2_dialog_line_input (_("rename"), prompt->str, src->str, extras, FALSE, &new_name);
		e2_filelist_disable_refresh ();

		if (result2 == OK)
		{
			if (strcmp((*iterator)->filename, new_name)==0)
			{
				g_free (new_name);
				continue;
			}
			g_string_printf(src, "%s%s", curr_view->dir, (*iterator)->filename);  //separator comes with dir
			g_string_printf(dest, "%s%s", curr_view->dir, new_name);
			slocal = F_FILENAME_TO_LOCALE (src->str);
			dlocal = F_FILENAME_TO_LOCALE (dest->str);

			gboolean exists = !e2_fs_access2 (dlocal);
			if (exists)
			{	//maybe new name exists, or maybe it's just the old name with some different case
				gchar *old_name_lc = g_utf8_casefold ((*iterator)->filename, -1);
				gchar *new_name_lc = g_utf8_casefold (new_name, -1);
				gboolean case_change = !strcmp (old_name_lc, new_name_lc);
				g_free (old_name_lc);
				g_free (new_name_lc);
				if (case_change)
				{	//we have just a case-difference that looks the same to the kernel
					//do interim name change to check, & also to ensure case-change happens
					gchar *tempname = e2_utils_get_tempname (slocal);
//					if (e2_task_backend_rename (slocal, tempname)
//						&& e2_task_backend_rename (tempname, dlocal))
					if (e2_task_backend_rename (slocal, tempname))
					{
						exists = !e2_fs_access2 (dlocal);
						if (exists)
						{	//the new name really exists
							//revert and process normally
							rename (tempname, slocal);
							g_free (tempname);
						}
						else
						{	//the former detection was fake, just do the rest of the rename
							//CHECKME this may be irrelevant
							e2_task_backend_rename (tempname, dlocal);
#ifdef E2_INCLIST
							_e2_task_treeview_line_rename ((*iterator)->ref, NOTFREEDnew_name);
#else
							gtk_tree_selection_unselect_path (selection, (*iterator)->path);
#endif
							g_free (tempname);
							continue;
						}
					}
					else
					{
						gchar *msg = g_strdup_printf (_("Cannot rename %s"), (*iterator)->filename);
						e2_output_print_error (msg, TRUE);
						g_free (tempname);
						continue;
					}
				}
			}
			if (check && exists)
			{
				e2_filelist_enable_refresh ();  //allow updates while we wait
				result = e2_dialog_ow_check (slocal, dlocal, extras);
				e2_filelist_disable_refresh ();
				if (result == OK)
				{
					if (e2_task_backend_rename (slocal, dlocal))
#ifdef E2_INCLIST
						_e2_task_treeview_line_rename ((*iterator)->ref, NOTFREEDnew_name);
#else
						gtk_tree_selection_unselect_path (selection, (*iterator)->path);
#endif
				}
				else if (result == NO_TO_ALL)
				{
					g_free (new_name);
					F_FREE (slocal);
					F_FREE (dlocal);
					break;
				}
			}
			else
			{
				if (e2_task_backend_rename (slocal, dlocal))
#ifdef E2_INCLIST
					_e2_task_treeview_line_rename ((*iterator)->ref, NOTFREEDnew_name);
#else
					gtk_tree_selection_unselect_path (selection, (*iterator)->path);
#endif
			}
			g_free (new_name);
			F_FREE (slocal);
			F_FREE (dlocal);
		}
		else if (result2 == NO_TO_ALL)
			break;
	}
	e2_fileview_clean_selected (names);
	g_string_free (prompt,TRUE);
	g_string_free (src,TRUE);
	g_string_free (dest,TRUE);

	//FIXME E2_INCLIST - just do other view if it's same
	e2_filelist_check_dirty (GINT_TO_POINTER(1));

//	gtk_widget_set_sensitive(app.main_window, TRUE);
	e2_widget_set_cursor (GDK_LEFT_PTR);
	e2_filelist_enable_refresh ();

	return TRUE;
}
/**
 * @brief execute the operation specified from the drag menu, on all dragged items
 * As the DnD process wants a result, the operation is not done in background
 * Item overwrite checks are performed if that option is in force.
 * @a uri_list is created in e2_filelist_drag_data_get () 
 *	For e2-sourced items, it is a string of the format:
 *	file://path/to/one_file\r\nfile://path/to/second_file\r\n
 * and so on for each selected item. The "file://" at the start of
 *	each element is ascii, the path/name data was converted to
 *	utf-8 if necessary
 * For externally-sourced items, it may be different eg
 *	file:/path/to/one_file\r\nfile:/path/to/second_file\r\n
 * It assumes all items have the same source path.
 * Nothing will be done if src and dest dirs are the same
 *
 * Processing of dropped items stops if any problem occurs
 *
 * @param func code for function to be performed on dragged items
 * @param uri_list utf-8 string, uri list of items to process with @a func
 * @param dest_dir utf-8 string, destination for items in @a uri_list, with trailing /
 *
 * @return TRUE if the items were all processed successfully
**/
gboolean e2_task_drop (E2_TaskType func, guchar *uri_list, gchar *dest_dir)
{
//	printd (DEBUG, "callback: drop task");
	gchar *s, *p, *filename, *slocal, *dlocal;

	gboolean retval = FALSE;
	gpointer funcname;
	switch (func)
	{
		case E2_TASK_LINK:
			funcname = e2_task_backend_link;
			break;
		case E2_TASK_MOVE:
			funcname = e2_task_backend_move;
			break;
		default:
//		case E2_TASK_COPY:
			funcname = e2_task_backend_copy;
			break;
	}
	//we overload the move and link funcs to allow for the preserve arg when copying
	gboolean (*op_func) (gchar *, gchar *, gboolean) = funcname;
	gchar **uris = g_strsplit ((gchar *)uri_list, "\r\n", -1);
	//get length to ignore at start of each uri
	s = uris[0];
/* internally-sourced directory items will not have a trailing /, because
	e2_fileview_drag_data_get_cb() translates the names afresh
	and externally sourced ones will probably not have one ??
	from the corresponding info->filename
	if (g_str_has_suffix (s, G_DIR_SEPARATOR_S))
	{
		skip_len = strlen (s) - 1;
		*(s+skip_len) = '\0';
			filename = strrchr (s, G_DIR_SEPARATOR);
		*(s+skip_len) = G_DIR_SEPARATOR;
	}
	else */
		filename = strrchr (s, G_DIR_SEPARATOR);
	filename++;
	//check for src path = dest path
	gchar save = *(filename);
	*(filename) = '\0';
	p = s;
	//look for the '/' that really starts the path, in context where the
	//no. of leading '/'s is not certain (various apps create different URIs)
	while (TRUE)
	{
		p = strchr (p, G_DIR_SEPARATOR);	//if always ascii /, don't need g_utf8_strchr()
		p++;
		if (*p != G_DIR_SEPARATOR)
		{
			p--;
			break;
		}
	}
	if (g_str_equal (p, dest_dir))
		return FALSE;

	gint start_skip = p - s;
	//remember the length to strip from each source path, to get item name
	gint skip_len = filename - s;	//assumes single-byte separator
	*(filename) = save;

	//count items, to allow appropriate dialog buttons display
	gchar **tmp = uris;
	gint ctr = 0;
	while (*tmp != NULL && **tmp != '\0')	//last string is empty
	{
		if (++ctr > 1) break;
		tmp++;
	}
	OW_ButtonFlags extras = (ctr > 1) ? BOTHALL : NONE;

	gboolean check = e2_option_bool_get ("confirm-overwrite");
	DialogButtons result;

	//walk the array of items
	tmp = uris;
	while (*tmp != NULL && **tmp != '\0')
	{
		//get src & dest paths
		s = *tmp + start_skip;	//skip over "file://" (ascii)
		slocal = F_FILENAME_TO_LOCALE (s);
		filename = *tmp + skip_len;
		p = g_strconcat (dest_dir, filename, NULL); //dest_dir comes with /
		dlocal = F_FILENAME_TO_LOCALE (p);

		if (check && (e2_fs_access2 (dlocal) == 0))
		{
//			gtk_widget_set_sensitive (app.main_window, FALSE);
			result = e2_dialog_ow_check (slocal, dlocal, extras);
//			gtk_widget_set_sensitive (app.main_window, TRUE);
			switch (result)
			{
				case YES_TO_ALL:
					check = FALSE;
				case OK:
					//can't unselect moved items on-the-fly, there is no path data here
					retval = op_func (slocal, dlocal, FALSE);
					if (!retval)
						result = NO_TO_ALL; //abort on failure
				case CANCEL:
					break;
				default:
					result = NO_TO_ALL;
					break;
			}
			if (result == NO_TO_ALL)
			{
				g_free (p);
				F_FREE (slocal);
				F_FREE (dlocal);
				break;
			}
		}
		else  //file doesn't exist, or don't care
			retval = op_func (slocal, dlocal, FALSE);

		g_free (p);
		F_FREE (slocal);
		F_FREE (dlocal);
		tmp++;
	}
	//cleanups
	g_strfreev (uris);
	//E2_INCLIST don't know what to refresh, so do it all
	e2_filelist_check_dirty (GINT_TO_POINTER(1));

	return retval;
}
/**
 * @brief change 'other' pane to be the same as 'current' pane
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_sync_dirs (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
#ifdef E2_NEWFILTER
	FIXME copy the store and change dir strings
#else
	e2_pane_change_dir (other_pane, curr_pane->path);
#endif
	return TRUE;
}
/**
 * @brief view content of 1st selected item in active pane
 *
 * The actual operation is performed by internal or extenal viewer,
 * depending on the option specfied
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget menu item ? UNUSED
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_view (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
	gchar *view_this = (gchar *) rt_data;	//CHECKME utf-8 needed
	if (view_this != NULL && *view_this != '\0')
	{  //we arrived here with a single file to be viewed
		//it could be a path or just a filename, and it might be in parentheses
		//which are a problem for building absolute paths if need be, so strip them
		if (*view_this == '"')
			view_this++;
		gchar *lng = view_this + strlen (view_this) - 1;
		if (*lng == '"')
			*lng = '\0';

		return (e2_task_backend_view (view_this));
	}
	//otherwise, process the active pane selection
	GPtrArray *names = e2_fileview_get_selected (curr_view);
	if (names == NULL) return FALSE;	//nothing selected
	guint count;
	E2_SelectedItemInfo **iterator = (E2_SelectedItemInfo **) names->pdata;

	e2_filelist_disable_refresh ();
	gtk_widget_set_sensitive (app.main_window, FALSE);

	for (count=0; count < names->len; count++, iterator++)
	{
//		if (g_str_equal(**iterator, "..")) continue;
		//this assumes that the current system dir is curr_view->dir
		e2_task_backend_view ((*iterator)->filename);
		break;	//int/ext viewer(s) don't have an 'abort' function, so stop after 1st view
	}
	e2_fileview_clean_selected (names);

	gtk_widget_set_sensitive (app.main_window, TRUE);
	gtk_widget_grab_focus (curr_view->treeview);
	e2_filelist_enable_refresh ();
	return TRUE;
}
/**
 * @brief perform the default action for the first selected item in the current pane
 *
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_edit (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
//	printd (DEBUG, "task: edit");
	//handle case of specific data instead of selection??
	FileInfo *info;
	if ((info = e2_fileview_get_selected_first_local (curr_view)) != NULL)
	{
		if (e2_option_bool_get ("use-external-editor"))
		{
			//CHECKME needs to be utf ?
			gchar *command = g_strdup_printf ("$[command-viewer] '%s'",	//no translation
				info->filename);
			e2_command_run (command, E2_COMMAND_RANGE_DEFAULT);
			g_free (command);
		}
		else
		{
			gchar *utf = F_FILENAME_FROM_LOCALE (info->filename);
			e2_edit_dialog_create (utf);
			F_FREE (utf);
		}
		return TRUE;
	}
	return FALSE;
}
/**
 * @brief toggle tag on selected item in active pane
 * 
 * Deleted after discussion with tooar
 *
 * @return
 */
/*void __e2_task_toggle_tag_cb (void)
{
//FIXME = logic of this is silly !!
	GList *tmp = e2_fileview_get_selection (curr_view);
	//deprecated
	FileInfo *info;
	gint row;

//	g_signal_emit_by_name (G_OBJECT (curr_view->clist), "end-selection");

	for (; tmp != NULL; tmp = tmp->next)
	{
/ *		row = GPOINTER_TO_INT (tmp->data);
		info = gtk_clist_get_row_data(GTK_CLIST(curr_view->clist), row);
		if (g_list_find (curr_view->tagged, info) != NULL)
		{
			curr_view->tagged = g_list_remove(curr_view->tagged, info);
			gtk_clist_set_background(GTK_CLIST(curr_view->clist), row, &LIST_COLOR);
		}
		else
		{
			curr_view->tagged = g_list_append(curr_view->tagged, info);
			gtk_clist_set_background(GTK_CLIST(curr_view->clist), row, e2_option_color_get ("color-tag"));
		}  * /
	//FIXME
	}
	e2_fileview_focus_row(curr_view, curr_view->row+1, FALSE, TRUE, TRUE, TRUE);
} */
/**
 * @brief refresh content of both pane views
 *
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_refresh (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
	e2_widget_set_cursor (GDK_WATCH);
	e2_filelist_disable_refresh ();
	e2_fileview_refresh_list (other_view);
	e2_fileview_refresh_list (curr_view);
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, refresh task");
#endif
	e2_filelist_enable_refresh ();
	e2_widget_set_cursor (GDK_LEFT_PTR);
	return TRUE;
}
/**
 * @brief revert all configuration settings to defailt values
 *
 *
 * @param widget always NULL, unused
 *
 * @return TRUE always
 */
gboolean e2_task_configure_default (GtkWidget *widget)
{
	//dump old data, don't re-read the config file, recreate screen
	e2_option_refresh (FALSE, TRUE);
	return TRUE;
}
/**
 * @brief initiate configuration change
 *
 *
 * @param widget always NULL, unused
 * @param page name string of 'page' in config dialog, or NULL
 *
 * @return TRUE, always
 */
gboolean e2_task_configure (GtkWidget *widget, gchar *page)
{  //put this in config dialog file
	gtk_widget_set_sensitive (app.main_window, FALSE);
	e2_config_dialog_create (page);
	gtk_widget_set_sensitive (app.main_window, TRUE);
	return TRUE;
}
/**
 * @brief find named item in current pane
 *
 * The search starts from the current line, forward to end, then cycles
 * back to the start
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data assigned when actionrt struct created
 * @param widget the button pressed to activate this task, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_find (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
	gchar *s, *pattern;
	gint i, last, rowcount;

	//find the no of rows in the store
	rowcount = gtk_tree_model_iter_n_children 
		(GTK_TREE_MODEL (curr_view->store), NULL);
	if (rowcount < 1)
		return FALSE;  //no message ...

//	prepare_dialog();
	e2_widget_set_cursor (GDK_WATCH);
//	gtk_widget_set_sensitive (app.main_window, FALSE);

	DialogButtons result2 = e2_dialog_line_input (_("find"), _("Enter a name or partial name to find:"),
		curr_view->last_find, 0, TRUE, &s);

	if (result2 == OK)
	{
		g_strlcpy (curr_view->last_find, s, sizeof(curr_view->last_find));
		pattern = g_strdup_printf("*%s*", s);

		GtkTreePath *path;
		GtkTreeIter iter;
		if (gtk_tree_model_iter_nth_child
			(GTK_TREE_MODEL (curr_view->store), &iter, NULL, curr_view->row))
		{
			path = gtk_tree_model_get_path
				(GTK_TREE_MODEL (curr_view->store), &iter);
		}
		else
		{
			//try fallback to the row of the current cursor position
			gtk_tree_view_get_cursor (GTK_TREE_VIEW (curr_view->treeview),
				&path, NULL);
		}

		if (path == NULL)
		{
			last = rowcount - 1;
			//no defined start pos, start from beginning
			i = 0;
		}
		else
		{
			last = *gtk_tree_path_get_indices (path);
			if (last == rowcount - 1)
				i = 0;	//roll around to the start
			else
				i = last + 1;	//start at the next line
			
			gtk_tree_path_free (path);
		}

		//iterate through the rows
		gchar *thisrowname;
		while (i != last)
		{
			thisrowname = e2_fileview_get_row_name (curr_view, i);
			if (g_pattern_match_simple (pattern, thisrowname))
			{
				e2_fileview_focus_row (curr_view, i, TRUE, TRUE, TRUE, TRUE);
				g_free (thisrowname);
				break;
			}
			g_free (thisrowname);
			if (++i >= rowcount)
				i = 0;  //cycle to start
		}
		if (i == last)
			//nothing found (or maybe found in the last row)
			e2_fileview_focus_row (curr_view, last, TRUE, TRUE, FALSE, TRUE);
		g_free(pattern);
		g_free(s);
	}
	else	//user cancelled
		e2_fileview_focus_row (curr_view, curr_view->row, FALSE, FALSE, FALSE, TRUE);

//	postpare_dialog (FALSE, NULL);
//	gtk_widget_set_sensitive (app.main_window, TRUE);
	e2_widget_set_cursor (GDK_LEFT_PTR);
	gtk_widget_grab_focus (curr_view->treeview);
	return TRUE;
}
/**
 * @brief perform the default action for the first selected item in the current pane
 *
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_open (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
//	printd (DEBUG, "task: open");
	//handle case of specific data instead of selection??
	FileInfo *info;
	if ((info = e2_fileview_get_selected_first_local (curr_view)) != NULL)
	{
		return (e2_task_backend_open (info));
	}
	return FALSE;
}
/**
 * @brief open item(s) in the current pane with a chosen application
 *
 * If the specified command does not include '%', all selected items
 * are concatenated, and sent as a single argument to the application.
 * Otherwise, the entered command is simply executed
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_open_with (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
	gchar *open_this = (gchar *) rt_data;	//CHECKME utf-8 needed
	if (open_this != NULL && *open_this != '\0')
	{  //we arrived here with a single file to be viewed
		//it could be a path or just a filename, and it might be in parentheses
		//which are a problem for building absolute paths if need be, so strip them
		if (*open_this == '"')
			open_this++;
		gchar *lng = open_this + strlen (open_this) - 1;
		if (*lng == '"')
			*lng = '\0';
	}

	gchar *command;

//	prepare_dialog();
	e2_widget_set_cursor (GDK_WATCH);
//	gtk_widget_set_sensitive (app.main_window, FALSE);

	DialogButtons result = e2_dialog_combo_input (_("open with"),
		_("Enter command:"), 0, &open_history, &command);

	gboolean retval = FALSE;
	if (result == OK)
	{
		retval = TRUE;
		//FIXME support macros other than with english letters
		if ((open_this == NULL || *open_this == '\0')
			&& strstr (command, "%f") == NULL)
				open_this = "%f";
		if (open_this != NULL)
		{
			gchar *freeme = command;
			command = g_strconcat (command, " ", open_this, NULL);
			g_free (freeme);
		}
		e2_command_run (command, E2_COMMAND_RANGE_DEFAULT);

		g_free(command);
	}

//	postpare_dialog (FALSE, NULL);
//	gtk_widget_set_sensitive (app.main_window, TRUE);
	e2_widget_set_cursor (GDK_LEFT_PTR);
	gtk_widget_grab_focus (curr_view->treeview);
	
	return retval;
}
/**
 * @brief open directory which is currently selected, in the other pane
 * 
 * There is no check whether the item is actually directory
 * The last-activated row in the other pane is focused - this is not
 * necessarily sensible
 *
 * @param action_data ptr to data assigned when action struct created at session start
 * @param rt_data ptr to data sent to this command, NULL to act on selection
 * @param widget always NULL, unused
 *
 * @return TRUE if task completed successfully, else FALSE
 */
gboolean e2_task_open_in_other_pane (gpointer action_data, gpointer rt_data, GtkWidget *widget)
{
	//handle case of specific data instead of selection??
	e2_filelist_disable_refresh ();
	gchar *name = e2_fileview_get_row_name (curr_view, curr_view->row);
	if (name == NULL)
	{
		e2_filelist_enable_refresh ();
		return FALSE;  //the model got mixed up, somehow
	}
	gchar *path = g_strdup_printf("%s%s", curr_view->dir, name); //separator comes with dir
	e2_pane_change_dir (other_pane, path);
	g_free (name);
	g_free (path);
	//CHECKME a more sensible row to goto ??
	e2_fileview_focus_row (other_view, other_view->row, FALSE, FALSE, TRUE, TRUE);
	e2_filelist_enable_refresh ();
	return TRUE;
}
/**
 * @brief register task-related actions
 * 
 * @return
 */
void e2_task_actions_register (void)
{
	E2_Action task_actions[] =
	{	//we don't bother explicitly setting each element's last-3 parameters
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_open, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_open_in_other_pane, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_open_with, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_view, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_edit, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_file_info, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_permissions, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_ownership, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_rename, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_copy, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_copy_as, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_move, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_move_as, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_symlink, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_symlink_as, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_delete, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_find, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_trashit, FALSE, 0 },
		{ NULL, E2_ACTION_TYPE_ITEM, e2_task_trashempty, FALSE, 0 }
	};
	task_actions[0].name = g_strconcat(_A(5),".",_A(56),NULL);
	task_actions[1].name = g_strconcat(_A(5),".",_A(57),NULL);
	task_actions[2].name = g_strconcat(_A(5),".",_A(58),NULL);
	task_actions[3].name = g_strconcat(_A(5),".",_A(86),NULL);
	task_actions[4].name = g_strconcat(_A(5),".",_A(36),NULL);
	task_actions[5].name = g_strconcat(_A(5),".",_A(47),NULL);
	task_actions[6].name = g_strconcat(_A(5),".",_A(62),NULL);
	task_actions[7].name = g_strconcat(_A(5),".",_A(59),NULL);
	task_actions[8].name = g_strconcat(_A(5),".",_A(68),NULL);
	task_actions[9].name = g_strconcat(_A(5),".",_A(31),NULL);
	task_actions[10].name = g_strconcat(_A(5),".",_A(32),NULL);
	task_actions[11].name = g_strconcat(_A(5),".",_A(54),NULL);
	task_actions[12].name = g_strconcat(_A(5),".",_A(55),NULL);
	task_actions[13].name = g_strconcat(_A(5),".",_A(78),NULL);
	task_actions[14].name = g_strconcat(_A(5),".",_A(79),NULL);
	task_actions[15].name = g_strconcat(_A(5),".",_A(35),NULL);
	task_actions[16].name = g_strconcat(_A(5),".",_A(6),NULL);
	task_actions[17].name = g_strconcat(_A(5),".",_A(84),NULL);
	task_actions[18].name = g_strconcat(_A(1),".",_A(85),NULL);	//not applied to selected items

	gint actions_count = sizeof (task_actions) / sizeof (E2_Action);
	gint i;
	for (i = 0; i < actions_count; i++)
	{
		e2_action_register (
			task_actions[i].name,
			task_actions[i].type,
			task_actions[i].func,
			NULL,
			task_actions[i].has_arg,
			task_actions[i].exclude,
			NULL);
	}
	//this one has data
	gchar *action = g_strconcat(_A(5),".",_A(33),NULL);
	e2_action_register (action, E2_ACTION_TYPE_ITEM, e2_task_copy,
		GINT_TO_POINTER (TRUE), FALSE, 0, NULL);
}
