/* $Id: e2_task_backend.c 1869 2006-04-27 01:58:02Z tpgww $
 *
 * Copyright (C) 2005-2006 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/e2_task_backend.c
 * @brief task backend functions
 *
 * This file contains functions used to handle different file operations
 * like copying, deleting, symlinking, etc.
 * The 'public' functions are all named like e2_task_<...>_backend ()
 * They return TRUE on successful completion
 * Private functions are named like _e2_task_backend_<...> ()
 * These functions are thread-safe, when E2_TREEWALK is defind
 */

#include "e2_task.h"
#include <string.h>
#include <fcntl.h>
#include "e2_filetype.h"
#include "e2_filelist.h"
#include "e2_dialog.h"

/**
 * @brief parse a symbolic permissions string
 * The string @a tmp1 would be like 'o+x' or 'g-r' etc
 * @param tmp1 permissions string to parse
 * @param setmask ptr to store for bits to be set in the mode flags
 * @param setmask ptr to store for bits to be cleared from the mode flags
 * @return
*/
static void _e2_task_backend_parse_flags (gchar *tmp1,
	mode_t *setmask, mode_t *clearmask)
{
	gint shifter;
	switch (tmp1[0])
	{	//get the scope
		case 'g':
			shifter=3;
			break;
		case 'u':
			shifter=6;
			break;
		case 'o':
		default:
			shifter=0;
			break;
	}
	gint mask=0;	//default
	switch (tmp1[2])
	{  //get the mode
		case  'r':
			mask=S_IROTH;
			break;
		case  'w':
			mask=S_IWOTH;
			break;
		case  'x':
			mask=S_IXOTH;
			break;
		case  's':
			if (tmp1[0] == 'u')
				mask=S_IROTH;
			else if (tmp1[0] == 'g')
				mask=S_IWOTH;
			break;
		case  't':
			mask=S_IXOTH;
			shifter = 9;
			break;
		default:
			break;
	}
	if (tmp1[1] == '+')
		*setmask |= mask << shifter;
	else if (tmp1[1] == '-')
		*clearmask	|= mask << shifter;
}
/**
 * @brief parse octal or symbolic permissions string @a mode
 * @param mode mode string
 * @param data pointer to data struct
 * @return nothing, the results are stored in global static variables
*/
static void _e2_task_backend_parse_mode (gchar *mode, E2_ChmodData *data)
{
	data->setmask = 0;
	data->clearmask = 0;
	if (strchr (mode, ',') != NULL)	//always ascii
	{	//multiple adds or removes
		gchar **tmp1, **tmp2;
		tmp1 = tmp2 = g_strsplit (mode, ",", -1);
		while (*tmp1 != NULL)
		{
			_e2_task_backend_parse_flags (*tmp1, &(data->setmask), &(data->clearmask));
			tmp1++;
		}
		g_strfreev (tmp2);
	}
	else
		if ((strchr (mode, '+') != NULL) || (strchr (mode, '-') != NULL))	//always ascii
			//single add or remove
			_e2_task_backend_parse_flags (mode, &(data->setmask), &(data->clearmask));
	else
	{	//octal string, 3 or 4 digits, no checking
		glong tmp = strtol (mode, NULL, 8);
		data->setmask = (mode_t) tmp;
		data->clearmask = ALLPERMS;
	}
	data->clearmask = ~data->clearmask;
}
/**
 * @brief set permissions of item @a path to @a mode
 * This is used when changing a single item, or recursively changing items.
 * Possible errors:
 * Name errors - see _e2_task_nftwfunc_delete documentation
 *ENOENT
 *  The named file doesn't exist.
 *EPERM
 *  This process does not have permission to change the access permissions
 * of this file. Only the file's owner (as judged by the effective user ID of the process)
 * or a privileged user can change them.
 *EROFS
 *  The file resides on a read-only file system.
 *EFTYPE
 * mode has the S_ISVTX ("sticky") bit set, and the named file is not a directory. 
 *
 * @param path localised string, absolute path of item to be processed
 * @param mode mode_t to be set
 * @return TRUE if succeeds
*/
static gboolean _e2_task_backend_chmod1 (gchar *path, mode_t mode)
{
	if (e2_fs_chmod (path, mode) && errno != ENOENT) //FIXME vfs
	{
		//CHECKME any other errors to ignore ??
		return (e2_task_error_local (_("Cannot change permissions of %s"), path));
	}
	return TRUE;
}
/**
 * @brief callback function for recursive directory chmods
 *
 * Tree is being walked breadth-first, not physical.
 * Dirs are made accessible and writable if not already
 * so and it's permitted, dirs are added to a list to be
 * processed after all the tree has been traversed, 
 * Other items are changed as requested (if possible)
 *
 * @param filename localised string, absolute path of item to change
 * @param statptr pointer to struct stat with info about the current item
 * @param status one of the treewalker-function status codes
 * @param user_data pointer to user-specified data
 * @return competion code: E2TW_CONTINUE if succeeds, others as appropriate
 */
static E2_TwResult _e2_task_twcb_chmod (const char *filename,
	const struct stat *statptr, E2_TwStatus status, E2_ChmodData *user_data)
{
	E2_TwResult retval = E2TW_CONTINUE;	//default error code = none
	mode_t mode, newmode;
	E2_DirEnt *data;
	switch (status)
	{
		case E2TW_DNR:
			//try to make the dir accessible
			mode = statptr->st_mode;
			newmode = (mode | S_IWUSR | S_IXUSR) & ALLPERMS;
			if (e2_fs_chmod (filename, newmode))
			{
				e2_task_error_local (_("Cannot access directory %s"), (gchar *)filename);
				retval = E2TW_FIXME;	//E2TW_SKIP_SUBTREE;
				break;
			}
		case E2TW_D:
			if (status == E2TW_D)
			{	//we didn't get here via a _DNR report
				mode = statptr->st_mode;
				//ensure dir is writable, if we can
				if (!(mode & S_IWUSR))
				{
					newmode = (mode | S_IWUSR) & ALLPERMS;
					if (e2_fs_chmod (filename, newmode))
					{
						e2_task_error_local (_("Cannot write to directory %s"), (gchar *)filename);
						retval = E2TW_FIXME;	//E2TW_SKIP_SUBTREE;
						break;
					}
				}
				else
					newmode = mode;	//set flag
			}
			if (newmode != mode)
			{	//add this dir to list of items to chmod after whole tree is scanned
				data = g_malloc (sizeof (E2_DirEnt));
				data->path = g_strdup (filename);
				if (user_data->scope != E2_RECURSE_OTHER)
					data->mode = (mode & user_data->clearmask) | user_data->setmask; //want to set this value
				else
					data->mode = mode & ALLPERMS;	//want to restore the original value
				user_data->dirdata = g_list_append (user_data->dirdata, data);
			}
		case E2TW_DP:	//ignore dir completion reports
		case E2TW_SL:	//no mode changes for links
		case E2TW_SLN:
			break;
		case E2TW_F:
			if (user_data->scope != E2_RECURSE_DIRS)
			{  //change mode now, if we can
				mode = (statptr->st_mode & user_data->clearmask) | user_data->setmask;
				if (!_e2_task_backend_chmod1 ((gchar *)filename, mode))
					retval = E2TW_FIXME;	//E2TW_STOP;  //prefer continue ??
			}
			break;
		case E2TW_DM:	//directory, not opened due to different file system
		case E2TW_DL:
		case E2TW_NS:	//un-statable item
			retval = E2TW_FIXME;
			break;		
		default:
			retval = E2TW_STOP;
			break;
	}
	if (retval == E2TW_FIXME)
	{
		user_data->continued_after_problem = TRUE;
		retval = E2TW_CONTINUE;	//continue after bad item
	}
	return retval;
}
/**
 * @brief set owners of item @a path
 * This is used when changing a single item, or recursively changing items.
 * Possible errors:
 * Name errors - see _e2_task_nftwfunc_delete documentation
 *EPERM
 *  This process lacks permission to make the requested change.
 *  Only privileged users or the file's owner can change the file's group.
 *  On most file systems, only privileged users can change the file owner;
 *  some file systems allow you to change the owner if you are currently the owner.
 *EROFS
 *   The file is on a read-only file system.
 *
 * @param path localised string, absolute path of item to be processed
 * @param statbuf pointer to a 'completed' stat struct for @a path
 * @param owner_id owner id to be set
 * @param group_id group id to be set
 * @return TRUE if succeeds
 */
gboolean _e2_task_backend_chown1 (gchar *path, struct stat *statbuf,
	uid_t owner_id, gid_t group_id)
{
	gint (*chown_func) () = (S_ISLNK (statbuf->st_mode)) ?
		e2_fs_lchown : e2_fs_chown;

	if (chown_func (path, owner_id, group_id) && errno != ENOENT)
	{
		return (e2_task_error_local (_("Cannot change ownership of %s"), path));
	}
	return TRUE;
}
/**
 * @brief helper function for recursive directory chowns
 * The tree is being scanned breadth-first, no link-through.
 * Dirs are made accessible if not already so and it's permitted,
 * dirs are added to a list to be processed after all the tree has
 * been traversed, other items are changed as requested (if possible)
 *
 * @param filename localised string, absolute path of item to change
 * @param statptr pointer to struct stat with info about the current item
 * @param status one of the treewalker-function status codes
 * @param user_data pointer to user-specified data
 * @return competion code: E2TW_CONTINUE if succeeds, others as appropriate
 */
static E2_TwResult  _e2_task_twcb_chown (const char *filename,
	const struct stat *statptr, E2_TwStatus status,  E2_ChownData *user_data)
{
	E2_TwResult retval = E2TW_CONTINUE;	//default error code = none
	mode_t mode, newmode;
	switch (status)
	{
		case E2TW_DNR:
			//try to make the dir accessible
			mode = statptr->st_mode;
			newmode = (mode | S_IXUSR | S_IWUSR) & ALLPERMS;
			if (e2_fs_chmod (filename, newmode))
			{
				e2_task_error_local (_("Cannot access directory %s"), (gchar *)filename);
				retval = E2TW_FIXME;	//E2TW_SKIP_SUBTREE;
				break;
			}
		case E2TW_D:
			if (status == E2TW_D)
			{	//we didn't get here via a _DNR report
				mode = statptr->st_mode;
				if (!(mode & S_IWUSR))
				{
					newmode = (mode | S_IWUSR) & ALLPERMS;
					if (e2_fs_chmod (filename, newmode))
					{
						e2_task_error_local (_("Cannot write to directory %s"), (gchar *)filename);
						retval = E2TW_FIXME;	//E2TW_SKIP_SUBTREE;
						break;
					}
				}
			}
			//add it to list of dirs to process after whole tree is scanned
			E2_DirEnt *data = g_malloc (sizeof (E2_DirEnt));
			data->path = g_strdup (filename);
			data->mode = statptr->st_mode & ALLPERMS;  //original mode, in case it was changed
			user_data->dirdata = g_list_append (user_data->dirdata, data);
 		case E2TW_DP:	//ignore dir completion reports
			break;
		case E2TW_F:
		case E2TW_SL:  //valid and invalid links
		case E2TW_SLN: //broken links (CHECKME not reported as E2TW_PHYS is used)
			if (!_e2_task_backend_chown1 ((gchar *)filename, (struct stat *) statptr,
					user_data->new_uid, user_data->new_gid))
				retval = E2TW_FIXME;	//E2TW_STOP;	//prefer continue ??
			break;
		case E2TW_DM:	//directory, not opened due to different file system
		case E2TW_DL:
		case E2TW_NS:	//un-statable item (note - this is a physical walk, no link-through problem here)
			retval = E2TW_FIXME;
			break;		
		default:
			retval = E2TW_STOP;
			break;
	}
	if (retval == E2TW_FIXME)
	{
		user_data->continued_after_problem = TRUE;
		retval = E2TW_CONTINUE;	//continue after bad item
	}
	return retval;
}
/**
 * @brief delete item as part of recursive directory delete
 * Possible errors:
 *EACCES
 * The process does not have search permission for a directory 
 * component of @a local_name.
 *ENAMETOOLONG
 * Either the total length of @a local_name is greater than PATH_MAX,
 * or an individual path element of @a local_name has a length greater
 * than NAME_MAX.
 *ENOENT
 * A path element in @a local_name doesn't exist, or is a symbolic
 * link whose target does not exist.
 *ENOTDIR
 * A path element in @a local_name exists, but it isn't a directory.
 *ELOOP
 * Too many symbolic links were resolved while trying to look up the file name.
 * For files only:
 *EACCES
 * No write permission for the directory from which the file is to be removed, 
 * or that directory has the sticky bit set and you do not own the file.
 *EBUSY
 * The file to be deleted is being used by the system in such a way that it 
 * can't be unlinked. For example, you might see this error if the file name
 * specifies the root directory or a mount point for a file system.
 *ENOENT
 * The file to be deleted doesn't exist.
 *EROFS
 * The directory containing the file to be deleted is on a read-only file system
 * For directories only:
 *ENOTEMPTY, EEXIST
 *  The directory to be deleted is not empty (GNU always uses ENOTEMPTY)
 *	
 * This is a callback for the treewalk function
 * @param local_name path of item reported by the walker, localised string
 * @param statbuf pointer to struct stat with data about @a local_name
 * @param status code from the walker, indicating what type of item is being reported
 * @param user_data UNUSED NULL pointer
 * @return E2TW_CONTINUE if delete succeeded, E2TW_STOP if not
 **/
static E2_TwResult _e2_task_twcb_delete (const gchar *local_name,
	const struct stat *statbuf, E2_TwStatus status, void *user_data)
{
	E2_TwResult result = E2TW_CONTINUE;

	switch (status)
	{
		case E2TW_D:	//directory
			//FIXME wrong test? - want to check for accessible and writable in general
//			if (!(statbuf->st_mode & S_IWUSR) || !(statbuf->st_mode & S_IXUSR))
			if (!(statbuf->st_mode & S_IXUSR))
			{
				mode_t thismode = (statbuf->st_mode | S_IXUSR) & ALLPERMS;
				//this will work if is writable in general?
				if (!_e2_task_backend_chmod1 ((gchar *)local_name, thismode))
					result = E2TW_STOP;
			}
			break;	//ignore directory-start reports
		case E2TW_SL:	//symbolic link
		case E2TW_SLN:	//symbolic link naming non-existing file
			if (!(statbuf->st_mode & S_IWUSR))
			{
				mode_t thismode = (statbuf->st_mode | S_IWUSR) & ALLPERMS;
				_e2_task_backend_chmod1 ((gchar *)local_name, thismode);
			}
			if (unlink (local_name))	//FIXME vfs
				result = E2TW_STOP;	//CHECKME which errors to report and/or ignore
			break;
		default:
/*		case E2TW_F:	//not directory or link
		case E2TW_DP:	//directory, all subdirs have been visited
		case E2TW_DL:	//directory, not opened due to tree-depth limit
		case E2TW_DM:	//directory, not opened due to different file system
		case E2TW_DNR:	//unreadable directory
		case E2TW_NS:	//un-statable item
*/
			if (!(statbuf->st_mode & S_IWUSR))
			{
				mode_t thismode = (statbuf->st_mode | S_IWUSR) & ALLPERMS;
				_e2_task_backend_chmod1 ((gchar *)local_name, thismode);
			}
			if (remove (local_name))	//FIXME vfs
				result = E2TW_STOP;	//CHECKME which errors to report and/or ignore
			break;
	}
	return result;
}
/**
 * @brief delete item @a itempath
 * This is for deleting any existing item that is to be replaced
 * as part of a copy, move, rename, link task.
 * It is called at the start of each of those tasks.
 * When applied to a link, that is NOT traversed, so if it is a link
 * to a directory (which requires a recursive delete) that dir
 * will be untouched (not unless/until processed in its own right, anyway).
 * For dirs, errors other than ENOTEMPTY return FALSE.
 * Not-empty dirs trigger a recursive delete, bottom-up,
 * again, no link traverses (dir's in the same tree will be processed
 * at some stage, in their own right)
 * No error processing other than ENOTEMPTY (OK?)
 *
 * @param itempath localised string, relative path of item to delete
 * @return TRUE if succeeds
 */
//CHECKME always return TRUE so the process tries to continue?
// return error code instead of FALSE ??
static gboolean _e2_task_backend_delete (gchar *itempath)
{
	//decide whether or not item is a dir, not looking through links
	struct stat statbuf;
	if (e2_fs_lstat (itempath, &statbuf))
		return (errno == ENOENT);	//ok if nothing to delete, otherwise error
	if (S_ISDIR (statbuf.st_mode))
	{	//dir
		//subject to permission, make sure we can traverse the dir
		//this fails if there is not sufficient write permission to delete the dir's contents
		if (!(statbuf.st_mode & S_IXUSR))
		{
			mode_t thismode = (statbuf.st_mode | S_IXUSR | S_IWUSR) & ALLPERMS;
			if (!_e2_task_backend_chmod1 (itempath, thismode))
				return FALSE;
		}
		//try for simple deletion first
		if (!rmdir (itempath))	//FIXME vfs
			return TRUE;
		else if (!(errno == ENOTEMPTY || errno == EEXIST))	//FIXME vfs
			return FALSE;	//error if failed for some reason other than not-empty
		//dir not empty, recursively delete its contents
		//flags for: no link follow, depth-first
		E2_TwFlags flags = E2TW_PHYS | E2TW_DEPTH;
		//FIXME which errors to report and/or ignore
		return (e2_fs_tw (itempath, _e2_task_twcb_delete, NULL, E2_DIRNEST_LIMIT, flags));
	}
	else	//not dir
		//FIXME which errors to report and/or ignore
		return (! unlink (itempath));	//FIXME vfs
}
#ifdef LATE_DELETE
/**
 * @brief replace @a replace with @a dest, with error message if not possible
 * @param dest localised path of item to be deleted
 * @param replace localised path of item to be replaced by @a dest
 * @return TRUE if the replacement was completed
 */
static gboolean _e2_task_backend_overwrite (gchar *dest, gchar *replace)
{
	if (!_e2_task_backend_delete (dest))
	{
		gchar *msg = (e2_fs_is_dir3 (dest)) ?
		_("Cannot delete part or all of existing %s"):
		_("Cannot delete existing %s");
		return (e2_task_error_local (msg, dest));
	}
	else
		rename (replace, dest);	//FIXME vfs

	return TRUE;
}
#endif
/**
 * @brief copy the regular file|fifo|socket|device @a src to @a dest
 * If the item is a regular file, this copies the actual file
 * - links are traversed to find the file
 * Prints error message if @a src can't be read, or @a dest can't
 * be created, then terminates
 * Content is funneled via a buffer on the stack.
 * @param src localised string, absolute path of item to copy
 * @param dest localised string, absolute path of copy destination
 * @param sametime TRUE to preserve the atime and ctime
 * @param srcstat pointer to stat struct for @a src
 * @return TRUE if succeeds
 */
static gboolean _e2_task_backend_filecopy (gchar *src, gchar *dest, 
	gboolean sametime, struct stat *srcstat)
{
	//FIXME vfs
//	struct stat src_sb;
	struct stat dest_sb;

#ifdef LATE_DELETE
	//work with a temp name only if the dest exists already
	gchar *temp = e2_task_tempname (dest);
	printd (DEBUG, "dest path is %s", temp);
#endif
	if (S_ISREG (srcstat->st_mode))
	{
		gint src_desc = TEMP_FAILURE_RETRY (open (src, O_RDONLY));  //O_NOATIME != 'cp', O_NOFOLLOW causes spurious error
		if (src_desc < 0)
		{
#ifdef LATE_DELETE
			if (temp != dest)
				g_free (temp);
#endif
			return (e2_task_error_local (_("Cannot open %s for reading"), src));
		}
		mode_t dest_mode = (srcstat->st_mode | S_IWUSR) & ALLPERMS;
		gint dest_desc =
#ifdef LATE_DELETE
		TEMP_FAILURE_RETRY (open (temp, O_WRONLY | O_CREAT | O_EXCL, dest_mode));
#else
		TEMP_FAILURE_RETRY (open (dest, O_WRONLY | O_CREAT | O_EXCL, dest_mode));
#endif
		if (dest_desc < 0)
		{
			e2_task_error_local (_("Cannot create file %s"), dest);
			TEMP_FAILURE_RETRY (close (src_desc));	//FIXME vfs
#ifdef LATE_DELETE
			if (temp != dest)
				g_free (temp);
#endif
			return FALSE;
		}
		if (e2_fs_fstat (dest_desc, &dest_sb))
		{
			e2_task_error_local (_("Cannot fstat %s"), dest);
			TEMP_FAILURE_RETRY (close (src_desc));	//FIXME vfs
			TEMP_FAILURE_RETRY (close (dest_desc)); //FIXME vfs
#ifdef LATE_DELETE
			if (temp != dest)
				g_free (temp);
#endif
			return FALSE;
		}

		blksize_t buf_size = dest_sb.st_blksize;  //or src ??
		gchar *buf = g_alloca (buf_size);
		//CHECKME suspend access-monitoirng here ...
		for (;;)
		{
			ssize_t n_read = read (src_desc, buf, buf_size); //FIXME vfs
			if (n_read < 0)
			{
				e2_task_error_local (_("Error reading file %s"), src);
				TEMP_FAILURE_RETRY (close (src_desc));	//FIXME vfs
				TEMP_FAILURE_RETRY (close (dest_desc));	//FIXME vfs
#ifdef LATE_DELETE
				unlink (temp);	//FIXME vfs
				if (temp != dest)
					g_free (temp);
#else
				unlink (dest);	//FIXME vfs
#endif
				return FALSE;
			}
			if (n_read == 0)
				break;

			ssize_t n_write = 0;
			while (n_write < n_read)
			{
				n_write = TEMP_FAILURE_RETRY (write (dest_desc, buf, n_read));	//FIXME vfs
				if (n_write == -1)
				{
					e2_task_error_local (_("Error writing file %s"), dest);
					TEMP_FAILURE_RETRY (close (src_desc));	//FIXME vfs
					TEMP_FAILURE_RETRY (close (dest_desc));	//FIXME vfs
#ifdef LATE_DELETE
					unlink (temp);	//FIXME vfs
					if (temp != dest)
						g_free (temp);
#else
					unlink (dest);	//FIXME vfs
#endif
					return FALSE;
				}
			}
		}
		TEMP_FAILURE_RETRY (close (src_desc));	//FIXME vfs
		TEMP_FAILURE_RETRY (close (dest_desc));	//FIXME vfs
	}

	//the following mimic the 'cp' app.
	//we do not clear the type-flags when creating these, as
	//that would revert the item to a regular file
	//but we do re-set the permissions explicitly, to circumvent
	//the mode mask
	else if (S_ISBLK (srcstat->st_mode)
			|| S_ISCHR (srcstat->st_mode)
			|| S_ISSOCK (srcstat->st_mode))	//CHECKME mknod() doco says nothing about handling sockets this way
	{
#ifdef LATE_DELETE
		if (mknod (temp, srcstat->st_mode, srcstat->st_rdev))	//FIXME vfs
		{
			if (temp != dest)
				g_free (temp);
#else
		if (mknod (dest, srcstat->st_mode, srcstat->st_rdev))	//FIXME vfs
		{
#endif
			return (e2_task_error_local (_("Cannot create special file %s"), dest));
		}
	}

	else if (S_ISFIFO (srcstat->st_mode))
	{
#ifdef LATE_DELETE
		if (mkfifo (temp, srcstat->st_mode))	//FIXME vfs
		{
			if (temp != dest)
				g_free (temp);
#else
		if (mkfifo (dest, srcstat->st_mode))	//FIXME vfs
		{
#endif
			return (e2_task_error_local (_("Cannot create FIFO %s"), dest));
		}
	}
#ifdef LATE_DELETE
	if (temp != dest && !_e2_task_backend_overwrite (dest, temp))
	{
		g_free (temp);
		return FALSE;
	}
	if (temp != dest)
		g_free (temp);
#endif
	if (S_ISREG (srcstat->st_mode))
		e2_fs_chmod (dest, srcstat->st_mode & ALLPERMS);
	else
		e2_fs_chmod (dest, srcstat->st_mode);
	if (sametime) // && !e2_fs_lstat (src, srcstat))
	{
		struct utimbuf tb;
		tb.modtime = srcstat->st_mtime;
		tb.actime = srcstat->st_atime;
//		e2_fs_utime (src, &tb);	//FIXME vfs
		e2_fs_utime (dest, &tb);
	}

	return TRUE;
}
/**
 * @brief 'copy' soft link itself (ie no traverse) during a recursive dir copy
 * This creates a new symlink to the original target (which may be relative).
 * Possible errors:
 *EEXIST
 *  There is already an existing item named newname
 *  (meaning, here, some unknown process created that item since we 
 * deleted one of that same name !) 
 *EROFS
 *  The new link item can't be created on a read-only file system.
 *ENOSPC
 *   The directory or file system cannot be extended to make the new link.
 *EIO
 *   A hardware error occurred while reading or writing data on the disk.
 *
 * It will print some error messages as needed
 * @param src localised string, absolute path of link to copy
 * @param dest localised string, absolute path of new link name
 * @return TRUE if succeeds
 */
static gboolean _e2_task_backend_linkcopy (gchar *src, gchar *dest)
{
	gchar linkto[PATH_MAX];	//+NAME_MAX];
	gint len = readlink (src, linkto, sizeof (linkto));	//FIXME vfs
	if (len > 0)
	{
		linkto[len] = '\0';
#ifdef LATE_DELETE
		gchar *temp = e2_task_tempname (dest);
		if (!symlink (linkto, temp))	//FIXME vfs
		{
			gboolean retval = TRUE;
			if (temp != dest)
			{
				retval = _e2_task_backend_overwrite (dest, temp);
				g_free (temp);
			}
			return retval;
		}
		if (temp != dest)
			g_free (temp);
#else
		if (!symlink (linkto, dest))	//FIXME vfs
			return TRUE;
#endif
		return (e2_task_error_local (_("Cannot create new link to %s"), linkto));
	}
	else
		return (e2_task_error_local (_("Cannot get target info for link %s"), src));
}
/**
 * @brief make a directory @a dest during a recursive dir copy task
 * This creates the dir always with o+x,o+w permissions, then
 * if need be stores the path and 'real' mode, for later setting of that 
 * real mode.
 * An error message is printed if the task fails for any reason.
 * Possible errors:
 * Name errors - see _e2_task_nftwfunc_delete documentation
 *EACCES
 * No write permission for the parent directory to which the new
 * directory is to be added.
 *EEXIST
 * A file named filename already exists (but in this context, it
 * should have been deleted already)
  *EMLINK
 * The parent directory has too many links (entries).
 *ENOSPC
 * The file system doesn't have enough room to create the
 * new directory.
 *EROFS
 * The parent directory of the directory being created is on
 * a read-only file system
 *
 * @param dest localised string, absolute path of dir to creae
 * @param mode stat mode flags of the new dir
 * @param dirdata list of data structs for created dirs
 * @return TRUE if succeeds 
 */
static gboolean _e2_task_backend_mkdir (gchar *dest, mode_t mode, GList **dirdata)
{
	//need to set USR bits for mkdir to accept the change ?
	mode_t thismode = (mode | S_IXUSR | S_IWUSR) & ALLPERMS;
	if (e2_fs_mkdir (dest, thismode))
	{
		return (e2_task_error_local (_("Cannot create new directory %s"), dest));
	}
//save mode of ALL dirs - something in the TW changes things ...
//	else
//		if (!(mode & S_IXUSR) || !(mode & S_IWUSR))
//		{
			//need to store data for the dir, with actual mode
			E2_DirEnt *data = g_malloc (sizeof (E2_DirEnt));
			data->path = g_strdup (dest);
			data->mode = mode & ALLPERMS;
			*dirdata = g_list_append (*dirdata, data);
//		}
	return TRUE;
}
/**
 * @brief callback function for recursive directory copying
 * This is called for each item in the directory to be copied.
 * Treewalk is breadth-first, not physical
 * @param filename localised string, path of item to copy
 * @param statptr pointer to struct stat with info about the current item
 * @param status one of the treewalker-function status codes
 * @param user_data pointer to user-specified data
 * @return competion code: E2TW_CONTINUE if succeeds, others as appropriate
 */
static E2_TwResult _e2_task_twcb_copy (const gchar *filename, 
	struct stat *statptr, E2_TwStatus status, E2_CopyData *user_data)
{
	E2_TwResult retval;
	gchar *dest;
	//no glib string functions due to localised strings
/*	if (strstr (filename, user_data->oldroot) == filename)
		dest = e2_utils_strcat (user_data->newroot, 
			(gchar *)filename+strlen (user_data->oldroot));
	else
		dest = strdup (filename);
*/
	dest = e2_utils_strcat (user_data->newroot, 
		(gchar *)filename+user_data->oldroot_len);

	switch (status)
	{
		case E2TW_F:  //file
			retval = (_e2_task_backend_filecopy ((gchar *)filename, dest,
					user_data->sametime, statptr)) ? 
				E2TW_CONTINUE : E2TW_FIXME;	//E2TW_STOP;	//CHECKME continue??
			break;
		case E2TW_D:  //dir
			retval = (_e2_task_backend_mkdir (dest, statptr->st_mode,
				&(user_data->dirdata))) ? E2TW_CONTINUE : E2TW_FIXME;	//E2TW_SKIP_SUBTREE;
			break;
		case E2TW_DP:	//directory, all subdirs have been visited
			retval = E2TW_CONTINUE;
			break;
		case E2TW_SL:  //link
		case E2TW_SLN:  //bad link CHECKME why bother trying to relink?
			retval = (_e2_task_backend_linkcopy ((gchar *)filename, dest)) ?
				E2TW_CONTINUE : E2TW_FIXME;	//FTW_STOP;	//CHECKME continue??
			break;
//		case E2TW_DL:	//directory, not opened due to tree-depth limit
//		case E2TW_DM:	//directory, not opened due to different file system
//			retval = E2TW_FIXME;
//			break;		
//		case E2TW_DNR:	//unreadable directory
//		case E2TW_NS:	//un-statable item
		default:
			retval = E2TW_FIXME;	//E2TW_STOP;
			break;
	}

	g_free (dest);

	if (retval != E2TW_CONTINUE)
		printd (DEBUG, "Error code for %s is %d", filename, errno);
//	any_error |= retval;
//	return FTW_CONTINUE;	//never abort the process
	if (retval == E2TW_FIXME)
	{
		user_data->continued_after_problem = TRUE;
		retval = E2TW_CONTINUE;	//continue after bad item
	}
	return retval;  //must report (some/all?) errors, else it keeps trying & hangs
}

  /************************/
 /******** public *********/
/************************/

/**
 * @brief copy item @a src to @a dest
 * Directories, links, others are handled distinctly
 * @param src localised string, absolute path of item to copy
 * @param dest localised string, absolute path of copy destination
 * @param preserve TRUE to preserve mtime, atime of copied file(s)
 * @return TRUE if copy succeeds
 */
gboolean e2_task_backend_copy (gchar *src, gchar *dest, gboolean preserve)
{
#ifndef LATE_DELETE
	if (e2_fs_access3 (src, R_OK))
	{
		return (e2_task_error_local (_("Cannot read %s"), src));
	}
	if (!_e2_task_backend_delete (dest))
	{
		return (e2_task_error_local (_("Cannot delete existing %s"), dest));
	}
#endif
	//decide whether src is a dir, link or other
	struct stat statbuf;
#ifdef LATE_DELETE
	if (e2_fs_lstat (src, &statbuf))
	{
		return (e2_task_error_local (_("Cannot read %s"), src));
	}
#else
	e2_fs_lstat (src, &statbuf);
#endif
	if (S_ISDIR (statbuf.st_mode))
	{	//recursive copy dir contents
		E2_CopyData data;
		data.sametime = preserve;
		data.continued_after_problem = FALSE;
		data.oldroot_len = strlen (src);
#ifdef LATE_DELETE
		//work with a temp name (top-level dir) only if the dest exists already
		data.newroot = e2_task_tempname (dest);
#else
		data.newroot = dest;
#endif
		data.dirdata = NULL;

		E2_TwFlags flags = E2TW_DEFAULT;	//breadth-first
		gboolean retval = e2_fs_tw (src, _e2_task_twcb_copy, &data, E2_DIRNEST_LIMIT, flags);

		//cleanups
		if (data.dirdata != NULL)
		{	//revert new dir permissions, LIFO order
			GList *base = g_list_last (data.dirdata);
			for (; base!=NULL ; base=base->prev)
			{
				E2_DirEnt *dirfix = base->data;
				e2_fs_chmod (dirfix->path, dirfix->mode);
				g_free (dirfix->path);
				g_free (dirfix);
			}
			g_list_free (data.dirdata);
		}
		if (data.continued_after_problem)
		{
			gchar *msg = g_strdup_printf (_("Cannot copy all of %s"), src);
			e2_output_print_error (msg, TRUE);
		}
		//after reverting modes (coz the temp names are stored in the list
#ifdef LATE_DELETE
		if (retval)
		{
			if (data.continued_after_problem)
			{
				if (!e2_fs_access2 (data.newroot))
				{
					gchar *temp2, *temp3;
					temp2 = g_strconcat (dest, "-",_("incomplete"), NULL);
					temp3 = e2_task_tempname (temp2);
					retval = e2_task_backend_rename (data.newroot, temp3);
					if (temp3 != temp2)
						g_free (temp3);
					g_free (temp2);
				}
			}
			else if (data.newroot != dest)
				retval = _e2_task_backend_overwrite (dest, data.newroot);
		}
		retval &= !data.continued_after_problem;
		if (data.newroot != dest)
			g_free (data.newroot);
#else
		retval &= !data.continued_after_problem;
#endif
		return retval;
	}
	else //not a dir
		if (S_ISLNK (statbuf.st_mode))
			//handle links separately, to avoid 'look-through'
			return (_e2_task_backend_linkcopy (src, dest));			
	else	//not dir or link
		return (_e2_task_backend_filecopy (src, dest, preserve, &statbuf));
}
/**
 * @brief move file @a src to @a dest
 * @param src localised string, absolute path of item to move
 * @param dest localised string, absolute path of move destination
 * @return TRUE if move succeeds
 */
gboolean e2_task_backend_move (gchar *src, gchar *dest)
{
#ifdef LATE_DELETE
	//work with a temp name only if the dest exists already
	gchar *temp = e2_task_tempname (dest);
	if (rename (src, temp))	//FIXME vfs
#else
	if (e2_fs_access3 (src, W_OK | R_OK))
	{
		return (e2_task_error_local (_("Cannot delete %s"), src));
	}
	if (!_e2_task_backend_delete (dest))
	{
		return (e2_task_error_local (_("Cannot delete existing %s"), dest));
	}
	if (rename (src, dest))	//FIXME vfs
#endif
	{
#ifdef LATE_DELETE
		if (dest != temp)
			g_free (temp);
#endif
		if (errno == EXDEV) //FIXME vfs
			return ( e2_task_backend_copy (src, dest, TRUE)
				&& e2_task_backend_delete (src) );
		else
		{
			gchar *utf = F_FILENAME_FROM_LOCALE (src);
			gchar *utf2 = F_FILENAME_FROM_LOCALE (dest);
			gchar *msg = g_strdup_printf (_("Cannot move %s to %s"), utf, utf2);
			e2_task_error (msg);
			F_FREE (utf);
			F_FREE (utf2);
			g_free (msg);
			return FALSE;
		}
	}
#ifdef LATE_DELETE
	else if (dest != temp && !_e2_task_backend_overwrite (dest, temp))
	{
		g_free (temp);
		return FALSE;
	}
	if (dest != temp)
		g_free (temp);
#endif
	return TRUE;
}
/**
 * @brief create soft link named @a name to @a target
 * @param target localised string, absolute path of item to link to
 * @param name localised string, absolute or relative path of link name
 * @return TRUE if creation succeeds
 */
gboolean e2_task_backend_link (gchar *target, gchar *name)
{
#ifdef LATE_DELETE
	//work with a temp name only if the dest exists already
	gchar *temp = e2_task_tempname (name);
	if (symlink (target, temp))	//FIXME vfs
#else
	if (e2_fs_access3 (target, R_OK))
	{
		return (e2_task_error_local (_("Cannot find %s"), target));
	}
	if (!_e2_task_backend_delete (name))
	{
		return (e2_task_error_local (_("Cannot delete existing %s"), name));
	}
	if (symlink (target, name))	//FIXME vfs
#endif
	{
#ifdef LATE_DELETE
		if (temp != name)
			g_free (temp);
#endif
		return (e2_task_error_local (_("Cannot create link to %s"), target));
	}
#ifdef LATE_DELETE
	else if (temp != name && !_e2_task_backend_overwrite (name, temp))
	{
		g_free (temp);
		return FALSE;
	}
	if (temp != name)
		g_free (temp);
#endif
	return TRUE;
}
/**
 * @brief rename item named @a oldsrc to @a newsrc
 * @param oldsrc localised string, absolute path of item to rename
 * @param newsrc localised string, absolute path of new name
 * @return TRUE if rename succeeds
 */
gboolean e2_task_backend_rename (gchar *oldsrc, gchar *newsrc)
{
#ifdef LATE_DELETE
	//work with a temp name only if the dest exists already
	gchar *temp = e2_task_tempname (newsrc);
	if (rename (oldsrc, temp))	//FIXME vfs
#else
	if (e2_fs_access3 (oldsrc, W_OK))
	{
		return (e2_task_error_local (_("Cannot rename %s"), oldsrc));
	}
	if (!_e2_task_backend_delete (newsrc))
	{
		return (e2_task_error_local (_("Cannot delete existing %s"), newsrc));
	}
	if (rename (oldsrc, newsrc))	//FIXME vfs
#endif
	{
#ifdef LATE_DELETE
		if (temp != newsrc)
			g_free (temp);
#endif
		gchar *utf = F_FILENAME_FROM_LOCALE (oldsrc);
		gchar *utf2 = F_FILENAME_FROM_LOCALE (newsrc);
		gchar *msg = g_strdup_printf (_("Cannot rename %s to %s"), utf, utf2);
		e2_task_error (msg);
		F_FREE (utf);
		F_FREE (utf2);
		g_free (msg);
		return FALSE;
	}
#ifdef LATE_DELETE
	else if (temp != newsrc && !_e2_task_backend_delete (newsrc))
	{
		g_free (temp);
		return (e2_task_error_local (_("Cannot delete existing %s"), newsrc));
	}
	if (rename (temp, newsrc))	//FIXME vfs
	{
		if (temp != newsrc)
			g_free (temp);
		return FALSE;
	}
	if (temp != newsrc)
		g_free (temp);
#endif
	return TRUE;
}
/**
 * @brief delete item @a path
 * @param path localised string, absolute path of item to delete
 * @return TRUE if delete succeeds
 */
gboolean e2_task_backend_delete (gchar *path)
{
//	if (e2_fs_access3 (path, W_OK))
//	{
//		return (e2_task_error_local (_("Cannot delete %s"), oldsrc));
//	}
	if (_e2_task_backend_delete (path))
		return TRUE;

	gchar *msg = (e2_fs_is_dir3 (path)) ?
		_("Cannot delete part or all of %s"):
		_("Cannot delete %s");
	return (e2_task_error_local (msg, path));
}
/**
 * @brief change permissions of item @a path to @a mode
 *
 * Only an item's owner (as judged by the effective uid of the process)
 * or a privileged user, can change item permissions.
 * If invoked on a non-dir, or a dir without recursion, it is processed
 * here. If recursion is invoked on a dir, a ntfw funtion is invoked. 
 * By that, all nested dir access permissons will be set to include x,
 * non-dir items will be have their new permissions set when
 * they are 'reported'. Finally, here, dirs will be set to their proper
 * permissions (bottom-up order) at the end of the process
 * Recursive chmod works on the host filesystem only.
 * Links are not affected, their target is processed. (OK ??)
 *
 * @param path localised string, absolute path of item to process
 * @param mode permissions string, symbolic or octal
 * @param recurse code for recurse type (ignored if @a path is not a dir)
 * @return TRUE if operation succeeds
*/
gboolean e2_task_backend_chmod (gchar *path, gchar *mode,
	E2_RecurseType recurse)
{
	E2_ChmodData data;
	if (recurse != E2_RECURSE_NONE)
	{
		//decide whether src is a dir or not
		struct stat statbuf;
		if (e2_fs_stat (path, &statbuf))  //looks _through_ links
		{
			//abort if we can't find the item
			return (e2_task_error_local (_("Cannot get current data for %s"), path));
		}
		if (S_ISDIR (statbuf.st_mode))
		{	//recursive chmod
			//park parameters where they can be accessed by helpers
			data.continued_after_problem = FALSE;
			data.scope = recurse;
			_e2_task_backend_parse_mode (mode, &data);
			//no processed dirs yet
			data.dirdata = NULL;

			E2_TwFlags flags = E2TW_MOUNT;	//this filesystem only, breadth-first
			gboolean retval = e2_fs_tw (path, _e2_task_twcb_chmod, &data, E2_DIRNEST_LIMIT, flags);

			if (data.dirdata != NULL)
			{	//change/revert dir permissions, LIFO (=bottom-up) order
				GList *base = g_list_last (data.dirdata);
				for (; base!=NULL ; base=base->prev)
				{
					E2_DirEnt *dirfix = base->data;
					_e2_task_backend_chmod1 (dirfix->path, dirfix->mode);
					g_free (dirfix->path);
					g_free (dirfix);
				}
				g_list_free (data.dirdata);
			}
			if (data.continued_after_problem)
			{
				gchar *msg = g_strdup_printf (_("Cannot change permissions of all of %s"), path);
				e2_output_print_error (msg, TRUE);
				retval = FALSE;
			}
			return retval;
		}
		else //not dir, handle without recurse
			recurse = E2_RECURSE_NONE;
	}
	if (recurse == E2_RECURSE_NONE)
	{
		//get current mode
		struct stat statbuf;
		if (e2_fs_lstat (path, &statbuf))  //no link pass-thru
		{
			return (e2_task_error_local (_("Cannot get current permissions of %s"), path));
		}
		if (S_ISLNK (statbuf.st_mode))
		{
			return FALSE;	//link permissions can't be changed
		}
		//decode the mode string
		_e2_task_backend_parse_mode (mode, &data);
		//get new mode
		mode_t new_mode = (statbuf.st_mode & data.clearmask) | data.setmask;

		return (_e2_task_backend_chmod1 (path, new_mode));
	}
	return FALSE; //warning prevention only
}
/**
 * @brief change ownership of item @a path to @a owner_id and @a group_id
 *
 * If invoked on a non-dir, or a dir without recursion, it is processed
 * here. If recursion is invoked on a dir, a ntfw funtion is invoked. 
 * By that, all nested dir access permissons will be set (breadth-first walk)
 * to include x, non-dir items will be have their new permissions set when
 * they are 'discovered'. Finally, here, dirs will be reverted to their proper
 * permissions (bottom-up order) at the end of the process
 * Recursive chown works on the host filesystem only.
 * Links are not affected, their target is processed. (CHECKME ok??)
 * Only an item's owner (as judged by the effective uid of the process)
 * or a privileged user, can change item permissions.
 *
 * @param path localised string, absolute path of item to process
 * @param owner_id uid to be applied
 * @param group_id gid to be applied
 * @param recurse_dirs TRUE to recurse (ignored if @a path is not a dir)
 * @return TRUE if the operation succeeds
 */
gboolean e2_task_backend_chown (gchar *path, uid_t owner_id,
	gid_t group_id, gboolean recurse_dirs)
{
	struct stat statbuf;
//	if (e2_fs_stat (path, &statbuf))  //looking _through_ links
	if (e2_fs_lstat (path, &statbuf))  //_not_ looking through links
	{
		return (e2_task_error_local (_("Cannot get current data for %s"), path));
	}
	if (recurse_dirs)
	{
		//decide whether or not item is dir
		if (S_ISDIR (statbuf.st_mode))
		{	//dir
			//park the data where they can be accessed by the helper
			E2_ChownData data = { FALSE, owner_id, group_id, NULL };

			//flags for: breadth-first, this filesystem only, no link pass-thru
			E2_TwFlags flags = E2TW_MOUNT | E2TW_PHYS;
			gboolean retval = e2_fs_tw (path, _e2_task_twcb_chown, &data, E2_DIRNEST_LIMIT, flags);

			//cleanups
			if (data.dirdata != NULL)
			{	//chmod and revert all dirs' permissions, LIFO order
				GList *base = g_list_last (data.dirdata);
				for (; base!=NULL ; base=base->prev)
				{
					E2_DirEnt *dirfix = base->data;
					_e2_task_backend_chown1 (dirfix->path, &statbuf, owner_id, group_id);
					_e2_task_backend_chmod1 (dirfix->path, dirfix->mode);
					g_free (dirfix->path);
					g_free (dirfix);
				}
				g_list_free (data.dirdata);
			}
			if (data.continued_after_problem)
						{
				gchar *msg = g_strdup_printf (_("Cannot change ownewship of all of %s"), path);
				e2_output_print_error (msg, TRUE);
				retval = FALSE;
			}
			return retval;
		}
		else	//not dir, ignore the recurse flag
			recurse_dirs = FALSE;
	}
	if (!recurse_dirs)
	{
		return (_e2_task_backend_chown1 (path, &statbuf, owner_id, group_id));
	}
	return FALSE;  //this can never happen, but it prevents warning
}
/**
 * @brief perform the default action for a specfied item
 *
 * If no default action exists, the user is invited to create one
 *
 * @param info data struct for the item to be 'actioned'
 *
 * @return TRUE if the operation succeeded
 */
gboolean e2_task_backend_open (FileInfo *info)
{
	gchar *ext, *freeme, *utf;
	e2_filelist_disable_refresh();
	utf = F_FILENAME_FROM_LOCALE (info->filename);
	if (e2_fs_is_dir (info))
		freeme = ext = g_strconcat (".", _("<directory>"), NULL);
	else if (e2_fs_is_executable (info))
		freeme = ext = g_strconcat (".", _("<executable>"), NULL);
	else
	{
		freeme = NULL; //not a constructed extension, no free needed
		ext = strchr (utf, '.');	//assumes '.' is ascii
	}
	if (ext == NULL) //no extension
	{
		if (e2_fs_is_executable (info))
		{
			gchar command_line[NAME_MAX+2];
			g_snprintf (command_line, sizeof(command_line), E2_COMMAND_PREFIX"%s", utf);
			e2_command_run (command_line, E2_COMMAND_RANGE_DEFAULT);
			ext = (gchar *)0x2;	//prevent dialog
		}
		else if (e2_fs_is_text (info->filename))
		{
			e2_task_backend_view (utf);
			ext = (gchar *)0x2;	//prevent dialog
		}
		else
			ext = (gchar *)0x1;	//force other dialog
	}
	else if (ext == info->filename) //it's a dot file
	{
		if (e2_fs_is_text (info->filename))
			e2_task_backend_view (utf);
		else
			ext = (gchar *)0x1;	//force other dialog
	}
	else // the file has an extension
	{
		gchar *action, *sep;
		do
		{
			ext++;	//NCHR(ext); //skip the . prefix
			action = e2_filetype_get_default_action (ext);
			if (action != NULL)
			{
				if ((sep = strchr (action, '@')) != NULL)	//if always ascii @, don't need g_utf8_strchr()
				  e2_filetype_exec_action (sep+1);
				else
				  e2_filetype_exec_action (action);
				g_free (action);
				break;
			}
		} while ((ext = strchr (ext, '.')) != NULL);	//if always ascii '.', don't need g_utf8_strchr()
	}

	gboolean retval= FALSE;
	if (ext == NULL) //no matching extension
		e2_filetype_dialog_create (utf, TRUE); //enter filetype, command or view
	else if (ext == GINT_TO_POINTER (1)) //no extension or hidden file
		e2_filetype_dialog_create (utf, FALSE); //enter command or view
	else
		retval = TRUE;

	if (freeme != NULL)
		g_free (freeme);
	F_FREE (utf);
	WAIT_FOR_EVENTS
	e2_filelist_enable_refresh();
	return retval;
}
/**
 * @brief view @a filename with the external viewer specified in the options
 * @param filename utf8 string, path of file to be viewed
 * @return TRUE always
 **/
gboolean e2_task_backend_view (gchar *filename)
{
	if (e2_option_bool_get ("use-external-viewer"))
	{
		gchar *command = g_strdup_printf ("$[command-viewer] '%s'",	//no translation
			filename);
		e2_command_run (command, E2_COMMAND_RANGE_DEFAULT);
		g_free (command);
	}
	else
		e2_view_dialog_show (filename);

	return TRUE;
}
