/* $Id: e2_fs.c 1868 2006-04-26 23:52:59Z tpgww $
 *
 * Copyright (C) 2005-2006 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/filesystem/e2_fs.c
 * @brief filesystem io functions
 **/

#include "emelfm2.h"
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <langinfo.h>
#include <pwd.h>
#include <grp.h>
#include "e2_fs.h"
#include "e2_utf8.h"
#include "e2_option_tree.h"

extern GMutex *task_mutex;

#if 0
/**
 * @brief check if any x permission flag is set in @a mode
 * @param mode
 * @return TRUE if any x flag is set
 **/
//FIXME this test should be for current user only, cf writable test
static gboolean _e2_fs_S_ISEXE (mode_t mode)
{
	if ((S_IXUSR & mode) || (S_IXGRP & mode) || (S_IXOTH & mode))
		return TRUE;
	return FALSE;
}
#endif //0

#ifndef FILES_UTF8ONLY
//check whether a character encoding is compatible with utf
static gboolean _e2_fs_charset_is_utf (gchar *filecoding)
{
	return (!strcmp (filecoding, "UTF-8") ||
			strstr (filecoding, "ASCII") != NULL ||
			!strcmp (filecoding, "ANSI_X3.4-1968") ||
			!strcmp (filecoding, "646") ||
			!strcmp (filecoding, "ISO646") ||
			!strcmp (filecoding, "ISO_646.IRV"));
}
//check whether the locale language is one of the english variants
static gboolean _e2_fs_language_is_english (void)
{
	const gchar *lang = g_getenv ("LANGUAGE");
	if (lang == NULL)
		lang = g_getenv ("LANG");
	return (lang != NULL && strstr (lang, "en_") != NULL);
}
/**
 * @brief check whether the local filesystem encoding is utf8 or ascii, and setup for conversion accordingly
 *
 * @return 
 **/
//FIXME vfs may need to do this for each displayed dir
void e2_fs_check_coding (void)
{
	gchar *filecoding;
	app.utf8_filenames = FALSE;
	//1st, check what the user specified about encoding
	if (e2_cl_options.encoding != NULL)
		app.utf8_filenames = _e2_fs_charset_is_utf (e2_cl_options.encoding);
	//otherwise, try to find something from the environment ...
	else
	{	
		//for glib, $G_FILENAME_ENCODING has priority over $G_BROKEN_FILENAMES
		filecoding = (gchar *) g_getenv ("G_FILENAME_ENCODING");
		if (filecoding == NULL)
		{
			//check what the locale data knows about the encoding
			filecoding = (gchar *) g_getenv ("G_BROKEN_FILENAMES");
			if (filecoding != NULL)
			{
				filecoding = nl_langinfo (CODESET);
				app.utf8_filenames = _e2_fs_charset_is_utf (filecoding);
			}
			else
				app.utf8_filenames = _e2_fs_language_is_english ();
		}
		else if (g_str_equal (filecoding, "@locale"))
		{
			const gchar *charset;
			app.utf8_filenames = g_get_charset (&charset)
			//we assume locales using english language have ascii filenames
			//(which are compatible with utf)
				|| _e2_fs_language_is_english ();
		}
		else
		{
		//FIXME maybe a list, check only its 1st member
			app.utf8_filenames = _e2_fs_charset_is_utf (filecoding)
				|| _e2_fs_language_is_english ();
		}
	}
	//set pointers to conversion functions
	if (app.utf8_filenames)
	{
 		e2_fname_to_locale = e2_fname_from_locale = e2_utf8_not_converted;
 		e2_fname_dupto_locale = e2_fname_dupfrom_locale = g_strdup;
		e2_fname_free = e2_utf8_not_freed;
	}		
	else
	{
		e2_fname_to_locale = e2_fname_dupto_locale = e2_utf8_filename_to_locale;
//#ifdef E2_CURRENTGTK_VERSION && GTK_CHECK_VERSION = (2, 6, 0)
		e2_fname_from_locale = e2_fname_dupfrom_locale = g_filename_display_name;
/*#else
		e2_fname_from_locale = e2_fname_dupfrom_locale = e2_utf8_filename_from_locale;
#endif */
		e2_fname_free = g_free;
	}
}
#endif

/**
 * @brief check whether @a filename exists
 * Can't use the function access(), which (for glibc at least)
 * looks through links
 * Errors other than non-readable are ignored
 * NOTE this does not work properly on FAT32 (etc??) FAT32 is case-insensitive
 * @param filename localised name of file to test
 *
 * @return -1 (TRUE) if @filename does not exist (like access(F_OK))
 **/
gint e2_fs_access2 (gchar *filename)
{
	struct stat statbuf;
	return (e2_fs_lstat (filename, &statbuf));
}
/**
 * @brief check whether @a filename can be 'used' in a manner consistent with flags in @a how
 * access() (for glibc at least) looks through links.
 * This mimics the glibc function access(), except that it does
 * not look through links
 *
 * @param filename localised name of file to test
 * @param how Or'd combination of R_OK, W_OK, X_OK and/or F_OK
 * @return -1 (TRUE) if all @how condition(s) are _not_ satisfied (this is the way access() works)
 **/
gint e2_fs_access3 (gchar *filename, gint how)
{
	static uid_t myuid = -1;
	static gid_t mygid = -1;
	struct stat statbuf;
	if (e2_fs_lstat (filename, &statbuf))
		return -1;
	else if (! S_ISLNK (statbuf.st_mode))
		return (access (filename, how));	//FIXME vfs

	if (task_mutex != NULL)
		g_mutex_lock (task_mutex);
	if (myuid == -1)
		myuid = getuid ();
	if (mygid == -1)
		mygid = getgid ();
	if (task_mutex != NULL)
		g_mutex_unlock (task_mutex);

	gboolean result = FALSE;

	if (how & R_OK)
	{
		if (statbuf.st_mode & S_IROTH)
			result = TRUE;
		else if (statbuf.st_mode & S_IRUSR)
			result = ((myuid == 0) || (myuid == statbuf.st_uid));
		else if (statbuf.st_mode & S_IRGRP)
			result = ((myuid == 0) || (mygid == statbuf.st_gid));
		if (!result)
			return -1;
	}
	if (how & W_OK)
	{
		if (statbuf.st_mode & S_IWOTH)
			result = TRUE;
		else if (statbuf.st_mode & S_IWUSR)
			result = ((myuid == 0) || (myuid == statbuf.st_uid));
		else if (statbuf.st_mode & S_IWGRP)
			result = ((myuid == 0) || (mygid == statbuf.st_gid));
		if (!result)
			return -1;
	}
	if (how & X_OK)
	{
		if (statbuf.st_mode & S_IXOTH)
			result = TRUE;
		else if (statbuf.st_mode & S_IXUSR)
			result = ((myuid == 0) || (myuid == statbuf.st_uid));
		else if (statbuf.st_mode & S_IXGRP)
			result = ((myuid == 0) || (mygid == statbuf.st_gid));
		if (!result)
			return -1;
	}
	//we already know the link exists, no need for F_OK test
	return 0;
}
/**
 * @brief execute the file (1) command and search its output for @a string
 *
 * Used in e2_fs_is_text and e2_fs_is_executable
 *
 * @param filename localised name of file to test
 * @param string ascii string for which to search, in the output from 'find'
 * @return TRUE if @a string is found
 **/
static gboolean _e2_fs_grep_file_output (gchar *filename, gchar *string)
{
	//do not use g_strdup_printf or g_snprintf because string is localised
	gchar line[PATH_MAX + 20];
	snprintf (line, sizeof (line), ("file \"%s\""), filename);	//do not translate
	E2_FILE *pipe = e2_fs_open_pipe (line);
	if (pipe == NULL)
		return FALSE;

	gboolean retval = FALSE;
	size_t bsize = E2_MAX_LEN;
	gchar *buf = (gchar *) g_malloc ((gulong) bsize);
#ifdef __USE_GNU
	ssize_t bytes_read;
	while ((bytes_read = getdelim (&buf, &bsize, '\n', pipe)) > 0)
#else
	gchar *tmpstr;
	while ((tmpstr = fgets (buf, bsize, pipe)) != NULL)
#endif
	{
		if (strstr (buf, string))
		{
			retval = TRUE;
			break;
		}
	}

	e2_fs_pipe_close (pipe);
	g_free (buf);
	return retval;
}
/**
 * @brief test whether @a filename is a text file
 * @param filename localised name of file to test
 * @return TRUE if it is a text file
 **/
gboolean e2_fs_is_text (gchar *filename)
{
	return _e2_fs_grep_file_output (filename, "text"); //do not translate
}
/**
 * @brief test whether @a info belongs to an executable item
 * If the item is a symlink, its target will be tested
 * @param info pointer to a FileInfo data struct
 * @return TRUE if system reports x permission, and the file (1) command says it's executable
 **/
gboolean e2_fs_is_executable (FileInfo *info)
{
	if (!e2_fs_access (info->filename, X_OK))
	{
		if (S_ISLNK (info->statbuf.st_mode))
			return TRUE;
		//non-link, double check ...
		return _e2_fs_grep_file_output (info->filename, "executable");  //string not to be translated
	}
	return FALSE;
}
/**
 * @brief check whether item in active pane is a directory or a link to one
 *
 * This assumes that the mode flags have been gathered using
 * lstat, not stat (ie no link look-through)
 *
 * Item needs to be in active pane because no path is added to info->filename
 *
 * @param info FileInfo structure
 *
 * @return TRUE if belongs to a directory or a link to a directory
 */
gboolean e2_fs_is_dir (FileInfo *info)
{
	if ( S_ISDIR (info->statbuf.st_mode) )
		return TRUE;
	if ( S_ISLNK(info->statbuf.st_mode) )
	{
		struct stat statbuf;
		if (!e2_fs_stat (info->filename, &statbuf)
			&& S_ISDIR (statbuf.st_mode))
		  return TRUE;
	}
	return FALSE;
}
/**
 * @brief get real path for @a path if its a link
 *
 * @param path pointer to freeable, absolute, localised, path string which may be a link name
 *
 * @return TRUE if a newly-allocated replacement path is provided in @a path
 */
#if defined(E2_FAM) && !defined(E2_FAM_KERNEL)
gboolean e2_fs_walk_link (gchar **path)
{
	struct stat sb;
	if (e2_fs_lstat (*path, &sb)
	 	&& S_ISLNK (sb.st_mode))
	{
		gchar *resolved_path = g_malloc (PATH_MAX);
		if (realpath (*path, resolved_path) != NULL)
		{
			g_free (*path);
			*path = resolved_path;
			return TRUE;
		}
		else
			g_free (resolved_path);
	}
	return FALSE;
}
#endif //defined(E2_FAM) && !defined(E2_FAM_KERNEL)
/**
 * @brief inline version of func that checks whether item in active pane is a directory or a link to one
 * NOW CONVERTED TO AN EQUIVALENT DEFINE
 * This assumes that the mode flags have been gathered using
 * lstat, not stat (ie no link look-through)
 *
 * Item needs to be in active pane because no path is added to info->filename
 *
 * @param info FileInfo structure
 * @param statbuf pointer to statbuf to use for statting @a info ->filename
 *
 * @return TRUE if belongs to a directory or a link to a directory
 */
/*inline gboolean e2_fs_is_dir_fast (FileInfo *info, struct stat *statbuf)
{
	if (S_ISDIR (info->statbuf.st_mode) )
		return TRUE;
	if (S_ISLNK(info->statbuf.st_mode) )
	{
		if (!stat (info->filename, statbuf)	//look thru the link
			&& S_ISDIR (statbuf->st_mode))
		  return TRUE;
	}
	return FALSE;
} */
/**
 * @brief check whether selected item in active pane is a directory or a link to one
 *
 * No path is added to info->filename, so must be in current dir
 *
 * @param info E2_SelectedItemInfo data structure
 *
 * @return TRUE if @a info belongs to a directory or a link to a directory
 */
gboolean e2_fs_is_dir2 (E2_SelectedItemInfo *info)
{
	struct stat statbuf;
	gchar *local = F_FILENAME_TO_LOCALE (info->filename);
	gboolean retval = (!e2_fs_stat (local, &statbuf) && S_ISDIR (statbuf.st_mode));
	F_FREE (local);
	return retval;
}
/**
 * @brief check whether @a path is a directory or a link to one
 *
 * @param path localised path string
 *
 * @return TRUE if @a info belongs to a directory or a link to a directory
 */
gboolean e2_fs_is_dir3 (gchar *local_path)
{
	struct stat statbuf;
	if (e2_fs_stat (local_path, &statbuf))
		return FALSE;
	return (S_ISDIR (statbuf.st_mode));
}
/**
 * @brief make directory @a local regardless of whether all ancestors exist
 * This only works for single-byte path-separator characters
 * @param path localised absolute path string
 *
 * @return FALSE if @a creation succeeds (like mkdir)
 */
gboolean e2_fs_recurse_mkdir (gchar *local, gint mode)
{
	struct stat sb;
	gchar c;
	gchar *s = local;	//skips check on fs root dir
	while ((s = strchr (s+1, G_DIR_SEPARATOR)) != NULL)
	{
		c = *s;
		*s = '\0';
		if (stat (local, &sb))	//thru links, not e2_fs_, FIXME vfs
		{
			if (errno != ENOENT || e2_fs_mkdir (local, mode))
			{
				*s = c;
				return TRUE;
			}
		}
		else if (! S_ISDIR (sb.st_mode))
		{
			*s = c;
			return TRUE;
		}
		*s = c;
	}
	//check last or only segment
	if (stat (local, &sb))	//thru links, not e2_fs_, FIXME vfs
	{
		if (errno != ENOENT || e2_fs_mkdir (local, mode))
			return TRUE;
	}
	return FALSE;
}
/**
 * @brief change system directory to @s path, if possible
 *
 * @a path may be a link, in which case it will be traversed
 *
 * @param path utf8 string with path to directory to be opened
 *
 * @return TRUE if @a path is opened ok
**/
gboolean e2_fs_chdir (gchar *path)
{
	gboolean retval = TRUE;
	gchar *local = F_FILENAME_TO_LOCALE (path);

	if (e2_fs_chdir_local (local))
	{	//go there failed
		e2_output_print_errno ();
		retval = FALSE;
	}
	F_FREE (local);
	return retval;
}
/**
 * @brief check whether the user is able to change to a specified directory
 *
 * This tests @a path for existence, is-a-directory, is-executable
 * If @a path is a link, it will be traversed
 *
 * @param path utf8 string with path to directory to be checked
 *
 * @return TRUE if @a path is ok to cd to
 **/
gboolean e2_fs_cd_isok (gchar *path)
{
	gchar *message;
	gboolean success = TRUE;
	gchar *local = F_FILENAME_TO_LOCALE (path);
	if (e2_fs_access (local, F_OK))
	{
		message = g_strdup_printf (_("The directory '%s' does not exist!"), path);
		success = FALSE;
	}
	else if (!e2_fs_is_dir3 (local))
	{
		message = g_strdup_printf (_("The file '%s' is not a directory!"), path);
		success = FALSE;
	}
	else if (e2_fs_access (local, X_OK))
	{
		message = g_strdup_printf (_("No permission to access directory '%s'!"), path);
		success = FALSE;
	}
	if (!success)
	{
		e2_output_print_error (message, TRUE);
	}
	F_FREE (local);
	return success;
}
/**
 * @brief check whether directory @a path is valid, and optionally, accessible, and if not, pick a fallback
 * If @a path refers to a link, it will be traversed.
 * If test(s) are not satisfied, @a path is set to a different path string,
 * pointing to a place up the tree branch that does pass the test(s)
 * If a different path string is returned, the old one is freed, of course
 *
 * @param path utf8 path string which may be invalid, must be freeable
 * @param accessible TRUE to find a path that is accessible, FALSE if don't care about that
 *
 * @return TRUE if @a path meets the specified test(s), FALSE if a replacement has been provided
 */
gboolean e2_fs_get_valid_path (gchar **path, gboolean accessible)
{
	gchar *local = F_FILENAME_TO_LOCALE (*path);
	if (e2_fs_is_dir3 (local)
		&& (!accessible || !e2_fs_access (local, R_OK | X_OK)))
	{
		F_FREE (local);
		return TRUE;
	}
	else	//walk back up the tree branch until we find an acceptable dir
	{
		gchar *p;
		while ((p = strrchr (local, G_DIR_SEPARATOR)) != NULL)
		{
			*p = '\0';	//CHECKME ok to change local (maybe == *path) ??
			if (e2_fs_is_dir3 (local)
				&& (!accessible || !e2_fs_access (local, R_OK | X_OK)))
			{	//replace the invalid path string with the valid one
				p = D_FILENAME_FROM_LOCALE (local);	//copy in case not changed by macro
				g_free (*path);
				*path = p;
				F_FREE (local);
				return FALSE;
			}
		}
		//whole branch bad, revert to default
		g_free (*path);
//		*path = g_strdup (G_DIR_SEPARATOR_S);
		*path = g_strconcat (g_get_home_dir (), G_DIR_SEPARATOR_S, NULL);
		F_FREE (local);
	}
	return FALSE;
}
/**
 * @brief stat item @a path, put results into @a buf
 *
 * This looks through links, but does not fail for hanging links
 *
 * @param path relative or absolute path of item to stat, localised string
 * @param buf pointer to buffer to hold the results
 *
 * @return 0 (FALSE) if the stat succeeded, else -1 (TRUE)
 **/
gint e2_fs_stat (const gchar *path, struct stat *buf)
{
	if (stat (path, buf))	//FIXME vfs
	{	//stat failed, but it could be just due to a relative symlink
		struct stat statbuf;	//don't disturb the caller's buffer
		if (!e2_fs_lstat (path, &statbuf)	//stat link succeeded
			&& (S_ISLNK (statbuf.st_mode)
			&& errno == ENOENT))	//CHECKME more checks for hanging link?
		{
			*buf = statbuf;
			return 0;
		}
		return -1;
	}
	return 0;
}
/**
 * @brief check whether the current user may modify a specified item
 *
 * This does not look through links (target may not exist)
 *
 * @param uid the current user's uid, or -1 to determine the latter
 * @param path utf string with path of item to check
 *
 * @return TRUE if the user has write authority for @a path
 **/
gboolean e2_fs_check_write_permission (gint uid, gchar *path)
{
	if (uid == -1)
		uid = getuid ();
	if (uid == 0)
		return TRUE; //user is root
	
	struct stat statbuf;
	gchar *local = F_FILENAME_TO_LOCALE (path);
	e2_fs_lstat (local, &statbuf);
	F_FREE (local);

	if ((uid == (gint) statbuf.st_uid)	//user is owner
//		&& (statbuf.st_mode & S_IWUSR )	//with write permission
	)
		return TRUE;

	if (statbuf.st_mode & S_IWOTH)
		return TRUE; //everyone has write permission

	struct group *groupdata = getgrgid (statbuf.st_gid);
	if (groupdata != NULL)
	{
		gchar **members = groupdata->gr_mem;
		while (*members != NULL)
		{
			struct passwd *pwbuf = getpwnam (*members);
			if ( pwbuf != NULL &&
				((gint) pwbuf->pw_uid == uid))
			{
				if (statbuf.st_mode & S_IWGRP)
						return TRUE; //user is group, with write permission
				else
					break;
			}
			members++;	
		}
	}

	return FALSE;
}
/**
 * @brief open file @a filepath for writing
 *
 * @param filepath path to file open
 *
 * @return pointer to data struct for opened file, or NULL if the open failed
 */
E2_FILE *e2_fs_open_file_to_write (gchar *filepath)
{
	gchar *local = F_FILENAME_TO_LOCALE (filepath);
	E2_FILE *f = fopen (local, "w");
	F_FREE (local);
	return f;
}
/**
 * @brief open pipe to read output from @a command
 *
 * @param command utf-8 command string
 * @return pointer to opened pipe, or NULL if open failed
 **/
E2_FILE *e2_fs_open_pipe (gchar *command)
{
	gchar *local = F_FILENAME_TO_LOCALE (command);
	E2_FILE *pipe;
	if ((pipe = popen (local, "r")) == NULL)	//FIXME vfs
	{
		gchar *s;
		if ((s = e2_utils_find_whitepace (command)) != NULL)
			*s = '\0';
		s = g_strdup_printf (_("Unable to open pipe for %s command"), command);
		e2_output_print_error (s, TRUE);
	}
	F_FREE (local);
	return pipe;
}
/**
 * @brief read into memory the contents of the directory associated with @a view
 * CWD is changed to the directory
 * The directory timestamp is upated, for refresh-management purposes
 * @param view data struct for the vieW
 *
 * @return list of FileInfo's for entries (maybe NULL), or -1 if a problem occurred
 */
GList *e2_fs_read_dir (FileView *view)
{
	//FIXME adapt for vfs
	//FIXME support a timout when vfs applied
	gchar *local = F_FILENAME_TO_LOCALE (view->dir);
//#ifndef E2_FAM
	//not necessary when a FAM is working, but benefit from
	//testing for that is not material ..
	struct stat sb;
	if (e2_fs_stat (local, &sb))	//stat before opendir
	{
		printd (WARN, "Unable to stat directory: %s", view->dir);
		F_FREE (local);
		return GINT_TO_POINTER (-1);
	}
//#endif

	DIR *dp = opendir (local);
	F_FREE (local);

	if (dp == NULL)
	{
		printd (WARN, "Unable to open directory: %s", view->dir);
		return GINT_TO_POINTER (-1);
	}

//#ifndef E2_FAM
	//not necessary when a FAM is working, but benefit from
	//testing for that is not material ..
	view->dir_mtime = sb.st_mtime;
	view->dir_ctime = sb.st_ctime;
//#endif
	
/*	would be nice to prevent access-time updates purely due to
	refreshing, but the atime really does change as a result
	of the refresh process, and if we were to revert it, the
	ctime would be changed instead
	times are immediately shown in the other pane if it's the
	parent of this one
	if we could suspend monitoring of such parent dir, we
	must not lose any prior reports for there
	QUERY can the changed atime itself trigger a report which
	causes a subsequent refresh, creating an endless cycle?
	especially with multiple threads and/or processors ?
	Assuming it can do so, we clear the reports 'queue'
	after the dir is opened
	Would be nice if we could just clear any report(s) due to
	the refresh per se - but the reporting is not that detailed */
/*#if defined (E2_FAM_INOTIFY) || defined (E2_FAM_DNOTIFY) || defined (E2_FAM_KQUEUE)
	if (view->refresh)
		e2_fs_FAM_cancel_monitor_dir (view->dir);
#endif */

/*#if defined (E2_FAM_DNOTIFY)
	//cut down on spurious context changes
	//block DNOTIFY_SIGNAL
	sigprocmask (SIG_BLOCK, &dnotify_signal_set, NULL);
#endif */

	//change current dir coz statted items in the dir have no path
	//no success-check, we've already done 2 of them, above
	e2_fs_chdir (view->dir);

	FileInfo *info;
	struct dirent entry;
	struct dirent *entryptr;
	GList *entries = NULL;
//	E2_BLOCK	//don't want this interrupted
	while (TRUE)
	{
		if (readdir_r (dp, &entry, &entryptr) || entryptr == NULL)	//FIXME vfs
			break;
#ifdef E2_NEWFILTER
		//one item we're not interested in
		if (entry.d_name[0] != '.' || entry.d_name[1] != '\0')
		{
#endif

#if defined(E2_CURRENTGTK_VERSION) && GLIB_CHECK_VERSION (2,10,0)
			info = g_slice_alloc (sizeof (FileInfo));
#else
			info = g_try_malloc (sizeof (FileInfo));
#endif
			if (info == NULL)
			{
#if defined(E2_CURRENTGTK_VERSION) && GLIB_CHECK_VERSION (2,10,0)
				g_list_free (entries);	//FIXME proper cleanup ?
#else
				e2_list_free_with_data (&entries);
#endif
				//FIXME warning
				entries = GINT_TO_POINTER (-1);
				break;
			}
			e2_fs_lstat (entry.d_name, &info->statbuf);
			//any truncation here means it will never be able to be stat'd anymore
			g_strlcpy (info->filename, entry.d_name, sizeof (info->filename));
			entries = g_list_append (entries, info);

#ifdef E2_NEWFILTER
		}
#endif
	}
	closedir (dp);
//	E2_UNBLOCK
	return entries;
}
/**
 * @brief read file contents into memory
 *
 * @param filepath path to file to be read
 * @param contents ptr to where the contents are to be stored
 *
 * @return TRUE if @a filepath was read successfully
 */
gboolean e2_fs_get_file_contents (gchar *filepath, gchar **contents)
{
	gchar *local = F_FILENAME_TO_LOCALE (filepath);
	//CHECKME ok to localise the name string  with this glib fn ?
	gboolean retval = g_file_get_contents (local, contents, NULL, NULL);
	F_FREE (local);
	return retval;
}
/**
 * @brief read piped command output into memory
 * A terminating \0 is always added
 * @param command command string to be executed
 * @param contents ptr to where the contents are to be stored
 *
 * @return TRUE if @a command result was read successfully
 */
gboolean e2_fs_get_command_output (gchar *command, gchar **contents)
{
	E2_FILE *pipe = e2_fs_open_pipe (command);
	if (pipe == NULL)
		return FALSE;
	size_t total = 0;
	size_t bsize = 4096;	//block size
	gchar *buf = (gchar *) g_malloc ((gulong) bsize);
	gchar *store = buf;
	ssize_t bytes_read;
	while ((bytes_read = fread (store, 1, bsize, pipe)) > 0)
	{
		if (bytes_read == bsize)
		{
			total += bsize;
			buf = realloc (buf, total+bsize);
			if (buf == NULL)
			{
				gchar *msg = g_strdup_printf (
					_("Command %s failed: not enough memory"), command);
				e2_output_print_error (msg, TRUE);
				e2_fs_pipe_close (pipe);
				return FALSE;
			}
			store = buf + total;
		}
		else
		{
			*(store+bytes_read) = '\0';
			bytes_read++;
			realloc (buf, total+bytes_read);
			break;
		}
	}
	e2_fs_pipe_close (pipe);
	*contents = buf;
	return TRUE;
}
/**
 * @brief blockwize read from stdin into memory
 * A terminating \0 is always added
 * @param contents ptr to set to where the input is stored
 *
 * @return TRUE if stdin was read successfully
 */
/*UNUSED
gboolean e2_fs_read_stdin (gchar **contents)
{
	fd_set fds;
	FD_ZERO (&fds);
	FD_SET (STDIN_FILENO, &fds);

	struct timeval wait;
	wait.tv_sec = 0;
	wait.tv_usec = 100000;

	if (select (STDIN_FILENO+1, &fds, NULL, NULL, &wait) != 1)
		return FALSE;

	size_t total = 0;
	size_t bsize = 1024;	//block size
	gchar *buf = (gchar *) g_malloc ((gulong) bsize);
	gchar *store = buf;
	ssize_t bytes_read;
	while (feof (stdin) == 0)
	{
		bytes_read = fread (buf, 1, bsize, stdin);
		if (ferror (stdin) != 0)
		{
			g_free (buf);
			return FALSE;	//error msg ??
		}
		if (bytes_read == bsize)
		{
			total += bsize;
			buf = realloc (buf, total+bsize);
			if (buf == NULL)
			{
				e2_output_print_error (_("Failed stdin read: not enough memory"), FALSE);
				fclose (stdin);
				return FALSE;
			}
			store = buf + total;
		}
		else
		{
			*(store+bytes_read) = '\0';
			bytes_read++;
			realloc (buf, total+bytes_read);
			break;
		}
	}
	fclose (stdin);
	*contents = buf;
	return TRUE;
} */
/**
 * @brief write a sequence of bytes from memory to stdout
 *
 * @param contents ptr to \0-terminated bytes to be written
 *
 * @return TRUE if @a contents was sent successfully
 */
/*UNUSED
gboolean e2_fs_write_stdout (gchar *contents)
{
	fd_set fds;
	FD_ZERO (&fds);
	FD_SET (STDOUT_FILENO, &fds);

	struct timeval wait;
	wait.tv_sec = 0;
	wait.tv_usec = 100000;

	if (select (STDOUT_FILENO+1, NULL, &fds, NULL, &wait) != 1)
		return FALSE;

	return (fputs (contents, stdout) != EOF);
} */
/**
 * @brief convert file mode_t to a readable permissions string
 * NOTE this handles only single-byte strings, and some parts are
 * not translated
 * String is 10 bytes long
 *
 * @param buf ptr to buffer where the result is to be stored
 * @param len integer length of @a buf
 * @param mode mode_t property to be parsed
 *
 * @return
 */
void e2_fs_get_perm_string (gchar *buf, gint len, mode_t mode)
{
	gchar *perm_sets[8];
	perm_sets[0] = _("---");
	perm_sets[1] = _("--x");
	perm_sets[2] = _("-w-");
	perm_sets[3] = _("-wx");
	perm_sets[4] = _("r--");
	perm_sets[5] = _("r-x");
	perm_sets[6] = _("rw-");
	perm_sets[7] = _("rwx");

	gint u, g, o;
	u = (mode & S_IRWXU) >> 6;
	g = (mode & S_IRWXG) >> 3;
	o = (mode & S_IRWXO);

	g_snprintf (buf, len, "-%s%s%s", perm_sets[u], perm_sets[g], perm_sets[o]);

	//_I( FIXME the things below do not suppport translation !!
	if (S_ISLNK(mode))
		buf[0] = 'l';
	else if (S_ISDIR(mode))
		buf[0] = 'd';
	else if (S_ISBLK(mode))
		buf[0] = 'b';
	else if (S_ISCHR(mode))
		buf[0] = 'c';
	else if (S_ISFIFO(mode))
		buf[0] = 'f';
	else if (S_ISSOCK(mode))
		buf[0] = 's';

	if (mode & S_ISVTX)
		buf[9] = (buf[9] == '-') ? 'T' : 't';
	if (mode & S_ISGID)
		buf[6] = (buf[6] == '-') ? 'S' : 's';
	if (mode & S_ISUID)
		buf[3] = (buf[3] == '-') ? 'S' : 's';
}
/**
 * @brief update m/a/c times of config dir
 *
 * This should be called after writing the config/cache files
 * to let any other instance know that it should re-read those
 * files (if that option is in force)
 *
 * @return
 **/
void e2_fs_touch_config_dir (void)
{
#ifdef E2_FAM
	if (app.monitor_type != E2_MONITOR_DEFAULT)
		return;
#endif
	struct stat stat_buf;

	gchar *local = F_FILENAME_TO_LOCALE (e2_cl_options.config_dir);
	if (e2_fs_utime (local, NULL) == 0)
	{
		if (!e2_fs_stat (local, &stat_buf))
			app.config_mtime = stat_buf.st_mtime;
		else
		{
			e2_output_print_errno();
			printd (WARN, "couldn't stat the config dir");
		}
	}
	else
	{
		e2_output_print_errno();
		printd (WARN, "couldn't touch the config dir");
	}
	F_FREE (local);
}
/**
 * @brief change time(s) of item @s path to current system time
 *
 * @param path utf8 string with absolute path of item to touch
 *
 * @return
**/
void e2_fs_touchnow (gchar *path)
{
	gchar *local = F_FILENAME_TO_LOCALE (curr_view->dir);
	e2_fs_utime (local, NULL);
	F_FREE (local);
}
