/* $Id: e2_fs.h 1856 2006-04-24 12:51:28Z tpgww $
 *
 * Copyright (C) 2005 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/filesystem/e2_fs.h
 * @brief header for filesystem io functions
 **/

#ifndef E2_FS_H
#define E2_FS_H

#include "emelfm2.h"
#include <unistd.h>
#include <utime.h>
#include <glib/gprintf.h>

#define E2_FILE FILE
#define e2_fs_fstat fstat
#define e2_fs_lstat lstat
//access() looks through links
//use e2_fs_access3 to prevent that
#define e2_fs_access access
#define e2_fs_mkdir mkdir
#define e2_fs_utime utime
#define e2_fs_file_write g_fprintf
#define e2_fs_file_put fputs
#define e2_fs_file_close fclose
#define e2_fs_pipe_close pclose
//no error check or report with this chdir
#define e2_fs_chdir_local chdir
#define e2_fs_chmod chmod
#define e2_fs_chown chown
#define e2_fs_lchown lchown

/*
for other dir-related fns, see:
_e2_filelist_store_refill ()
disk_usage () in cpbar plugin
disk_usage () in mvbar plugin
_e2p_du_add_usage () in du plugin
g_dir_open in _e2_complete_path_check_dir()
g_dir_close in ditto
g_dir_read_name in ditto
pipe usages inlcude getdelim()
*/

void e2_fs_check_coding (void);
gint e2_fs_stat (const gchar *path, struct stat *buf);
gint e2_fs_access2 (gchar *filename);
gint e2_fs_access3 (gchar *filename, gint how);
gboolean e2_fs_is_text (gchar *filename);
gboolean e2_fs_is_executable (FileInfo *info);
gboolean e2_fs_is_dir (FileInfo *info);
//inline gboolean e2_fs_is_dir_fast (FileInfo *info, struct stat *statbuf);
//make this always inline ...
//here we use stat() because we want to traverse the link
#define e2_fs_is_dir_fast(a,b) \
  S_ISDIR (((FileInfo *)a)->statbuf.st_mode) \
	|| (S_ISLNK (((FileInfo *)a)->statbuf.st_mode) \
		&& !stat (((FileInfo *)a)->filename, (struct stat *)b) \
		&& S_ISDIR (((struct stat *)b)->st_mode))
gboolean e2_fs_is_dir2 (E2_SelectedItemInfo *info);
gboolean e2_fs_is_dir3 (gchar *local_path);
gboolean e2_fs_recurse_mkdir (gchar *local, gint mode);
#if defined(E2_FAM) && !defined(E2_FAM_KERNEL)
gboolean e2_fs_walk_link (gchar **path);
#endif
gboolean e2_fs_chdir (gchar *path);
gboolean e2_fs_get_valid_path (gchar **path, gboolean accessible);
gboolean e2_fs_cd_isok (gchar *path);
gboolean e2_fs_check_write_permission (gint uid, gchar *path);
E2_FILE *e2_fs_open_pipe (gchar *command);
GList *e2_fs_read_dir (FileView *view);
void e2_fs_rel_path (GString *src, GString *dest);
void e2_fs_get_perm_string (gchar *buf, gint len, mode_t mode);
gboolean e2_fs_get_file_contents (gchar *filepath, gchar **contents);
gboolean e2_fs_get_command_output (gchar *command, gchar **contents);
E2_FILE *e2_fs_open_file_to_write (gchar *filepath);
void e2_fs_touch_config_dir (void);
void e2_fs_touchnow (gchar *path);

//away-from-normal declaration here, to avoid build problems !!
void e2_option_tree_write_to_file (E2_FILE *f, E2_OptionSet *set,
	GtkTreeIter *iter, gint level);

#ifdef E2_FAM
//these apply whether or not kernel-based fam is being used
void e2_fs_FAM_connect (void);
void e2_fs_FAM_disconnect (void);
void e2_fs_FAM_change (gchar *olddir, E2_PaneRuntime *rt);
gboolean e2_fs_FAM_monitor_config (void);
gboolean e2_fs_FAM_cancel_monitor_config (void);
void e2_fs_FAM_check_dirty (gboolean *p1dirty, gboolean *p2dirty, gboolean *cfgdirty);
#else
void e2_fs_FAM_check_dirty (gboolean *p1dirty, gboolean *p2dirty);
void e2_fs_FAM_config_stamp (void);
#endif
#if defined (E2_FAM_INOTIFY) || defined (E2_FAM_DNOTIFY) || defined (E2_FAM_KQUEUE)
//void e2_fs_FAM_cancel_monitor_dir (gchar *path);
//void e2_fs_FAM_monitor_dir (gchar *path);
void e2_fs_FAM_clean_reports (gchar *path);
#endif

//flags for instructing e2_fs_tw how to perform its function
typedef enum
{
	E2TW_DEFAULT = 0,
	E2TW_PHYS = 1,	//perform physical walk, don't look though symlinks
	E2TW_MOUNT = 1 << 1,	//report only items on same file system as the argument
	E2TW_CHDIR = 1 << 2,	//chdir to each subdir after it is opened
	E2TW_DEPTH = 1 << 3,	//depth-first walk (actually, depth when a dir is found)
} E2_TwFlags;

//codes for status reports to tw callback functions
typedef enum
{
	E2TW_F,		//not-directory or link
	E2TW_SL,	//symbolic link
	E2TW_SLN,	//ymbolic link naming non-existing file
	E2TW_D,	//directory
	E2TW_DL,	//directory, not opened due to tree-depth limit
	E2TW_DM,	//directory, not opened due to different file system
	E2TW_DP,	//directory, all subdirs have been visited
	E2TW_DNR,	//unreadable directory
	E2TW_NS	//un-statable item
} E2_TwStatus;

//codes for feedback from tw callback functions
typedef enum
{
	E2TW_CONTINUE,	//continue with next sibling or for E2TW_D with the first child
	E2TW_STOP,			//abort the walk, return FALSE from e2_fs_tw();
	E2TW_SKIP_SUBTREE,	//only meaningful for E2TW_D: Don't walk through the
									//subtree, instead just continue with its next sibling
//	E2TW_SKIP_SIBLINGS,	//continue with FTW_DP callback for current directory
									//(if FTW_DEPTH) and then its siblings.
	E2TW_FAILED,		//for internal (not callback fn) use - subdir walk returned a FALSE result
	E2TW_FIXME			//interim value which needs to be changed later
} E2_TwResult;

gboolean e2_fs_tw (gchar *start_item, E2_TwResult (*callback) (),
	void *user_data, gint max_depth, E2_TwFlags exec_flags);

#endif //ndef E2_FS_H
