/* $Id: e2_fs_walk.c 1774 2006-03-18 21:39:46Z tpgww $
 *
 * Copyright (C) 2005-2006 tooar <tooar@gmx.net>
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/filesystem/e2_fs_walk.c
 * @brief tree-walk functions
 **/

#include "e2_fs.h"
#include <string.h>
#include <dirent.h>

/* twstat pointer to stat() or lstat(), to be used for stat oprations
 * cbfunc pointer to the function to be called for each located item
 * user_data void pointer to user-specified data to be supplied 
 *      to cbunc each time it is called
 * max_depth the largest number of tree levels to descend, or -1 for no limit
 * exec_flags flags indicating the way the walk is to be performed
 * device the system on which the starting item resides
 * original_path newly-allocated localised path string, = cwd when the walk is started
 */
typedef struct _E2_TwData
{
	gint (*twstat) (const gchar *, struct stat *);
	E2_TwResult (*cbfunc) (const gchar *, const struct stat *, E2_TwStatus, void *);
	void *user_data;
	gint max_depth;
	E2_TwFlags exec_flags;
	dev_t device;
//	gchar *original_path;
} E2_TwData;

/**
 * @brief determine the appropriate status report code for link @a local
 *
 * With relative symlinks, stat() generally fails, and anyway, can't
 * distinguish valid links from hanging ones. Nor can lstat().
 * So this uses a kludgy but robust alternative approach
 *
 * @param local absolute path of link-item to process, localised string
 *
 * @return callbck status code in accordance with whether the link is valid
 **/
static E2_TwResult _e2_fs_tw_link (gchar *local)
{
	E2_TwStatus code;
	gchar buf[PATH_MAX];
	gint read = readlink (local, buf, sizeof (buf));	//FIXME vfs
	if (read == -1)
	{
		code = E2TW_SLN;
	}
	else
	{
		buf[read] = '\0';
		gchar *converted1 = F_FILENAME_FROM_LOCALE (local);
		gchar *converted2 = F_FILENAME_FROM_LOCALE (buf);
		gchar *targetpath = e2_utils_translate_relative_path (converted1, converted2);
		F_FREE (converted1);
		F_FREE (converted2);
		//targetpath has a trailing /, with which access() can't cope
		gint len = strlen (targetpath) - 1;
		*(targetpath+len) = '\0';

		converted1 = F_FILENAME_TO_LOCALE (targetpath);
		//access() needs to look through links
		code = (e2_fs_access (converted1, F_OK)) ? E2TW_SLN : E2TW_SL;
		g_free (targetpath);
		F_FREE (converted1);
	}
	return code;
}
/**
 * @brief process 'non-directory' item @a local, as part of a treewalk
 *
 * This determines the appropriate status report code for the callback,
 * then calls it
 * @a local may be the path of a link, regular file, etc
 * @a local will have been stat'd or lstat'd (depending on walk-flag 
 * E2TW_PHYS) into @a statbuf, before this function is called
 *
 * @param local absolute path of item to process, localised string
 * @param statbuf pointer to struct stat used for stating @a local
 * @param twdata pointer to struct with 'constant' data for each call here
 *
 * @return result code in accordance with callback-function response
 **/
static E2_TwResult _e2_fs_tw_nondir (gchar *local,
	struct stat *statbuf, E2_TwData *twdata)
{
	E2_TwStatus code;
	struct stat statbuf2;	//don't disturb statbuf
	if (!(twdata->exec_flags & E2TW_PHYS))
	{	//the prior stat() processed a non-link, or might, or might not,
		//have successfully processed a link
		//check if it was a link
		if (e2_fs_lstat (local, &statbuf2) 	//stat link succeeded
			&& S_ISLNK (statbuf2.st_mode))
				code = E2TW_SL;
		else if (S_ISLNK (statbuf2.st_mode))
			//in general, relative links (valid or not) will end up here,
			//unless E2TW_CHDIR is invoked
			code = _e2_fs_tw_link (local);
		else
			code = E2TW_F;
	}
	else	//the prior lstat did not look through a link
		if (S_ISLNK (statbuf->st_mode))
			code = _e2_fs_tw_link (local);
	else
		code = E2TW_F;

	E2_TwResult result = (*twdata->cbfunc) (local, statbuf, code, twdata->user_data);
	return result;
}
/**
 * @brief process directory @a local, as part of a treewalk
 *
 * if possible, and in accord with the walk-mode determined by flags
 * in @a twdata, @a local is opened and its contents are each
 * 'reported' to the callback function recorded in @a twdata. 
 * Any subdir found among the contents will recurse to this
 * function.
 * Non-dirs are handled by another function, before coming
 * here or when found.
 * As appropriate, status report code(s) provided to the specified
 * callback function will be:
 * E2TW_F item is not a directory or link
 * E2TW_SL symbolic link
 * E2TW_SLN symbolic link whose target is non-existent
 * E2TW_D	directory
 * E2TW_DL directory, not opened due to tree-depth limit
 * E2TW_DM directory, not opened due to different file system
 * E2TW_DNR unreadable directory
 * E2TW_DP directory processed (all its contents have been reported)
 * E2TW_NS un-statable item
 *
 * E2TW_D and E2TW_DP will both be reported for any opened
 * directory (unless, of course, processing is aborted by a message
 * from the callback, between the two reports)
 *
 * @param local absolute path of _directory_ to process, localised string, with or without trailing "/"
 * @param depth enumator of current parent's level in the tree (relative
 *      to the original parent), 0 to max_depth (if the latter is not -1)
 * @param twdata pointer to struct with 'constant' data for each call here
 *
 * @return result code in accordance with callback-function responses, E2TW_STOP or E2TW_CONTINUE if all was well
 **/
static E2_TwResult _e2_fs_tw (const gchar *local, gint depth, E2_TwData *twdata)
{
	struct stat statbuf;
	E2_TwResult result;
	GList *parked_dirs, *tmp;
	gint newdepth;
	//populate the local statbuf, if we can
	if ((*twdata->twstat) (local, &statbuf))
	{	//failed (statbuf will be garbage for this call)
		result = (*twdata->cbfunc) (local, &statbuf, E2TW_NS, twdata->user_data);
		return ((result == E2TW_CONTINUE) ? result : E2TW_STOP);	//for an un-statable item, only continue or stop
	}
	//check whether it is ok to open the dir
	newdepth = depth + 1;
	if (twdata->max_depth != -1 && newdepth > twdata->max_depth)
	{
		result = (*twdata->cbfunc) (local, &statbuf, E2TW_DL, twdata->user_data);
		return ((result == E2TW_STOP) ? result : E2TW_CONTINUE); //max depth reached, would expect normally to continue
	}

	else if ((twdata->exec_flags & E2TW_MOUNT) &&
		twdata->device != statbuf.st_dev)
	{
		result = (*twdata->cbfunc) (local, &statbuf, E2TW_DM, twdata->user_data);
		return ((result == E2TW_STOP) ? result : E2TW_CONTINUE); //unwanted filesystem, would expect normally to continue
	}
	//check whether it is possible to open the dir
	DIR *dp;
	if ((dp = opendir (local)) == NULL)
	{
		result = (*twdata->cbfunc) (local, &statbuf, E2TW_DNR, twdata->user_data);
		//continue if no problem or a problem down the tree
		//any problem is handled by the relevant tw func and its continued-after-problem flag
		return (((result == E2TW_SKIP_SUBTREE && newdepth > 1)
					|| result == E2TW_CONTINUE) ?
//		return ((result == E2TW_CONTINUE) ?
					E2TW_CONTINUE : E2TW_STOP); //continue if so ordered
	}
	else	//this dir opened successfully
	{
		//if dir-start report for this dir not issued before, do so now
		if ((twdata->exec_flags & E2TW_DEPTH)	//depth-1st
			|| newdepth == 1)	//breadth-1st, but this is the 1st call
		{
			result = (*twdata->cbfunc) (local, &statbuf, E2TW_D, twdata->user_data);

			if (result == E2TW_SKIP_SUBTREE)
			{
				closedir (dp);
				//continue if the problem is down the tree
				//any problem is handled by the relevant tw func and its continued-after-problem flag
				return ((newdepth > 1) ? E2TW_CONTINUE : E2TW_STOP);
//				return E2TW_STOP;
			}
			else if (result == E2TW_STOP)
			{
				closedir (dp);
				return result;
			}
		}

		if (twdata->exec_flags & E2TW_CHDIR)
			e2_fs_chdir_local (local);	//CHECKME error check needed?

//======= DIR OPEN, CWD NOW LOCAL
		parked_dirs = NULL;
		struct dirent entry_store;
		struct dirent *entry;
		gint len1 = strlen (local);
		gint len2, len3;
		//walk it (tree-depth now = newdepth)
		while (readdir_r (dp, &entry_store, &entry) == 0 && entry != NULL)
		{
			if (strcmp (entry->d_name, ".") == 0 || strcmp (entry->d_name, "..") == 0)
				continue;

			//construct item's full path
			//mimic g_strdup_printf() with potentially-non-utf8 strings
			len2 = strlen (entry->d_name);
			len3 = len1 + len2 + sizeof (gchar);
			//space for strings, with joining "/" and trailing '\0'
//			gchar *path = g_malloc ((gulong) (len1 + len2 + sizeof (gchar) + 1));
			gchar *path = g_malloc ((gulong) (len3 + 1));
#ifdef __USE_GNU
			gchar *next = mempcpy (path, local, len1);
			if (*(next-1) != G_DIR_SEPARATOR)	//ascii check is sufficient
				next = mempcpy (next, G_DIR_SEPARATOR_S, sizeof (gchar));
			next = mempcpy (next, entry->d_name, len2);
			*next = '\0';
#else
			memcpy (path, local, len1);
			if (*(path+len1-sizeof (gchar)) == G_DIR_SEPARATOR)
				len1 -= sizeof (gchar);	//fudge next copy
			else
				memcpy (path+len1, G_DIR_SEPARATOR_S, sizeof (gchar));
			memcpy (path+len1+sizeof (gchar), entry->d_name, len2);
			*(path+len3) = '\0';
#endif
			if ((*twdata->twstat) (path, &statbuf))
			{	//process stat failure
				result = (*twdata->cbfunc) (path, &statbuf, E2TW_NS, twdata->user_data);
				if (result == E2TW_STOP)
				{
					if (! (twdata->exec_flags & E2TW_DEPTH)
						&& parked_dirs != NULL)
							e2_list_free_with_data (&parked_dirs);
					closedir (dp);
					g_free (path);
					return result;
				}
			}

			else if (S_ISDIR (statbuf.st_mode))
			{
				if (!(twdata->exec_flags & E2TW_DEPTH))
				{	//breadth-first walk
					//issue the dir-start report here
					result = (*twdata->cbfunc) (path, &statbuf, E2TW_D, twdata->user_data);

					if (result == E2TW_SKIP_SUBTREE)
						result = E2TW_STOP;
					if (result == E2TW_STOP)
					{
						closedir (dp);
						g_free (path);
						return result;
					}

					//log the dir, for later opening
					parked_dirs = g_list_append (parked_dirs, strdup (path));
				}
				else
				{	//depth-first walk, open the dir now
					//(start- and end-reports issued in the recursed function)
					if (_e2_fs_tw (path, newdepth, twdata) != E2TW_CONTINUE)
					{
						//no parks to cleanup if depth-1st walk
						closedir (dp);
						g_free (path);
						return E2TW_STOP;
					}
					//restore system dir to where we are
					if (twdata->exec_flags & E2TW_CHDIR)
						e2_fs_chdir_local (local);	//CHECKME do error-check here in case dir is gone since last opened ?
				}
			}
			else	//not a dir
			{
				if (_e2_fs_tw_nondir (path, &statbuf, twdata) != E2TW_CONTINUE)
				{
					if (! (twdata->exec_flags & E2TW_DEPTH)
						&& parked_dirs != NULL)
							e2_list_free_with_data (&parked_dirs);
					closedir (dp);
					g_free (path);
					return E2TW_STOP;
				}
			}
			g_free (path);
		}
		closedir (dp);
//========= DIR CLOSED, CWD STILL LOCAL
	}

	if (!(twdata->exec_flags & E2TW_DEPTH)
		&& parked_dirs != NULL)
	{	//breadth-first walk
		//process the dirs that haven't yet been opened
		for (tmp = parked_dirs ; tmp != NULL ; tmp = g_list_next (tmp))
		{
			gchar *path = (gchar *) tmp->data;
			//remember ! start-report already issued, do not repeat inside the recurse
			if (_e2_fs_tw (path, newdepth, twdata) != E2TW_CONTINUE)
			{
				e2_list_free_with_data (&parked_dirs);
				//CHECKME dir end-report here, if aborted ?
				return E2TW_STOP;
			}
		}
		e2_list_free_with_data (&parked_dirs);
	}

	(*twdata->twstat) (local, &statbuf);	//freshen the statbuf FIXME another copy instead of refresh ??
	//issue dir-end report
	result = (*twdata->cbfunc) (local, &statbuf, E2TW_DP, twdata->user_data);
	return (result == E2TW_STOP) ? result : E2TW_CONTINUE;
}
/**
 * @brief locate and report all items in directory tree under @a start_dir
 *
 * This is a replacement for nftw(), with user data (to avoid global variables
 * in callback functions, so threads can be used) and extra types of status-
 * report provided to callback functions, and to enable vfs support
 *
 * @param start_item absoute path of item (dir or non-dir) in which to start the walk, localised string
 * @param callback pointer to the function to be called for each located item
 * @param user_data pointer to user-specified data to be supplied to each call to @a callback
 * @param max_depth the largest number of tree levels to descend, or -1 for no limit
 * @param exec_flags flags indicating the way the walk is to be performed
 *
 * @return TRUE if all of the tree was processed in way(s) that @a callback considered ok
 **/
gboolean e2_fs_tw (gchar *start_item, E2_TwResult (*callback) (),
	void *user_data, gint max_depth, E2_TwFlags exec_flags)
{
	gchar *utf, *msg;
	struct stat statbuf;
	//note: stat(), not lstat(), because _e2_fs_tw_nondir() (below)
	//needs to follow a stat()
	if (e2_fs_stat (start_item, &statbuf))
	{
		//handle stat error
		utf = F_FILENAME_FROM_LOCALE (start_item);
		msg = g_strdup_printf (_("Cannot proceed - stat %s failed"), utf);
		e2_output_print_error (msg, TRUE);
		F_FREE (utf);
		return FALSE;
	}

	//park the 'constant' data in the heap where all nested functions can get to it
	E2_TwData *data = g_malloc (sizeof (E2_TwData));

	data->twstat = (exec_flags & E2TW_PHYS) ? e2_fs_lstat : e2_fs_stat;
	data->cbfunc = callback;
	data->user_data = user_data;
	data->max_depth = max_depth;
	data->exec_flags = exec_flags;
//	if (exec_flags & E2TW_MOUNT)
//	{
		data->device = statbuf.st_dev;
//	}

	E2_TwResult result;
	//quick exit if selected item not a dir (or a link to one)
	if (!S_ISDIR (statbuf.st_mode))
	{
		result = _e2_fs_tw_nondir (start_item,
			&statbuf, data);
		g_free (data);
		return (result == E2TW_CONTINUE);
	}

	if (exec_flags & E2TW_CHDIR)
	{
		gchar *current_path;
		gchar *s = NULL;
		if (S_ISDIR (statbuf.st_mode))
			current_path = start_item;
		else
		{
			s = strrchr (start_item, G_DIR_SEPARATOR);
			if (s != start_item)
				*s = '\0';
			else
			{
				current_path = G_DIR_SEPARATOR_S;
				s = NULL;
			}
		}
		gchar *original_path = getcwd (NULL, 0);	//CHECKME error test ?? FIXME vfs
		e2_fs_chdir_local (current_path);
		if (s != NULL)
			*s = G_DIR_SEPARATOR;
		result = _e2_fs_tw (start_item, 0, data);	//DO IT
		e2_fs_chdir_local (original_path);
		g_free (original_path);
	}
	else
		//JUST DO IT
		result = _e2_fs_tw (start_item, 0, data);

	g_free (data);
	return (result == E2TW_CONTINUE);
}
